/*
 * Copyright © 2019 - 2023 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/statvfs.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3cap/enatsc3cap.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <librzrsdk2/redzone_c_api.h>
#include <librzrsdk2/props_api.h>

#include "enatsc3_demod_defs.h"
#include "enatsc3source_defs.h"
#include "enatsc3source.h"
#include "airwavzrzrsdk2source.h"

static char *ensys = "atsc32mpts";
static char *ensubsys = "bbrzrsdk2src";

#define MAX_BASEBAND_PAYLOAD_SIZE 65536

#define MAX_BASEBAND_FILE_SOURCES 32

typedef struct
{
    struct _ENATSC3Source base;

    RedZoneCaptureHandle_t hRedZoneCapture;

    int32_t fOpened;
    ENATSC3BroadcastSystem broadcastSystem;
    ENATSC3SourceWriteCallback ATSC1WriteCallback;
    void *pATSC1WriteCallbackUserData;

    ENVerbosity verbosity;

    enum RedZoneOperatingMode operatingMode;
    int32_t frequencyKHz;

    // The currently tuned PLPs and their slot assignemnts. Most ATSC 3 demods
    // can handle only 4 PLPs, as that is all that is required by spec.
    uint8_t vPLPIds[4];
    uint64_t plpIdMask;

    int32_t fPLPIdMaskChanged;
    pthread_mutex_t plpIdMaskChangedLock;
    ENATSC3ALPHandle_t hATSC3ALP;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    uint64_t basebandPacketNumber;

    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback;
    void *pCallbackUserData;

    struct _stats
    {
        uint8_t masterLock;
        uint64_t plpValidMask;
        uint64_t llsValidMask;
        int32_t snr;
        int32_t rssi;

        // Total packets received
        uint32_t nBasebandPackets;
        // Total bits received
        uint32_t nBytes;
        // Number of packets containing uncorrected errors.
        uint32_t nErroredPackets;

        // ALP statistics
        uint64_t nALPPackets;

        struct _plpStats
        {
            ENATSC3Modulation modulation;
            ENATSC3Coding coding;

            // PLP baseband packets received
            uint32_t nBasebandPackets;
            // PLP ALP packets received
            uint32_t nALPPackets;
            // PLP bits received
            uint32_t nBytes;
            // Number of continuity counter sequence discontinuities detected
            uint32_t nContinuityCounterErrors;
            // Number of packets containing uncorrected errors.
            uint32_t nErroredPackets;
            int32_t lastContinuityCounter;
        } vPLPStats[64];
    } stats;
    int32_t statsCounter;

    pthread_mutex_t statsLock;

    // Frozen capture detection
    uint32_t basebandPacketCountFrozen;
    uint32_t lastBasebandPacketCount;

    struct _basebandPacketListElement
    {
        struct _basebandPacketListElement *pNext;
        uint32_t plpId;
        struct timespec packetTimeUTC; 
        uint64_t packetNumber;
        int32_t fError;
        int32_t fDiscontinuity;
        ENATCS3L1Time l1Time;
        int32_t payloadLength;
        uint8_t vPayload[0];
    } *pBasebandPacketList;
    pthread_mutex_t basebandPacketListLock;

    pthread_mutex_t *pRunningLock;
    pthread_cond_t *pRunningCond;

    pthread_mutex_t lock;
} ENATSC3BasebandRZRSDK2Source;

struct
{
    int32_t fInitialized;

    ENATSC3BasebandRZRSDK2Source vSources[MAX_BASEBAND_FILE_SOURCES];
    int32_t nSources;
} atsc3BasebandRZRSDK2SourceGlobalsG =
{
    .fInitialized = 0,
};

void
ATSC3BasebandRZRSDK2SourceBasebandPacketCallback(
    RedZoneCaptureBasebandPacket *pPacket, void *pUserData)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)pUserData;
    struct _plpStats *pPLPStats;
    int32_t fCCError = 0;
    int32_t payloadLength = 0;
    int32_t i;

    for (i = 0; i < pPacket->v0.nVecs; i++)
    {
        payloadLength += pPacket->v0.vVecs[i].length;
    }

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (!pSource->fOpened)
    {
        ERRORPRINT("Baseband packet on closed source?\n");
        goto out;
    }

    if (pPacket->v0.plpId > 64)
    {
        ERRORPRINT("Invalid PLP Id %d\n", pPacket->v0.plpId);
        pSource->stats.nBasebandPackets++;
        pSource->stats.nErroredPackets++;
        for (i = 0; i < pPacket->v0.nVecs; i++)
        {
            pSource->stats.nBytes += pPacket->v0.vVecs[i].length;
        }
        goto out;
    }

    pSource->stats.nBasebandPackets++;
    pSource->stats.nBytes += payloadLength;

    pPLPStats = pSource->stats.vPLPStats + pPacket->v0.plpId;
    pPLPStats->nBasebandPackets++;
    pPLPStats->nBytes += payloadLength;
    if (pPLPStats->lastContinuityCounter != -1 &&
        ((pPLPStats->lastContinuityCounter + 1) & 0xff) !=
        pPacket->v0.continuityCounter)
    {
        pPLPStats->nContinuityCounterErrors++;
        fCCError = 1;
    }
    pPLPStats->lastContinuityCounter = pPacket->v0.continuityCounter;
    if (pPacket->v0.fError)
    {
        pPLPStats->nErroredPackets++;
        pSource->stats.nErroredPackets++;
    }

    if (pSource->plpIdMask & (1ULL << pPacket->v0.plpId))
    {
        struct _basebandPacketListElement *pBasebandPacketListElement = NULL;

        pBasebandPacketListElement = (struct _basebandPacketListElement *)
            calloc(1, 
            sizeof(struct _basebandPacketListElement) + payloadLength);
        if (!pBasebandPacketListElement)
        {
            ERRORPRINT("Out of memory\n");
            goto out;
        }

        clock_gettime(CLOCK_REALTIME, 
            &pBasebandPacketListElement->packetTimeUTC);

        for (i = 0; i < pPacket->v0.nVecs; i++)
        {
            memcpy(pBasebandPacketListElement->vPayload + 
                pBasebandPacketListElement->payloadLength,
                pPacket->v0.vVecs[i].pData, pPacket->v0.vVecs[i].length);
            pBasebandPacketListElement->payloadLength += 
                pPacket->v0.vVecs[i].length;
        }

        pBasebandPacketListElement->l1Time.precision = 
            pPacket->v0.L1TimePrecision;
        pBasebandPacketListElement->l1Time.seconds = 
            pPacket->v0.L1TimeSeconds;
        pBasebandPacketListElement->l1Time.milliseconds = 
            pPacket->v0.L1TimeMilliseconds;
        pBasebandPacketListElement->l1Time.microseconds = 
            pPacket->v0.L1TimeMicroseconds;
        pBasebandPacketListElement->l1Time.nanoseconds = 
            pPacket->v0.L1TimeNanoseconds;
        pBasebandPacketListElement->plpId = pPacket->v0.plpId;
        pBasebandPacketListElement->packetNumber = 
            pSource->basebandPacketNumber;
        pBasebandPacketListElement->fError = pPacket->v0.fError;
        pBasebandPacketListElement->fDiscontinuity = fCCError;
        pBasebandPacketListElement->payloadLength = payloadLength;

        if (pthread_mutex_lock(&pSource->basebandPacketListLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        if (!pSource->pBasebandPacketList)
        {
            pSource->pBasebandPacketList = pBasebandPacketListElement;
        }
        else
        {
            struct _basebandPacketListElement *pIter = 
                pSource->pBasebandPacketList;
            int count = 0;

            while (pIter->pNext)
            {
                pIter = pIter->pNext;
                count++;
            }
            pIter->pNext = pBasebandPacketListElement;
            if (count && ((count % 256) == 0))
            {
                LOGPRINT_NOFILE("WARN: %d packets queued\n", count);
            }
        }

        if (pthread_mutex_unlock(&pSource->basebandPacketListLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        if (pthread_mutex_lock(pSource->pRunningLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        if (pthread_cond_signal(pSource->pRunningCond))
        {
            ERRORPRINT("Failed to signal\n");
        }

        if (pthread_mutex_unlock(pSource->pRunningLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    pSource->basebandPacketNumber++;

out:
    if (pthread_mutex_unlock(&pSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return;

}

void
ATSC3BasebandRZRSDK2SourceALPPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)pCbUserData;

    pSource->stats.nALPPackets++;
    if (plpId < numof(pSource->stats.vPLPStats))
    {
        pSource->stats.vPLPStats[plpId].nALPPackets++;
    }

    if (ENATSC3ALPProcessALPPacket(pSource->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        if (pSource->verbosity > ENVerbosity_Terse)
        {
            LOGPRINT("Failed to process ALP packet\n");
        }
    }
}

static ENATSC3Modulation
RZRToENTASC3Modulation(uint8_t mod)
{
    switch (mod)
    {
        case RZR_ATSC3_PLP_MOD_QPSK:
            return ENATSC3Modulation_QPSK;
        case RZR_ATSC3_PLP_MOD_16QAM:
            return ENATSC3Modulation_16QAM;
        case RZR_ATSC3_PLP_MOD_64QAM:
            return ENATSC3Modulation_64QAM;
        case RZR_ATSC3_PLP_MOD_256QAM:
            return ENATSC3Modulation_256QAM;
        case RZR_ATSC3_PLP_MOD_1024QAM:
            return ENATSC3Modulation_1024QAM;
        case RZR_ATSC3_PLP_MOD_4096QAM:
            return ENATSC3Modulation_4096QAM;
    }

    return ENATSC3Modulation_Unknown;
}

static uint8_t
RZRToENTASC3Coding(uint8_t cod)
{
    switch (cod)
    {
        case RZR_ATSC3_PLP_COD_2_15:
            return 2;
        case RZR_ATSC3_PLP_COD_3_15:
            return 3;
        case RZR_ATSC3_PLP_COD_4_15:
            return 4;
        case RZR_ATSC3_PLP_COD_5_15:
            return 5;
        case RZR_ATSC3_PLP_COD_6_15:
            return 6;
        case RZR_ATSC3_PLP_COD_7_15:
            return 7;
        case RZR_ATSC3_PLP_COD_8_15:
            return 8;
        case RZR_ATSC3_PLP_COD_9_15:
            return 9;
        case RZR_ATSC3_PLP_COD_10_15:
            return 10;
        case RZR_ATSC3_PLP_COD_11_15:
            return 11;
        case RZR_ATSC3_PLP_COD_12_15:
            return 12;
        case RZR_ATSC3_PLP_COD_13_15:
            return 13;
    }

    return 0;
}

static inline void
GetStatistics(ENATSC3BasebandRZRSDK2Source *pSource)
{
    struct _stats stats = 
    {
        .masterLock = 0
    };
    int32_t fHaveMasterLock = 0;
    int32_t fHavePLPValidMask = 0;
    int32_t fHaveLLSValidMask = 0;
    int32_t fHaveSNR = 0;
    int32_t fHaveRSSI = 0;
    uint64_t plpStatsMask = 0;
    struct RedZonePLPSpecificInfo3 vPLPInfo[64];
    uint64_t mask;
    int32_t i;

    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZoneMasterLockProp, &stats.masterLock,
        sizeof(&stats.masterLock) != RZR_SUCCESS))
    {
        ERRORPRINT("Failed to get master lock\n");
    }
    else
    {
        fHaveMasterLock = 1;
    }

    if (stats.masterLock)
    {
        uint64_t plpValidMask;
        uint64_t llsValidMask;
        int32_t snr;
        int32_t rssi;

        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZonePLPValidBitmaskProp, &plpValidMask, sizeof(plpValidMask)) != 
            RZR_SUCCESS)
        {
            ERRORPRINT("Failed to get plp valid mask\n");
        }
        else
        {
            fHavePLPValidMask = 1;
            stats.plpValidMask = plpValidMask;
        }

        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZoneLLSValidBitmaskProp, &llsValidMask, sizeof(llsValidMask)) != 
            RZR_SUCCESS)
        {
            ERRORPRINT("Failed to get lls plps mask\n");
        }
        else
        {
            fHaveLLSValidMask = 1;
            stats.llsValidMask = llsValidMask;
        }

        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZoneSNRProp, &snr, sizeof(snr)) != RZR_SUCCESS)
        {
            ERRORPRINT("Failed to get snr\n");
        }
        else
        {
            fHaveSNR = 1;
            stats.snr = snr;
        }

        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, RedZoneRSSIProp, 
            &rssi, sizeof(rssi)) != RZR_SUCCESS)
        {
            ERRORPRINT("Failed to get rssi\n");
        }
        else
        {
            fHaveRSSI = 1;
            stats.rssi = rssi;
        }

        memset(vPLPInfo, 0, sizeof(vPLPInfo));
        for (i = 0; i < numof(pSource->vPLPIds); i++)
        {
            if (pSource->vPLPIds[i] != 255)
            {
                int32_t plpId = pSource->vPLPIds[i];
                vPLPInfo[plpId].index = 255;
                vPLPInfo[plpId].plp_id = plpId;
                if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
                    RedZonePLPSpecificInfoProp, vPLPInfo + plpId, 
                    sizeof(vPLPInfo[0])))
                {
                    ERRORPRINT("Failed to get plp %d info\n", plpId);
                    continue;
                }
                plpStatsMask |= 1ULL << plpId;
            }
        }
    }

    RedZoneCaptureStatistics statistics;
    RedZoneCaptureGetStatistics(pSource->hRedZoneCapture, &statistics);

    if ((pSource->statsCounter & 3) == 0)
    {
        LOGPRINT_NOFILE("lock: %d snr: %d rssi: %d Baseband frames: %"PRId64" ALP Packets: %"PRId64" RF errors: %"PRId64" Continuity errors: %"PRId64" Framing errors:%"PRId64"\n",
            stats.masterLock, stats.snr, stats.rssi, statistics.nPackets, 
            pSource->stats.nALPPackets, statistics.nUncorrectedRFErrors,
            statistics.nContinuityCounterErrors, statistics.nFramingErrors);
    }
    pSource->statsCounter++;

    if (pthread_mutex_lock(&pSource->statsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    // Check for frozen capture and restart if needed
    if (stats.masterLock && 
        pSource->lastBasebandPacketCount == statistics.nPackets)
    {
        pSource->basebandPacketCountFrozen++;
        if (pSource->basebandPacketCountFrozen == 3)
        {
            // Force restart of the capture
            ERRORPRINT("Frozen capture detected, restarting\n");
            pSource->fPLPIdMaskChanged = 1;
            pSource->basebandPacketCountFrozen = 0;
        }
    }
    else
    {
        pSource->basebandPacketCountFrozen = 0;
    }
    pSource->lastBasebandPacketCount = statistics.nPackets;

    if (fHaveMasterLock)
    {
       pSource->stats.masterLock = stats.masterLock;
    }

    if (fHavePLPValidMask)
    {
        pSource->stats.plpValidMask = stats.plpValidMask;
    }

    if (fHaveLLSValidMask)
    {
        pSource->stats.llsValidMask = stats.llsValidMask;
    }

    if (fHaveSNR)
    {
        pSource->stats.snr = stats.snr;
    }

    if (fHaveRSSI)
    {
        pSource->stats.rssi = stats.rssi;
    }

    for (i = 0, mask = 1; i < 64; i++, mask <<= 1)
    {
        if (plpStatsMask & mask)
        {
            pSource->stats.vPLPStats[i].modulation = 
                RZRToENTASC3Modulation(vPLPInfo[i].mod);
            pSource->stats.vPLPStats[i].coding = 
                RZRToENTASC3Coding(vPLPInfo[i].cod);
        }
    }

    if (pthread_mutex_unlock(&pSource->statsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }
}

void
ATSC3BasebandRZRSDK2SourceSetVerbosity(ENATSC3Source hSource,
    ENVerbosity verbosity)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;

    pSource->verbosity = verbosity;
}

ENStatus
ATSC3BasebandRZRSDK2SourceRun(ENATSC3Source hSource, int32_t *pFRunning,
    pthread_mutex_t *pRunningLock, pthread_cond_t *pRunningCond)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;
    uint64_t plpIdMask;
    int32_t i;
    int32_t j;
    uint64_t llsValidMask = 0;
    struct timespec tNextStats = {.tv_sec = 0, .tv_nsec = 0};
    ENStatus rval = ENStatus_Success;

    pSource->pRunningLock = pRunningLock;
    pSource->pRunningCond = pRunningCond;

    memset(&pSource->stats, 0, sizeof(pSource->stats));

    plpIdMask =
        (1ULL << (pSource->vPLPIds[0] == 255 ? 0 : pSource->vPLPIds[0])) |
        (1ULL << (pSource->vPLPIds[1] == 255 ? 0 : pSource->vPLPIds[1])) |
        (1ULL << (pSource->vPLPIds[2] == 255 ? 0 : pSource->vPLPIds[2])) |
        (1ULL << (pSource->vPLPIds[3] == 255 ? 0 : pSource->vPLPIds[3]));
    if (ENATSC3BasebandSetPLPs(pSource->hATSC3Baseband, plpIdMask))
    {
        ERRORPRINT("Failed to set baseband processor PLP mask %x\n", 
            plpIdMask);
    }

    if (RedZoneCaptureStart(pSource->hRedZoneCapture))
    {
        ERRORPRINT("Failed to start RedZoneCapture\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(pRunningLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    while (*pFRunning)
    {
        int ptret;
        struct timespec timeout;
        struct timespec mononow;

        clock_gettime(CLOCK_MONOTONIC, &timeout);
        timeout.tv_nsec += NANOSECONDS_PER_SECOND / 10;
        if (timeout.tv_nsec >= NANOSECONDS_PER_SECOND)
        {
            timeout.tv_nsec -= NANOSECONDS_PER_SECOND;
            timeout.tv_sec++;
        }

        ptret = pthread_cond_timedwait(pRunningCond, pRunningLock, &timeout);
        if (ptret && ptret != ETIMEDOUT)
        {
            ERRORPRINT("Failed to wait: %s\n", strerror(ptret));
        }

        if (!*pFRunning)
        {
            break;
        }

        if (pthread_mutex_lock(&pSource->basebandPacketListLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        while (pSource->pBasebandPacketList)
        {
            struct _basebandPacketListElement *pElement;

            pElement = pSource->pBasebandPacketList;
            pSource->pBasebandPacketList = pSource->pBasebandPacketList->pNext;

            if (pthread_mutex_unlock(&pSource->basebandPacketListLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            if (ENATSC3BasebandProcessBasebandPacket(pSource->hATSC3Baseband, 
                pElement->plpId, &pElement->packetTimeUTC, 
                pElement->packetNumber, pElement->fError, 
                pElement->fDiscontinuity, &pElement->l1Time, 
                pElement->vPayload, pElement->payloadLength))
            {
                if (pSource->verbosity > ENVerbosity_Terse)
                {
                    LOGPRINT("Failed to process baseband packet\n");
                }
            }

            free(pElement);

            if (pthread_mutex_lock(&pSource->basebandPacketListLock))
            {
                ERRORPRINT("Failed to lock\n");
            }
        }

        if (pthread_mutex_unlock(&pSource->basebandPacketListLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        clock_gettime(CLOCK_MONOTONIC, &mononow);
        if (mononow.tv_sec > tNextStats.tv_sec ||
            (mononow.tv_sec == tNextStats.tv_sec &&
             mononow.tv_nsec >= tNextStats.tv_nsec))
        {
            GetStatistics(pSource);
            tNextStats = mononow;
            tNextStats.tv_sec++;

            if (pSource->stats.llsValidMask  != llsValidMask)
            {
                struct timespec rtnow;

                clock_gettime(CLOCK_REALTIME, &rtnow);

                llsValidMask = pSource->stats.llsValidMask;
                pSource->LLSPLPsCallback(llsValidMask, 
                    &rtnow, pSource->basebandPacketNumber,
                    pSource->pCallbackUserData);
            }
        }

        if (pSource->fPLPIdMaskChanged)
        {
            uint64_t plpIdMask;
            uint64_t plpIdMaskTemp;
            int32_t fChanged = 0;

            // We're going to snapshot the mask since processing the change 
            // could take a while and we don't want to hold the lock that
            // long. If someone changes the mask again we'll catch it the next
            // time around.
            if (pthread_mutex_lock(&pSource->plpIdMaskChangedLock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            plpIdMask = pSource->plpIdMask;
            pSource->fPLPIdMaskChanged = 0;

            if (pthread_mutex_unlock(&pSource->plpIdMaskChangedLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            // First reconcile the current PLPs with the mask.
            plpIdMaskTemp = plpIdMask;
            for (i = 0; i < numof(pSource->vPLPIds); i++)
            {
                if (pSource->vPLPIds[i] != 255)
                {
                    if (plpIdMaskTemp & (1ULL << pSource->vPLPIds[i]))
                    {
                        // Already set, remove from mask
                        plpIdMaskTemp &= ~1ULL << pSource->vPLPIds[i];
                    }
                    else
                    {
                        // No longer set, free up slot
                        pSource->vPLPIds[i] = 255;
                        fChanged = 1;
                    }
                }
            }

            // Next assign new PLPs to slots
            for (i = 0; i < 64; i++)
            {
                if (plpIdMaskTemp & 1)
                {
                    for (j = 0; j < numof(pSource->vPLPIds); j++)
                    {
                        if (pSource->vPLPIds[j] == 255)
                        {
                            pSource->vPLPIds[j] = i;
                            fChanged = 1;
                            break;
                        }
                    }

                    if (j == numof(pSource->vPLPIds))
                    {
                        ERRORPRINT("Too many plps: couldn't set %d, mask %x, slots %d %d %d %d\n", 
                            i, plpIdMask, pSource->vPLPIds[0], 
                            pSource->vPLPIds[1], pSource->vPLPIds[2], 
                            pSource->vPLPIds[3]);
                    }
                }

                plpIdMaskTemp >>= 1;
            }

            if (fChanged)
            {
                LOGPRINT("Setting PLP mask 0x%x slots %d %d %d %d\n",
                    plpIdMask, pSource->vPLPIds[0], pSource->vPLPIds[1], 
                    pSource->vPLPIds[2], pSource->vPLPIds[3]); 

                RedZonePLPSet val;

                val.plp0_id = pSource->vPLPIds[0];
                val.plp1_id = pSource->vPLPIds[1];
                val.plp2_id = pSource->vPLPIds[2];
                val.plp3_id = pSource->vPLPIds[3];

                if (RedZoneCaptureStop(pSource->hRedZoneCapture))
                {
                    ERRORPRINT("Failed to stop RedZoneCapture\n");
                }

                if (RedZoneCaptureSetProp(pSource->hRedZoneCapture,
                    RedZonePLPSelectionProp, &val, sizeof(val)))
                {
                    ERRORPRINT("Failed to set PLP selection\n");
                }

                if (ENATSC3BasebandSetPLPs(pSource->hATSC3Baseband, 
                    plpIdMask))
                {
                    ERRORPRINT("Failed to set baseband processor PLP mask %x\n",
                        plpIdMask);
                }

                if (RedZoneCaptureStart(pSource->hRedZoneCapture))
                {
                    ERRORPRINT("Failed to start RedZoneCapture\n");
                }
            }
            else
            {
                LOGPRINT("No change to PLP mask 0x%x slots %d %d %d %d\n",
                    plpIdMask, pSource->vPLPIds[0], pSource->vPLPIds[1], 
                    pSource->vPLPIds[2], pSource->vPLPIds[3]); 
            }
        }
    }

    if (pthread_mutex_unlock(pRunningLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (RedZoneCaptureStop(pSource->hRedZoneCapture))
    {
        ERRORPRINT("Failed to stop RedZoneCapture\n");
        rval = ENStatus_Failure;
        goto out;
    }

out:
    return rval;
}

void
ENATSC3BasebandRZRSDK2SourceMPEGTSPacketCallback(
    RedZoneCaptureBasebandPacket *pPacket, void *pUserData)
{
    ENATSC3BasebandRZRSDK2Source *pSource = 
        (ENATSC3BasebandRZRSDK2Source *)pUserData;
    int32_t i;

#if 0
    // XXX Version doesn't appear to be set for ATSC 1.0?
    if (pPacket->version != 0)
    {
        ERRORPRINT("Unhandled RedZoneCaptureBasebandPacket version %d\n",
            pPacket->version);
        goto out;
    }
#endif

    for (i = 0; i < pPacket->v0.nVecs; i++)
    {
        pSource->ATSC1WriteCallback(pPacket->v0.vVecs[i].pData, 
            pPacket->v0.vVecs[i].length, pSource->pATSC1WriteCallbackUserData);
    }

    return;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceOpen_ATSC10(uint64_t tuid,
    int32_t frequencyHz, ENATSC3SourceWriteCallback ATSC1WriteCallback,
    void *pATSC1WriteCallbackUserData, ENATSC3Source *ppSource)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t fSourceLocked = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
    for (i = 0; 
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (pSource->base.tuid == tuid)
        {
            break;
        }
    }

    if (i == atsc3BasebandRZRSDK2SourceGlobalsG.nSources)
    {
        ERRORPRINT("Could't find source %"PRIx64"n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }
    fSourceLocked = 1;

    if (pSource->fOpened)
    {
        ERRORPRINT("Source %"PRIx64" already opened\n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->verbosity = ENVerbosity_Terse;
    pSource->frequencyKHz = frequencyHz/1000 + 3000;
    pSource->operatingMode = RedZoneOperatingModeATSC;

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneOperatingModeProp, &pSource->operatingMode, 
        sizeof(pSource->operatingMode)))
    {
        ERRORPRINT("Failed to set RedZoneOperatingModeProp to %d\n",
            pSource->operatingMode);
        rval = ENStatus_Failure;
        goto errout;
    }

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneFrequencyKHzProp, &pSource->frequencyKHz, 
        sizeof(pSource->frequencyKHz)))
    {
        ERRORPRINT("Failed to set frequency %d KHz\n", pSource->frequencyKHz);
        rval = ENStatus_Failure;
        goto errout;
    }

    if (RedZoneCaptureRegisterCallbacks(pSource->hRedZoneCapture,
        ENATSC3BasebandRZRSDK2SourceMPEGTSPacketCallback, pSource))
    {
        ERRORPRINT("Failed to register MPEGTS packet callback\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->fOpened = 1;
    pSource->broadcastSystem = ENATSC3BroadcastSystem_ATSC10;
    pSource->ATSC1WriteCallback = ATSC1WriteCallback;
    pSource->pATSC1WriteCallbackUserData = pATSC1WriteCallbackUserData;

    *ppSource = (ENATSC3Source)pSource;

out:
    if (fSourceLocked)
    {
        if (pthread_mutex_unlock(&pSource->lock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    return rval;

errout:
    goto out;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceOpen_ATSC30(uint64_t tuid,
    int32_t frequencyHz, ENATSC3ALPHandle_t hATSC3ALP, ENATSC3Source *ppSource)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t fSourceLocked = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
    for (i = 0; 
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (pSource->base.tuid == tuid)
        {
            break;
        }
    }

    if (i == atsc3BasebandRZRSDK2SourceGlobalsG.nSources)
    {
        ERRORPRINT("Could't find source %"PRIx64"n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }
    fSourceLocked = 1;

    if (pSource->fOpened)
    {
        ERRORPRINT("Source %"PRIx64" already opened\n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->broadcastSystem = ENATSC3BroadcastSystem_ATSC30;
    pSource->basebandPacketNumber = 0;
    pSource->hATSC3ALP = hATSC3ALP;
    pSource->plpIdMask = 0;
    pSource->fPLPIdMaskChanged = 0;
    pSource->LLSPLPsCallback = NULL;
    pSource->pCallbackUserData = NULL;
    pSource->verbosity = ENVerbosity_Terse;

    pSource->pBasebandPacketList = NULL;
    pSource->hATSC3Baseband = NULL;
    pSource->frequencyKHz = frequencyHz/1000 + 3000;

    // Set all PLPs invalid. Note that we will actually tune the first slot
    // to PLP 0, since one PLP must be set when initially tuning and 0 seems
    // to always be present in emissions.
    pSource->vPLPIds[0] = 0;
    pSource->vPLPIds[1] = 255;
    pSource->vPLPIds[2] = 255;
    pSource->vPLPIds[3] = 255;

    pSource->operatingMode = RedZoneOperatingModeATSC3;

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneOperatingModeProp, &pSource->operatingMode, 
        sizeof(pSource->operatingMode)))
    {
        ERRORPRINT("Failed to set RedZoneOperatingModeProp to %d\n",
            pSource->operatingMode);
        rval = ENStatus_Failure;
        goto errout;
    }


    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneFrequencyKHzProp, &pSource->frequencyKHz, 
        sizeof(pSource->frequencyKHz)))
    {
        ERRORPRINT("Failed to set frequency %d KHz\n", pSource->frequencyKHz);
        rval = ENStatus_Failure;
        goto errout;
    }

    // Some demods require that we tune at least one PLP, even though at this
    // time we may not know which PLPs exist in the emission. So we'll tune to 
    // PLP 0, which is usually present. The app will eventually select 
    // the PLPs its interested in and a new set of PLPs will be set.
    RedZonePLPSet plpSetVal =
    {
        .plp0_id = pSource->vPLPIds[0],
        .plp1_id = pSource->vPLPIds[1],
        .plp2_id = pSource->vPLPIds[2],
        .plp3_id = pSource->vPLPIds[3],
    };

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZonePLPSelectionProp, &plpSetVal, sizeof(plpSetVal)))
    {
        ERRORPRINT("Failed to set PLP selection\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (RedZoneCaptureRegisterCallbacks(pSource->hRedZoneCapture,
        ATSC3BasebandRZRSDK2SourceBasebandPacketCallback, pSource))
    {
        ERRORPRINT("Failed to register baseband packet callback\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (ENATSC3BasebandOpen(&pSource->hATSC3Baseband))
    {
        ERRORPRINT("Failed to open baseband processor\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (ENATSC3BasebandRegisterCallbacks(pSource->hATSC3Baseband, 
        ATSC3BasebandRZRSDK2SourceALPPacketCallback, pSource))
    {
        ERRORPRINT("Failed to register baseband callbacks\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->fOpened = 1;

    *ppSource = (ENATSC3Source)pSource;

out:
    if (fSourceLocked)
    {
        if (pthread_mutex_unlock(&pSource->lock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    return rval;

errout:
    if (pSource->hATSC3Baseband)
    {
        if (ENATSC3BasebandClose(pSource->hATSC3Baseband))
        {
            ERRORPRINT("Failed to close baseband processor\n");
        }
        pSource->hATSC3Baseband = NULL;
    }

    goto out;
}

ENStatus
ATSC3BasebandRZRSDK2SourceClose(ENATSC3Source hSource)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (RedZoneCaptureRegisterCallbacks(pSource->hRedZoneCapture, NULL, NULL))
    {
        ERRORPRINT("WARN: Failed to unregister baseband packet callback\n");
    }

    if (ENATSC3BasebandClose(pSource->hATSC3Baseband))
    {
        ERRORPRINT("Failed to close baseband processor\n");
    }
    pSource->hATSC3Baseband = NULL;

    while (pSource->pBasebandPacketList)
    {
        struct _basebandPacketListElement *pNext;

        pNext = pSource->pBasebandPacketList->pNext;
        free(pSource->pBasebandPacketList);
        pSource->pBasebandPacketList = pNext;
    }

    pSource->fOpened = 0;

    if (pthread_mutex_unlock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandRZRSDK2SourceSetPLPIds(ENATSC3Source hSource, uint64_t plpIdMask)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->plpIdMaskChangedLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->plpIdMask = plpIdMask;
    pSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pSource->plpIdMaskChangedLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandRZRSDK2SourceRegisterCallbacks(ENATSC3Source pSource,
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback, void *pUserData)
{
    ENATSC3BasebandRZRSDK2Source *pBasebandRZRSDK2Source =
        (ENATSC3BasebandRZRSDK2Source *)pSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pBasebandRZRSDK2Source->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandRZRSDK2Source->LLSPLPsCallback = LLSPLPsCallback;
    pBasebandRZRSDK2Source->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pBasebandRZRSDK2Source->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandRZRDKSourceProbe(ENATSC3BasebandRZRSDK2Source *pSource, 
    DeviceFilter_t *pDeviceFilter)
{
    RedZoneCaptureHandle_t hRedZoneCapture = NULL;
    int32_t fHavePLPIdMaskChangedLock = 0;
    int32_t fHaveStatsLock = 0;
    int32_t fHaveLock = 0;
    int32_t fHaveBasebandPacketListLock = 0;
    int32_t rzrret;
    ENStatus rval = ENStatus_Success;

    rzrret = RedZoneCaptureOpen(&hRedZoneCapture);
    if (rzrret)
    {
        ERRORPRINT("RedZoneCaptureOpen failed %d\n", rzrret);
        rval = ENStatus_Failure;
        goto errout;
    }

    rzrret = RedZoneCaptureInitSelect(hRedZoneCapture, pDeviceFilter);
    if (rzrret)
    {
        rval = ENStatus_NotFound;
        goto errout;
    }

    pSource->hRedZoneCapture = hRedZoneCapture;

    if (RedZoneCaptureGetProp(hRedZoneCapture, 
        RedZoneModelNumberStringProp, pSource->base.strModel, 
        sizeof(pSource->base.strModel)) != RZR_SUCCESS)
    {
        ERRORPRINT("Failed to get model\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    if (RedZoneCaptureGetProp(hRedZoneCapture, 
        RedZoneSerialNumberStringProp, pSource->base.strSerial,
        sizeof(pSource->base.strSerial)) != RZR_SUCCESS)
    {
        ERRORPRINT("Failed to get serial\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    LOGPRINT("RZR %s / %s\n", pSource->base.strModel, pSource->base.strSerial);

    pSource->base.tuid = ((uint64_t)ENATSC3SourceType_BasebandRZRSDK2 << 56) |
       atoll(pSource->base.strSerial);
    if (snprintf(pSource->base.strName, sizeof(pSource->base.strName), "%s:%s",
        pSource->base.strModel, pSource->base.strSerial) >= 
        sizeof(pSource->base.strName))
    {
        ERRORPRINT("Name too long\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    pSource->base.Run = ATSC3BasebandRZRSDK2SourceRun;
    pSource->base.Close = ATSC3BasebandRZRSDK2SourceClose;
    pSource->base.SetPLPIds = ATSC3BasebandRZRSDK2SourceSetPLPIds;
    pSource->base.RegisterCallbacks = 
        ATSC3BasebandRZRSDK2SourceRegisterCallbacks;
    pSource->base.SetVerbosity = ATSC3BasebandRZRSDK2SourceSetVerbosity;

    pSource->fOpened = 0;
    if (pthread_mutex_init(&pSource->lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    if (pthread_mutex_init(&pSource->basebandPacketListLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveBasebandPacketListLock = 1;

    if (pthread_mutex_init(&pSource->plpIdMaskChangedLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHavePLPIdMaskChangedLock = 1;

    if (pthread_mutex_init(&pSource->statsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveStatsLock = 1;

out:
    return rval;

errout:
    if (fHaveStatsLock)
    {
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }
    if (fHavePLPIdMaskChangedLock)
    {
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }
    if (fHaveBasebandPacketListLock)
    {
        if (pthread_mutex_destroy(&pSource->basebandPacketListLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }
    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }
    if (hRedZoneCapture)
    {
        if (RedZoneCaptureClose(hRedZoneCapture))
        {
            ERRORPRINT("Failed to close RedZoneCapture\n");
        }
    }

    goto out;
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-label"
ENStatus
ENATSC3BasebandRZRSDK2SourceInitialize(void)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t i;
    DeviceFilter_t deviceFilter = 
    {
        .model = MODEL_RZR_1200,
        .instance = 0
    };
    ENStatus rval = ENStatus_Success;

    atsc3BasebandRZRSDK2SourceGlobalsG.nSources = 0;

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
    atsc3BasebandRZRSDK2SourceGlobalsG.nSources = 0;

    while (1)
    {
        ENStatus lret;

        if (atsc3BasebandRZRSDK2SourceGlobalsG.nSources ==
            numof(atsc3BasebandRZRSDK2SourceGlobalsG.vSources))
        {
            ERRORPRINT("Too many sources\n");
            break;
        }

        lret = ENATSC3BasebandRZRDKSourceProbe(pSource, &deviceFilter);
        if (lret == ENStatus_NotFound)
        {
            if (deviceFilter.model == MODEL_RZR_1200)
            {
                deviceFilter.model = MODEL_RZR_1400;
                deviceFilter.instance = 0;
                continue;
            }

            break;
        }

        atsc3BasebandRZRSDK2SourceGlobalsG.nSources++;
        pSource++;

        deviceFilter.instance++;
    }

out:
    return rval;

errout:
    for (i = 0, pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->basebandPacketListLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (RedZoneCaptureClose(pSource->hRedZoneCapture))
        {
            ERRORPRINT("Failed to close RedZoneCapture\n");
            pSource->hRedZoneCapture = NULL;
        }
    }

    goto out;
}
#pragma GCC diagnostic pop

ENStatus
ENATSC3BasebandRZRSDK2SourceShutdown(void)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources; 
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (RedZoneCaptureClose(pSource->hRedZoneCapture))
        {
            ERRORPRINT("Failed to close RedZoneCapture\n");
        }
        pSource->hRedZoneCapture = NULL;

        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    return rval;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceQuerySources(ENATSC3SourceDescription *pSources,
    int32_t nSources, int32_t *pNSourcesUsed)
{
    ENATSC3BasebandRZRSDK2Source *pLocalSources;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    *pNSourcesUsed = 0;
    pLocalSources = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
    for (i = 0; 
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources && nSources; 
         i++, nSources--, pLocalSources++, pSources++)
    {
        pSources->tuid = pLocalSources->base.tuid;
        strncpy(pSources->strSerial, pLocalSources->base.strSerial, 
            sizeof(pSources->strSerial));
        pSources->strSerial[sizeof(pSources->strSerial) - 1] = 0;
        strncpy(pSources->strName, pLocalSources->base.strName, 
            sizeof(pSources->strName));
        pSources->strName[sizeof(pSources->strName) - 1] = 0;
        strncpy(pSources->strModel, pLocalSources->base.strModel, 
            sizeof(pSources->strModel));
        pSources->strModel[sizeof(pSources->strModel) - 1] = 0;
    }

    if (i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources)
    {
        ERRORPRINT("WARN: Sources truncated\n");
        rval = ENStatus_BufferTooSmall;
    }

    *pNSourcesUsed = i;

    return rval;
}
