/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <enatsc3cap/enatsc3cap.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include "enatsc3_demod_defs.h"
#include "enatsc3source_defs.h"
#include "enatsc3source.h"
#include "basebandfilesource.h"

static char *ensys = "atsc32mpts";
static char *ensubsys = "bbfilesrc";

#define MAX_PAYLOAD_SIZE (348 * 188)

#define MAX_BASEBAND_FILE_SOURCES 4

typedef struct
{
    struct _ENATSC3Source base;

    int32_t fOpened;

    ENVerbosity verbosity;

    ENATSC3BroadcastSystem broadcastSystem;
    ENATSC3SourceWriteCallback ATSC1WriteCallback;
    void *pATSC1WriteCallbackUserData;
    int ifd;
    int32_t fRealtime;
    uint64_t plpIdMask;
    int32_t fPLPIdMaskChanged;
    ENATSC3ALPHandle_t hATSC3ALP;

    uint64_t packetNumber;
    uint64_t nBytes;

    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback;
    void *pCallbackUserData;

    pthread_mutex_t lock;
} ENATSC3BasebandFileSource;

struct
{
    int32_t fInitialized;
    ENATSC3BasebandFileSource vSources[MAX_BASEBAND_FILE_SOURCES];

    pthread_mutex_t lock;

} atsc3BasebandFileSourceGlobalsG =
{
    .fInitialized = 0,
};

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    ENATSC3BasebandFileSource *pUserData = 
        (ENATSC3BasebandFileSource *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        if (pUserData->verbosity > ENVerbosity_Terse)
        {
            LOGPRINT("Failed to process ALP packet\n");
        }
    }
}

static ENStatus
ATSC3BasebandFileSourceRunATSC1(ENATSC3BasebandFileSource *pSource, 
    int32_t *pFRunning)
{
    int64_t tOffsetSeconds;
    int64_t tOffsetNanoseconds;
    int32_t fHaveTimeOffset = 0;
    uint8_t *pPayload = NULL;
    int32_t fDone = 0;
    ENATSC3CapFileHeader fileHeader;
    int nb;
    ENStatus rval = ENStatus_Success;

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto out;
    }

    nb = read(pSource->ifd, &fileHeader, sizeof(fileHeader));
    if (nb != sizeof(fileHeader))
    {
        ERRORPRINT("Failed to read file header: %s\n", strerror(errno));
        rval = ENStatus_Failure;
        goto out;
    }

    if (ntohs(fileHeader.magic) != 0x656e ||
        fileHeader.format != ENATSC3CapFormat_V2 ||
        fileHeader.type != ENATSC3CapType_ATSC1_MPEGTS)
    {
        ERRORPRINT("bad file header (magic %x format %d type %d)\n",
            ntohs(fileHeader.magic), fileHeader.format, fileHeader.type);
        rval = ENStatus_Failure;
        goto out;
    }

    while (*pFRunning && !fDone)
    {
        ENATSC3CapRecordHeaderV2 header;

        nb = read(pSource->ifd, &header, sizeof(header));
        if (nb != sizeof(header))
        {
            if (nb == 0)
            {
                break;
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading MPEGTS packet header: %s\n",
                    strerror(errno));
                break;
            }
            else
            {
                ERRORPRINT("Short read of MPEGTS packet header (%d != %zd)\n",
                    nb, sizeof(header));
                break;
            }
        }

        header.length = ntohl(header.length);
        header.rtwhen.sec = ENUtils_ntohll(header.rtwhen.sec);
        header.rtwhen.nsec = ntohl(header.rtwhen.nsec);
        header.monowhen.sec = ENUtils_ntohll(header.monowhen.sec);
        header.monowhen.nsec = ntohl(header.monowhen.nsec);

        if (pSource->fRealtime)
        {
            if (!fHaveTimeOffset)
            {
                struct timespec mononow;

                clock_gettime(CLOCK_MONOTONIC_RAW, &mononow);
                tOffsetSeconds = mononow.tv_sec;
                tOffsetSeconds -= header.monowhen.sec;
                tOffsetNanoseconds = mononow.tv_nsec;
                tOffsetNanoseconds -= header.monowhen.nsec;

                fHaveTimeOffset = 1;
            }
            else
            {
                struct timespec mononow;
                int64_t delta;

                while (*pFRunning)
                {
                    clock_gettime(CLOCK_MONOTONIC_RAW, &mononow);
                    delta = tOffsetSeconds - ((int64_t)mononow.tv_sec -
                        (int64_t)header.monowhen.sec);
                    delta *= NANOSECONDS_PER_SECOND;
                    delta += tOffsetNanoseconds - (mononow.tv_nsec -
                        header.monowhen.nsec);

                    if (delta >= NANOSECONDS_PER_SECOND / 100 / 2)
                    {
                        usleep(10000);
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        if (!*pFRunning)
        {
            break;
        }

        while (header.length)
        {
            int32_t nToWrite = MIN(header.length, MAX_PAYLOAD_SIZE);

            nb = read(pSource->ifd, pPayload, nToWrite);
            if (nb != nToWrite)
            {
                if (nb == 0)
                {
                    ERRORPRINT("EOF reading packet payload\n");
                }
                else if (nb < 0)
                {
                    ERRORPRINT("Error reading ts data: %s\n", strerror(errno));
                }
                else
                {
                    ERRORPRINT("Short read of ts data (%d != %zd)\n",
                        nb, sizeof(header));
                }
                fDone = 1;
                break;
            }

            if (header.type == ENATSC3CapPayloadType_ATSC1_MPEGTS)
            {
                if (pSource->ATSC1WriteCallback(pPayload, nToWrite,
                    pSource->pATSC1WriteCallbackUserData))
                {
                    ERRORPRINT("WARN: Failed to write payload\n");
                }
            }

            header.length -= nToWrite;
        }
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

static ENStatus
ATSC3BasebandFileSourceRunATSC3(ENATSC3BasebandFileSource *pSource, 
    int32_t *pFRunning)
{
    int64_t tOffsetSeconds;
    int64_t tOffsetNanoseconds;
    int32_t fHaveTimeOffset = 0;
    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3CapFileHeader fileHeader;
    uint8_t *pPayload = NULL;
    uint64_t plpIdMaskLocal;
    uint64_t llsValidMask = 0;
    int32_t lastCC = -1;
    ENStatus rval = ENStatus_Success;

    pSource->packetNumber = 0;
    pSource->nBytes = 0;

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENATSC3BasebandOpen(&hATSC3Baseband))
    {
        ERRORPRINT("Failed to open baseband processor\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENATSC3BasebandRegisterCallbacks(hATSC3Baseband, ALPPacketCallback, 
        pSource))
    {
        ERRORPRINT("Failed to register baseband callbacks\n");
        rval = ENStatus_Failure;
        goto out;
    }

    plpIdMaskLocal = pSource->plpIdMask;
    if (ENATSC3BasebandSetPLPs(hATSC3Baseband, plpIdMaskLocal))
    {
        ERRORPRINT("Failed to set baseband processor PLP mask to %x\n",
            plpIdMaskLocal);
        rval = ENStatus_Failure;
        goto out;
    }

    if (read(pSource->ifd, &fileHeader, sizeof(fileHeader)) !=
        sizeof(fileHeader))
    {
        ERRORPRINT("Failed to read file header\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fileHeader.type != ENATSC3CapType_ATSC3_Baseband)
    {
        ERRORPRINT("Unhandled format %d or type %d\n",
            fileHeader.format, fileHeader.type);
        rval = ENStatus_Failure;
        goto out;
    }

    while (*pFRunning)
    {
        ENATSC3CapRecordHeaderV2 header;
        int nb;
        int32_t fCCError = 0;
        int32_t payloadLength;
        struct timespec packetTime;

        nb = read(pSource->ifd, &header, sizeof(header));
        if (nb != sizeof(header))
        {
            if (nb == 0)
            {
                LOGPRINT_NOFILE("End of file\n");
                break;
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading rzrcap packet header: %s\n",
                    strerror(errno));
                rval = ENStatus_Failure;
                break;
            }
            else
            {
                ERRORPRINT("Short read of rzrcap packet header (%d != %zd)\n",
                    nb, sizeof(header));
                rval = ENStatus_Failure;
                break;
            }
        }

        header.length = ntohl(header.length);
        header.rtwhen.sec = ENUtils_ntohll(header.rtwhen.sec);
        header.rtwhen.nsec = ntohl(header.rtwhen.nsec);
        header.monowhen.sec = ENUtils_ntohll(header.monowhen.sec);
        header.monowhen.nsec = ntohl(header.monowhen.nsec);

        if (lastCC != -1)
        {
            if (header.continuityCounter != ((lastCC + 1) & 0xff))
            {
                fCCError = 1;
            }
        }
        lastCC = header.continuityCounter;

        if (header.length > MAX_PAYLOAD_SIZE)
        {
            ERRORPRINT("Baseband packet too large (%d > %d)\n",
                header.length, MAX_PAYLOAD_SIZE);
            rval = ENStatus_Failure;
            break;
        }

        payloadLength = header.length;
        packetTime.tv_sec = header.rtwhen.sec;
        packetTime.tv_nsec = header.rtwhen.nsec;

        nb = read(pSource->ifd, pPayload, header.length);
        if (nb != header.length)
        {
            if (nb == 0)
            {
                ERRORPRINT("EOF reading packet payload\n");
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading baseband packet: %s\n",
                    strerror(errno));
            }
            else
            {
                ERRORPRINT("Short read of baseband packet (%d != %zd)\n",
                    nb, sizeof(header));
            }

            rval = ENStatus_Failure;
            break;
        }

        if (pSource->fRealtime)
        {
            if (!fHaveTimeOffset)
            {
                struct timespec mononow;

                clock_gettime(CLOCK_MONOTONIC_RAW, &mononow);
                tOffsetSeconds = mononow.tv_sec;
                tOffsetSeconds -= header.monowhen.sec;
                tOffsetNanoseconds = mononow.tv_nsec;
                tOffsetNanoseconds -= header.monowhen.nsec; 

                fHaveTimeOffset = 1;
            }
            else
            {
                struct timespec mononow;
                int64_t delta;

                while (1)
                {
                    clock_gettime(CLOCK_MONOTONIC_RAW, &mononow);
                    delta = tOffsetSeconds - ((int64_t)mononow.tv_sec -
                        (int64_t)header.monowhen.sec);
                    delta *= NANOSECONDS_PER_SECOND;
                    delta += tOffsetNanoseconds - (mononow.tv_nsec -
                        header.monowhen.nsec);

                    if (delta >= NANOSECONDS_PER_SECOND / 100 / 2)
                    {
                        usleep(10000);
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        if (header.type == ENATSC3CapPayloadType_Demod_V2)
        {
            uint64_t currentLLSValidMask;

            if (fileHeader.format == ENATSC3CapFormat_V2)
            {
                ENATSC3CapDemodulatorRecordV2 *pDemodRecord;

                // This is a tuner/demod statistics packet
                pDemodRecord = (ENATSC3CapDemodulatorRecordV2 *)pPayload;
                currentLLSValidMask =
                    ENUtils_ntohll(pDemodRecord->atsc3.llsPresentMask);
            }
            else
            {
                ERRORPRINT("Unhandled file format %d\n", fileHeader.format);
                continue;
            }

            if (currentLLSValidMask != llsValidMask)
            {
                if (pSource->LLSPLPsCallback)
                {
                    pSource->LLSPLPsCallback(currentLLSValidMask,
                        &packetTime, pSource->packetNumber,
                        pSource->pCallbackUserData);
                }
                llsValidMask = currentLLSValidMask;
            }
            continue;
        }

        if (header.type != ENATSC3CapPayloadType_ATSC3_Baseband)
        {
            ERRORPRINT("Unexpected capture record type %d\n",
                header.type);
            continue;
        }

        // See if the plp mask has changed
        if (pSource->fPLPIdMaskChanged)
        {
            if (pthread_mutex_lock(&pSource->lock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            plpIdMaskLocal = pSource->plpIdMask;
            pSource->fPLPIdMaskChanged = 0;

            if (pthread_mutex_unlock(&pSource->lock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            if (ENATSC3BasebandSetPLPs(hATSC3Baseband, plpIdMaskLocal))
            {
                ERRORPRINT(
                    "Failed to update baseband processor PLP mask to %x\n",
                    plpIdMaskLocal);
            }
        }

        if (plpIdMaskLocal & (1ULL << header.plpId))
        {
            ENATCS3L1Time l1time;

            l1time.precision = header.broadcastTime.L1B_time_info_flag;
            l1time.seconds = htonl(header.broadcastTime.L1D_time_sec);
            l1time.milliseconds = htonl(header.broadcastTime.L1D_time_msec);
            l1time.microseconds = htonl(header.broadcastTime.L1D_time_usec);
            l1time.nanoseconds = htonl(header.broadcastTime.L1D_time_nsec);

            if (ENATSC3BasebandProcessBasebandPacket(hATSC3Baseband, 
                header.plpId, &packetTime, pSource->packetNumber, 
                header.fError, fCCError, &l1time, pPayload, payloadLength))
            {
                ERRORPRINT("Failed to process baseband packet\n");
            }
        }

        pSource->packetNumber++;
        pSource->nBytes += payloadLength;
    }

    if (ENATSC3BasebandSetPLPs(hATSC3Baseband, 0))
    {
        ERRORPRINT("Failed to set baseband processor PLP mask to 0\n");
        rval = ENStatus_Failure;
        goto out;
    }

out:
    if (hATSC3Baseband)
    {
        ENATSC3BasebandClose(hATSC3Baseband);
    }

    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

static ENStatus
ATSC3BasebandFileSourceRun(ENATSC3Source hSource, int32_t *pFRunning,
    pthread_mutex_t *pRunningLock, pthread_cond_t *pRunningCond)
{
    ENATSC3BasebandFileSource *pSource = (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    switch (pSource->broadcastSystem)
    {
        case ENATSC3BroadcastSystem_ATSC10:
            rval = ATSC3BasebandFileSourceRunATSC1(pSource, pFRunning); 
            break;
        case ENATSC3BroadcastSystem_ATSC30:
            rval = ATSC3BasebandFileSourceRunATSC3(pSource, pFRunning); 
            break;
        default:
            ERRORPRINT("Unhandled broadcast system %d\n", 
                pSource->broadcastSystem);
            rval = ENStatus_Failure;
            break;
    }

    return rval;
}

ENStatus
ATSC3BasebandFileSourceClose(ENATSC3Source hSource)
{
    ENATSC3BasebandFileSource *pBasebandFileSource = 
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandFileSource->fOpened = 0;

    if (pthread_mutex_unlock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandFileSourceSetPLPIds(ENATSC3Source hSource, uint64_t plpIdMask)
{
    ENATSC3BasebandFileSource *pBasebandFileSource =
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandFileSource->plpIdMask = plpIdMask;
    pBasebandFileSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandFileSourceRegisterCallbacks(ENATSC3Source hSource,
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback, void *pUserData)
{
    ENATSC3BasebandFileSource *pBasebandFileSource =
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandFileSource->LLSPLPsCallback = LLSPLPsCallback;
    pBasebandFileSource->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

void
ATSC3BasebandFileSourceSetVerbosity(ENATSC3Source hSource,
    ENVerbosity verbosity)
{
    ENATSC3BasebandFileSource *pSource = (ENATSC3BasebandFileSource *)hSource;

    pSource->verbosity = verbosity;
}

ENStatus
ENATSC3BasebandFileSourceInitialize(void)
{
    int32_t i = 0;
    int32_t fHaveLock = 0;
    ENStatus rval = ENStatus_Success;
    if (pthread_mutex_init(&atsc3BasebandFileSourceGlobalsG.lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    for (i = 0; i < numof(atsc3BasebandFileSourceGlobalsG.vSources); i++)
    {
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.tuid =
            ((uint64_t)ENATSC3SourceType_BasebandFile << 56) | i;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.Run =
            ATSC3BasebandFileSourceRun;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.Close =
            ATSC3BasebandFileSourceClose;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.SetPLPIds =
            ATSC3BasebandFileSourceSetPLPIds;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.RegisterCallbacks =
            ATSC3BasebandFileSourceRegisterCallbacks;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.SetVerbosity =
            ATSC3BasebandFileSourceSetVerbosity;
        strncpy(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel,
            "bbfile",
            sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel) - 1);
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel[sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel) - 1] = 0;
        if (snprintf(
            atsc3BasebandFileSourceGlobalsG.vSources[i].base.strSerial,
            sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strSerial),
            "%d", i) >= sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strSerial))
        {
            ENUtils_CrashAssert(0 && "Serial too long");
        };
        if (snprintf(
            atsc3BasebandFileSourceGlobalsG.vSources[i].base.strName,
            sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strName),
            "bbandfts%d", i) >= sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strName))
        {
            ENUtils_CrashAssert(0 && "Name too long");
        };
        atsc3BasebandFileSourceGlobalsG.vSources[i].fOpened = 0;
        if (pthread_mutex_init(
            &atsc3BasebandFileSourceGlobalsG.vSources[i].lock, NULL))
        {
            ERRORPRINT("Failed to init lock\n");
            for (i--; i >= 0; i--)
            {
                if (pthread_mutex_destroy(
                    &atsc3BasebandFileSourceGlobalsG.vSources[i].lock))
                {
                    ERRORPRINT("Failed to destroy lock\n");
                }
            }
            goto errout;
        }
    }

out:
    return rval;

errout:
    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&atsc3BasebandFileSourceGlobalsG.lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    goto out;
}

ENStatus
ENATSC3BasebandFileSourceShutdown(void)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < numof(atsc3BasebandFileSourceGlobalsG.vSources); i++)
    {
        if (pthread_mutex_destroy(
            &atsc3BasebandFileSourceGlobalsG.vSources[i].lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (pthread_mutex_destroy(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandFileSourceQuerySources(ENATSC3SourceDescription *pSources,
    int32_t nSources, int32_t *pNSourcesUsed)
{
    ENATSC3BasebandFileSource *pLocalSources;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    *pNSourcesUsed = 0;
    pLocalSources = atsc3BasebandFileSourceGlobalsG.vSources;
    for (i = 0;
         i < numof(atsc3BasebandFileSourceGlobalsG.vSources) && nSources;
         i++, nSources--, pLocalSources++, pSources++)
    {
        pSources->tuid = pLocalSources->base.tuid;
        strncpy(pSources->strSerial, pLocalSources->base.strSerial,
            sizeof(pSources->strSerial));
        pSources->strSerial[sizeof(pSources->strSerial) - 1] = 0;
        strncpy(pSources->strName, pLocalSources->base.strName,
            sizeof(pSources->strName));
        pSources->strName[sizeof(pSources->strName) - 1] = 0;
        strncpy(pSources->strModel, pLocalSources->base.strModel,
            sizeof(pSources->strModel));
        pSources->strModel[sizeof(pSources->strModel) - 1] = 0;
    }

    if (i < numof(atsc3BasebandFileSourceGlobalsG.vSources))
    {
        ERRORPRINT("WARN: Sources truncated\n");
        rval = ENStatus_BufferTooSmall;
    }

    *pNSourcesUsed = i;

    return rval;
}

static ENStatus
ENATSC3BasebandFileSourceOpen_Internal(uint64_t tuid, 
    ENATSC3BroadcastSystem broadcastSystem, int ifd, int32_t fRealtime, 
    ENATSC3ALPHandle_t hATSC3ALP, 
    ENATSC3SourceWriteCallback ATSC1WriteCallback, 
    void *pATSC1WriteCallbackUserData, ENATSC3Source *ppSource)
{
    ENATSC3BasebandFileSource *pSource;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pSource = atsc3BasebandFileSourceGlobalsG.vSources;
    for (i = 0;
         i < numof(atsc3BasebandFileSourceGlobalsG.vSources);
         i++, pSource++)
    {
        if (pSource->base.tuid == tuid)
        {
            break;
        }
    }

    if (i == numof(atsc3BasebandFileSourceGlobalsG.vSources))
    {
        ERRORPRINT("Could't find source %"PRIx64"n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pSource->fOpened)
    {
        ERRORPRINT("Source %"PRIx64" already opened\n", tuid);
        rval = ENStatus_Failure;
        goto unlockout;
    }

    if (lseek(ifd, 0, SEEK_SET) != 0)
    {
        ERRORPRINT("Failed to seek input file: %s\n", strerror(errno));
        rval = ENStatus_Failure;
        goto unlockout;
    }

    pSource->fOpened = 1;
    pSource->broadcastSystem = broadcastSystem;
    pSource->ATSC1WriteCallback = ATSC1WriteCallback;
    pSource->pATSC1WriteCallbackUserData = pATSC1WriteCallbackUserData;
    pSource->ifd = ifd;
    pSource->fRealtime = fRealtime;
    pSource->hATSC3ALP = hATSC3ALP;
    pSource->plpIdMask = 0;
    pSource->fPLPIdMaskChanged = 0;
    pSource->LLSPLPsCallback = NULL;
    pSource->pCallbackUserData = NULL;

    pSource->verbosity = ENVerbosity_Terse;

    *ppSource = (ENATSC3Source)pSource;

unlockout:
    if (pthread_mutex_unlock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

out:
    return rval;
}

ENStatus
ENATSC3BasebandFileSourceOpen_ATSC10(uint64_t tuid, int ifd, int32_t fRealtime, 
    ENATSC3SourceWriteCallback ATSC1WriteCallback, 
    void *pATSC1WriteCallbackUserData, ENATSC3Source *ppSource)
{
    ENStatus rval;

    rval = ENATSC3BasebandFileSourceOpen_Internal(tuid, 
        ENATSC3BroadcastSystem_ATSC10, ifd, fRealtime, NULL, 
        ATSC1WriteCallback, pATSC1WriteCallbackUserData, ppSource);

    return rval;
}

ENStatus
ENATSC3BasebandFileSourceOpen_ATSC30(uint64_t tuid, int ifd, int32_t fRealtime, 
    ENATSC3ALPHandle_t hATSC3ALP, ENATSC3Source *ppSource)
{
    ENStatus rval;

    rval = ENATSC3BasebandFileSourceOpen_Internal(tuid, 
        ENATSC3BroadcastSystem_ATSC30, ifd, fRealtime, hATSC3ALP, NULL, NULL, 
        ppSource);

    return rval;
}
