/*
 * Copyright © 2022-2023 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of
 * the following permission added to Section 15 as permitted in Section 7(a):
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/statvfs.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <curl/curl.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3cap/enatsc3cap.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3j288alp/libenatsc3j288alp.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <libhdhomerun/hdhomerun.h>

#include "enatsc3_demod_defs.h"
#include "enatsc3source_defs.h"
#include "enatsc3source.h"

#include "hdhrj288alpsource.h"

static char *ensys = "atsc32mpts";
static char *ensubsys = "j288alphdhrsrc";


#define MAX_BASEBAND_PAYLOAD_SIZE 65536

#define MAX_SOURCES 32

#define INVALID_PLP_ID 255

#define HDHR_MAX_TUNERS 128
#define DISCOVER_DATA_EXPIRATION_SECONDS 5

static struct 
{
    char *pString;
    int32_t sString;
    ENATSC3Modulation mod;

} vModTableG[] = 
{
    {"qpsk", sizeof("qpsk") - 1, ENATSC3Modulation_QPSK},
    {"qam16", sizeof("qam16") - 1, ENATSC3Modulation_16QAM},
    {"qam64", sizeof("qam64") - 1, ENATSC3Modulation_64QAM},
    {"qam256", sizeof("qam256") - 1, ENATSC3Modulation_256QAM},
    {"qam1024", sizeof("qam1024") - 1, ENATSC3Modulation_1024QAM},
    {"qam4096", sizeof("qam4096") - 1, ENATSC3Modulation_4096QAM},
};

static struct 
{
    char *pString;
    int32_t sString;
    ENATSC3Coding cod;

} vCodTableG[] = 
{
    {"2/15", sizeof("2/15") - 1, ENATSC3Coding_2_15},
    {"3/15", sizeof("3/15") - 1, ENATSC3Coding_3_15},
    {"4/15", sizeof("4/15") - 1, ENATSC3Coding_4_15},
    {"5/15", sizeof("5/15") - 1, ENATSC3Coding_5_15},
    {"6/15", sizeof("6/15") - 1, ENATSC3Coding_6_15},
    {"7/15", sizeof("7/15") - 1, ENATSC3Coding_7_15},
    {"8/15", sizeof("8/15") - 1, ENATSC3Coding_8_15},
    {"9/15", sizeof("9/15") - 1, ENATSC3Coding_9_15},
    {"10/15", sizeof("10/15") - 1, ENATSC3Coding_10_15},
    {"11/15", sizeof("11/15") - 1, ENATSC3Coding_11_15},
    {"12/15", sizeof("12/15") - 1, ENATSC3Coding_12_15},
    {"13/15", sizeof("13/15") - 1, ENATSC3Coding_13_15},
};

typedef struct
{
    struct _ENATSC3Source base;

    int32_t fOpened;
    int32_t fRunning;

    ENATSC3BroadcastSystem broadcastSystem;

    // ATSC1 variables
    struct
    {
        ENATSC3SourceWriteCallback ATSC1WriteCallback;
        void *pATSC1WriteCallbackUserData;
        // 174 TS packets is the maximum supported by the entangle atsc30rtp
        // TS Source
        uint8_t vTSPackets[188 * 174 * 4];
        int32_t nTSPacketBytesUsed;
        struct timespec tLastWrite;
        pthread_mutex_t writeCbLock;
    } atsc1;

    int32_t fRestartTransferThread;
    pthread_mutex_t restartTransferThreadLock;

    ENVerbosity verbosity;

    struct hdhomerun_discover_device_t hdhr;

    // the tuner instance on the HDHR
    int32_t tunerNum;

    int32_t frequencyKHz;
    int32_t rfChannelNumber;

    // The currently tuned PLPs and their slot assignemnts. Most ATSC 3 demods
    // can handle only 4 PLPs, as that is all that is required by spec.
    uint8_t vPLPIds[4];
    uint64_t plpIdMask;
    uint64_t requestedPlpIdMask;
    int32_t singlePLPId;

    int32_t fPLPIdMaskChanged;
    pthread_mutex_t plpIdMaskChangedLock;
    ENATSC3ALPHandle_t hATSC3ALP;

    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback;
    void *pCallbackUserData;

    pthread_t transferThread;
    pthread_t plpStatusThread;

    ENATSC3J288ALPHandle_t hATSC3J288ALP;

    // It seems that after a retune the HDHR can have about a baseband packet's
    // worth of stale data which it will emit on the next tune. We set a 
    // countdown to discard the first n bytes, and also a timeout which will
    // fire if we don't receive that number of bytes within the timeout.
    int32_t discardCountdown;
    struct timespec discardMonoTimeout;

    struct _stats
    {
        uint8_t masterLock;
        uint64_t plpValidMask;
        uint64_t plpLockedMask;
        uint64_t llsValidMask;
        int32_t snr;
        int32_t rssi;

        // Total bits received
        uint32_t nBytes;
        // Number of packets containing uncorrected errors.
        uint32_t nErroredPackets;

        // ALP statistics
        uint64_t nALPPackets;

        struct _plpStats
        {
            ENATSC3Modulation modulation;
            ENATSC3Coding coding;

            // PLP ALP packets received
            uint32_t nALPPackets;
            // PLP bits received
            uint32_t nBytes;
            // Number of continuity counter sequence discontinuities detected
            uint32_t nContinuityCounterErrors;
            // Number of packets containing uncorrected errors.
            uint32_t nErroredPackets;
            int32_t lastContinuityCounter;
        } vPLPStats[64];
    } stats;

    pthread_mutex_t statsLock;

    ENATSC3J288PTPInfo ptpInfo;
    int32_t fHavePTPInfo;
    pthread_mutex_t ptpInfoLock;

    int curlSocket;
    pthread_mutex_t curlSocketLock;

    pthread_mutex_t lock;
} ENATSC3J288ALPHDHRSource;

struct
{
    int32_t fInitialized;

    ENATSC3J288ALPHDHRSource vSources[MAX_SOURCES];
    int32_t nSources;
} atsc3J288ALPHDHRSourceGlobalsG =
{
    .fInitialized = 0,
};

void 
ATSC3J288ALPHDHRSourcePTPCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3J288PTPInfo *pPTPInfo, void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;

    if (pthread_mutex_lock(&pSource->ptpInfoLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->ptpInfo = *pPTPInfo;
    pSource->fHavePTPInfo = 1;

    if (pthread_mutex_unlock(&pSource->ptpInfoLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }
}

void
ATSC3J288ALPHDHRSourceALPPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;

    // If only one PLP is requested then the sony demod does not include the 
    // extension header indicating the source PLP Id, so it arrives as 255.
    // We need to figure it out here.
    if (plpId == INVALID_PLP_ID)
    {
        if (pSource->singlePLPId != INVALID_PLP_ID)
        {
            plpId = pSource->singlePLPId;
        }
    }

    pSource->stats.nALPPackets++;
    pSource->stats.nBytes += sPacket;
    if (plpId < numof(pSource->stats.vPLPStats))
    {
        pSource->stats.vPLPStats[plpId].nALPPackets++;
        pSource->stats.vPLPStats[plpId].nBytes += sPacket;
    }

    if (ENATSC3ALPProcessALPPacket(pSource->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        if (pSource->verbosity > ENVerbosity_Terse)
        {
            LOGPRINT("Failed to process ALP packet\n");
        }
    }
}

void
ATSC3J288ALPHDHRSourceBasebandPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
}

void
ATSC3J288ALPHDHRSourceSetVerbosity(ENATSC3Source hSource,
    ENVerbosity verbosity)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;

    pSource->verbosity = verbosity;
}

static curl_socket_t 
storeCurlSocket(void *pCbUserData, curl_socket_t curlfd, curlsocktype purpose)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;

    if (pthread_mutex_lock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pSource->fRunning)
    {
        pSource->curlSocket = curlfd;
        unsigned int rx_size = 1024 * 1024;
        if (setsockopt(pSource->curlSocket, SOL_SOCKET, SO_RCVBUFFORCE, 
            &rx_size, sizeof(rx_size)) < 0)
        {
            if (errno == EPERM)
            {
                if (setsockopt(pSource->curlSocket, SOL_SOCKET, SO_RCVBUF, 
                    &rx_size, sizeof(rx_size)) < 0)
                {
                    ERRORPRINT("WARN: Failed to set SO_RCVBUF: %s\n",
                        strerror(errno));
                }
            }
            else
            {
                ERRORPRINT("WARN: Failed to set SO_RCVBUFFORCE: %s\n",
                    strerror(errno));
            }
        }

        socklen_t len = sizeof(rx_size);
        if (getsockopt(pSource->curlSocket, SOL_SOCKET, SO_RCVBUF,
            (void *)&rx_size, &len) < 0)
        {
            ERRORPRINT("WARN: Failed to get SO_RCVBUF: %s\n", strerror(errno));
        }

        if (rx_size != 2 * 1024 * 1024)
        {
            ERRORPRINT("Failed to set RCVBUF, SO_RCVBUF %d\n", rx_size);
        }
    }
    else
    {
        ERRORPRINT("NOT RUNNING\n");
    }

    if (pthread_mutex_unlock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return CURL_SOCKOPT_OK;
}

static void
closeCurlSocket(ENATSC3J288ALPHDHRSource *pSource)
{
    if (pthread_mutex_lock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pSource->curlSocket != -1)
    {
        close(pSource->curlSocket);
        pSource->curlSocket = -1;
    }

    if (pthread_mutex_unlock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to lock\n");
    }
}

static size_t
curlWriteFunction_ATSC10(void *ptr, size_t size, size_t nmemb, 
    void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;
    uint8_t *pData = (uint8_t *)ptr;
    int32_t nb = size * nmemb;
    int32_t nToCopy;
    struct timespec mononow;
    int64_t deltans;
    size_t rval = size * nmemb;

    if (pthread_mutex_lock(&pSource->atsc1.writeCbLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    clock_gettime(CLOCK_MONOTONIC, &mononow);
    deltans = mononow.tv_sec - pSource->atsc1.tLastWrite.tv_sec;
    deltans *= NANOSECONDS_PER_SECOND;
    deltans += mononow.tv_nsec - pSource->atsc1.tLastWrite.tv_nsec;

    do
    {
        if (nb)
        {
            nToCopy = MIN(
                sizeof(pSource->atsc1.vTSPackets) - 
                pSource->atsc1.nTSPacketBytesUsed, nb);
            memcpy(pSource->atsc1.vTSPackets + 
                pSource->atsc1.nTSPacketBytesUsed, pData, nToCopy);
            pData += nToCopy;
            nb -= nToCopy;
            pSource->atsc1.nTSPacketBytesUsed += nToCopy;
        }

        if (pSource->atsc1.nTSPacketBytesUsed == 
            sizeof(pSource->atsc1.vTSPackets) || 
            deltans > NANOSECONDS_PER_SECOND / 10)
        {
            nToCopy = (pSource->atsc1.nTSPacketBytesUsed / 188) * 188;
            pSource->atsc1.ATSC1WriteCallback(pSource->atsc1.vTSPackets, 
                nToCopy, pSource->atsc1.pATSC1WriteCallbackUserData);
            if (nToCopy < pSource->atsc1.nTSPacketBytesUsed)
            {
                memmove(pSource->atsc1.vTSPackets, pSource->atsc1.vTSPackets + 
                    nToCopy, pSource->atsc1.nTSPacketBytesUsed - nToCopy);
            }
            pSource->atsc1.nTSPacketBytesUsed -= nToCopy;
            pSource->atsc1.tLastWrite = mononow;
            deltans = 0;
        }
    } while (nb);

    if (pthread_mutex_unlock(&pSource->atsc1.writeCbLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

static size_t
curlWriteFunction_ATSC30(void *ptr, size_t size, size_t nmemb, 
    void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;
    struct timespec utcnow;
    size_t rval = size * nmemb;

    if (pSource->discardCountdown)
    {
        pSource->discardCountdown -= size * nmemb;
        {
            if (pSource->discardCountdown < 0)
            {
                pSource->discardCountdown = 0;
            }
        }

        if (pSource->discardCountdown)
        {
            struct timespec mononow;

            clock_gettime(CLOCK_MONOTONIC, &mononow);
            if (mononow.tv_sec > pSource->discardMonoTimeout.tv_sec ||
                (mononow.tv_sec == pSource->discardMonoTimeout.tv_sec &&
                 mononow.tv_nsec > pSource->discardMonoTimeout.tv_nsec))
            {
                LOGPRINT("Timed out waiting for stale data to drain\n");
                pSource->discardCountdown = 0;
            }
        }
        goto out;
    }

    clock_gettime(CLOCK_REALTIME, &utcnow);
    if (ENATSC3J288ALPProcessJ288Packets(pSource->hATSC3J288ALP, 
        &utcnow, (uint8_t *)ptr, size * nmemb))
    {
        ERRORPRINT("Failed to process J288 packets\n");
    }

out:
    return rval;
}

void *
ATSC3J288ALPHDHRSourceTransferThread_ATSC10(void *arg)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)arg;
    CURLcode res;
    char strError[CURL_ERROR_SIZE];
    char strURL[128];
    CURL *curl = NULL;

    pSource->atsc1.nTSPacketBytesUsed = 0;
    pSource->atsc1.tLastWrite.tv_sec = 0;
    pSource->atsc1.tLastWrite.tv_nsec = 0;

    if (snprintf(strURL, sizeof(strURL),
        "http://%d.%d.%d.%d:5004/tuner%d/ch%d?duration=311040000",
        (pSource->hdhr.ip_addr >> 24) & 0xff, 
        (pSource->hdhr.ip_addr >> 16) & 0xff, 
        (pSource->hdhr.ip_addr >> 8) & 0xff, 
        (pSource->hdhr.ip_addr) & 0xff, 
        pSource->tunerNum, pSource->rfChannelNumber ) >=
        sizeof(strURL))
    {
        ERRORPRINT("URL too long\n");
        goto out;
    }

    curl = curl_easy_init();
    if (!curl)
    {
        ERRORPRINT("Out of memory\n");
        goto out;
    }

    if (curl_easy_setopt(curl, CURLOPT_SOCKOPTFUNCTION, storeCurlSocket) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set opensocketfunction\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_SOCKOPTDATA, pSource) != CURLE_OK)
    {
        ERRORPRINT("Failed to set opensocketdata\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_URL, strURL) != CURLE_OK)
    {
        ERRORPRINT("Failed to set URL\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_USERAGENT, "encurl/7.29.0") != CURLE_OK)
    {
        ERRORPRINT("Failed to set useragent\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "") != CURLE_OK)
    {
        ERRORPRINT("Failed to set encoding\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_HTTPGET, 1L) != CURLE_OK)
    {
        ERRORPRINT("Failed to set httpget\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set httpver\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, 1400 * 188) != CURLE_OK)
    {
        ERRORPRINT("Failed to set buffersize\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, 
        curlWriteFunction_ATSC10) != CURLE_OK)
    {
        ERRORPRINT("Failed to set writefunc\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_WRITEDATA, pSource) != CURLE_OK)
    {
        ERRORPRINT("Failed to set userdata\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, strError) != CURLE_OK)
    {
        ERRORPRINT("Failed to set errorbuffer\n");
        goto out;
    }

    res = curl_easy_perform(curl);
    if (res != CURLE_OK)
    {
        int curlSocket;

        if (pthread_mutex_lock(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        curlSocket = pSource->curlSocket;

        if (pthread_mutex_unlock(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        if (curlSocket != -1)
        {
            ERRORPRINT("CURL %s failed with %d (%s)\n", strURL, res, 
                curl_easy_strerror(res));
        }
    }

out:
    if (curl)
    {
        curl_easy_cleanup(curl);
    }

    return NULL;
}

void *
ATSC3J288ALPHDHRSourceTransferThread_ATSC30(void *arg)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)arg;
    char strPLPIds[32] = "";
    int32_t off = 0;
    CURLcode res;
    char strError[CURL_ERROR_SIZE];
    char strURL[128];
    CURL *curl = NULL;
    int32_t i;

    for (i = 0; i < numof(pSource->vPLPIds); i++)
    {
        if (pSource->vPLPIds[i] != INVALID_PLP_ID)
        {
            int ret;

            ret = snprintf(strPLPIds + off, sizeof(strPLPIds) - off, "p%d",
                pSource->vPLPIds[i]);
            if (ret < 0)
            {
                ERRORPRINT("Error adding PLP %d\n", pSource->vPLPIds[i]);
                goto out;
            }
            off += ret;
            if (off >= sizeof(strPLPIds))
            {
                ERRORPRINT("PLP ID string too long\n");
                goto out;
            }
        }
    }
    
    if (strPLPIds[0] == 0)
    {
        ERRORPRINT("No PLPs tuned, forcing 0\n");
        strcpy(strPLPIds, "p0");
    }

    if (snprintf(strURL, sizeof(strURL),
        "http://%d.%d.%d.%d:5004/tuner%d/ch%d%s?duration=311040000&format=j288",
        (pSource->hdhr.ip_addr >> 24) & 0xff, 
        (pSource->hdhr.ip_addr >> 16) & 0xff, 
        (pSource->hdhr.ip_addr >> 8) & 0xff, 
        (pSource->hdhr.ip_addr) & 0xff, 
        pSource->tunerNum, pSource->rfChannelNumber, 
        strPLPIds) >=
        sizeof(strURL))
    {
        ERRORPRINT("URL too long\n");
        goto out;
    }

    curl = curl_easy_init();
    if (!curl)
    {
        ERRORPRINT("Out of memory\n");
        goto out;
    }

    if (curl_easy_setopt(curl, CURLOPT_SOCKOPTFUNCTION, storeCurlSocket) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set sockoptfunction\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_SOCKOPTDATA, pSource) != CURLE_OK)
    {
        ERRORPRINT("Failed to set sockoptdata\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_URL, strURL) != CURLE_OK)
    {
        ERRORPRINT("Failed to set URL\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_USERAGENT, "encurl/7.29.0") != CURLE_OK)
    {
        ERRORPRINT("Failed to set useragent\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "") != CURLE_OK)
    {
        ERRORPRINT("Failed to set encoding\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_HTTPGET, 1L) != CURLE_OK)
    {
        ERRORPRINT("Failed to set httpget\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set httpver\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, 1400 * 188) != CURLE_OK)
    {
        ERRORPRINT("Failed to set buffersize\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, 
        curlWriteFunction_ATSC30) != CURLE_OK)
    {
        ERRORPRINT("Failed to set writefunc\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_WRITEDATA, pSource) != CURLE_OK)
    {
        ERRORPRINT("Failed to set userdata\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, strError) != CURLE_OK)
    {
        ERRORPRINT("Failed to set errorbuffer\n");
        goto out;
    }

    if (ENATSC3J288ALPOpen(&pSource->hATSC3J288ALP))
    {
        ERRORPRINT("Failed to open J288 processor\n");
        goto out;
    }

    if (ENATSC3J288ALPRegisterCallbacks(pSource->hATSC3J288ALP,
        ATSC3J288ALPHDHRSourceALPPacketCallback, 
        ATSC3J288ALPHDHRSourceBasebandPacketCallback,
        ATSC3J288ALPHDHRSourcePTPCallback, pSource))
    {
        ERRORPRINT("Failed to register J288 callbacks\n");
    }

    // Drop four baseband packets worth of data, but limit the timeout to
    // 1 second from now. 
    pSource->discardCountdown = 8192 * 4;
    clock_gettime(CLOCK_MONOTONIC, &pSource->discardMonoTimeout);
    pSource->discardMonoTimeout.tv_sec++;

    res = curl_easy_perform(curl);
    if (res != CURLE_OK)
    {
        int curlSocket;

        if (pthread_mutex_lock(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        curlSocket = pSource->curlSocket;

        if (pthread_mutex_unlock(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        if (curlSocket != -1)
        {
            ERRORPRINT("CURL %s failed with %d (%s)\n", strURL, res, 
                curl_easy_strerror(res));
        }
    }

out:
    if (curl)
    {
        curl_easy_cleanup(curl);
    }

    if (pSource->hATSC3J288ALP)
    {
        if (ENATSC3J288ALPClose(pSource->hATSC3J288ALP))
        {
            ERRORPRINT("Failed to close J288 processor\n");
        }
        pSource->hATSC3J288ALP = NULL;
    }

    return NULL;
}

void *
PLPStatusThread(void *arg)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)arg;
    struct hdhomerun_device_t *pHDHomerun = NULL;
    char strPLPInfoPath[32];
    char strDebugPath[32];
    char strHDHR[32];
    int32_t hdhrErrorCount = 0;
    int32_t hdhrDetuneCount = 0;
    int32_t hdhrFrozenCount = 0;

    if (snprintf(strPLPInfoPath, sizeof(strPLPInfoPath), "/tuner%d/plpinfo",
        pSource->tunerNum) >= sizeof(strPLPInfoPath))
    {
        ERRORPRINT("Path too long\n");
        goto out;
    }

    if (snprintf(strDebugPath, sizeof(strDebugPath), "/tuner%d/debug",
        pSource->tunerNum) >= sizeof(strDebugPath))
    {
        ERRORPRINT("Path too long\n");
        goto out;
    }

    if (snprintf(strHDHR, sizeof(strHDHR), "%s-%d", pSource->base.strSerial,
        pSource->tunerNum) >= sizeof(strHDHR))
    {
        ERRORPRINT("HDHR name too long\n");
        goto out;
    }

    while (pSource->fRunning)
    {
        char *pDebugString;
        char *pTunerStatusString;
        char *pDeviceStatusString;
        char *pTSStatusString;
        char *pNetStatusString;
        char *pLockStatus;
        int32_t masterLock = 0;
        char *pPLPInfoString;
        char *pPLPInfoIter;
        char *pLLS;
        char *pPLPLock;
        char *pPLPMod;
        char *pPLPCod;
        char *pSNQ;
        char *pSS;
        char *pTE;
        char *pBPS;
        int32_t ss = 0;
        int32_t snq = 0;
        uint32_t te = 0;
        uint32_t bps = 0;
        ENATSC3Modulation vPLPMod[64];
        ENATSC3Coding vPLPCod[64];
        int32_t plpId;
        uint64_t plpValidMask = 0;
        uint64_t plpLockedMask = 0;
        uint64_t llsValidMask = 0;
        int32_t fParseError = 0;
        int32_t i;
        int hdret;

        if (!pHDHomerun)
        {
            pHDHomerun = hdhomerun_device_create_from_str(strHDHR, NULL);
            if (!pHDHomerun)
            {
                ERRORPRINT("Failed to create HDHR device for %s\n", strHDHR);
                sleep(1);
                continue;
            }
            hdhrErrorCount = 0;
            hdhrDetuneCount = 0;
            hdhrFrozenCount = 0;
        }

        if (hdhrErrorCount >= 5 || hdhrDetuneCount >= 2 || 
            hdhrFrozenCount >= 5)
        {
            if (hdhrErrorCount >= 5)
            {
                ERRORPRINT("HDHomerun %s not responding, restarting\n", 
                    pSource->base.strSerial);
            }
            if (hdhrDetuneCount >= 2)
            {
                ERRORPRINT("HDHomerun %s detuned, restarting\n",
                    pSource->base.strSerial);
            }

            hdhomerun_device_destroy(pHDHomerun);
            pHDHomerun = NULL;

            if (pthread_mutex_lock(&pSource->restartTransferThreadLock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            pSource->fRestartTransferThread = 1;

            if (pthread_mutex_unlock(&pSource->restartTransferThreadLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            sleep(1);
            continue;
        }

        hdret = hdhomerun_device_get_var(pHDHomerun, strDebugPath, 
            &pDebugString, NULL);
        if (hdret != 1)
        {
            ERRORPRINT("%s: Failed to get tuner status (%d)\n", 
                pSource->base.strSerial, hdret);
            hdhrErrorCount++;
            sleep(1);
            continue;
        }

        // Example tuner status:
        // tun: ch=atsc3:85000000:0 lock=atsc3:84750000 ss=81(-60dBm) snq=64(16.9dB) seq=100 dbg=8-88/-195121
        // dev: bps=3080192 resync=0 overflow=0
        // ts:  bps=0 te=345892 crc=0
        // net: bps=3083520 pps=264 err=0 stop=0

        pTunerStatusString = strstr(pDebugString, "tun: ");
        pDeviceStatusString = strstr(pTunerStatusString, "dev: ");
        pTSStatusString = strstr(pDeviceStatusString, "ts: ");
        pNetStatusString = strstr(pTSStatusString, "net: ");
        if (!pTunerStatusString || !pDeviceStatusString || 
            !pTSStatusString || !pNetStatusString)
        {
            ERRORPRINT("%s: Failed to parse debug status %s\n", 
                pSource->base.strSerial, pDebugString);
            sleep(1);
            continue;
        }
        pTunerStatusString += 5;
        pDeviceStatusString += 5;
        pTSStatusString += 4;
        pNetStatusString += 5;

        if (strncmp(pTunerStatusString, "ch=", 3))
        {
            ERRORPRINT("%s: Unexpected status %s\n", pSource->base.strSerial, 
                pTunerStatusString);
            hdhrErrorCount++;
            sleep(1);
            continue;
        }

        if (!strncmp(pTunerStatusString + 3, "none", 4))
        {
            ERRORPRINT("%s: HDHR %s detuned\n", pSource->base.strSerial, 
                strHDHR);
            hdhrDetuneCount++;
            sleep(1);
            continue;
        }

        hdhrDetuneCount = 0;

        pLockStatus = strstr(pTunerStatusString, "lock=");
        if (pLockStatus && pLockStatus < pDeviceStatusString &&
            !strncmp(pLockStatus + 5, "atsc3", 5))
        {
            masterLock = 1;
        }

        pSS = strstr(pTunerStatusString, "ss=");
        if (pSS && pSS < pDeviceStatusString)
        {
            pSS += 3;
            while (isdigit(*pSS)) 
            {
                pSS++;
            }
            if (*pSS == '(')
            {
                ss = atoi(pSS + 1);
            }
        }

        pSNQ = strstr(pTunerStatusString, "snq=");
        if (pSNQ && pSNQ < pDeviceStatusString)
        {
            pSNQ += 4;
            while (isdigit(*pSNQ)) 
            {
                pSNQ++;
            }
            if (*pSNQ == '(')
            {
                snq = atoi(pSNQ + 1);
            }
        }

        pBPS = strstr(pDeviceStatusString, "bps=");
        if (pBPS && pBPS < pTSStatusString)
        {
            bps = atoi(pBPS + 4);
        }

        pTE = strstr(pTSStatusString, "te=");
        if (pTE && pTE < pNetStatusString)
        {
            te = atoi(pTE + 3);
        }

        if (masterLock && bps == 0)
        {
            hdhrFrozenCount++;
        }
        else
        {
            hdhrFrozenCount = 0;
        }

        hdret = hdhomerun_device_get_var(pHDHomerun, strPLPInfoPath, 
            &pPLPInfoString, NULL);
        if (hdret != 1)
        {
            ERRORPRINT("%s: Failed to get plp info (%d)\n", 
                pSource->base.strSerial, hdret);
            hdhrErrorCount++;
            sleep(1);
            continue;
        }
        hdhrErrorCount = 0;

        if (pPLPInfoString[0])
        {
            // Example plp info:
            //  1: sfi=0 mod=qam256 cod=9/15 layer=core ti=none lls=0

            pPLPInfoIter = pPLPInfoString;
            while (*pPLPInfoIter != '\n' && *pPLPInfoIter)
            {
                char *pPLPInfoEnd;

                pPLPInfoEnd = strchr(pPLPInfoIter, '\n');
                if (!pPLPInfoEnd)
                {
                    ERRORPRINT("%s: Failed to parse tuner status string {%s}\n",
                        pSource->base.strSerial, pPLPInfoIter);
                    fParseError = 1;
                    break;
                }

                plpId = atol(pPLPInfoIter);
                if (plpId < 0 || plpId > 63)
                {
                    ERRORPRINT("%s: Invalid plp %d (%s)\n", 
                        pSource->base.strSerial, plpId,
                        pPLPInfoString);
                }
                else
                {
                    plpValidMask |= ((int64_t)1 << plpId);

                    pLLS = (char *)memmem(pPLPInfoIter,
                        (size_t)(pPLPInfoEnd - pPLPInfoIter), "lls=", 4);
                    if (!pLLS)
                    {
                        ERRORPRINT("%s: Failed to find lls info {%s}\n",
                            pSource->base.strSerial, pPLPInfoString);
                    }
                    else
                    {
                        if (pLLS[4] == '0')
                        {
                            // Nothing to do
                        }
                        else if (pLLS[4] == '1')
                        {
                            llsValidMask |= ((int64_t)1 << plpId);
                        }
                        else
                        {
                            ERRORPRINT("%s: Unexpected lls value %s (%s)\n",
                                pSource->base.strSerial, pLLS, pPLPInfoString);
                        }
                    }

                    pPLPLock = memmem(pPLPInfoIter,
                        (size_t)(pPLPInfoEnd - pPLPInfoIter), "lock=", 5);
                    if (!pPLPLock)
                    {
                        // Lack of a "lock=" means the PLP is not tuned. 
                        // Implicitly it has no lock.
                    }
                    else
                    {
                        if (pPLPLock[5] == '0')
                        {
                            // Nothing to do
                        }
                        else if (pPLPLock[5] == '1')
                        {
                            plpLockedMask |= ((int64_t)1 << plpId);
                        }
                        else
                        {
                            ERRORPRINT("%s: Unexpected lock value %s (%s)\n",
                                pSource->base.strSerial, pLLS, pPLPInfoString);
                        }
                    }

                    pPLPMod = memmem(pPLPInfoIter,
                        (size_t)(pPLPInfoEnd - pPLPInfoIter), "mod=", 4);
                    if (!pPLPMod)
                    {
                        ERRORPRINT("%s: Failed to find modulation info {%s}\n",
                            pSource->base.strSerial, pPLPInfoString);
                    }
                    else
                    {
                        for (i = 0; i < numof(vModTableG); i++)
                        {
                            if (!strncmp(pPLPMod + 4, vModTableG[i].pString, 
                                vModTableG[i].sString))
                            {
                                vPLPMod[plpId] = vModTableG[i].mod;
                                break;
                            }
                        }

                        if (i == numof(vModTableG))
                        {
                            ERRORPRINT("%s: Unexpected mod value %s (%s)\n",
                                pSource->base.strSerial, pPLPMod, 
                                pPLPInfoString);
                            vPLPMod[plpId] = ENATSC3Modulation_Unknown;
                        }
                    }

                    pPLPCod = memmem(pPLPInfoIter,
                        (size_t)(pPLPInfoEnd - pPLPInfoIter), "cod=", 4);
                    if (!pPLPCod)
                    {
                        ERRORPRINT("%s: Failed to find coding info {%s}\n",
                            pSource->base.strSerial, pPLPInfoString);
                    }
                    else
                    {
                        for (i = 0; i < numof(vCodTableG); i++)
                        {
                            if (!strncmp(pPLPCod + 4, vCodTableG[i].pString, 
                                vCodTableG[i].sString))
                            {
                                vPLPCod[plpId] = vCodTableG[i].cod;
                                break;
                            }
                        }

                        if (i == numof(vCodTableG))
                        {
                            ERRORPRINT("%s: Unexpected cod value %s (%s)\n",
                                pSource->base.strSerial, pPLPCod, 
                                pPLPInfoString);
                            vPLPCod[plpId] = ENATSC3Coding_Unknown;
                        }
                    }
                }

                pPLPInfoIter = pPLPInfoEnd + 1;
            }
        }
        else
        {
            // Empty plp info can happen when there isn't lock, and also for
            // a time as a transfer is restarted.
            uint64_t mask;

            plpValidMask = pSource->stats.plpValidMask;
            llsValidMask = pSource->stats.llsValidMask;
            plpLockedMask = 0;

            for (i = 0, mask = plpValidMask; i < 64; i++, plpValidMask <<= 1)
            {
                if (mask & 1)
                {
                    vPLPMod[i] = pSource->stats.vPLPStats[i].modulation;
                    vPLPCod[i] = pSource->stats.vPLPStats[i].coding;
                }
            }
        }

        if (!fParseError)
        {
            uint64_t mask;
            int32_t i;

            if (pthread_mutex_lock(&pSource->statsLock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            pSource->stats.masterLock = masterLock;
            pSource->stats.plpValidMask = plpValidMask;
            pSource->stats.plpLockedMask = plpLockedMask;
            pSource->stats.llsValidMask = llsValidMask;
            pSource->stats.snr = snq * 100;
            pSource->stats.rssi = ss * 100;
            pSource->stats.nErroredPackets = te;

            for (i = 0, mask = plpValidMask; i < 64; i++, plpValidMask <<= 1)
            {
                if (mask & 1)
                {
                    pSource->stats.vPLPStats[i].modulation = vPLPMod[i];
                    pSource->stats.vPLPStats[i].coding = vPLPCod[i];
                }
            }

            if (pthread_mutex_unlock(&pSource->statsLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }
        }

        sleep(1);
    }

out:
    if (pHDHomerun)
    {
        hdhomerun_device_destroy(pHDHomerun);
    }

    return NULL;
}

ENStatus
ATSC3J288ALPHDHRSourceRun_ATSC10(ENATSC3J288ALPHDHRSource *pSource,
    int32_t *pFRunning, pthread_mutex_t *pRunningLock, 
    pthread_cond_t *pRunningCond)
{
    struct sched_param sched_param;
    pthread_attr_t threadAttr;
    int32_t fHaveTransferThread = 0;
    int32_t fHaveThreadAttr = 0;
    ENStatus rval = ENStatus_Success;

    memset(&pSource->stats, 0, sizeof(pSource->stats));

    if (pthread_attr_init(&threadAttr))
    {
        ERRORPRINT("Failed to init thread attr\n");
        rval = ENStatus_Failure;
        goto out;
    }
    fHaveThreadAttr = 1;

    if (pthread_attr_setinheritsched(&threadAttr, PTHREAD_EXPLICIT_SCHED))
    {
        ERRORPRINT("WARN: Failed to disable inherited scheduling policy\n");
    }

    if (pthread_attr_setschedpolicy(&threadAttr, SCHED_FIFO))
    {
        ERRORPRINT("WARN: Failed to set RT FIFO scheduling policy\n");
    }

    sched_param.sched_priority = ENUtils_Scheduling_Realtime_High - 1;
    if (pthread_attr_setschedparam(&threadAttr, &sched_param))
    {
        ERRORPRINT("WARN: Failed to set RT FIFO scheduling prority\n");
    }

    pSource->fRestartTransferThread = 0;
    pSource->fRunning = 1;

    while (*pFRunning)
    {
        if (!fHaveTransferThread)
        {
            if (pthread_create(&pSource->transferThread, &threadAttr, 
                ATSC3J288ALPHDHRSourceTransferThread_ATSC10, pSource))
            {
                ERRORPRINT("Failed to start transfer thread\n");
                rval = ENStatus_Failure;
                goto out;
            }
            ENUtils_SetThreadName(pSource->transferThread, 
                "TransferThread1");
            fHaveTransferThread = 1;
            pSource->fRestartTransferThread = 0;
        }

        if (pthread_mutex_lock(pRunningLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        while (*pFRunning && fHaveTransferThread)
        {
            int ptret;
            struct timespec timeout;

            clock_gettime(CLOCK_MONOTONIC, &timeout);
            timeout.tv_nsec += NANOSECONDS_PER_SECOND / 10;
            if (timeout.tv_nsec >= NANOSECONDS_PER_SECOND)
            {
                timeout.tv_nsec -= NANOSECONDS_PER_SECOND;
                timeout.tv_sec++;
            }

            ptret = pthread_cond_timedwait(pRunningCond, pRunningLock, 
                &timeout);
            if (ptret && ptret != ETIMEDOUT)
            {
                ERRORPRINT("Failed to wait: %s\n", strerror(ptret));
            }

            if (!*pFRunning)
            {
                break;
            }

            if (pSource->fRestartTransferThread)
            {
                LOGPRINT_NOFILE("Restarting HDHR transfer thread\n");
                if (pthread_mutex_unlock(&pSource->restartTransferThreadLock))
                {
                    ERRORPRINT("Failed to lock\n");
                }

                closeCurlSocket(pSource);

                if (pthread_join(pSource->transferThread, NULL))
                {
                    ERRORPRINT("Failed to join transfer thread\n");
                }

                fHaveTransferThread = 0;

                if (pthread_mutex_unlock(&pSource->restartTransferThreadLock))
                {
                    ERRORPRINT("Failed to lock\n");
                }
            }
        }

        if (pthread_mutex_unlock(pRunningLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

    }

out:
    pSource->fRunning = 0;

    if (fHaveTransferThread)
    {
        closeCurlSocket(pSource);

        if (pthread_join(pSource->transferThread, NULL))
        {
            ERRORPRINT("Failed to join transfer thread\n");
        }

        fHaveTransferThread = 0;
    }

    if (fHaveThreadAttr)
    {
        if (pthread_attr_destroy(&threadAttr))
        {
            ERRORPRINT("Failed to destroy thread attr\n");
        }
    }

    return rval;
}

ENStatus
ATSC3J288ALPHDHRSourceRun_ATSC30(ENATSC3J288ALPHDHRSource *pSource, 
    int32_t *pFRunning, pthread_mutex_t *pRunningLock, 
    pthread_cond_t *pRunningCond)
{
    int32_t i;
    int32_t j;
    uint64_t llsValidMask = 0;
    int32_t fHaveTransferThread = 0;
    int32_t fHavePLPStatusThread = 0;
    ENStatus rval = ENStatus_Success;

    memset(&pSource->stats, 0, sizeof(pSource->stats));

    for (i = 0; i < numof(pSource->stats.vPLPStats); i++)
    {
        pSource->stats.vPLPStats[i].lastContinuityCounter = -1;
    }

    pSource->fRestartTransferThread = 0;
    pSource->fRunning = 1;

    // We start off tuned to PLP 0
    pSource->vPLPIds[0] = 0;
    for (i = 1; i < numof(pSource->vPLPIds); i++)
    {
        pSource->vPLPIds[i] = INVALID_PLP_ID;
    }

    if (pthread_create(&pSource->plpStatusThread, NULL, PLPStatusThread,
        pSource))
    {
        ERRORPRINT("%s: Failed to start PLP status thread\n",
            pSource->base.strSerial);
        rval = ENStatus_Failure;
        goto out;
    }
    fHavePLPStatusThread = 1;

    while (*pFRunning)
    {
        if (!fHaveTransferThread)
        {
            if (pthread_create(&pSource->transferThread, NULL, 
                ATSC3J288ALPHDHRSourceTransferThread_ATSC30, pSource))
            {
                ERRORPRINT("%s: Failed to start traansfer thread\n",
                    pSource->base.strSerial);
                rval = ENStatus_Failure;
                goto out;
            }
            ENUtils_SetThreadName(pSource->transferThread,
                "TransferThread3");
            fHaveTransferThread = 1;
            pSource->fRestartTransferThread = 0;
        }

        if (pthread_mutex_lock(pRunningLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        while (*pFRunning && fHaveTransferThread)
        {
            int ptret;
            struct timespec timeout;

            clock_gettime(CLOCK_MONOTONIC, &timeout);
            timeout.tv_nsec += NANOSECONDS_PER_SECOND / 10;
            if (timeout.tv_nsec >= NANOSECONDS_PER_SECOND)
            {
                timeout.tv_nsec -= NANOSECONDS_PER_SECOND;
                timeout.tv_sec++;
            }

            ptret = pthread_cond_timedwait(pRunningCond, pRunningLock, 
                &timeout);
            if (ptret && ptret != ETIMEDOUT)
            {
                ERRORPRINT("Failed to wait: %s\n", strerror(ptret));
            }

            if (pSource->stats.llsValidMask != llsValidMask)
            {
                struct timespec rtnow;

                clock_gettime(CLOCK_REALTIME, &rtnow);

                llsValidMask = pSource->stats.llsValidMask;
                pSource->LLSPLPsCallback(llsValidMask, 
                    &rtnow, pSource->stats.nALPPackets,
                    pSource->pCallbackUserData);
            }

            if (pSource->fPLPIdMaskChanged)
            {
                int32_t fPLPsChanged = 0;
                uint64_t plpIdMaskTemp;

                // We're going to snapshot the mask since processing the change 
                // could take a while and we don't want to hold the lock that
                // long. If someone changes the mask again we'll catch it the 
                // next time around.
                if (pthread_mutex_lock(&pSource->plpIdMaskChangedLock))
                {
                    ERRORPRINT("Failed to lock\n");
                }

                pSource->plpIdMask = pSource->requestedPlpIdMask;
                pSource->fPLPIdMaskChanged = 0;

                // First reconcile the current PLPs with the mask.
                plpIdMaskTemp = pSource->plpIdMask;

                // We need at least one PLP
                if (plpIdMaskTemp == 0)
                {
                    plpIdMaskTemp = 1;
                }

                for (i = 0; i < numof(pSource->vPLPIds); i++)
                {
                    if (pSource->vPLPIds[i] != INVALID_PLP_ID)
                    {
                        if (plpIdMaskTemp & 1ULL << pSource->vPLPIds[i])
                        {
                            // Already set, remove from mask
                            plpIdMaskTemp &= ~1ULL << pSource->vPLPIds[i];
                        }
                        else
                        {
                            // No longer set, free up slot
                            pSource->vPLPIds[i] = INVALID_PLP_ID;
                            fPLPsChanged = 1;
                        }
                    }
                }

                // Next assign new PLPs to slots
                for (i = 0; i < 64; i++)
                {
                    if (plpIdMaskTemp & 1)
                    {
                        for (j = 0; j < numof(pSource->vPLPIds); j++)
                        {
                            if (pSource->vPLPIds[j] == INVALID_PLP_ID)
                            {
                                pSource->vPLPIds[j] = i;
                                break;
                            }
                        }

                        if (j == numof(pSource->vPLPIds))
                        {
                            ERRORPRINT("%s: Too many plps: couldn't set %d, mask %x, slots %d %d %d %d\n", 
                                pSource->base.strSerial, i, pSource->plpIdMask, 
                                pSource->vPLPIds[0], pSource->vPLPIds[1], 
                                pSource->vPLPIds[2], pSource->vPLPIds[3]);
                        }
                        fPLPsChanged = 1;
                    }

                    plpIdMaskTemp >>=1;
                }

                // See if only a single PLP is active. This is required to
                // determine the PLP Id of ALP and Baseband packets. See the
                // ALP and Baseband packet callback functions.
                pSource->singlePLPId = INVALID_PLP_ID;
                for (i = 0, plpIdMaskTemp = pSource->plpIdMask; 
                     plpIdMaskTemp;
                     i++, plpIdMaskTemp >>= 1)
                {
                    if (plpIdMaskTemp & 1)
                    {
                        if (pSource->singlePLPId == INVALID_PLP_ID)
                        {
                            pSource->singlePLPId = i;
                        }
                        else
                        {
                            // More than one PLP has been requested.
                            pSource->singlePLPId = INVALID_PLP_ID;
                            break;
                        }
                    }
                }

                LOGPRINT("%s: Setting PLP mask 0x%x slots %d %d %d %d\n",
                    pSource->base.strSerial, pSource->plpIdMask, 
                    pSource->vPLPIds[0], pSource->vPLPIds[1], 
                    pSource->vPLPIds[2], pSource->vPLPIds[3]); 

                if (fPLPsChanged)
                {
                    if (pthread_mutex_lock(
                        &pSource->restartTransferThreadLock))
                    {
                        ERRORPRINT("Failed to lock\n");
                    }

                    pSource->fRestartTransferThread = 1;

                    if (pthread_mutex_unlock(
                        &pSource->restartTransferThreadLock))
                    {
                        ERRORPRINT("Failed to lock\n");
                    }
                }
                else
                {
                    LOGPRINT("%s: No change to selected PLPs, ignoring\n",
                        pSource->base.strSerial);
                }

                if (pthread_mutex_unlock(&pSource->plpIdMaskChangedLock))
                {
                    ERRORPRINT("Failed to unlock\n");
                }
            }

            if (pSource->fRestartTransferThread)
            {
                LOGPRINT_NOFILE("%s: Restarting HDHR transfer thread\n",
                    pSource->base.strSerial);
                if (pthread_mutex_unlock(&pSource->restartTransferThreadLock))
                {
                    ERRORPRINT("Failed to lock\n");
                }

                closeCurlSocket(pSource);

                if (pthread_join(pSource->transferThread, NULL))
                {
                    ERRORPRINT("%s: Failed to join transfer thread\n",
                        pSource->base.strSerial);
                }

                fHaveTransferThread = 0;

                if (pthread_mutex_unlock(&pSource->restartTransferThreadLock))
                {
                    ERRORPRINT("Failed to lock\n");
                }
            }
        }

        if (pthread_mutex_unlock(pRunningLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

out:
    pSource->fRunning = 0;

    if (fHaveTransferThread)
    {
        closeCurlSocket(pSource);

        if (pthread_join(pSource->transferThread, NULL))
        {
            ERRORPRINT("%s: Failed to join transfer thread\n", 
                pSource->base.strSerial);
        }

        fHaveTransferThread = 0;
    }

    if (fHavePLPStatusThread)
    {
        if (pthread_join(pSource->plpStatusThread, NULL))
        {
            ERRORPRINT("Failed to join PLP status thread\n");
        }
    }

    return rval;
}

ENStatus
ATSC3J288ALPHDHRSourceRun(ENATSC3Source hSource, int32_t *pFRunning,
    pthread_mutex_t *pRunningLock, pthread_cond_t *pRunningCond)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;
    ENStatus rval;

    switch (pSource->broadcastSystem)
    {
        case ENATSC3BroadcastSystem_ATSC10:
            rval = ATSC3J288ALPHDHRSourceRun_ATSC10(pSource, pFRunning,
                pRunningLock, pRunningCond);
            break;
        case ENATSC3BroadcastSystem_ATSC30:
            rval = ATSC3J288ALPHDHRSourceRun_ATSC30(pSource, pFRunning,
                pRunningLock, pRunningCond);
            break;
        default:
            ERRORPRINT("Unhandled broadcast system %d\n", 
                pSource->broadcastSystem);
            rval = ENStatus_Failure;
            break;
    }

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceOpen_ATSC10(uint64_t tuid, int32_t frequencyHz, 
    ENATSC3SourceWriteCallback ATSC1WriteCallback,
    void *pATSC1WriteCallbackUserData, ENATSC3Source *ppSource)
{
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t fSourceLocked = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources;
    for (i = 0;
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pSource->base.tuid == tuid)
        {
            break;
        }
    }

    if (i == atsc3J288ALPHDHRSourceGlobalsG.nSources)
    {
        ERRORPRINT("Could't find source %"PRIx64"n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }
    fSourceLocked = 1;

    if (pSource->fOpened)
    {
        ERRORPRINT("Source %"PRIx64" already opened\n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->verbosity = ENVerbosity_Terse;
    pSource->broadcastSystem = ENATSC3BroadcastSystem_ATSC10;
    pSource->atsc1.ATSC1WriteCallback = ATSC1WriteCallback;
    pSource->atsc1.pATSC1WriteCallbackUserData = pATSC1WriteCallbackUserData;

    pSource->frequencyKHz = frequencyHz/1000 + 3000;

    // channel lower edge frequency to channel number
    if (frequencyHz >= 470000000 && frequencyHz <= 884000000)
    {
        // UHF
        pSource->rfChannelNumber = 
            14 + ((frequencyHz - 470000000) / 6000000);
    }
    else if (frequencyHz >= 174000000 && frequencyHz <= 210000000)
    {
        // VHF Hi
        pSource->rfChannelNumber = 
            7 + ((frequencyHz - 174000000) / 6000000);
    }
    else if (frequencyHz >= 54000000 && frequencyHz <= 82000000)
    {
        // VHF Lo
        pSource->rfChannelNumber = 
            2 + ((frequencyHz - 54000000) / 6000000);
    }
    else
    {
        ERRORPRINT("Invalid frequency %d\n", frequencyHz);
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->curlSocket = -1;

    pSource->fOpened = 1;

    *ppSource = (ENATSC3Source)pSource;

out:
    if (fSourceLocked)
    {
        if (pthread_mutex_unlock(&pSource->lock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    return rval;

errout:
    goto out;
}

ENStatus
ENATSC3J288ALPHDHRSourceOpen_ATSC30(uint64_t tuid, int32_t frequencyHz, 
    ENATSC3ALPHandle_t hATSC3ALP, ENATSC3Source *ppSource)
{
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t fSourceLocked = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources;
    for (i = 0;
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pSource->base.tuid == tuid)
        {
            break;
        }
    }

    if (i == atsc3J288ALPHDHRSourceGlobalsG.nSources)
    {
        ERRORPRINT("Could't find source %"PRIx64"n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }
    fSourceLocked = 1;

    if (pSource->fOpened)
    {
        ERRORPRINT("Source %"PRIx64" already opened\n", tuid);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->verbosity = ENVerbosity_Terse;
    pSource->broadcastSystem = ENATSC3BroadcastSystem_ATSC30;

    pSource->frequencyKHz = frequencyHz/1000 + 3000;

    // Set all PLPs invalid. Note that we will actually tune the first slot
    // to PLP 0, since one PLP must be set when initially tuning and 0 seems
    // to always be present in emissions.
    pSource->vPLPIds[0] = INVALID_PLP_ID;
    pSource->vPLPIds[1] = INVALID_PLP_ID;
    pSource->vPLPIds[2] = INVALID_PLP_ID;
    pSource->vPLPIds[3] = INVALID_PLP_ID;

    // channel lower edge frequency to channel number
    if (frequencyHz >= 470000000 && frequencyHz <= 884000000)
    {
        // UHF
        pSource->rfChannelNumber = 
            14 + ((frequencyHz - 470000000) / 6000000);
    }
    else if (frequencyHz >= 174000000 && frequencyHz <= 210000000)
    {
        // VHF Hi
        pSource->rfChannelNumber = 
            7 + ((frequencyHz - 174000000) / 6000000);
    }
    else if (frequencyHz >= 54000000 && frequencyHz <= 82000000)
    {
        // VHF Lo
        pSource->rfChannelNumber = 
            2 + ((frequencyHz - 54000000) / 6000000);
    }
    else
    {
        ERRORPRINT("Invalid frequency %d\n", frequencyHz);
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->hATSC3ALP = hATSC3ALP;
    pSource->requestedPlpIdMask = 0;
    pSource->plpIdMask = 0;
    pSource->singlePLPId = INVALID_PLP_ID;
    pSource->fPLPIdMaskChanged = 0;
    pSource->LLSPLPsCallback = NULL;
    pSource->pCallbackUserData = NULL;
    pSource->curlSocket = -1;
    pSource->hATSC3J288ALP = NULL;

    pSource->fOpened = 1;

    *ppSource = (ENATSC3Source)pSource;

out:
    if (fSourceLocked)
    {
        if (pthread_mutex_unlock(&pSource->lock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    return rval;

errout:
    goto out;
}

ENStatus
ATSC3J288ALPHDHRSourceClose(ENATSC3Source hSource)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->fOpened = 0;

    if (pthread_mutex_unlock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    return rval;
}

ENStatus
ATSC3J288ALPHDHRSourceSetPLPIds(ENATSC3Source hSource, uint64_t plpIdMask)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->plpIdMaskChangedLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->requestedPlpIdMask = plpIdMask;
    pSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pSource->plpIdMaskChangedLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3J288ALPHDHRSourceRegisterCallbacks(ENATSC3Source pSource,
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback, void *pUserData)
{
    ENATSC3J288ALPHDHRSource *pJ288ALPHDHRSource =
        (ENATSC3J288ALPHDHRSource *)pSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pJ288ALPHDHRSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pJ288ALPHDHRSource->LLSPLPsCallback = LLSPLPsCallback;
    pJ288ALPHDHRSource->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pJ288ALPHDHRSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceInitializeSource(ENATSC3J288ALPHDHRSource *pSource,
    struct hdhomerun_discover_device_t *pHDHR, int32_t tunerNum)
{
    int32_t fHaveLock = 0;
    int32_t fHaveRestartTransferThreadLock = 0;
    int32_t fHavePLPIdMaskChangedLock = 0;
    int32_t fHaveWriteCbLock = 0;
    int32_t fHaveStatsLock = 0;
    int32_t fHavePTPInfoLock = 0;
    int32_t fHaveCurlSocketLock = 0;
    ENStatus rval = ENStatus_Success;

    pSource->hdhr = *pHDHR;
    pSource->base.tuid = ((uint64_t)ENATSC3SourceType_J288ALPHDHR << 56) | 
        ((uint64_t)pHDHR->device_id << 8) | tunerNum;
    pSource->base.Run = ATSC3J288ALPHDHRSourceRun;
    pSource->base.Close = ATSC3J288ALPHDHRSourceClose;
    pSource->base.SetPLPIds = ATSC3J288ALPHDHRSourceSetPLPIds;
    pSource->base.RegisterCallbacks = ATSC3J288ALPHDHRSourceRegisterCallbacks;
    pSource->base.SetVerbosity = ATSC3J288ALPHDHRSourceSetVerbosity;
    strncpy(pSource->base.strModel, "HDHR",
        sizeof(pSource->base.strModel) - 1);
    pSource->base.strModel[sizeof(pSource->base.strModel) - 1] = 0;
    snprintf(pSource->base.strSerial, sizeof(pSource->base.strSerial), 
        "%08X:%d", pHDHR->device_id, tunerNum);
    pSource->base.strSerial[sizeof(pSource->base.strSerial) - 1] = 0;
    snprintf(pSource->base.strName, sizeof(pSource->base.strName), 
        "hdhratsc3:%08X:%d", pHDHR->device_id, tunerNum);
    pSource->base.strName[sizeof(pSource->base.strName) - 1] = 0;
    pSource->tunerNum = tunerNum;
    pSource->fOpened = 0;

    if (pthread_mutex_init(&pSource->lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    if (pthread_mutex_init(&pSource->restartTransferThreadLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveRestartTransferThreadLock = 1;

    if (pthread_mutex_init(&pSource->atsc1.writeCbLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveWriteCbLock = 1;

    if (pthread_mutex_init(&pSource->plpIdMaskChangedLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHavePLPIdMaskChangedLock = 1;

    if (pthread_mutex_init(&pSource->statsLock, NULL))
    {
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveStatsLock = 1;

    if (pthread_mutex_init(&pSource->ptpInfoLock, NULL))
    {
        rval = ENStatus_Failure;
        goto errout;
    }
    fHavePTPInfoLock = 1;

    if (pthread_mutex_init(&pSource->curlSocketLock, NULL))
    {
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveCurlSocketLock = 1;

out:
    return rval;

errout:
    if (fHaveCurlSocketLock)
    {
        if (pthread_mutex_destroy(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHavePTPInfoLock)
    {
        if (pthread_mutex_destroy(&pSource->ptpInfoLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHaveStatsLock)
    {
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHavePLPIdMaskChangedLock)
    {
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHaveWriteCbLock)
    {
        if (pthread_mutex_destroy(&pSource->atsc1.writeCbLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHaveRestartTransferThreadLock)
    {
        if (pthread_mutex_destroy(&pSource->restartTransferThreadLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    goto out;
}

static ENStatus
HDHRGetInfo(struct hdhomerun_discover_device_t *pHDHR, 
    int32_t *pFModulationATSC3, char *pModel, int32_t sModel)
{
    struct hdhomerun_device_t *pHDHomerun = NULL;
    char *pModelString;
    char *pFeatureString;
    char *pFeatureIter;
    int hdret;
    ENStatus rval = ENStatus_Success;
    
    *pFModulationATSC3 = 0;

    pHDHomerun = hdhomerun_device_create(pHDHR->device_id, pHDHR->ip_addr, 0,
        NULL);
    if (!pHDHomerun)
    {
        ERRORPRINT("Failed to create HDHR device for %08X\n", pHDHR->device_id);
        rval = ENStatus_Failure;
        goto out;
    }

    hdret = hdhomerun_device_get_var(pHDHomerun, "/sys/hwmodel",
        &pModelString, NULL);
    if (hdret != 1)
    {
        ERRORPRINT("Failed to get model for %08X (%d)\n", pHDHR->device_id, 
            hdret);
        rval = ENStatus_Failure;
        goto out;
    }
    strncpy(pModel, pModelString, sModel);
    pModel[sModel - 1] = 0;

    hdret = hdhomerun_device_get_var(pHDHomerun, "/sys/features",
        &pFeatureString, NULL);
    if (hdret != 1)
    {
        ERRORPRINT("Failed to get feature info for %08X (%d)\n", 
            pHDHR->device_id, hdret);
        rval = ENStatus_Failure;
        goto out;
    }

    pFeatureIter = pFeatureString;
    while (1)
    {
        char *pIterEnd;

        pIterEnd = strchr(pFeatureIter, '\n');
        if (!pIterEnd)
        {
            break;
        }
        if (!strncmp(pFeatureIter, "modulation: ", sizeof("modulation: ") - 1))
        {
            char *pModulation = pFeatureIter + sizeof("modulation: ") - 1;
            while (pModulation < pIterEnd)
            {
                if (!strncmp(pModulation, "atsc3", 5) && 
                    (pModulation[5] == '\n' || pModulation[5] == ' '))
                {
                    *pFModulationATSC3 = 1;
                }
                
                while (*pModulation != ' ' && *pModulation != '\n')
                {
                    pModulation++;
                }
                pModulation++;
            }
        }

        pFeatureIter = pIterEnd + 1;
    }

out:
    if (pHDHomerun)
    {
        hdhomerun_device_destroy(pHDHomerun);
    }

    return rval;
}

static ENStatus
ATSC3J288ALPHDHRSourceFindHDHomeruns(void)
{
    struct hdhomerun_discover_device_t vHDHR[HDHR_MAX_TUNERS];
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t nHDHR;
    int32_t i;
    int32_t j;
    ENStatus rval = ENStatus_Success;

    nHDHR = hdhomerun_discover_find_devices_custom_v2(0,
        HDHOMERUN_DEVICE_TYPE_TUNER, HDHOMERUN_DEVICE_ID_WILDCARD,
        vHDHR, numof(vHDHR));
    if (nHDHR < 0)
    {
        ERRORPRINT("Failed to send discover\n");
        rval = ENStatus_Failure;
        goto out;
    }

    LOGPRINT_NOCONSOLE("Found %d HDHomerun devices\n", nHDHR);

    pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources;
    for (i = 0; i < nHDHR; i++)
    {
        int32_t fModulationATSC3;
        char strModel[64];

        if (HDHRGetInfo(vHDHR + i, &fModulationATSC3, strModel, 
            sizeof(strModel)))
        {
            ERRORPRINT("Failed to query features for %08X\n",  
                vHDHR[i].device_id);
        }

        if (!fModulationATSC3)
        {
            continue;
        }

        // There doesn't seem to be a way to differentiate between ATSC-3 
        // capable tuners vs ATSC-1 only. We know on HDHR5-4K-DEV and 
        // HDFX-4K-DEV that tuners 0 and 1 are ATSC 1 & ATSC3 while tuners 2 
        // and 3 are ATSC-1 only.
        if (!strcmp(strModel, "HDHR5-4K-DEV") || 
            !strcmp(strModel, "HDFX-4K-DEV"))
        {
            vHDHR[i].tuner_count = 2;
        }
        else
        {
            vHDHR[i].tuner_count = 0;
        }

        for (j = 0; j < vHDHR[i].tuner_count; j++)
        {
            if (atsc3J288ALPHDHRSourceGlobalsG.nSources == 
                numof(atsc3J288ALPHDHRSourceGlobalsG.vSources))
            {
                ERRORPRINT("Too many sources, not adding %08X:%d\n", 
                    vHDHR[i].device_id, j);
            }
            else
            {
                LOGPRINT_NOFILE("ADDING %08X:%d\n",
                    vHDHR[i].device_id, j);
                if (ENATSC3J288ALPHDHRSourceInitializeSource(pSource, 
                    vHDHR + i, j))
                {
                    ERRORPRINT("Failed to add %08X:%d\n", 
                        vHDHR[i].device_id, j);
                }
                else
                {
                    pSource++;
                    atsc3J288ALPHDHRSourceGlobalsG.nSources++;
                }
            }
        }
    }

out:
    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceInitialize(void)
{
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t fCurlInitialized = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (curl_global_init(CURL_GLOBAL_ALL) != CURLE_OK)
    {
        ERRORPRINT("Failed to init curl\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fCurlInitialized = 1;

    atsc3J288ALPHDHRSourceGlobalsG.nSources = 0;

    ATSC3J288ALPHDHRSourceFindHDHomeruns();

out:
    return rval;

errout:
    for (i = 0, pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources;
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->ptpInfoLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->atsc1.writeCbLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->restartTransferThreadLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fCurlInitialized)
    {
        curl_global_cleanup();
    }

    goto out;
}

ENStatus
ENATSC3J288ALPHDHRSourceShutdown(void)
{
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources; 
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->atsc1.writeCbLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }
    
    curl_global_cleanup();

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceQuerySources(ENATSC3SourceDescription *pSources, 
    int32_t nSources, int32_t *pNSourcesUsed)
{
    ENATSC3J288ALPHDHRSource *pLocalSources;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    *pNSourcesUsed = 0;
    pLocalSources = atsc3J288ALPHDHRSourceGlobalsG.vSources;
    for (i = 0;
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources && nSources;
         i++, nSources--, pLocalSources++, pSources++)
    {
        pSources->tuid = pLocalSources->base.tuid;
        strncpy(pSources->strSerial, pLocalSources->base.strSerial,
            sizeof(pSources->strSerial));
        pSources->strSerial[sizeof(pSources->strSerial) - 1] = 0;
        strncpy(pSources->strName, pLocalSources->base.strName,
            sizeof(pSources->strName));
        pSources->strName[sizeof(pSources->strName) - 1] = 0;
        strncpy(pSources->strModel, pLocalSources->base.strModel,
            sizeof(pSources->strModel));
        pSources->strModel[sizeof(pSources->strModel) - 1] = 0;
    }

    if (i < atsc3J288ALPHDHRSourceGlobalsG.nSources)
    {
        ERRORPRINT("WARN: Sources truncated\n");
        rval = ENStatus_BufferTooSmall;
    }

    *pNSourcesUsed = i;

    return rval;
}
