/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <assert.h>

#include <zlib.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>

#include <libenatsc3lls/atsc3lls_slt.h>
#include <libenatsc3lls/atsc3lls_systemtime.h>
#include <libenatsc3lls/atsc3lls_aeat.h>
#include <libenatsc3lls/atsc3lls_rrt.h>
#include <libenatsc3lls/atsc3lls_onscreenmessage.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

#define SLT_MAX_LENGTH 65536
#define AEAT_MAX_LENGTH 65536
#define CERTIFICATION_DATA_MAX_LENGTH 65536
#define SIGNED_MULTITABLE_MAX_LENGTH 65536

struct UserData
{
    ENVerbosity verbosity;
    int32_t counter;
    int32_t fIgnoreChecksumErrors;
    int ifd;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;
    ENATSC3IPV4Handle_t hATSC3IPV4;
    char *pOutPath;

    uint64_t plpIds;

    struct
    {
        int32_t version;
        int32_t length;
        uint8_t *pLMT;
        int32_t nAllocated;
    } vLMTs[64];

    int32_t vSLTVersion[256];
    int32_t vSLTLength[256];
    uint8_t *vpSLT[256];

    int32_t vAEATVersion[256];
    int32_t vAEATLength[256];
    uint8_t *vpAEAT[256];

    int32_t vCertificationDataVersion[256];
    int32_t vCertificationDataLength[256];
    uint8_t *vpCertificationData[256];

    int32_t vSignedMultiTableVersion[256];
    int32_t vSignedMultiTableLength[256];
    uint8_t *vpSignedMultiTable[256];
};

typedef struct
{
    uint8_t lls_table_id;
    uint8_t lls_group_id;
    uint8_t group_count_minus1;
    uint8_t lls_table_version;
} LLSInfo;

int32_t fRunningG = 1;

// Forward declarations
int32_t
ProcessLLSPacket(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData);

void
PrintSLT(ATSC3LLSSLT_SLTNode *pParsedSLT, LLSInfo *pLLSInfo)
{
    int32_t i;
    int32_t j;
    int32_t k;

    printf("SLT group %d version %d\n", pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version);
    for (i = 0; i < pParsedSLT->services.nOffsetsUsed; i++)
    {
        ATSC3LLSSLT_ServiceNode *pService;

        printf("  Service %d\n", i);

        pService = (ATSC3LLSSLT_ServiceNode *)
            EnXMLParserDataOffsetArrayGetPointer(&pParsedSLT->services, i);

        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_serviceId)
        {
            printf("    serviceId: %"PRIu32"\n", pService->serviceId);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_sltSvcSeqNum)
        {
            printf("    sltSvcSeqNum: %"PRIu32"\n", pService->sltSvcSeqNum);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_protected)
        {
            printf("    protected: %d\n", pService->protected);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_majorChannelNo)
        {
            printf("    majorChannelNo: %"PRIu32"\n", pService->majorChannelNo);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_minorChannelNo)
        {
            printf("    minorChannelNo: %"PRIu32"\n", pService->minorChannelNo);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_serviceCategory)
        {
            printf("    serviceCategory: %"PRIu32"\n", 
                pService->serviceCategory);
        }
        if (pService->base.presentFlags &
            ATSC3LLSSLTServiceNodeAttribute_shortServiceName)
        {
            printf("    shortServiceName: %s\n",
                (char *)EnXMLParserDataBufferGetPointer(
                    &pParsedSLT->dataBuffer, 
                    pService->shortServiceNameDataOffset));
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_hidden)
        {
            printf("    hidden: %d\n", pService->hidden);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_broadbandAccessRequired)
        {
            printf("    broadbandAccessRequired: %d\n", 
                pService->broadbandAccessRequired);
        }
#if 0
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_essential)
        {
            printf("    essential: %d\n", pService->essential);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_simulcastTSID)
        {
            printf("    simulcastTSID: %"PRIu16"\n", pService->simulcastTSID);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_simulcastMajorChannelNo)
        {
            printf("    simulcastMajorChannelNo: %"PRIu32"\n", 
                pService->simulcastMajorChannelNo);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_simulcastMinorChannelNo)
        {
            printf("    simulcastMinorChannelNo: %"PRIu32"\n", 
                pService->simulcastMinorChannelNo);
        }
#endif
        if (pService->base.presentFlags &
            ATSC3LLSSLTServiceNodeChildNode_BroadcastSvcSignaling)
        {
            ATSC3LLSSLT_BroadcastSvcSignalingNode *pBroadcastSvcSignalingNode = 
                (ATSC3LLSSLT_BroadcastSvcSignalingNode *)
                &pService->broadcastSvcSignaling;
            printf("    BroadcastSvcSignaling\n");
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsProtocol)
            {
                printf("      slsProtocol: %d\n",
                    pBroadcastSvcSignalingNode->slsProtocol);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMajorProtocolVersion)
            {
                printf("      slsMajorProtocolVersion: %d\n",
                    pBroadcastSvcSignalingNode->slsMajorProtocolVersion);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMinorProtocolVersion)
            {
                printf("      slsMinorProtocolVersion: %d\n",
                    pBroadcastSvcSignalingNode->slsMinorProtocolVersion);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress)
            {
                printf("      slsDestinationIpAddress: %s\n",
                    pBroadcastSvcSignalingNode->slsDestinationIpAddress);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort)
            {
                printf("      slsDestinationUdpPort: %d\n",
                    pBroadcastSvcSignalingNode->slsDestinationUdpPort);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsSourceIpAddress)
            {
                printf("      slsSourceIpAddress: %s\n",
                    pBroadcastSvcSignalingNode->slsSourceIpAddress);
            }
        }
        if (pService->base.presentFlags &
            ATSC3LLSSLTServiceNodeChildNode_svcInetUrl)
        {
            ATSC3LLSSLTInetUrlNode *pSvcInetUrl = 
                (ATSC3LLSSLTInetUrlNode *)&pService->svcInetUrl;
            printf("    svcInetUrl: %s type %d\n", 
                (char *)EnXMLParserDataBufferGetPointer(
                    &pParsedSLT->dataBuffer, pSvcInetUrl->urlDataOffset),
                pSvcInetUrl->urlType);
        }
        printf("    OtherBsids (%d)\n", pService->otherBsids.nOffsetsUsed);
        for (j = 0; j < pService->otherBsids.nOffsetsUsed; j++)
        {
            ATSC3LLSSLT_OtherBsidNode *pOtherBsid;
            pOtherBsid = (ATSC3LLSSLT_OtherBsidNode *)
                EnXMLParserDataOffsetArrayGetPointer(&pService->otherBsids, j);
            printf("    %d: ", j);
            for (k = 0; k < pOtherBsid->otherBsids.nOffsetsUsed; k++)
            {
                printf("%d ", 
                    EnXMLParserArrayGetValue(&pOtherBsid->otherBsids, k));
            }
            printf("type %d\n", pOtherBsid->type);
        }
    }
}

int32_t
ProcessSLT(LLSInfo *pLLSInfo, int64_t udpPacketNumber, const uint8_t *pPayload, 
    int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSSLT_SLTNode parsedSLT;
    int32_t fHaveParsedSLT = 0;
    uint8_t groupId;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    groupId = pLLSInfo->lls_group_id;

    if (pUserData->vSLTVersion[groupId] == pLLSInfo->lls_table_version)
    {
        if (pUserData->vpSLT[groupId])
        {
            if (pUserData->vSLTLength[groupId] == payloadLength &&
                !memcmp(pUserData->vpSLT[groupId], pPayload, payloadLength))
            {
                goto out;
            }
            fprintf(stderr, "Unexpected change in SLT %d/%d %d/%d %u/%u\n", 
                pUserData->vSLTVersion[groupId], pLLSInfo->lls_table_version,
                pUserData->vSLTLength[groupId], payloadLength,
                ENUtils_CalcCRC32(pUserData->vpSLT[groupId], payloadLength), 
                ENUtils_CalcCRC32(pPayload, payloadLength));
        }
    }

    if (!pUserData->vpSLT[groupId])
    {
        pUserData->vpSLT[groupId] = (uint8_t *)malloc(SLT_MAX_LENGTH);
        if (!pUserData->vpSLT[groupId])
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > SLT_MAX_LENGTH)
    {
        fprintf(stderr, "SLT too large\n");
        goto out;
    }
    pUserData->vSLTLength[groupId] = payloadLength;
    memcpy(pUserData->vpSLT[groupId], pPayload, payloadLength);

    pUserData->vSLTVersion[groupId] = pLLSInfo->lls_table_version;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/slt-%.3d-%.3d-%.5d",
        pUserData->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version, pUserData->counter++) > 
        sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s: %s\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseSLT(strXML, xmlLength,  &parsedSLT))
    {
        printf("%"PRId64" Failed to parse SLT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedSLT = 1;

    PrintSLT(&parsedSLT, pLLSInfo);

out:
    if (fHaveParsedSLT)
    {
        if (ATSC3LLSFreeParsedSLT(&parsedSLT))
        {
            fprintf(stderr, "Failed to free parsed SLT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

void
PrintRRT(ATSC3LLSRRT_RatingRegionTablesNode *pParsedRRT)
{
    int32_t i;
    int32_t j;
    int32_t k;
    int32_t l;

    printf("Region Rating Table\n");
    for (i = 0; i < pParsedRRT->ratingRegionTables.nOffsetsUsed; i++)
    {
        ATSC3LLSRRT_RatingRegionTableNode *pRRT;

        pRRT = (ATSC3LLSRRT_RatingRegionTableNode *)
            EnXMLParserDataOffsetArrayGetPointer(
            &pParsedRRT->ratingRegionTables, i);

        if (pRRT->base.presentFlags & 
            ATSC3LLSRRTRatingRegionTableNodeAttribute_regionIdentifier)
        {
            printf("  regionIdentifier: %d\n", pRRT->regionIdentifier);
        }

        if (pRRT->base.presentFlags &
            ATSC3LLSRRTRatingRegionTableChildNode_RegionIdText)
        {
            ATSC3LLSRRT_TextTypeNode *pRegionIdText;

            printf("  RegionIdText\n");
            for (j = 0; j < pRRT->regionIdTexts.nOffsetsUsed; j++)
            {
                pRegionIdText = (ATSC3LLSRRT_TextTypeNode *)
                    EnXMLParserDataOffsetArrayGetPointer(
                    &pRRT->regionIdTexts, j);
                printf("    text: %s",
                    (char *)EnXMLParserDataBufferGetPointer(
                        &pParsedRRT->dataBuffer, 
                        pRegionIdText->stringDataOffset)); 
                if (pRegionIdText->base.presentFlags & 
                    ATSC3LLSRRTTextTypeNodeAttribute_lang)
                {
                    printf(" lang: %s", 
                        (char *)EnXMLParserDataBufferGetPointer(
                            &pParsedRRT->dataBuffer, 
                            pRegionIdText->langDataOffset)); 
                }
                printf("\n");
            }
        }

        if (pRRT->base.presentFlags &
            ATSC3LLSRRTRatingRegionTableChildNode_Dimension)
        {
            ATSC3LLSRRT_DimensionNode *pDimension;

            for (j = 0; j < pRRT->dimensions.nOffsetsUsed; j++)
            {
                printf("  Dimension %d\n", j);
                pDimension = (ATSC3LLSRRT_DimensionNode *)
                    EnXMLParserDataOffsetArrayGetPointer(
                    &pRRT->dimensions, j);
                if (pDimension->base.presentFlags & 
                    ATSC3LLSRRTDimensionNodeAttribute_dimensionLevels)
                {
                    printf("    dimensionLevels: %u\n", 
                        pDimension->dimensionLevels);
                }
                if (pDimension->base.presentFlags & 
                    ATSC3LLSRRTDimensionNodeAttribute_dimensionGraduated)
                {
                    printf("    dimensionGraduated: %d\n", 
                        pDimension->dimensionGraduated);
                }

                for (k = 0; k < pDimension->dimensionTitles.nOffsetsUsed; k++)
                {
                    ATSC3LLSRRT_TextTypeNode *pDimensionTitle;

                    pDimensionTitle = (ATSC3LLSRRT_TextTypeNode *)
                        EnXMLParserDataOffsetArrayGetPointer(
                        &pDimension->dimensionTitles, k);
                    printf("    DimensionTitle %d: %s", k,
                        (char *)EnXMLParserDataBufferGetPointer(
                            &pParsedRRT->dataBuffer, 
                            pDimensionTitle->stringDataOffset)); 
                    if (pDimensionTitle->base.presentFlags & 
                        ATSC3LLSRRTTextTypeNodeAttribute_lang)
                    {
                        printf(" lang: %s", 
                            (char *)EnXMLParserDataBufferGetPointer(
                                &pParsedRRT->dataBuffer, 
                                pDimensionTitle->langDataOffset)); 
                    }
                    printf("\n");
                }

                for (k = 0; k < pDimension->ratings.nOffsetsUsed; k++)
                {
                    ATSC3LLSRRT_RatingNode *pRating;

                    pRating = (ATSC3LLSRRT_RatingNode *)
                        EnXMLParserDataOffsetArrayGetPointer(
                        &pDimension->ratings, k);
                    printf("    Rating %d: value %d\n", k,
                        pRating->ratingValue);
                    printf("      RatingValueAbbrevs: {");

                    for (l = 0; 
                         l < pRating->ratingValueAbbrevs.nOffsetsUsed; 
                         l++)
                    {
                        ATSC3LLSRRT_TextTypeNode *pAbbrev;

                        pAbbrev = (ATSC3LLSRRT_TextTypeNode *)
                            EnXMLParserDataOffsetArrayGetPointer(
                            &pRating->ratingValueAbbrevs, l);
                        printf(" %s ", 
                            (char *)EnXMLParserDataBufferGetPointer(
                                &pParsedRRT->dataBuffer,
                                pAbbrev->stringDataOffset));
                        if (pAbbrev->base.presentFlags &
                            ATSC3LLSRRTTextTypeNodeAttribute_lang)
                        {
                            printf(" (%s) ", 
                                (char *)EnXMLParserDataBufferGetPointer(
                                    &pParsedRRT->dataBuffer,
                                    pAbbrev->langDataOffset));
                        }
                    }
                    printf("}\n");

                    printf("      RatingValueStrings: {");
                    for (l = 0; 
                         l < pRating->ratingValueStrings.nOffsetsUsed; 
                         l++)
                    {
                        ATSC3LLSRRT_TextTypeNode *pString;

                        pString = (ATSC3LLSRRT_TextTypeNode *)
                            EnXMLParserDataOffsetArrayGetPointer(
                            &pRating->ratingValueStrings, l);
                        printf(" %s ", 
                            (char *)EnXMLParserDataBufferGetPointer(
                                &pParsedRRT->dataBuffer,
                                pString->stringDataOffset));
                        if (pString->base.presentFlags &
                            ATSC3LLSRRTTextTypeNodeAttribute_lang)
                        {
                            printf(" (%s) ", 
                                (char *)EnXMLParserDataBufferGetPointer(
                                    &pParsedRRT->dataBuffer,
                                    pString->langDataOffset));
                        }
                    }
                    printf("}\n");
                }
            }
        }
    }
}

int32_t
ProcessRRT(LLSInfo *pLLSInfo, int64_t udpPacketNumber, const uint8_t *pPayload, 
    int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSRRT_RatingRegionTablesNode parsedRatingRegionTables;
    int32_t fHaveParsedRRT = 0;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/rrt-%.3d-%.3d-%.5d",
        pUserData->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version, pUserData->counter++) > 
        sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseRRT(strXML, xmlLength, &parsedRatingRegionTables))
    {
        printf("%"PRId64" Failed to parse RRT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedRRT = 1;

    PrintRRT(&parsedRatingRegionTables);

out:
    if (fHaveParsedRRT)
    {
        if (ATSC3LLSFreeParsedRRT(&parsedRatingRegionTables))
        {
            fprintf(stderr,"Failed to free RRT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessSystemTime(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSSystemTime_SystemTimeNode parsedSystemTime;
    int32_t fHaveParsedSystemTime;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/systime-%.3d-%.3d-%.5d",
        pUserData->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version, pUserData->counter++) > 
        sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseSystemTime(strXML, xmlLength,  &parsedSystemTime))
    {
        printf("%"PRId64" Failed to parse SystemTime\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedSystemTime = 1;

out:
    if (fHaveParsedSystemTime)
    {
        if (ATSC3LLSFreeParsedSystemTime(&parsedSystemTime))
        {
            fprintf(stderr, "Failed to free parsed system time\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}
int32_t
ProcessAEAT(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSAEAT_AEATNode parsedATSC3LLSAEAT;
    int32_t fHaveParsedAEAT;
    uint8_t groupId;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    groupId = pLLSInfo->lls_group_id;

    if (pUserData->vAEATVersion[groupId] == pLLSInfo->lls_table_version)
    {
        if (pUserData->vpAEAT[groupId])
        {
            if (pUserData->vAEATLength[groupId] == payloadLength &&
                !memcmp(pUserData->vpAEAT[groupId], pPayload, payloadLength))
            {
                goto out;
            }
            fprintf(stderr, "Unexpected change in AEAT %d/%d %d/%d %u/%u\n", 
                pUserData->vAEATVersion[groupId], pLLSInfo->lls_table_version,
                pUserData->vAEATLength[groupId], payloadLength,
                ENUtils_CalcCRC32(pUserData->vpAEAT[groupId], payloadLength), 
                ENUtils_CalcCRC32(pPayload, payloadLength));
        }
    }

    if (!pUserData->vpAEAT[groupId])
    {
        pUserData->vpAEAT[groupId] = (uint8_t *)malloc(AEAT_MAX_LENGTH);
        if (!pUserData->vpAEAT[groupId])
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > AEAT_MAX_LENGTH)
    {
        fprintf(stderr, "AEAT too large\n");
        goto out;
    }
    pUserData->vAEATLength[groupId] = payloadLength;
    memcpy(pUserData->vpAEAT[groupId], pPayload, payloadLength);

    pUserData->vAEATVersion[groupId] = pLLSInfo->lls_table_version;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/aeat-%.3d-%.3d-%.5d",
        pUserData->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version, pUserData->counter++) > 
        sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseAEAT(strXML, xmlLength,  &parsedATSC3LLSAEAT))
    {
        printf("%"PRId64" Failed to parse AEAT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedAEAT = 1;
    printf("%"PRId64" AEAT parsed successfully\n", udpPacketNumber);

out:
    if (fHaveParsedAEAT)
    {
        if (ATSC3LLSFreeParsedAEAT(&parsedATSC3LLSAEAT))
        {
            fprintf(stderr, "Failed to free parsed AEAT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessOnscreenMessageNotification(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSOSMN_OnscreenMessageNotificationNode parsedOnscreenMessageNotification;
    int32_t fHaveParsedOSMN = 0;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char*)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), 
        "%s/onscreenmessage-%.3d-%.3d-%.5d", pUserData->pOutPath, 
        pLLSInfo->lls_group_id, pLLSInfo->lls_table_version, 
        pUserData->counter++) > sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseOnscreenMessageNotification(strXML, xmlLength,
        &parsedOnscreenMessageNotification))
    {
        printf("%"PRId64" Failed to parse OnscreenMessageNotification\n",
            udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedOSMN = 1;

out:
    if (fHaveParsedOSMN)
    {
        if (ATSC3LLSFreeParsedOnscreenMessageNotification(
                &parsedOnscreenMessageNotification))
        {
            fprintf(stderr, "Failed to free parsed OSMN\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessCertificationData(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
#if 0
    ATSC3LLSCertificationData_CertificationDataNode parsedCertificationData;
    int32_t fHaveParsedCertificationData = 0;
#endif
    uint8_t groupId;
    int fd = -1;
    static int counter = 0;
    int ret;
    int32_t rval = 0;

    groupId = pLLSInfo->lls_group_id;

    if (pUserData->vCertificationDataVersion[groupId] == 
        pLLSInfo->lls_table_version)
    {
        if (pUserData->vpCertificationData[groupId])
        {
            if (pUserData->vCertificationDataLength[groupId] == payloadLength &&
                !memcmp(pUserData->vpCertificationData[groupId], pPayload, 
                    payloadLength))
            {
                goto out;
            }
            fprintf(stderr, "Unexpected change in CertificationData\n");
        }
    }

    if (!pUserData->vpCertificationData[groupId])
    {
        pUserData->vpCertificationData[groupId] = 
            (uint8_t *)malloc(CERTIFICATION_DATA_MAX_LENGTH);
        if (!pUserData->vpCertificationData[groupId])
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > CERTIFICATION_DATA_MAX_LENGTH)
    {
        fprintf(stderr, "CertificationData too large\n");
        goto out;
    }
    pUserData->vCertificationDataLength[groupId] = payloadLength;
    memcpy(pUserData->vpCertificationData[groupId], pPayload, payloadLength);

    pUserData->vCertificationDataVersion[groupId] = pLLSInfo->lls_table_version;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/cdt-%.3d-%.3d-%.3d-%.5d",
        pUserData->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version, counter++, pUserData->counter++) > 
        sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s: %s\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

#if 0
    if (ATSC3LLSParseCertificationData(strXML, xmlLength,  &parsedCertificationData))
    {
        printf("%"PRId64" Failed to parse signed LLS\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedCertificationData = 1;

    PrintCertificationData(&parsedCertificationData, pLLSInfo);
#endif

out:
#if 0
    if (fHaveParsedCertificationData)
    {
        if (ATSC3LLSFreeParsedCertificationData(&parsedCertificationData))
        {
            fprintf(stderr, "Failed to free parsed signed LLS\n");
        }
    }
#endif

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessSignedMultiTable(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    uint8_t groupId;
    uint8_t lls_payload_count;
    int fd = -1;
    static int counter = 0;
    int32_t i;
    int32_t rval = 0;

    groupId = pLLSInfo->lls_group_id;

    if (pUserData->vSignedMultiTableVersion[groupId] == 
        pLLSInfo->lls_table_version)
    {
        if (pUserData->vpSignedMultiTable[groupId])
        {
            if (pUserData->vSignedMultiTableLength[groupId] == payloadLength &&
                !memcmp(pUserData->vpSignedMultiTable[groupId], pPayload, 
                    payloadLength))
            {
                goto out;
            }
            fprintf(stderr, "Unexpected change in SignedMultiTable\n");
        }
    }

    if (!pUserData->vpSignedMultiTable[groupId])
    {
        pUserData->vpSignedMultiTable[groupId] = 
            (uint8_t *)malloc(SIGNED_MULTITABLE_MAX_LENGTH);
        if (!pUserData->vpSignedMultiTable[groupId])
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > SIGNED_MULTITABLE_MAX_LENGTH)
    {
        fprintf(stderr, "SignedMultitable too large\n");
        goto out;
    }
    pUserData->vSignedMultiTableLength[groupId] = payloadLength;
    memcpy(pUserData->vpSignedMultiTable[groupId], pPayload, payloadLength);

    pUserData->vSignedMultiTableVersion[groupId] = pLLSInfo->lls_table_version;

    if (snprintf(strFilename, sizeof(strFilename), 
        "%s/SignedMultiTable-%.3d-%.3d-%.3d-%.5d", pUserData->pOutPath, 
        pLLSInfo->lls_group_id, pLLSInfo->lls_table_version, 
        counter++, pUserData->counter++) > sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, pPayload, payloadLength) != payloadLength)
    {
        printf("%"PRId64" Failed to write to %s: %s\n", udpPacketNumber,
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (!payloadLength)
    {
        fprintf(stderr, "SignedMultiTable too short\n");
        goto out;
    }
    lls_payload_count = pPayload[0];
    payloadLength--;
    pPayload++;

    for (i = 0; i < lls_payload_count; i++)
    {
        LLSInfo llsInfo;
        int32_t lls_payload_length;

        llsInfo = *pLLSInfo;
        if (payloadLength < 4)
        {
            fprintf(stderr, "SignedMultiTable: table %d header truncated\n", i);
            goto out;
        }
        llsInfo.lls_table_id = pPayload[0];
        llsInfo.lls_table_version = pPayload[1];
        lls_payload_length = (pPayload[2] << 8) | pPayload[3];

        payloadLength -= 4;
        pPayload += 4;

        if (payloadLength < lls_payload_length)
        {
            fprintf(stderr, "SignedMultiTable: payload %d trucated\n", i);
        }

        rval = ProcessLLSPacket(&llsInfo, udpPacketNumber, pPayload, 
            lls_payload_length, pUserData);
        if (rval)
        {
            goto out;
        }

        payloadLength -= lls_payload_length;
        pPayload += lls_payload_length;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessLLSPacket(LLSInfo *pLLSInfo, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    int32_t rval = 0;

    switch(pLLSInfo->lls_table_id)
    {
        case 0x01: // SLT
            rval = ProcessSLT(pLLSInfo, udpPacketNumber, pPayload,
                payloadLength, pUserData);
            break;
        case 0x02: // RRT
            rval = ProcessRRT(pLLSInfo, udpPacketNumber, pPayload,
                payloadLength, pUserData);
            break;
        case 0x03: // SystemTime
          rval = ProcessSystemTime(pLLSInfo, udpPacketNumber, pPayload,
              payloadLength, pUserData);
          break;
        case 0x04: // AEAT
            rval = ProcessAEAT(pLLSInfo, udpPacketNumber, pPayload,
                payloadLength, pUserData);
            break;
        case 0x05: // OnscreenMessageNotification
            rval = ProcessOnscreenMessageNotification(pLLSInfo, udpPacketNumber,
                pPayload, payloadLength, pUserData);
            break;
        case 0x06: // CertificationData
            rval = ProcessCertificationData(pLLSInfo, udpPacketNumber,
                pPayload, payloadLength, pUserData);
            break;
        case 0xfe: // SignedMultiTable 
            rval = ProcessSignedMultiTable(pLLSInfo, udpPacketNumber,
                pPayload, payloadLength, pUserData);
            break;
        default:
            printf("%"PRId64" Unhandled LLS table id %d\n", udpPacketNumber,
                pLLSInfo->lls_table_id);
            rval = 1;
            break;
    }

    return rval;
}

int32_t
ProcessUDPPacket(int64_t udpPacketNumber, const uint8_t *pPayload,
    int32_t payloadLength, struct UserData *pUserData)
{
    LLSInfo llsInfo;
    int32_t offset;
    int32_t rval = 0;

   if (payloadLength < 4)
   {
       printf("%"PRId64" LLS payload too short (%d)\n", udpPacketNumber,
           payloadLength);
       rval = 1;
       goto out;
   }

   llsInfo.lls_table_id = pPayload[0];
   llsInfo.lls_group_id = pPayload[1];
   llsInfo.group_count_minus1 = pPayload[2];
   llsInfo.lls_table_version = pPayload[3];
   offset = 4;

   rval = ProcessLLSPacket(&llsInfo, udpPacketNumber, pPayload + offset, 
       payloadLength - offset, pUserData);

out:
    return rval;
}

int32_t
HandleUDPPacket(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, struct UserData *pUserData)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (packetLength < 8)
    {
        fprintf(stderr, "UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength < length)
    {
        fprintf(stderr, "Not enough data for UDP packet: %d < %d\n",
            packetLength, length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    if (destinationAddress != 0xE000173C || destPort != 4937)
    {
        goto out;
    }

    if (checksum)
    {
        struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader =
        {
            .sourceAddress = htonl(sourceAddress),
            .destinationAddress = htonl(destinationAddress),
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        if (pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("UDP\n");
            printf("  Source port: %d\n", sourcePort);
            printf("  Dest port: %d\n", destPort);
            printf("  Length: %d\n", length);
            printf("  Checksum: %x\n", checksum);
        }

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);
    }
    else
    {
        sum = 0;
    }

    if (sum == 0)
    {
        if (pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("  Checksum verified\n");
        }
    }
    else
    {
        if (!pUserData->fIgnoreChecksumErrors || 
            pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("  UDP CHECKSUM ERROR: %x\n", sum);
        }
    }

    if (sum == 0 || pUserData->fIgnoreChecksumErrors)
    {
        if (ProcessUDPPacket(packetNumber, pPacket + 8, packetLength - 8,
            pUserData))
        {
            fprintf(stderr, "Failed to process UDP packet\n");
            rval = 1;
        }
    }

out:
    return rval;
}

void
TransportPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    switch (protocol)
    {
        case 17:
            HandleUDPPacket(pPacketTimeUTC, packetNumber, sourceAddress, 
                destinationAddress, pPacket, sPacket, pUserData);
            break;
        default:
            fprintf(stderr, "Unhandled protocol %d\n", protocol);
            break;
    }
}

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, pPacketTimeUTC, 
        packetNumber, pPayload, payloadLength);

    return;
}

void
HandleLinkMappingTable(ENATSC3ALPLinkLayerSignalingHeader *pLLSHeader, 
    uint32_t plpId, struct timespec *pPacketTimeUTC, const uint8_t *pPayload, 
    int32_t payloadLength, struct UserData *pUserData)
{
    uint32_t numPLPs;
    int i;
    int j;

    if (pUserData->vLMTs[plpId].version == pLLSHeader->signaling_version)
    {
        if (pUserData->vLMTs[plpId].length != payloadLength ||
            memcmp(pUserData->vLMTs[plpId].pLMT + 5, pPayload + 5, 
            payloadLength - 5))
        {
            fprintf(stderr, "LMT changed unexpectedly\n");
        }
        else
        {
            goto out;
        }
    }

    pUserData->vLMTs[plpId].version = pLLSHeader->signaling_version;
    if (payloadLength > pUserData->vLMTs[plpId].nAllocated)
    {
        uint8_t *pNew;

        pNew = realloc(pUserData->vLMTs[plpId].pLMT, payloadLength + 4096);
        if (!pNew)
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
        pUserData->vLMTs[plpId].pLMT = pNew;
        pUserData->vLMTs[plpId].nAllocated = payloadLength + 4096;
    }

    pUserData->vLMTs[plpId].length = payloadLength;
    memcpy(pUserData->vLMTs[plpId].pLMT, pPayload, payloadLength);

    // Skip over signaling header
    pPayload += 5;

    if (!payloadLength)
    {
        printf("SHORT LMT PACKET\n");
        goto out;
    }
    numPLPs = (pPayload[0] >> 2) + 1;
    pPayload++;
    payloadLength--;

    printf("LMT plp %d version %d\n", plpId, pLLSHeader->signaling_version);
    printf("  num PLPs: %d\n", numPLPs);

    for (i = 0; i < numPLPs; i++)
    {
        uint32_t plpID;
        uint32_t numMulticasts;

        if (payloadLength < 2)
        {
            printf("SHORT LMT PACKET\n");
            goto out;
        }
        plpID = pPayload[0] >> 2;
        numMulticasts = pPayload[1];
        pPayload += 2;
        payloadLength -= 2;

        printf("    plp: %d\n", plpID);
        printf("      multicasts: %d\n", numMulticasts);

        for (j = 0; j < numMulticasts; j++)
        {
            uint32_t srcIPAddr;
            uint32_t dstIPAddr;
            uint32_t srcUDPPort;
            uint32_t dstUDPPort;
            uint32_t sidFlag;
            uint32_t compressedFlag;
            uint32_t SID;
            uint32_t contextId;

            if (payloadLength < 13)
            {
                printf("SHORT LMT PACKET\n");
                goto out;
            }

            srcIPAddr = (pPayload[0] << 24) | (pPayload[1] << 16) | 
                (pPayload[2] << 8) | pPayload[3]; 
            dstIPAddr = (pPayload[4] << 24) | (pPayload[5] << 16) | 
                (pPayload[6] << 8) | pPayload[7]; 
            srcUDPPort = (pPayload[8] << 8) | pPayload[9];
            dstUDPPort = (pPayload[10] << 8) | pPayload[11];
            sidFlag = pPayload[12] & 0x80;
            compressedFlag = pPayload[12] & 0x40;

            pPayload += 13;
            payloadLength -= 13;

            if (sidFlag)
            {
                if (payloadLength < 1)
                {
                    printf("SHORT LMT PACKET\n");
                    goto out;
                }
                SID = pPayload[0];
                pPayload++;
                payloadLength--;
            }
            else
            {
                SID = 0;
            }

            if (compressedFlag)
            {
                if (payloadLength < 1)
                {
                    printf("SHORT LMT PACKET\n");
                    goto out;
                }
                contextId = pPayload[0];
                pPayload++;
                payloadLength--;
            }
            else
            {
                contextId = 0;
            }

            printf("        Multicast %d\n", j);
            printf("          src: %d.%d.%d.%d:%d\n", 
                (srcIPAddr >> 24)  & 0xff, (srcIPAddr >> 16)  & 0xff,
                (srcIPAddr >> 8)  & 0xff, srcIPAddr  & 0xff,
                srcUDPPort);
            printf("          dst: %d.%d.%d.%d:%d\n", 
                (dstIPAddr >> 24)  & 0xff, (dstIPAddr >> 16)  & 0xff,
                (dstIPAddr >> 8)  & 0xff, dstIPAddr  & 0xff,
                dstUDPPort);
            printf("          SID flag: %d\n", sidFlag);
            if (sidFlag)
            {
                printf("          SID: %d\n", SID);
            }
            printf("          compressed flag: %d\n", compressedFlag);
            if (compressedFlag)
            {
                printf("          context id: %d\n", contextId);
            }
        }
    }

out:
    return;
}

void
LinkLevelSignalingPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, int32_t fError, 
    int32_t fDiscontinuity, const uint8_t *pPayload, int32_t payloadLength, 
    void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    ENATSC3ALPLinkLayerSignalingHeader llsheader;

    llsheader.signaling_type = pPayload[0];
    llsheader.signaling_type_extension = (pPayload[1] << 8) | pPayload[2];
    llsheader.signaling_version = pPayload[3];
    llsheader.signaling_format = (pPayload[4] >> 6) & 0x3;
    llsheader.signaling_encoding = (pPayload[4] >> 4) & 0x3;

    switch (llsheader.signaling_type)
    {
        case ENATSC3ALPSignalingType_LinkMappingTable:
            HandleLinkMappingTable(&llsheader, plpId, pPacketTimeUTC, 
                pPayload, payloadLength, pUserData);
            break;
        default:
            printf("Unhandled LLS signaling type %d\n", 
                llsheader.signaling_type);
            break;
    }

    return;
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, 
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC,
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }

}

void
FileReaderBasebandPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const ENATCS3L1Time *pL1Time, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pL1Time, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3FileType *pInFileType, char **ppOutPath, int32_t *pTOffsetSeconds, 
    uint64_t *pPLPIds, ENVerbosity *pVerbosity, 
    int32_t *pFIgnoreChecksumErrors)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;
    *pPLPIds = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPIds |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignorechecksum"))
        {
            *pFIgnoreChecksumErrors = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutPath == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s [-b alpcapturefile] [-bb basebandcapturefile] -o path -plp plpid [-plp plpid ...] [-offset seconds] [-v verbosity] [-ignorechecksum]\n", argv[0]);
        err = 1;
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    struct UserData userdata = 
    {
        .verbosity = ENVerbosity_Terse,
        .fIgnoreChecksumErrors = 0,
        .hATSC3ALP = NULL,
        .hATSC3IPV4 = NULL,
        .pOutPath = NULL,
        .ifd = -1,
    };
    int32_t tOffsetSeconds = 0;
    struct sigaction act;
    int32_t i;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    for (i = 0; i < numof(userdata.vSLTVersion); i++)
    {
        userdata.vSLTVersion[i] = -1;
    }
    for (i = 0; i < numof(userdata.vSLTLength); i++)
    {
        userdata.vSLTLength[i] = 0;
    }
    for (i = 0; i < numof(userdata.vpSLT); i++)
    {
        userdata.vpSLT[i] = NULL;
    }

    for (i = 0; i < numof(userdata.vCertificationDataVersion); i++)
    {
        userdata.vCertificationDataVersion[i] = -1;
    }
    for (i = 0; i < numof(userdata.vCertificationDataLength); i++)
    {
        userdata.vCertificationDataLength[i] = 0;
    }
    for (i = 0; i < numof(userdata.vpCertificationData); i++)
    {
        userdata.vpCertificationData[i] = NULL;
    }

    for (i = 0; i < numof(userdata.vSignedMultiTableVersion); i++)
    {
        userdata.vSignedMultiTableVersion[i] = -1;
    }
    for (i = 0; i < numof(userdata.vSignedMultiTableLength); i++)
    {
        userdata.vSignedMultiTableLength[i] = 0;
    }
    for (i = 0; i < numof(userdata.vpSignedMultiTable); i++)
    {
        userdata.vpSignedMultiTable[i] = NULL;
    }

    for (i = 0; i < numof(userdata.vLMTs); i++)
    {
        userdata.vLMTs[i].version = -1;
        userdata.vLMTs[i].length = 0;
        userdata.vLMTs[i].pLMT = NULL;
        userdata.vLMTs[i].nAllocated = 0;
    }

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &inFileType, &userdata.pOutPath, 
        &tOffsetSeconds, &userdata.plpIds, &userdata.verbosity, 
        &userdata.fIgnoreChecksumErrors))
    {
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband,
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandSetVerbosity(userdata.hATSC3Baseband,
            userdata.verbosity))
        {
            fprintf(stderr, "Failed to set baseband verbosity\n");
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
    {
        fprintf(stderr, "Failed to open ALP processor\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, IPV4PacketCallback, 
        LinkLevelSignalingPacketCallback, &userdata))
    {
        fprintf(stderr, "Failed to register callbacks\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ALPSetVerbosity(userdata.hATSC3ALP, userdata.verbosity))
    {
        fprintf(stderr, "Failed to set alp verbosity\n");
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband,
            userdata.plpIds))
        {
            fprintf(stderr, "Failed to sert baseband PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, userdata.plpIds))
    {
        fprintf(stderr, "Failed to set ALP PLP mask %"PRIx64"\n",
            userdata.plpIds);
        rval = 1;
        goto out;
    }

    if (ENATSC3IPV4Open(&userdata.hATSC3IPV4))
    {
        fprintf(stderr, "Failed to open ATSC3IPV4 session\n");
        rval = 1;
        goto out;
    }

    ENATSC3IPV4SetVerbosity(userdata.hATSC3IPV4, userdata.verbosity);

    if (ENATSC3IPV4RegisterCallbacks(userdata.hATSC3IPV4,
        TransportPacketCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ATSC3IPV4 callback\n");
        rval = 1;
        goto out;
    }

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds, FileReaderBasebandPacketCallback, NULL, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, 0))
    {
        fprintf(stderr, "Failed to clear ALP PLP mask\n");
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0))
        {
            fprintf(stderr, "Failed to clear baseband PLP Mask\n");
            rval = 1;
            goto out;
        }
    }
    
out:
    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (userdata.hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(userdata.hATSC3IPV4))
        {
            fprintf(stderr, "Failed to close ATSC3IPV4 session\n");
        }
    }

    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
