/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <assert.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3cap/enatsc3cap.h>
#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

#define MAX_PAYLOAD_SIZE 65536

struct UserData
{
    int ifd;
    int ofd;
    int32_t verbosity;

    ENATSC3BasebandHandle_t hATSC3Baseband;

    uint64_t plpIds;
};

int32_t fRunningG = 1;

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    ENATSC3CapRecordHeaderV2 header;

    header.type = ENATSC3CapPayloadType_ATSC3_ALP;
    header.plpId = plpId;
    header.length = htonl(sPacket);
    header.monowhen.sec = ENUtils_htonll(pPacketTimeUTC->tv_sec);
    header.monowhen.nsec = htonl(pPacketTimeUTC->tv_nsec);
    header.rtwhen.sec = ENUtils_htonll(pPacketTimeUTC->tv_sec);
    header.rtwhen.nsec = htonl(pPacketTimeUTC->tv_nsec);
    header.broadcastTime.L1B_time_info_flag = 0;

    if (write(pUserData->ofd, &header, sizeof(header)) != sizeof(header))
    {
        fprintf(stderr, "Failed to write record header: %s\n", 
            strerror(errno));
        goto out;
    }

    if (write(pUserData->ofd, pPacket, sPacket) != sPacket)
    {
        fprintf(stderr, "Failed to write packet: %s\n", strerror(errno));
        goto out;
    }

out:
    return;
}

void
FileReaderBasebandPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity,
    const ENATCS3L1Time *pL1Time, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pL1Time, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, char **ppOutFilename, 
    uint64_t *pPLPIds, int32_t *pVerbosity)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutFilename = NULL;
    *pPLPIds = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPIds |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutFilename == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b basebandcapturefile -o alpfile -plp plpid [-plp plpid ..] [-v level]\n", argv[0]);
        err = 1;
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    char *pInFilename;
    char *pOutFilename;
    struct UserData userdata = 
    {
        .ifd = -1,
        .ofd = -1,
        .verbosity = 0,
        .hATSC3Baseband = NULL,
    };
    struct sigaction act;
    ENATSC3CapFileHeader fileHeader;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &pOutFilename, 
        &userdata.plpIds, &userdata.verbosity))
    {
        rval = 1;
        goto out;
    }

    if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
    {
        fprintf(stderr, "Failed to open baseband processor\n");
        rval = 1;
        goto out;
    }

    if (userdata.verbosity)
    {
        ENATSC3BasebandSetVerbosity(userdata.hATSC3Baseband, 
            userdata.verbosity);
    }

    if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband, 
        ALPPacketCallback,  &userdata))
    {
        fprintf(stderr, "Failed to register callbacks\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, userdata.plpIds))
    {
        fprintf(stderr, "Failed to set baseband PLP mask %"PRIx64"\n", 
            userdata.plpIds);
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    userdata.ofd = open(pOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (userdata.ofd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n",
            pOutFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    fileHeader.format = ENATSC3CapFormat_V2;
    fileHeader.type = ENATSC3CapType_ATSC3_ALP;
    fileHeader.magic = htons(ENATSC3CAP_HEADER_MAGIC);
    if (write(userdata.ofd, &fileHeader, sizeof(fileHeader)) !=
        sizeof(fileHeader))
    {
        fprintf(stderr, "Failed to write file header: %s\n", strerror(errno));
        rval = 1;
        goto out;
    }

    rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 0, 
        FileReaderBasebandPacketCallback, NULL, &fRunningG, &userdata);

    if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0)) 
    {
        fprintf(stderr, "Failed to clear baseband PLP mask\n");
        rval = 1;
        goto out;
    }
    
out:
    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.ofd >= 0)
    {
        close(userdata.ofd);
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
