/*
 * Copyright © 2018-2023 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <assert.h>

#include <zlib.h>
#include <libxml2/libxml/xmlreader.h>

#include <openssl/x509.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>

#include <libenatsc3lls/libenatsc3lls.h>
#include <libenatsc3lls/atsc3lls_slt.h>
#include <libenatsc3lls/atsc3lls_systemtime.h>
#include <libenatsc3lls/atsc3lls_aeat.h>
#include <libenatsc3lls/atsc3lls_rrt.h>
#include <libenatsc3lls/atsc3lls_onscreenmessage.h>
#include <libenatsc3lls/atsc3lls_cdt.h>

#include <libenatsc3ssp/libenatsc3ssp.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

#define SLT_MAX_LENGTH 65536
#define SMT_MAX_LENGTH 65536
#define CDT_MAX_LENGTH 65536

struct UserData
{
    ENVerbosity verbosity;
    int32_t fIgnoreChecksumErrors;
    int ifd;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;
    ENATSC3IPV4Handle_t hATSC3IPV4;
    char *pOutPath;

    // Mask of the PLPs currently being received for LLS processing
    uint64_t llsPLPs;

    // Mask of the PLPs currently being received for service reception.
    uint64_t servicePLPs;

    struct
    {
        int32_t version;
        int32_t length;
        uint8_t *pLMT;
        int32_t nAllocated;
    } vLMTs[64];
    int32_t fLMTChanged;

    struct
    {
        int32_t version;
        int32_t length;
        uint8_t *pSLT;
    } vSLT[256];

    struct
    {
        int32_t version;
        int32_t length;
        uint8_t *pSMT;
    } vSMT[256];

    struct
    {
        int32_t version;
        int32_t length;
        uint8_t *pCDT;
    } vCDT[256];

    ENATSC3SSPHandle_t hSSP;
};

typedef struct
{
    uint8_t lls_table_id;
    uint8_t lls_group_id;
    uint8_t group_count_minus1;
    uint8_t lls_table_version;
} LLSInfo;

int32_t fRunningG = 1;

int32_t
ProcessLLSTable(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3LLSTableHeader *pLLSTableHeader, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t fSigned, 
    void *pUserData);

void
WriteToFile(int64_t udpPacketNumber, char *pFilename, void *pData, 
    int32_t sData)
{
    int fd;

    fd = open(pFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n",
           udpPacketNumber, pFilename, strerror(errno));
        return;
    }

    if (write(fd, pData, sData) != sData)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", 
            udpPacketNumber, pFilename, strerror(errno));
    }
    
    close(fd);
}

int32_t
FindPLPForMulticastAddress(uint32_t multicastDestIPAddr, 
    uint32_t multicastDestPort, int32_t *pPLPId, struct UserData *pUserData)
{
    const uint8_t *pPayload;
    int32_t payloadLength;
    uint32_t numPLPs;
    int32_t plpId;
    int i;
    int j;
    int32_t rval = 1;

    for (plpId = 0; plpId < 64; plpId++)
    {
        pPayload = pUserData->vLMTs[plpId].pLMT;
        payloadLength = pUserData->vLMTs[plpId].length;

        if (!payloadLength)
        {
            continue;
        }

        // Skip over signaling header
        pPayload += 5;

        if (payloadLength <= 1)
        {
            goto out;
        }

        numPLPs = (pPayload[0] >> 2) + 1;
        pPayload++;
        payloadLength--;

        for (i = 0; i < numPLPs; i++)
        {
            uint32_t plpId;
            uint32_t numMulticasts;

            if (payloadLength < 2)
            {
                goto out;
            }
            plpId = pPayload[0] >> 2;
            numMulticasts = pPayload[1];
            pPayload += 2;
            payloadLength -= 2;

            for (j = 0; j < numMulticasts; j++)
            {
                uint32_t destIPAddr;
                uint32_t destUDPPort;
                uint32_t sidFlag;
                uint32_t compressedFlag;

                if (payloadLength < 13)
                {
                    goto out;
                }

                destIPAddr = (pPayload[4] << 24) | (pPayload[5] << 16) | 
                    (pPayload[6] << 8) | pPayload[7]; 
                destUDPPort = (pPayload[10] << 8) | pPayload[11];
                sidFlag = pPayload[12] & 0x80;
                compressedFlag = pPayload[12] & 0x40;

                pPayload += 13;
                payloadLength -= 13;

                if (sidFlag)
                {
                    if (payloadLength < 1)
                    {
                        goto out;
                    }
                    pPayload++;
                    payloadLength--;
                }

                if (compressedFlag)
                {
                    if (payloadLength < 1)
                    {
                        goto out;
                    }
                    pPayload++;
                    payloadLength--;
                }

                if (multicastDestIPAddr == destIPAddr && 
                    multicastDestPort == destUDPPort)
                {
                    *pPLPId = plpId;
                    rval = 0;
                    goto out;
                }
            }
        }
    }

out:
    return rval;
}

int32_t
PrintSLT(ATSC3LLSSLT_SLTNode *pParsedSLT, uint32_t lls_group_id, 
    uint32_t lls_table_version, struct UserData *pUserData)
{
    int32_t i;
    int32_t j;
    int32_t k;
    int32_t rval = 0;

    printf("SLT group %d version %d\n", lls_group_id, lls_table_version);
    printf("  bsids: ");
    for (i = 0; i < pParsedSLT->bsids.nOffsetsUsed; i++)
    {
        printf("%d ", EnXMLParserArrayGetValue(&pParsedSLT->bsids, i));
    }
    printf("\n");

    for (i = 0; i < pParsedSLT->services.nOffsetsUsed; i++)
    {
        ATSC3LLSSLT_ServiceNode *pService;

        printf("  Service %d\n", i);

        pService = (ATSC3LLSSLT_ServiceNode *)
            EnXMLParserDataOffsetArrayGetPointer(&pParsedSLT->services, i);

        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_serviceId)
        {
            printf("    serviceId: %"PRIu32"\n", pService->serviceId);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_sltSvcSeqNum)
        {
            printf("    sltSvcSeqNum: %"PRIu32"\n", pService->sltSvcSeqNum);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_protected)
        {
            printf("    protected: %d\n", pService->protected);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_majorChannelNo)
        {
            printf("    majorChannelNo: %"PRIu32"\n", pService->majorChannelNo);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_minorChannelNo)
        {
            printf("    minorChannelNo: %"PRIu32"\n", pService->minorChannelNo);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_serviceCategory)
        {
            printf("    serviceCategory: %"PRIu32"\n", 
                pService->serviceCategory);
        }
        if (pService->base.presentFlags &
            ATSC3LLSSLTServiceNodeAttribute_shortServiceName)
        {
            printf("    shortServiceName: %s\n",
                (char *)EnXMLParserDataBufferGetPointer(
                    &pParsedSLT->dataBuffer, 
                    pService->shortServiceNameDataOffset));
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_hidden)
        {
            printf("    hidden: %d\n", pService->hidden);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_broadbandAccessRequired)
        {
            printf("    broadbandAccessRequired: %d\n", 
                pService->broadbandAccessRequired);
        }
#if 0
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_essential)
        {
            printf("    essential: %d\n", pService->essential);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_simulcastTSID)
        {
            printf("    simulcastTSID: %"PRIu16"\n", pService->simulcastTSID);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_simulcastMajorChannelNo)
        {
            printf("    simulcastMajorChannelNo: %"PRIu32"\n", 
                pService->simulcastMajorChannelNo);
        }
        if (pService->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_simulcastMinorChannelNo)
        {
            printf("    simulcastMinorChannelNo: %"PRIu32"\n", 
                pService->simulcastMinorChannelNo);
        }
#endif
        if (pService->base.presentFlags &
            ATSC3LLSSLTServiceNodeChildNode_BroadcastSvcSignaling)
        {
            ATSC3LLSSLT_BroadcastSvcSignalingNode *pBroadcastSvcSignalingNode = 
                (ATSC3LLSSLT_BroadcastSvcSignalingNode *)
                &pService->broadcastSvcSignaling;
            printf("    BroadcastSvcSignaling\n");
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsProtocol)
            {
                printf("      slsProtocol: %d\n",
                    pBroadcastSvcSignalingNode->slsProtocol);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMajorProtocolVersion)
            {
                printf("      slsMajorProtocolVersion: %d\n",
                    pBroadcastSvcSignalingNode->slsMajorProtocolVersion);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMinorProtocolVersion)
            {
                printf("      slsMinorProtocolVersion: %d\n",
                    pBroadcastSvcSignalingNode->slsMinorProtocolVersion);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress)
            {
                printf("      slsDestinationIpAddress: %s\n",
                    pBroadcastSvcSignalingNode->slsDestinationIpAddress);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort)
            {
                printf("      slsDestinationUdpPort: %d\n",
                    pBroadcastSvcSignalingNode->slsDestinationUdpPort);
            }
            if (pBroadcastSvcSignalingNode->base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsSourceIpAddress)
            {
                printf("      slsSourceIpAddress: %s\n",
                    pBroadcastSvcSignalingNode->slsSourceIpAddress);
            }

            if ((pBroadcastSvcSignalingNode->base.presentFlags & 
                 ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress) &&
                (pBroadcastSvcSignalingNode->base.presentFlags & 
                 ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort))
            {
                int32_t plpId;
                struct in_addr destAddr;

                if (!inet_aton(
                    pBroadcastSvcSignalingNode->slsDestinationIpAddress, 
                    &destAddr))
                {
                    printf("      plpId: Failed to parse dest addr\n");
                }

                else if (FindPLPForMulticastAddress(htonl(destAddr.s_addr), 
                    pBroadcastSvcSignalingNode->slsDestinationUdpPort,
                    &plpId, pUserData))
                {
                    printf("      plpId: %s:%d not in LMT\n",
                        pBroadcastSvcSignalingNode->slsDestinationIpAddress,
                        pBroadcastSvcSignalingNode->slsDestinationUdpPort);
                    rval = 1;
                }
                else
                {
                    printf("      plpId: %d (from LMT)\n", plpId);
                }
            }
            else
            {
                printf("      plpId: dest addr or port missing\n");
            }
        }
        if (pService->base.presentFlags &
            ATSC3LLSSLTServiceNodeChildNode_svcInetUrl)
        {
            ATSC3LLSSLTInetUrlNode *pSvcInetUrl = 
                (ATSC3LLSSLTInetUrlNode *)&pService->svcInetUrl;
            printf("    svcInetUrl: %s type %d\n", 
                (char *)EnXMLParserDataBufferGetPointer(
                    &pParsedSLT->dataBuffer, pSvcInetUrl->urlDataOffset),
                pSvcInetUrl->urlType);
        }
        printf("    OtherBsids (%d)\n", pService->otherBsids.nOffsetsUsed);
        for (j = 0; j < pService->otherBsids.nOffsetsUsed; j++)
        {
            ATSC3LLSSLT_OtherBsidNode *pOtherBsid;
            pOtherBsid = (ATSC3LLSSLT_OtherBsidNode *)
                EnXMLParserDataOffsetArrayGetPointer(&pService->otherBsids, j);
            printf("    %d: ", j);
            for (k = 0; k < pOtherBsid->otherBsids.nOffsetsUsed; k++)
            {
                printf("%d ",
                    EnXMLParserArrayGetValue(&pOtherBsid->otherBsids, k));
            }
            printf("type %d\n", pOtherBsid->type);
        }
    }

    return rval;
}

int32_t
ProcessSLT(int32_t plpId, ENATSC3LLSTableHeader *pLLSTableHeader, 
    int64_t udpPacketNumber, const uint8_t *pPayload, int32_t payloadLength, 
    struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSSLT_SLTNode parsedSLT;
    int32_t fHaveParsedSLT = 0;
    uint8_t groupId;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    groupId = pLLSTableHeader->lls_group_id;

    if (pUserData->vSLT[groupId].version == pLLSTableHeader->lls_table_version)
    {
        if (pUserData->vSLT[groupId].pSLT)
        {
            if (pUserData->vSLT[groupId].length == payloadLength &&
                payloadLength > 10 && 
                !memcmp(pUserData->vSLT[groupId].pSLT + 10, pPayload + 10, payloadLength - 10))
            {
                if (!pUserData->fLMTChanged)
                {
                    goto out;
                }
                printf("LMT changed, redisplaying SLT\n");
            }
            else
            {
                fprintf(stderr, "Unexpected change in SLT\n");
            }
        }
    }

    if (!pUserData->vSLT[groupId].pSLT)
    {
        pUserData->vSLT[groupId].pSLT = (uint8_t *)malloc(SLT_MAX_LENGTH);
        if (!pUserData->vSLT[groupId].pSLT)
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > SLT_MAX_LENGTH)
    {
        fprintf(stderr, "SLT too large\n");
        goto out;
    }
    pUserData->vSLT[groupId].length = payloadLength;
    memcpy(pUserData->vSLT[groupId].pSLT, pPayload, payloadLength);

    pUserData->vSLT[groupId].version = pLLSTableHeader->lls_table_version;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/slt-%.3d-%.3d",
            pUserData->pOutPath, pLLSTableHeader->lls_group_id, 
            pLLSTableHeader->lls_table_version) > sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               udpPacketNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, strXML, xmlLength) != xmlLength)
        {
            printf("%"PRId64" Failed to write to %s: %s\n", udpPacketNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ATSC3LLSParseSLT(strXML, xmlLength,  &parsedSLT))
    {
        printf("%"PRId64" Failed to parse SLT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }

    if (PrintSLT(&parsedSLT, pLLSTableHeader->lls_group_id, 
        pLLSTableHeader->lls_table_version, pUserData))
    {
        goto out;
    }

    fHaveParsedSLT = 1;
    pUserData->fLMTChanged = 0;

out:
    if (fHaveParsedSLT)
    {
        if (ATSC3LLSFreeParsedSLT(&parsedSLT))
        {
            fprintf(stderr, "Failed to free parsed SLT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

void
PrintRRT(ATSC3LLSRRT_RatingRegionTablesNode *pParsedRRT, uint32_t lls_group_id,
    uint32_t lls_table_version)
{
    int32_t i;
    int32_t j;
    int32_t k;
    int32_t l;

    printf("RRT group %d version %d\n", lls_group_id, lls_table_version);

    for (i = 0; i < pParsedRRT->ratingRegionTables.nOffsetsUsed; i++)
    {
        ATSC3LLSRRT_RatingRegionTableNode *pRRT;

        pRRT = (ATSC3LLSRRT_RatingRegionTableNode *)
            EnXMLParserDataOffsetArrayGetPointer(
            &pParsedRRT->ratingRegionTables, i);

        if (pRRT->base.presentFlags & 
            ATSC3LLSRRTRatingRegionTableNodeAttribute_regionIdentifier)
        {
            printf("  regionIdentifier: %d\n", pRRT->regionIdentifier);
        }

        if (pRRT->base.presentFlags &
            ATSC3LLSRRTRatingRegionTableChildNode_RegionIdText)
        {
            ATSC3LLSRRT_TextTypeNode *pRegionIdText;

            printf("  RegionIdText\n");
            for (j = 0; j < pRRT->regionIdTexts.nOffsetsUsed; j++)
            {
                pRegionIdText = (ATSC3LLSRRT_TextTypeNode *)
                    EnXMLParserDataOffsetArrayGetPointer(
                    &pRRT->regionIdTexts, j);
                printf("    text: %s",
                    (char *)EnXMLParserDataBufferGetPointer(
                        &pParsedRRT->dataBuffer, 
                        pRegionIdText->stringDataOffset)); 
                if (pRegionIdText->base.presentFlags & 
                    ATSC3LLSRRTTextTypeNodeAttribute_lang)
                {
                    printf(" lang: %s", 
                        (char *)EnXMLParserDataBufferGetPointer(
                            &pParsedRRT->dataBuffer, 
                            pRegionIdText->langDataOffset)); 
                }
                printf("\n");
            }
        }

        if (pRRT->base.presentFlags &
            ATSC3LLSRRTRatingRegionTableChildNode_Dimension)
        {
            ATSC3LLSRRT_DimensionNode *pDimension;

            for (j = 0; j < pRRT->dimensions.nOffsetsUsed; j++)
            {
                printf("  Dimension %d\n", j);
                pDimension = (ATSC3LLSRRT_DimensionNode *)
                    EnXMLParserDataOffsetArrayGetPointer(
                    &pRRT->dimensions, j);
                if (pDimension->base.presentFlags & 
                    ATSC3LLSRRTDimensionNodeAttribute_dimensionLevels)
                {
                    printf("    dimensionLevels: %u\n", 
                        pDimension->dimensionLevels);
                }
                if (pDimension->base.presentFlags & 
                    ATSC3LLSRRTDimensionNodeAttribute_dimensionGraduated)
                {
                    printf("    dimensionGraduated: %d\n", 
                        pDimension->dimensionGraduated);
                }

                for (k = 0; k < pDimension->dimensionTitles.nOffsetsUsed; k++)
                {
                    ATSC3LLSRRT_TextTypeNode *pDimensionTitle;

                    pDimensionTitle = (ATSC3LLSRRT_TextTypeNode *)
                        EnXMLParserDataOffsetArrayGetPointer(
                        &pDimension->dimensionTitles, k);
                    printf("    DimensionTitle %d: %s", k,
                        (char *)EnXMLParserDataBufferGetPointer(
                            &pParsedRRT->dataBuffer, 
                            pDimensionTitle->stringDataOffset)); 
                    if (pDimensionTitle->base.presentFlags & 
                        ATSC3LLSRRTTextTypeNodeAttribute_lang)
                    {
                        printf(" lang: %s", 
                            (char *)EnXMLParserDataBufferGetPointer(
                                &pParsedRRT->dataBuffer, 
                                pDimensionTitle->langDataOffset)); 
                    }
                    printf("\n");
                }

                for (k = 0; k < pDimension->ratings.nOffsetsUsed; k++)
                {
                    ATSC3LLSRRT_RatingNode *pRating;

                    pRating = (ATSC3LLSRRT_RatingNode *)
                        EnXMLParserDataOffsetArrayGetPointer(
                        &pDimension->ratings, k);
                    printf("    Rating %d: value %d\n", k,
                        pRating->ratingValue);
                    printf("      RatingValueAbbrevs: {");

                    for (l = 0; 
                         l < pRating->ratingValueAbbrevs.nOffsetsUsed; 
                         l++)
                    {
                        ATSC3LLSRRT_TextTypeNode *pAbbrev;

                        pAbbrev = (ATSC3LLSRRT_TextTypeNode *)
                            EnXMLParserDataOffsetArrayGetPointer(
                            &pRating->ratingValueAbbrevs, l);
                        printf(" %s ", 
                            (char *)EnXMLParserDataBufferGetPointer(
                                &pParsedRRT->dataBuffer,
                                pAbbrev->stringDataOffset));
                        if (pAbbrev->base.presentFlags &
                            ATSC3LLSRRTTextTypeNodeAttribute_lang)
                        {
                            printf(" (%s) ", 
                                (char *)EnXMLParserDataBufferGetPointer(
                                    &pParsedRRT->dataBuffer,
                                    pAbbrev->langDataOffset));
                        }
                    }
                    printf("}\n");

                    printf("      RatingValueStrings: {");
                    for (l = 0; 
                         l < pRating->ratingValueStrings.nOffsetsUsed; 
                         l++)
                    {
                        ATSC3LLSRRT_TextTypeNode *pString;

                        pString = (ATSC3LLSRRT_TextTypeNode *)
                            EnXMLParserDataOffsetArrayGetPointer(
                            &pRating->ratingValueStrings, l);
                        printf(" %s ", 
                            (char *)EnXMLParserDataBufferGetPointer(
                                &pParsedRRT->dataBuffer,
                                pString->stringDataOffset));
                        if (pString->base.presentFlags &
                            ATSC3LLSRRTTextTypeNodeAttribute_lang)
                        {
                            printf(" (%s) ", 
                                (char *)EnXMLParserDataBufferGetPointer(
                                    &pParsedRRT->dataBuffer,
                                    pString->langDataOffset));
                        }
                    }
                    printf("}\n");
                }
            }
        }
    }
}

int32_t
ProcessRRT(int32_t plpId, ENATSC3LLSTableHeader *pLLSTableHeader, 
    int64_t udpPacketNumber, const uint8_t *pPayload, int32_t payloadLength, 
    struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSRRT_RatingRegionTablesNode parsedRatingRegionTables;
    int32_t fHaveParsedRRT = 0;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/rrt-%.3d-%.3d",
            pUserData->pOutPath, pLLSTableHeader->lls_group_id, 
            pLLSTableHeader->lls_table_version) > sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               udpPacketNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, strXML, xmlLength) != xmlLength)
        {
            printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ATSC3LLSParseRRT(strXML, xmlLength, &parsedRatingRegionTables))
    {
        printf("%"PRId64" Failed to parse RRT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedRRT = 1;

    PrintRRT(&parsedRatingRegionTables, pLLSTableHeader->lls_group_id,
        pLLSTableHeader->lls_table_version);

out:
    if (fHaveParsedRRT)
    {
        if (ATSC3LLSFreeParsedRRT(&parsedRatingRegionTables))
        {
            fprintf(stderr,"Failed to free RRT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessSystemTime(int32_t plpId, ENATSC3LLSTableHeader *pLLSTableHeader, 
    int64_t udpPacketNumber, const uint8_t *pPayload, int32_t payloadLength, 
    struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSSystemTime_SystemTimeNode parsedSystemTime;
    int32_t fHaveParsedSystemTime;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/systime-%.3d-%.3d",
            pUserData->pOutPath, pLLSTableHeader->lls_group_id, 
            pLLSTableHeader->lls_table_version) > sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               udpPacketNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, strXML, xmlLength) != xmlLength)
        {
            printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ATSC3LLSParseSystemTime(strXML, xmlLength,  &parsedSystemTime))
    {
        printf("%"PRId64" Failed to parse SystemTime\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedSystemTime = 1;

out:
    if (fHaveParsedSystemTime)
    {
        if (ATSC3LLSFreeParsedSystemTime(&parsedSystemTime))
        {
            fprintf(stderr, "Failed to free parsed system time\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessAEAT(int32_t plpId, ENATSC3LLSTableHeader *pLLSTableHeader, 
    int64_t udpPacketNumber, const uint8_t *pPayload, int32_t payloadLength, 
    struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSAEAT_AEATNode parsedATSC3LLSAEAT;
    int32_t fHaveParsedAEAT;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/aeat-%.3d-%.3d",
            pUserData->pOutPath, pLLSTableHeader->lls_group_id, 
            pLLSTableHeader->lls_table_version) > sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               udpPacketNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, strXML, xmlLength) != xmlLength)
        {
            printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ATSC3LLSParseAEAT(strXML, xmlLength,  &parsedATSC3LLSAEAT))
    {
        printf("%"PRId64" Failed to parse AEAT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedAEAT = 1;

out:
    if (fHaveParsedAEAT)
    {
        if (ATSC3LLSFreeParsedAEAT(&parsedATSC3LLSAEAT))
        {
            fprintf(stderr, "Failed to free parsed AEAT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessOnscreenMessageNotification(int32_t plpId, 
    ENATSC3LLSTableHeader *pLLSTableHeader, int64_t udpPacketNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSOSMN_OnscreenMessageNotificationNode parsedOnscreenMessageNotification;
    int32_t fHaveParsedOSMN = 0;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char*)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename), 
            "%s/onscreenmessage-%.3d-%.3d", pUserData->pOutPath, 
            pLLSTableHeader->lls_group_id, pLLSTableHeader->lls_table_version) >
            sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               udpPacketNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, strXML, xmlLength) != xmlLength)
        {
            printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ATSC3LLSParseOnscreenMessageNotification(strXML, xmlLength,
        &parsedOnscreenMessageNotification))
    {
        printf("%"PRId64" Failed to parse OnscreenMessageNotification\n",
            udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedOSMN = 1;

out:
    if (fHaveParsedOSMN)
    {
        if (ATSC3LLSFreeParsedOnscreenMessageNotification(
                &parsedOnscreenMessageNotification))
        {
            fprintf(stderr, "Failed to free parsed OSMN\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

void
SignedLLSTableCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3LLSTableHeader *pLLSTableHeader, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t fSigned, void *pUserData)
{
    ProcessLLSTable(plpId, pPacketTimeUTC, packetNumber, pLLSTableHeader,
        pPayload, payloadLength, fSigned, pUserData);
}

int32_t
ProcessSMT(int32_t plpId, ENATSC3LLSTableHeader *pLLSTableHeader, 
    struct timespec *pPacketTimeUTC, int64_t packetNumber, 
    const uint8_t *pPayload, int32_t payloadLength, struct UserData *pUserData)
{
    char strFilename[256];
    uint8_t groupId;
    int fd = -1;
    int32_t rval = 0;

    groupId = pLLSTableHeader->lls_group_id;

    if (pUserData->vSMT[groupId].version == pLLSTableHeader->lls_table_version)
    {
        if (pUserData->vSMT[groupId].pSMT)
        {
            if (pUserData->vSMT[groupId].length== payloadLength &&
                payloadLength > 10 && 
                !memcmp(pUserData->vSMT[groupId].pSMT + 10, pPayload + 10, payloadLength - 10))
            {
                if (!pUserData->fLMTChanged)
                {
                    goto out;
                }
                printf("LMT changed, redisplaying SMT\n");
            }
            else
            {
                fprintf(stderr, "Unexpected change in SMT\n");
            }
        }
    }

    if (!pUserData->vSMT[groupId].pSMT)
    {
        pUserData->vSMT[groupId].pSMT = (uint8_t *)malloc(SMT_MAX_LENGTH);
        if (!pUserData->vSMT[groupId].pSMT)
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > SMT_MAX_LENGTH)
    {
        fprintf(stderr, "SMT too large\n");
        goto out;
    }
    pUserData->vSMT[groupId].length = payloadLength;
    memcpy(pUserData->vSMT[groupId].pSMT, pPayload, payloadLength);

    pUserData->vSMT[groupId].version = pLLSTableHeader->lls_table_version;

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename),
            "%s/signedmultitable-%.3d-%.3d", pUserData->pOutPath,
            pLLSTableHeader->lls_group_id, pLLSTableHeader->lls_table_version) >
            sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               packetNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, pPayload, payloadLength) != payloadLength)
        {
            printf("%"PRId64" Failed to write to %s (%s)\n", packetNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3SSPVerifySignedMultitableSignature(pUserData->hSSP, pPayload, 
        payloadLength, 1 /*fDontVerifySigningCertificate */))
    {
        fprintf(stderr, "Failed to verify SignedMultitable\n");
        rval = 1;
        goto out;
    }

    rval = ENATSC3LLSProcessSignedMultiTable(plpId, pPacketTimeUTC,
        packetNumber, pLLSTableHeader, pPayload, payloadLength,
        SignedLLSTableCallback, pUserData);

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessCertificationDataTable(int32_t plpId, 
    ENATSC3LLSTableHeader *pLLSTableHeader, struct timespec *pPacketTimeUTC, 
    int64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    struct UserData *pUserData)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSCDT_CertificationDataNode parsedCertificationData;
    int32_t fHaveParsedCertificationData = 0;
    struct timespec tMonoTimeout;
    uint8_t groupId;
    int fd = -1;
    int ret;
    int32_t rval = 0;

    groupId = pLLSTableHeader->lls_group_id;

    if (pUserData->vCDT[groupId].version == pLLSTableHeader->lls_table_version)
    {
        if (pUserData->vCDT[groupId].pCDT)
        {
            if (pUserData->vCDT[groupId].length == payloadLength &&
                payloadLength > 10 && 
                !memcmp(pUserData->vCDT[groupId].pCDT + 10, pPayload + 10, payloadLength - 10))
            {
                goto out;
            }
            else
            {
                fprintf(stderr, "Unexpected change in CDT\n");
            }
        }
    }

    if (!pUserData->vCDT[groupId].pCDT)
    {
        pUserData->vCDT[groupId].pCDT = (uint8_t *)malloc(CDT_MAX_LENGTH);
        if (!pUserData->vCDT[groupId].pCDT)
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
    }

    if (payloadLength > CDT_MAX_LENGTH)
    {
        fprintf(stderr, "CDT too large\n");
        goto out;
    }
    pUserData->vCDT[groupId].length = payloadLength;
    memcpy(pUserData->vCDT[groupId].pCDT, pPayload, payloadLength);

    pUserData->vCDT[groupId].version = pLLSTableHeader->lls_table_version;

    ret = ENUtils_GunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char*)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (pUserData->pOutPath)
    {
        if (snprintf(strFilename, sizeof(strFilename), 
            "%s/certificationdata-%.3d-%.3d", pUserData->pOutPath, 
            pLLSTableHeader->lls_group_id, pLLSTableHeader->lls_table_version) >
            sizeof(strFilename))
        {
            printf("Filename too long\n");
            rval = 1;
            goto out;
        }

        fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0)
        {
            printf("%"PRId64" Failed to open %s for writing (%s)\n",
               packetNumber, strFilename, strerror(errno));
            rval = 1;
            goto out;
        }

        if (write(fd, strXML, xmlLength) != xmlLength)
        {
            printf("%"PRId64" Failed to write to %s (%s)\n", packetNumber,
                strFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (ATSC3LLSParseCDT(strXML, xmlLength, &parsedCertificationData))
    {
        printf("%"PRId64" Failed to parse CertificationData\n",
            packetNumber);
        rval = 1;
        goto out;
    }

    fHaveParsedCertificationData = 1;

    clock_gettime(CLOCK_MONOTONIC, &tMonoTimeout);
    tMonoTimeout.tv_sec += 5;
    if (ENATSC3SSPVerifyAndLoadCertificationData(pUserData->hSSP,
        strXML, xmlLength, &parsedCertificationData, &tMonoTimeout, 
        1 /*fDontVerifySigningCertificate */, 
        1 /*fIgnoreOCSPResponseExpired */))
    {
        printf("Failed to verify and load certification data\n");
        rval = 1;
        goto out;
    }

out:
    if (fHaveParsedCertificationData)
    {
        if (ATSC3LLSFreeParsedCDT(&parsedCertificationData))
        {
            fprintf(stderr, "Failed to free parsed CDT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessLLSTable(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3LLSTableHeader *pLLSTableHeader, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t fSigned, 
    void *pUserData)
{
    int32_t rval = 0;

    switch(pLLSTableHeader->lls_table_id)
    {
        case 0x01: // SLT
            rval = ProcessSLT(plpId, pLLSTableHeader, packetNumber, 
                pPayload, payloadLength, pUserData);
           break;
        case 0x02: // RRT
            rval = ProcessRRT(plpId, pLLSTableHeader, packetNumber, 
                pPayload, payloadLength, pUserData);
           break;
        case 0x03: // SystemTime
            rval = ProcessSystemTime(plpId, pLLSTableHeader, packetNumber, 
                pPayload, payloadLength, pUserData);
           break;
        case 0x04: // AEAT
            rval = ProcessAEAT(plpId, pLLSTableHeader, packetNumber, 
                pPayload, payloadLength, pUserData);
           break;
        case 0x05: // OnscreenMessageNotification
            rval = ProcessOnscreenMessageNotification(plpId, pLLSTableHeader, 
                packetNumber, pPayload, payloadLength, pUserData);
           break;
        case 0x6: // CertificationData
            rval = ProcessCertificationDataTable(plpId, pLLSTableHeader, 
                pPacketTimeUTC, packetNumber, pPayload, payloadLength, 
                pUserData);
            break;
        case 0xfe: // SignedMultiTable
            rval = ProcessSMT(plpId, pLLSTableHeader, pPacketTimeUTC, 
                packetNumber, pPayload, payloadLength, pUserData);
            break;
        default:
            printf("%"PRId64" Unhandled LLS table id %d\n", packetNumber,
                pLLSTableHeader->lls_table_id);
            rval = 1;
            break;
    }
  
    return rval;
}

int32_t
ProcessLLSUDPPacket(int32_t plpId, struct timespec *pPacketTimeUTC, 
    int64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    struct UserData *pUserData)
{
    ENATSC3LLSTableHeader *pLLSTableHeader;
    int32_t rval = 0;

    if (payloadLength < 4)
    {
        printf("%"PRId64" LLS payload too short (%d)\n", packetNumber,
            payloadLength);
        rval = 1;
        goto out;
    }

    pLLSTableHeader = (ENATSC3LLSTableHeader *)pPayload;

    rval = ProcessLLSTable(plpId, pPacketTimeUTC, packetNumber, pLLSTableHeader,
        pPayload + 4, payloadLength - 4, 0 /* unsigned */, pUserData);

out:
    return rval;
}

int32_t
HandleUDPPacket(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, struct UserData *pUserData)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (packetLength < 8)
    {
        fprintf(stderr, "UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength < length)
    {
        fprintf(stderr, "Not enough data for UDP packet: %d < %d\n",
            packetLength, length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    // We only care about LLS packets
    if (destinationAddress != 0xE000173C || destPort != 4937)
    {
        goto out;
    }

    if (checksum)
    {
       struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader =
        {
            .sourceAddress = htonl(sourceAddress),
            .destinationAddress = htonl(destinationAddress),
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        if (pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("UDP\n");
            printf("  Source port: %d\n", sourcePort);
            printf("  Dest port: %d\n", destPort);
            printf("  Length: %d\n", length);
            printf("  Checksum: %x\n", checksum);
        }

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);
    }
    else
    {
        sum = 0;
    }

    if (sum == 0)
    {
        if (pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("  Checksum verified\n");
        }
    }
    else
    {
        if (!pUserData->fIgnoreChecksumErrors || 
            pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("  UDP CHECKSUM ERROR: %x\n", sum);
        }
    }

    if (sum == 0 || pUserData->fIgnoreChecksumErrors)
    {
        if (ProcessLLSUDPPacket(plpId, pPacketTimeUTC, packetNumber, 
            pPacket + 8, packetLength - 8, pUserData))
        {
            fprintf(stderr, "Failed to process UDP packet\n");
            rval = 1;
        }
    }

out:
    return rval;
}

void
TransportPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    switch (protocol)
    {
        case 17:
            HandleUDPPacket(plpId, pPacketTimeUTC, packetNumber, sourceAddress, 
                destinationAddress, pPacket, sPacket, pUserData);
            break;
        default:
            fprintf(stderr, "Unhandled protocol %d\n", protocol);
            break;
    }
}

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, pPacketTimeUTC, 
        packetNumber, pPayload, payloadLength);

    return;
}

void
HandleLinkMappingTable(ENATSC3ALPLinkLayerSignalingHeader *pLLSHeader, 
    uint32_t plpId, struct timespec *pPacketTimeUTC, const uint8_t *pPayload, 
    int32_t payloadLength, struct UserData *pUserData)
{
    uint32_t numPLPs;
    int i;
    int j;

    if (pUserData->vLMTs[plpId].version == pLLSHeader->signaling_version)
    {
        if (pUserData->vLMTs[plpId].length != payloadLength ||
            memcmp(pUserData->vLMTs[plpId].pLMT + 5, pPayload + 5, 
            payloadLength - 5))
        {
            fprintf(stderr, "LMT changed unexpectedly\n");
        }
        else
        {
            goto out;
        }
    }

    pUserData->vLMTs[plpId].version = pLLSHeader->signaling_version;
    if (payloadLength > pUserData->vLMTs[plpId].nAllocated)
    {
        uint8_t *pNew;

        pNew = realloc(pUserData->vLMTs[plpId].pLMT, payloadLength + 4096);
        if (!pNew)
        {
            fprintf(stderr, "Out of memory\n");
            goto out;
        }
        pUserData->vLMTs[plpId].pLMT = pNew;
        pUserData->vLMTs[plpId].nAllocated = payloadLength + 4096;
    }

    pUserData->vLMTs[plpId].length = payloadLength;
    memcpy(pUserData->vLMTs[plpId].pLMT, pPayload, payloadLength);

    // Skip over signaling header
    pPayload += 5;

    if (!payloadLength)
    {
        printf("SHORT LMT PACKET\n");
        goto out;
    }
    numPLPs = (pPayload[0] >> 2) + 1;
    pPayload++;
    payloadLength--;

    printf("LMT version %d\n", pLLSHeader->signaling_version);
    printf("  num PLPs: %d\n", numPLPs);

    for (i = 0; i < numPLPs; i++)
    {
        uint32_t plpId;
        uint32_t numMulticasts;

        if (payloadLength < 2)
        {
            printf("SHORT LMT PACKET\n");
            goto out;
        }
        plpId = pPayload[0] >> 2;
        numMulticasts = pPayload[1];
        pPayload += 2;
        payloadLength -= 2;

        printf("  plp: %d\n", plpId);
        printf("    multicasts: %d\n", numMulticasts);

        for (j = 0; j < numMulticasts; j++)
        {
            uint32_t srcIPAddr;
            uint32_t destIPAddr;
            uint32_t srcUDPPort;
            uint32_t destUDPPort;
            uint32_t sidFlag;
            uint32_t compressedFlag;
            uint32_t SID;
            uint32_t contextId;

            if (payloadLength < 13)
            {
                printf("SHORT LMT PACKET\n");
                goto out;
            }

            srcIPAddr = (pPayload[0] << 24) | (pPayload[1] << 16) | 
                (pPayload[2] << 8) | pPayload[3]; 
            destIPAddr = (pPayload[4] << 24) | (pPayload[5] << 16) | 
                (pPayload[6] << 8) | pPayload[7]; 
            srcUDPPort = (pPayload[8] << 8) | pPayload[9];
            destUDPPort = (pPayload[10] << 8) | pPayload[11];
            sidFlag = pPayload[12] & 0x80;
            compressedFlag = pPayload[12] & 0x40;

            pPayload += 13;
            payloadLength -= 13;

            if (sidFlag)
            {
                if (payloadLength < 1)
                {
                    printf("SHORT LMT PACKET\n");
                    goto out;
                }
                SID = pPayload[0];
                pPayload++;
                payloadLength--;
            }
            else
            {
                SID = 0;
            }

            if (compressedFlag)
            {
                if (payloadLength < 1)
                {
                    printf("SHORT LMT PACKET\n");
                    goto out;
                }
                contextId = pPayload[0];
                pPayload++;
                payloadLength--;
            }
            else
            {
                contextId = 0;
            }

            printf("      Multicast %d\n", j);
            printf("        src: %d.%d.%d.%d:%d\n", 
                (srcIPAddr >> 24)  & 0xff, (srcIPAddr >> 16)  & 0xff,
                (srcIPAddr >> 8)  & 0xff, srcIPAddr  & 0xff,
                srcUDPPort);
            printf("        dest: %d.%d.%d.%d:%d\n", 
                (destIPAddr >> 24)  & 0xff, (destIPAddr >> 16)  & 0xff,
                (destIPAddr >> 8)  & 0xff, destIPAddr  & 0xff,
                destUDPPort);
            printf("        SID flag: %d\n", sidFlag);
            if (sidFlag)
            {
                printf("        SID: %d\n", SID);
            }
            printf("        compressed flag: %d\n", compressedFlag);
            if (compressedFlag)
            {
                printf("        context id: %d\n", contextId);
            }
        }
    }

    pUserData->fLMTChanged = 1;

out:
    return;
}

void
LinkLayerSignalingPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPayload, 
    int32_t payloadLength, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    ENATSC3ALPLinkLayerSignalingHeader llsheader;

    llsheader.signaling_type = pPayload[0];
    llsheader.signaling_type_extension = (pPayload[1] << 8) | pPayload[2];
    llsheader.signaling_version = pPayload[3];
    llsheader.signaling_format = (pPayload[4] >> 6) & 0x3;
    llsheader.signaling_encoding = (pPayload[4] >> 4) & 0x3;

    switch (llsheader.signaling_type)
    {
        case ENATSC3ALPSignalingType_LinkMappingTable:
            HandleLinkMappingTable(&llsheader, plpId, pPacketTimeUTC, 
                pPayload, payloadLength, pUserData);
            break;
        default:
            printf("Unhandled LLS signaling type %d\n", 
                llsheader.signaling_type);
            break;
    }

    return;
}

void
LLSPLPsChangedCallback(uint64_t llsPLPs, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, void *pCbUserData)
{
    uint64_t activePLPs;
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    printf("lls plp mask: %"PRIx64"\n", llsPLPs);

    // Determine which PLPs we need. This is the union of the PLPs carrying LLS
    // and the PLPs which are in use for receiving services.
    activePLPs = llsPLPs | pUserData->servicePLPs;

    printf("Setting PLP mask %"PRIx64"\n", activePLPs);

    if (pUserData->hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(pUserData->hATSC3Baseband, activePLPs))
        {
            fprintf(stderr, 
                "WARN: Failed to set baseband PLP mask %"PRIx64"\n", 
                    activePLPs);
            goto out;
        }
    }

    if (pUserData->hATSC3ALP)
    {
        if (ENATSC3ALPSetPLPs(pUserData->hATSC3ALP, activePLPs))
        {
            fprintf(stderr, "WARN: Failed to set ALP PLP mask %"PRIx64"\n",
                activePLPs);
            goto out;
        }
    }

    pUserData->llsPLPs = llsPLPs;

out:
    return;
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, 
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
FileReaderIPV4PacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, 
        pPacketTimeUTC, packetNumber, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process IPV4 packet\n");
    }
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC,
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }

}

void
FileReaderBasebandPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const ENATCS3L1Time *pL1Time, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pL1Time, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3FileType *pInFileType, char **ppOutPath, int32_t *pTOffsetSeconds, 
    ENVerbosity *pVerbosity, int32_t *pFIgnoreChecksumErrors)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-pcapipv4"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_PCAPIPV4;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignorechecksum"))
        {
            *pFIgnoreChecksumErrors = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b alpcapturefile | -bb basebandcapturefile | -j288alp j288alpcapturefile | -pcapipv4 pcapfile [ -o path ] [-offset seconds] [-v verbosity] [-ignorechecksum]\n", argv[0]);
        err = 1;
    }

    return err;
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    struct UserData userdata = 
    {
        .verbosity = ENVerbosity_Terse,
        .fIgnoreChecksumErrors = 0,
        .hATSC3ALP = NULL,
        .hATSC3IPV4 = NULL,
        .pOutPath = NULL,
        .ifd = -1,
        .hSSP = NULL,
    };
    int32_t tOffsetSeconds = 0;
    int32_t i;
    int rval = 0;

    for (i = 0; i < numof(userdata.vSLT); i++)
    {
        userdata.vSLT[i].version = -1;
        userdata.vSLT[i].length = 0;
        userdata.vSLT[i].pSLT = NULL;
    }

    for (i = 0; i < numof(userdata.vSMT); i++)
    {
        userdata.vSMT[i].version = -1;
        userdata.vSMT[i].length = 0;
        userdata.vSMT[i].pSMT = NULL;
    }

    for (i = 0; i < numof(userdata.vCDT); i++)
    {
        userdata.vCDT[i].version = -1;
        userdata.vCDT[i].length = 0;
        userdata.vCDT[i].pCDT = NULL;
    }

    for (i = 0; i < numof(userdata.vLMTs); i++)
    {
        userdata.vLMTs[i].version = -1;
        userdata.vLMTs[i].length = 0;
        userdata.vLMTs[i].pLMT = NULL;
        userdata.vLMTs[i].nAllocated = 0;
    }

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (ENATSC3SSPCreate(NULL, NULL, &userdata.hSSP))
    {
        fprintf(stderr, "Failed to create ssp\n");
        rval = 1;
        goto out;
    }

    if (parseArgs(argc, argv, &pInFilename, &inFileType, &userdata.pOutPath, 
        &tOffsetSeconds, &userdata.verbosity, &userdata.fIgnoreChecksumErrors))
    {
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband, 
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandSetVerbosity(userdata.hATSC3Baseband,
            userdata.verbosity))
        {
            fprintf(stderr, "Failed to set baseband verbosity\n");
            rval = 1;
            goto out;
        }
    }

    if (inFileType != ENATSC3FileType_PCAPIPV4)
    {
        if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
        {
            fprintf(stderr, "Failed to open ALP processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, IPV4PacketCallback, 
            LinkLayerSignalingPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register callbacks\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3ALPSetVerbosity(userdata.hATSC3ALP, userdata.verbosity))
        {
            fprintf(stderr, "Failed to set baseband verbosity\n");
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3IPV4Open(&userdata.hATSC3IPV4))
    {
        fprintf(stderr, "Failed to open ATSC3IPV4 session\n");
        rval = 1;
        goto out;
    }

    ENATSC3IPV4SetVerbosity(userdata.hATSC3IPV4, userdata.verbosity);

    if (ENATSC3IPV4RegisterCallbacks(userdata.hATSC3IPV4,
        TransportPacketCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ATSC3IPV4 callback\n");
        rval = 1;
        goto out;
    }

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, LLSPLPsChangedCallback, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds,  FileReaderBasebandPacketCallback, 
                LLSPLPsChangedCallback, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, 
                LLSPLPsChangedCallback, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_PCAPIPV4:
            rval = ENATSC3FileReaderReadPCAPIPV4File(userdata.ifd, 
                tOffsetSeconds, FileReaderIPV4PacketCallback, 
                LLSPLPsChangedCallback, &fRunningG, &userdata);
            break;
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

out:
    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (userdata.hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(userdata.hATSC3IPV4))
        {
            fprintf(stderr, "Failed to close ATSC3IPV4 session\n");
        }
    }

    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.hSSP)
    {
        if (ENATSC3SSPDestroy(userdata.hSSP))
        {
            fprintf(stderr, "Failed to destroy ssp\n");
        }
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    for (i = 0; i < numof(userdata.vLMTs); i++)
    {
        free(userdata.vLMTs[i].pLMT);
    }

    for (i = 0; i < numof(userdata.vSLT); i++)
    {
        free(userdata.vSLT[i].pSLT);
    }

    for (i = 0; i < numof(userdata.vSMT); i++)
    {
        free(userdata.vSMT[i].pSMT);
    }

    for (i = 0; i < numof(userdata.vCDT); i++)
    {
        free(userdata.vCDT[i].pCDT);
    }

    return rval;
}
