/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <assert.h>

#include <zlib.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include <libenatsc3lls/atsc3lls_slt.h>
#include <libenatsc3lls/atsc3lls_systemtime.h>
#include <libenatsc3lls/atsc3lls_aeat.h>
#include <libenatsc3lls/atsc3lls_rrt.h>
#include <libenatsc3lls/atsc3lls_onscreenmessage.h>

#include <enatsc3cap/enudpcap.h>

#define MAX_PAYLOAD_SIZE 65536
#define NANOSECONDS_PER_SECOND 1000000000

typedef struct
{
    off_t curpos;
    char *pOutPath;
}  ParserInfo;

struct
{
    int32_t fInitialized;
    int32_t verbose;
    uint8_t vPadding[2048];
} udpToLLSGlobalsG =
{
    .fInitialized = 0,
    .verbose = 1
};

typedef struct
{
    uint8_t lls_table_id;
    uint8_t lls_group_id;
    uint8_t group_count_minus1;
    uint8_t lls_table_version;
} LLSInfo;

int32_t
ProcessSLT(int64_t udpPacketNumber, uint8_t *pPayload, 
    int32_t payloadLength, ParserInfo *pParserInfo, LLSInfo *pLLSInfo)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSSLT_SLTNode parsedSLT;
    int32_t fHaveParsedSLT = 0;
    int fd = -1; 
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe(pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/slt-%.3d-%.3d", 
        pParserInfo->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version) > sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n", 
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s: %s\n", udpPacketNumber, 
            strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseSLT(strXML, xmlLength,  &parsedSLT))
    {
        printf("%"PRId64" Failed to parse SLT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedSLT = 1;

out:
    if (fHaveParsedSLT)
    {
        if (ATSC3LLSFreeParsedSLT(&parsedSLT))
        {
            fprintf(stderr, "Failed to free parsed SLT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessRRT(int64_t udpPacketNumber, uint8_t *pPayload, 
    int32_t payloadLength, ParserInfo *pParserInfo, LLSInfo *pLLSInfo)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSRRT_RatingRegionTablesNode parsedRatingRegionTables;
    int32_t fHaveParsedRRT = 0;
    int fd = -1; 
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe(pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/rrt-%.3d-%.3d", 
        pParserInfo->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version) > sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n", 
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber, 
            strFilename, strerror(errno)); 
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseRRT(strXML, xmlLength, &parsedRatingRegionTables))
    {
        printf("%"PRId64" Failed to parse RRT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedRRT = 1;

out:
    if (fHaveParsedRRT)
    {
        if (ATSC3LLSFreeParsedRRT(&parsedRatingRegionTables))
        {
            fprintf(stderr,"Failed to free RRT\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessSystemTime(int64_t udpPacketNumber, uint8_t *pPayload, 
    int32_t payloadLength, ParserInfo *pParserInfo, LLSInfo *pLLSInfo)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSSystemTime_SystemTimeNode parsedSystemTime;
    int fd = -1; 
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe(pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/systime-%.3d-%.3d", 
        pParserInfo->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version) > sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n", 
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber, 
            strFilename, strerror(errno)); 
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseSystemTime(strXML, xmlLength,  &parsedSystemTime))
    {
        printf("%"PRId64" Failed to parse SystemTime\n", udpPacketNumber);
        rval = 1;
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessAEAT(int64_t udpPacketNumber, uint8_t *pPayload, 
    int32_t payloadLength, ParserInfo *pParserInfo, LLSInfo *pLLSInfo)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSAEAT_AEATNode parsedATSC3LLSAEAT;
    int fd = -1; 
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe(pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), "%s/aeat-%.3d-%.3d", 
        pParserInfo->pOutPath, pLLSInfo->lls_group_id, 
        pLLSInfo->lls_table_version) > sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n", 
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber, 
            strFilename, strerror(errno)); 
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseAEAT(strXML, xmlLength,  &parsedATSC3LLSAEAT))
    {
        printf("%"PRId64" Failed to parse AEAT\n", udpPacketNumber);
        rval = 1;
        goto out;
    }
out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessOnscreenMessageNotification(int64_t udpPacketNumber, uint8_t *pPayload, 
    int32_t payloadLength, ParserInfo *pParserInfo, LLSInfo *pLLSInfo)
{
    char strFilename[256];
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSOSMN_OnscreenMessageNotificationNode parsedOnscreenMessageNotification;
    int fd = -1; 
    int ret;
    int32_t rval = 0;

    ret = ENUtils_GunzipUnpipe(pPayload, payloadLength, (unsigned char*)strXML, 
        sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (snprintf(strFilename, sizeof(strFilename), 
        "%s/onscreenmessage-%.3d-%.3d", pParserInfo->pOutPath, 
        pLLSInfo->lls_group_id, pLLSInfo->lls_table_version) > 
        sizeof(strFilename))
    {
        printf("Filename too long\n");
        rval = 1;
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        printf("%"PRId64" Failed to open %s for writing (%s)\n", 
           udpPacketNumber, strFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (write(fd, strXML, xmlLength) != xmlLength)
    {
        printf("%"PRId64" Failed to write to %s (%s)\n", udpPacketNumber, 
            strFilename, strerror(errno)); 
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseOnscreenMessageNotification(strXML, xmlLength, 
        &parsedOnscreenMessageNotification))
    {
        printf("%"PRId64" Failed to parse OnscreenMessageNotification\n",
            udpPacketNumber);
        rval = 1;
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

int32_t
ProcessUDPPacket(int64_t udpPacketNumber, uint8_t *pPayload, 
    int32_t payloadLength, ParserInfo *pParserInfo)
{
    LLSInfo llsInfo;
    int32_t offset;
    int32_t rval = 0;

   if (payloadLength < 4)
   {
       printf("%"PRId64" LLS payload too short (%d)\n", udpPacketNumber,
           payloadLength);
       rval = 1;
       goto out;
   }

   llsInfo.lls_table_id = pPayload[0];
   llsInfo.lls_group_id = pPayload[1];
   llsInfo.group_count_minus1 = pPayload[2];
   llsInfo.lls_table_version = pPayload[3];
   offset = 4;

   switch(llsInfo.lls_table_id)
   {
       case 0x01: // SLT
          rval = ProcessSLT(udpPacketNumber, pPayload + offset, 
              payloadLength - offset, pParserInfo, &llsInfo);
          break;
       case 0x02: // RRT
          fprintf(stderr, "NOT PROCESSING RRT NODE - REDUCE PARSED STRUCTURE SIZE\n");
          rval = ProcessRRT(udpPacketNumber, pPayload + offset, 
              payloadLength - offset, pParserInfo, &llsInfo);
          break;
       case 0x03: // SystemTime
          rval = ProcessSystemTime(udpPacketNumber, pPayload + offset, 
              payloadLength - offset, pParserInfo, &llsInfo);
          break;
       case 0x04: // AEAT
          rval = ProcessAEAT(udpPacketNumber, pPayload + offset, 
              payloadLength - offset, pParserInfo, &llsInfo);
          break;
       case 0x05: // OnscreenMessageNotification
          rval = ProcessOnscreenMessageNotification(udpPacketNumber, 
              pPayload + offset, payloadLength - offset, pParserInfo, 
              &llsInfo);
          break;
       default:
           printf("%"PRId64" Unhandld LLS table id %d\n", udpPacketNumber, 
               llsInfo.lls_table_id);
           rval = 1;
           break;
   }

out:
    return rval;
}

int32_t
UDPToLLSInitialize(void)
{
    int32_t rval = 0;

    memset(udpToLLSGlobalsG.vPadding, 0, sizeof(udpToLLSGlobalsG.vPadding));

    udpToLLSGlobalsG.fInitialized = 1;

    return rval;
}

int32_t
UDPToLLSShutdown(void)
{
    int32_t rval = 0;

    udpToLLSGlobalsG.fInitialized = 0;

    return rval;
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, char **ppOutPath)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (!*ppInFilename)
    {
        err = 1;
    }

    if (!*ppOutPath)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b udpfilename -o outpath\n", argv[0]);
    }

    return err;
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    int32_t fUDPToLLSInitialized = 0;
    char *pInFilename;
    int fd = -1;
    uint8_t *pPayload = NULL;
    ParserInfo parserInfo;
    int64_t udpPacketNumber = 0;
    int rval = 0;

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (UDPToLLSInitialize())
    {
        fprintf(stderr, "Failed to init udp to mmt\n");
        rval = 1;
        goto out;
    }

    fUDPToLLSInitialized = 1;

    memset(&parserInfo, 0, sizeof(parserInfo));

    if (parseArgs(argc, argv, &pInFilename, &parserInfo.pOutPath))
    {
        rval = 1;
        goto out;
    }

    fd =  open(pInFilename, O_RDONLY);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", pInFilename, 
            strerror(errno));
        rval = 1;
        goto out;
    }

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    while (1)
    {
        EntangleCapHeader header;
        int nb;

        parserInfo.curpos = lseek(fd, 0, SEEK_CUR);
        nb = read(fd, &header, sizeof(header));
        if (nb == 0)
        {
            break;
        }
        else if (nb < 0)
        {
            fprintf(stderr, "Failed to read udp file: %s\n", strerror(errno));
            rval = 1;
            break;
        }
        else if (nb != sizeof(header))
        {
            fprintf(stderr, "Partial read of UDP header, exiting\n");
            rval = 1;
            break;
        }

        header.length = ntohs(header.length);
        if (header.length > MAX_PAYLOAD_SIZE)
        {
            fprintf(stderr, "Payload too long: %d > %d\n", 
                header.length, MAX_PAYLOAD_SIZE);
            rval = 1;
            break;
        }

        nb = read(fd, pPayload, header.length);
        if (nb < 0)
        {
            fprintf(stderr, "Error reading payload: %s\n", strerror(errno));
            rval = 1;
            break;
        }
        else if (nb != header.length)
        {
            fprintf(stderr, "Partial read of UDP payload, exiting\n");
            rval = 1;
            break;
        }

        // LLS packets arrive on 224.0.23.60 port 4937
        if (ntohl(header.ipv4addr) == 0xE000173C &&
            ntohs(header.port) == 4937)
        {
            if (ProcessUDPPacket(udpPacketNumber, pPayload, header.length, 
                &parserInfo))
            {
                printf("%"PRId64" Failed to process UDP packet\n", 
                    udpPacketNumber);
            }
        }

        udpPacketNumber++;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    if (pPayload)
    {
        free(pPayload);
    }

    if (fUDPToLLSInitialized)
    {
        if (UDPToLLSShutdown())
        {
            fprintf(stderr, "Failed to shut down UDP to LLS\n");
            rval = 1;
        }
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
