/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 * 
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <assert.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <libxml2/libxml/xmlreader.h>

#include <libenatsc3utils/enxml.h>
#include <libenatsc3utils/enipfilterset.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3sls/atsc3sls_held.h>
#include <libenatsc3mmtp/libenatsc3mmtp.h>
#include <libenatsc3mmtp/atsc3mmtpsls_usbd.h>

#include <enatsc3cap/enudpcap.h>

#define MAX_PAYLOAD_SIZE 65536

struct UserData
{
    EnIPFilterSet *pIPFilterSet;

    char *pOutPath;
    uint8_t lastMPTSVersion;
    int32_t audio_packet_id;
    int32_t video_packet_id;
    int audioFd;
    int videoFd;
};

void
USBDAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const char *pUSBD, 
    int32_t sUSBD, int32_t version, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    ATSC3MMTPSLSUSBDBundleDescriptionNode parsedUSBD;
    int32_t fHaveParsedUSBD = 0;
    char strFilename[128];
    ssize_t nb;
    int fd = -1;

    if (ENATSC3MMTPSLSUSBDParse(pUSBD, sUSBD, &parsedUSBD))
    {
        fprintf(stderr, "WARN: Failed to parse USBD\n");
    }
    else
    {
        fHaveParsedUSBD = 1;
    }

    if (snprintf(strFilename, sizeof(strFilename),
        "%s/%s-%d-usbd-%.3d", pUserData->pOutPath, pSignalingInfo->pAssetId,
        pSignalingInfo->packet_id, pSignalingInfo->version) >
        sizeof(strFilename))
    {
        fprintf(stderr, "Filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_CREAT | O_WRONLY | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pUSBD, sUSBD);
    if (nb != sUSBD)
    {
        if (nb == -1)
        {
            fprintf(stderr, "Failed to write payload of %d bytes: %s\n",
                sUSBD, strerror(errno));
        }
        else
        {
            fprintf(stderr, "Short write %zd != %d\n", nb, sUSBD);
        }
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    if (fHaveParsedUSBD)
    {
        if (ENATSC3MMTPSLSUSBDFree(&parsedUSBD))
        {
            fprintf(stderr, "Failed to free parsed USBD\n");
        }
    }

    return;
}

void
HELDAvailableCallback(struct timespec *pPacketTimeUTC,  uint64_t packetNumber,
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const char *pHELD, 
    int32_t sHELD, int32_t version, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    ENATSC3SLSHELDNode parsedHELD;
    int32_t fHaveParsedHELD = 0;
    char strFilename[128];
    ssize_t nb;
    int fd = -1;

    if (ENATSC3SLSHELDParse(pHELD, sHELD, &parsedHELD))
    {
        fprintf(stderr, "WARN: Failed to parse HELD\n");
    }
    else
    {
        fHaveParsedHELD = 1;
    }

    if (snprintf(strFilename, sizeof(strFilename),
        "%s/%s-%d-held-%.3d", pUserData->pOutPath, pSignalingInfo->pAssetId,
        pSignalingInfo->packet_id, pSignalingInfo->version) >
        sizeof(strFilename))
    {
        fprintf(stderr, "Filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_CREAT | O_WRONLY | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pHELD, sHELD);
    if (nb != sHELD)
    {
        if (nb == -1)
        {
            fprintf(stderr, "Failed to write payload of %d bytes: %s\n",
                sHELD, strerror(errno));
        }
        else
        {
            fprintf(stderr, "Short write %zd != %d\n", nb, sHELD);
        }
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    if (fHaveParsedHELD)
    {
        if (ENATSC3SLSHELDFree(&parsedHELD))
        {
            fprintf(stderr, "Failed to free parsed HELD\n");
        }
    }

    return;
}

void 
MPTSignalingMessageCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const ENATSC3MMTPMPTSignalingInfo *pSignalingInfo, 
    void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    int32_t i;
    int32_t j;

    if (pUserData->lastMPTSVersion == pSignalingInfo->message_version)
    {
        goto out;
    }

    if (((pUserData->lastMPTSVersion + 1) & 0xff) != 
        pSignalingInfo->message_version &&
        pUserData->lastMPTSVersion != -1)
    {
        fprintf(stderr, 
            "WARN: MPTS message version discrepancy last %d current %d\n",
            pUserData->lastMPTSVersion, pSignalingInfo->message_version);
    }
    pUserData->lastMPTSVersion = pSignalingInfo->message_version;


    for (i = 0; i < pSignalingInfo->nAssets; i++)
    {
        for (j = 0; j < pSignalingInfo->vAssets[i].nLocations; j++)
        {
            if (pSignalingInfo->vAssets[i].vLocations[0].location_type == 0)
            {
                break;
            }
        }

        if (j == pSignalingInfo->vAssets[i].nLocations)
        {
            fprintf(stderr, "Couldn't find packet_id for asset %d\n", i);
            continue;
        }

        printf("Asset %d: type %.4s packet_id 0x%x\n", i, 
            pSignalingInfo->vAssets[i].asset_type, 
            pSignalingInfo->vAssets[i].vLocations[j].packet_id);

        if (!strncmp((char *)pSignalingInfo->vAssets[i].asset_type, "ac-4", 4))
        {
            if (pUserData->audio_packet_id == -1)
            {
                pUserData->audio_packet_id = 
                    pSignalingInfo->vAssets[i].vLocations[j].packet_id;
            }
        }
        else if (!strncmp((char *)pSignalingInfo->vAssets[i].asset_type, 
            "hev1", 4))
        {
            if (pUserData->video_packet_id == -1)
            {
                pUserData->video_packet_id = 
                    pSignalingInfo->vAssets[i].vLocations[j].packet_id;
            }
        }
    }

out:
    return;
}

void 
MediaDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t packet_id, const char *pAssetId, int32_t fragment_type, 
    uint32_t sequence_number, uint64_t mpuPresentationTime, 
    uint32_t mpuDuration, uint64_t timeScale, const uint8_t *pData, 
    int32_t sData, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    int fd = -1;
    char strFilename[128];

    if (snprintf(strFilename, sizeof(strFilename), "%s/%s-%d-%.6d-%d", 
        pUserData->pOutPath, pAssetId, packet_id, sequence_number, 
        fragment_type) > sizeof(strFilename))
    {
        fprintf(stderr, "Filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_CREAT | O_WRONLY, 0666);
    if (fd < 0)
    {
        fprintf(stderr ,"Failed to open %s for writing: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    if (lseek(fd, 0, SEEK_END) < 0)
    {
        fprintf(stderr, "Failed to seek to end of file: %s\n", strerror(errno));
        goto out;
    }

    int nb;
    nb = write(fd, pData, sData);
    if (nb != sData)
    {
        if (nb == -1)
        {
            fprintf(stderr, "Failed to write payload of %d bytes: %s\n",
                sData, strerror(errno));
        }
        else
        {
            fprintf(stderr, "Short write %d != %d\n", nb, sData);
        }
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }
}

void
AddIPAddress(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetAddFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to add filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
RemoveIPAddress(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetRemoveFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to remove filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, char **ppOutPath,
    uint32_t *pDestinationAddr, uint16_t *pDestinationPort)
{
    int32_t err = 0;
    int32_t fHaveDestinationAddr = 0;
    int32_t fHaveDestinationPort = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-port"))
        {
            if (i + 1 < argc)
            {
                *pDestinationPort = atol(argv[i + 1]);
                fHaveDestinationPort = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-addr"))
        {
            if (i + 1 < argc)
            {
                struct in_addr addr;
                if (inet_aton(argv[i + 1], &addr))
                {
                    *pDestinationAddr = htonl(addr.s_addr);
                    fHaveDestinationAddr = 1;
                }
                else err = 1;
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (!*ppInFilename || !*ppOutPath || !fHaveDestinationAddr ||
        !fHaveDestinationPort)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b udpfilename -o outpath -addr multicastaddr -port port\n", argv[0]);
    }

    return err;
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0; 
    int32_t fXML2Initialized = 0;
    char *pInFilename;
    uint32_t destinationAddr;
    uint16_t destinationPort;
    int fd = -1;
    uint8_t *pPayload = NULL;
    MMTPSessionHandle_t hMMTPSession = NULL;
    int64_t udpPacketNumber = 0;
    struct UserData userdata;
    int rval = 0;

    memset(&userdata, 0, sizeof(userdata));
    userdata.lastMPTSVersion = -1;
    userdata.audio_packet_id = -1;
    userdata.video_packet_id = -1;

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &userdata.pOutPath, 
        &destinationAddr, &destinationPort))
    {
        rval = 1;
        goto out;
    }

    if (EnIPFilterSetAllocate(64, &userdata.pIPFilterSet))
    {
        fprintf(stderr, "Failed to allocate IP filter set\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3MMTPSessionAllocate(destinationAddr, destinationPort, 
        AddIPAddress, RemoveIPAddress, &userdata, "test", &hMMTPSession))
    {
        fprintf(stderr, "Failed to allocate MMTP session\n");
        rval = 1;
        goto out;
    }

    ENATSC3MMTPSessionRegisterCallbacks(hMMTPSession, USBDAvailableCallback, 
        HELDAvailableCallback, NULL, NULL, NULL, MPTSignalingMessageCallback, 
        MediaDataCallback, &userdata);

    fd = open(pInFilename, O_RDONLY);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", pInFilename, 
            strerror(errno));
        rval = 1;
        goto out;
    }

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    while (1)
    {
        EntangleCapHeader header;
        int32_t payloadLength;
        struct timespec packetTime;
        int nb;

        nb = read(fd, &header, sizeof(header));
        if (nb == 0)
        {
            break;
        }
        else if (nb < 0)
        {
            fprintf(stderr, "Failed to read udp file: %s\n", strerror(errno));
            rval = 1;
            break;
        }
        else if (nb != sizeof(header))
        {
            fprintf(stderr, "Partial read of UDP header, exiting\n");
            rval = 1;
            break;
        }

        payloadLength = ntohs(header.length);
        packetTime.tv_sec = ENUtils_ntohll(header.sec);
        packetTime.tv_nsec = ntohl(header.nsec);
        if (payloadLength > MAX_PAYLOAD_SIZE)
        {
            fprintf(stderr, "Payload too long: %d > %d\n",
                payloadLength, MAX_PAYLOAD_SIZE);
            rval = 1;
            break;
        }

        nb = read(fd, pPayload, payloadLength);
        if (nb < 0)
        {
            fprintf(stderr, "Error reading payload: %s\n", strerror(errno));
            rval = 1;
            break;
        }
        else if (nb != payloadLength)
        {
            fprintf(stderr, "Partial read of UDP payload, exiting\n");
            rval = 1;
            break;
        }

        if (header.ipv4addr == destinationAddr &&
            ntohs(header.port) == destinationPort)
        {
            ENATSC3MMTPSessionProcessUDPPacket(hMMTPSession, 
                &packetTime, udpPacketNumber, pPayload, payloadLength);
        }

        udpPacketNumber++;
    }

out:
    if (hMMTPSession)
    {
        if (ENATSC3MMTPSessionFree(hMMTPSession))
        {
            fprintf(stderr, "Failed to free MMTP session\n");
        }
    }

    if (userdata.pIPFilterSet)
    {
        if (EnIPFilterSetFree(userdata.pIPFilterSet))
        {
            fprintf(stderr, "Failed to free IP filter set\n");
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    if (pPayload)
    {
        free(pPayload);
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
