/*
 * Copyright © 2020-2023 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <assert.h>

#include <curl/curl.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libhdhomerun/hdhomerun.h>

#include "enatsc3cap.h"

struct UserData
{
    int ofd;

    char *pHDHR;
    char *pAddr;
    int32_t channel;
    int32_t vPLPIds[4];
    int32_t nPLPIds;
    int32_t tunerNum;

    ENATSC3CapRecordHeaderV2 header;
    uint8_t vBuffer[188 * 348];
    int32_t nBufferUsed;

    uint64_t llsValidMask;
    uint64_t plpValidMask;
    pthread_mutex_t plpMaskLock;

    uint8_t continuityCounter;
};

int32_t fRunningG = 1;

void *
PLPStatusThread(void *arg)
{
    struct UserData *pUserData = (struct UserData *)arg;
    struct hdhomerun_device_t *pHDHomerun = NULL;
    char strPath[32];
    char strHDHR[32];

    if (snprintf(strPath, sizeof(strPath), "/tuner%d/plpinfo",
        pUserData->tunerNum) >= sizeof(strPath))
    {
        fprintf(stderr, "Path too long\n");
        goto out;
    }

    if (snprintf(strHDHR, sizeof(strHDHR), "%s-%d", pUserData->pHDHR,
        pUserData->tunerNum) >= sizeof(strHDHR))
    {
        fprintf(stderr, "HDHR name too long\n");
        goto out;
    }

    while (fRunningG)
    {
        char *pPLPInfoString;
        char *pPLPInfoIter;
        char *pLLS;
        int32_t plpId;
        uint64_t plpValidMask = 0;
        uint64_t llsValidMask = 0;
        int32_t fParseError = 0;
        int hdret;

        if (!pHDHomerun)
        {
            pHDHomerun = hdhomerun_device_create_from_str(strHDHR, NULL);
            if (!pHDHomerun)
            {
                fprintf(stderr, "Failed to create HDHR device for %s\n", 
                    strHDHR);
                sleep(1);
                continue;
            }
        }

        hdret = hdhomerun_device_get_var(pHDHomerun, strPath, 
            &pPLPInfoString, NULL);

        if (hdret != 1)
        {
            fprintf(stderr, "Failed to get tuner status (%d)\n", hdret);
            pHDHomerun = NULL;
            continue;
        }

        // Example plp info:
        //  1: sfi=0 mod=qam256 cod=9/15 layer=core ti=none lls=0

        printf("plpinfo: %s\n", pPLPInfoString);
        pPLPInfoIter = pPLPInfoString;
        while (*pPLPInfoIter != '\n' && *pPLPInfoIter)
        {
            char *pPLPInfoEnd;

            pPLPInfoEnd = strchr(pPLPInfoIter, '\n');
            if (!pPLPInfoEnd)
            {
                fprintf(stderr, "Failed to parse tuner status string {%s}\n",
                    pPLPInfoIter);
                fParseError = 1;
                break;
            }

            printf("Processing %.*s\n", (int)(pPLPInfoEnd - pPLPInfoIter), pPLPInfoIter);

            plpId = atol(pPLPInfoIter);
            if (plpId < 0 || plpId > 63)
            {
                fprintf(stderr, "Invalid plp %d (%s)\n", plpId, 
                    pPLPInfoString);
            }
            else
            {
                plpValidMask |= ((int64_t)1 << plpId);
                pLLS = memmem(pPLPInfoIter, 
                    (size_t)(pPLPInfoEnd - pPLPInfoIter), "lls=", 4);
                if (!pLLS)
                {
                    fprintf(stderr, "Failed to find lls info {%s}\n",
                        pPLPInfoString);
                }
                else
                {
                    if (pLLS[4] == '0')
                    {
                        // Nothing to do
                    }
                    else if (pLLS[4] == '1')
                    {
                        llsValidMask |= ((int64_t)1 << plpId);
                    }
                    else
                    {
                        fprintf(stderr, "Unexpected lls value %s (%s)\n", 
                            pLLS, pPLPInfoString);
                    }
                }
            }

            pPLPInfoIter = pPLPInfoEnd + 1;
        }

        if (!fParseError)
        {
            printf("plp present 0x%"PRIX64" lls 0x%"PRIX64"\n", 
                plpValidMask, llsValidMask);

            if (pthread_mutex_lock(&pUserData->plpMaskLock))
            {
                fprintf(stderr, "Failed to lock\n");
            }

            pUserData->plpValidMask = plpValidMask;
            pUserData->llsValidMask = llsValidMask;

            if (pthread_mutex_unlock(&pUserData->plpMaskLock))
            {
                fprintf(stderr, "Failed to unlock\n");
            }
        }

        usleep(500000);
    }

out:
    if (pHDHomerun)
    {
        hdhomerun_device_destroy(pHDHomerun);
    }

    return NULL;
}

static size_t
curlWriteFunction(void *ptr, size_t size, size_t nmemb, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    struct timespec mononow;
    struct timespec rtnow;
    size_t nb = size * nmemb;
    uint8_t *pData = (uint8_t *)ptr;
    size_t rval = size * nmemb;

    while (fRunningG && nb)
    {
        int32_t nToCopy;

        // Initialize the header when the buffer is empty so the timestamps
        // are correct
        if (pUserData->nBufferUsed == 0)
        {
            clock_gettime(CLOCK_MONOTONIC, &mononow);
            clock_gettime(CLOCK_REALTIME, &rtnow);

            pUserData->header.type = ENATSC3CapPayloadType_ATSC3_J288ALP;
            pUserData->header.continuityCounter = 
                pUserData->continuityCounter++;
            pUserData->header.fError = 0;
            pUserData->header.plpId = 0;
            pUserData->header.length = sizeof(pUserData->vBuffer);
            pUserData->header.length = htonl(pUserData->header.length);
            pUserData->header.monowhen.sec = mononow.tv_sec;
            pUserData->header.monowhen.sec = 
                ENUtils_htonll(pUserData->header.monowhen.sec);
            pUserData->header.monowhen.nsec = mononow.tv_nsec;
            pUserData->header.monowhen.nsec = 
                htonl(pUserData->header.monowhen.nsec);
            pUserData->header.rtwhen.sec = rtnow.tv_sec;
            pUserData->header.rtwhen.sec = 
                ENUtils_htonll(pUserData->header.rtwhen.sec);
            pUserData->header.rtwhen.nsec = rtnow.tv_nsec;
            pUserData->header.rtwhen.nsec = 
                htonl(pUserData->header.rtwhen.nsec);
            pUserData->header.broadcastTime.L1B_time_info_flag = 0;

            if (pUserData->nPLPIds > 1)
            {
                pUserData->header.plpId = 255;
            }
            else
            {
                pUserData->header.plpId = pUserData->vPLPIds[0];
            }
        }

        nToCopy = MIN(nb, sizeof(pUserData->vBuffer) - pUserData->nBufferUsed);
        memcpy(pUserData->vBuffer + pUserData->nBufferUsed, pData, nToCopy);
        pData += nToCopy;
        pUserData->nBufferUsed += nToCopy;
        nb -= nToCopy;

        if (pUserData->nBufferUsed == sizeof(pUserData->vBuffer))
        {
            ENATSC3CapRecordHeaderV2 demodRecordHeader;
            ENATSC3CapDemodulatorRecordV2 demodRecord;

            // The demod record header is the same as a regular header except 
            // for the type and length.
            demodRecordHeader = pUserData->header;
            demodRecordHeader.type = ENATSC3CapPayloadType_Demod_V2;
            demodRecordHeader.length = htonl(sizeof(demodRecord));

            memset(&demodRecord, 0, sizeof(demodRecord));

            if (pthread_mutex_lock(&pUserData->plpMaskLock))
            {
                fprintf(stderr, "Failed to lock\n");
            }

            demodRecord.atsc3.llsPresentMask = pUserData->llsValidMask;
            demodRecord.atsc3.plpPresentMask = pUserData->plpValidMask;
            demodRecord.atsc3.plpEnabledMask = pUserData->plpValidMask;

            if (pthread_mutex_unlock(&pUserData->plpMaskLock))
            {
                fprintf(stderr, "Failed to unlock\n");
            }

            if (write(pUserData->ofd, &demodRecordHeader, 
                sizeof(demodRecordHeader)) != sizeof(demodRecordHeader))
            {
                fprintf(stderr, "Failed to write demod record header: %s\n", 
                    strerror(errno));
                rval = 0;
                goto out;
            }

            if (write(pUserData->ofd, &demodRecord, sizeof(demodRecord)) != 
                sizeof(demodRecord))
            {
                fprintf(stderr, "Failed to write demod record payload: %s\n",
                    strerror(errno));
                rval = 0;
                goto out;
            }

            if (write(pUserData->ofd, &pUserData->header, 
                sizeof(pUserData->header)) != sizeof(pUserData->header))
            {
                fprintf(stderr, "Failed to write record header: %s\n", strerror(errno));
                rval = 0;
                goto out;
            }

            if (write(pUserData->ofd, pUserData->vBuffer, 
                sizeof(pUserData->vBuffer)) != sizeof(pUserData->vBuffer))
            {
                fprintf(stderr, "Failed to write record payload: %s\n", 
                    strerror(errno));
                rval = 0;
                goto out;
            }

            pUserData->nBufferUsed = 0;
        }
    }

out:
    if (!fRunningG)
    {
        rval = 0;
    }

    return rval;
}

int32_t
parseArgs(int argc, char **argv, char **ppOutFilename, char **ppHDHR, 
    char **ppAddr, int32_t *pTuner, int32_t *pChannel, int32_t *pPLPIds, 
    int32_t sPLPIds, int32_t *pNPLPIds)
{
    int32_t err = 0;
    int32_t i;

    *ppOutFilename = NULL;
    *ppHDHR = NULL;
    *ppAddr = NULL;
    *pChannel = -1;
    *pTuner = -1;
    *pNPLPIds = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ch"))
        {
            if (i + 1 < argc)
            {
                *pChannel = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                if (*pNPLPIds < sPLPIds)
                {
                    *(pPLPIds++) = atol(argv[i+1]);
                    (*pNPLPIds)++;
                }
                else 
                {
                    fprintf(stderr, "Too many PLPs\n");
                    err = 1;
                }
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-hdhr"))
        {
            if (i + 1 < argc)
            {
                *ppHDHR = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-addr"))
        {
            if (i + 1 < argc)
            {
                *ppAddr = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-tuner"))
        {
            if (i + 1 < argc)
            {
                *pTuner = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }
    
    if (!*ppOutFilename || *pChannel == -1 || *ppHDHR == NULL|| *ppAddr == NULL || *pTuner == -1)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -o outfilename -hdhr deviceid -addr hdhraddr -tuner num -ch channel -plp plpId [-plp plpId ...]\n", argv[0]);
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    char *pOutFilename;
    CURL *curl = NULL;
    CURLcode res;
    char strError[CURL_ERROR_SIZE];
    char strURL[128];
    struct sigaction act;
    struct UserData userdata = 
    {
        .ofd = -1,
        .nBufferUsed = 0,
        .llsValidMask = 0,
        .plpValidMask = 0,
    };
    ENATSC3CapFileHeader fileHeader;
    pthread_t plpThread;
    int32_t fHavePLPThread = 0;
    int32_t fHavePLPMaskLock = 0;
    int32_t i;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (parseArgs(argc, argv, &pOutFilename, &userdata.pHDHR, 
        &userdata.pAddr, &userdata.tunerNum, &userdata.channel, 
        userdata.vPLPIds, numof(userdata.vPLPIds), &userdata.nPLPIds))
    {
        rval = 1;
        goto out;
    }

    if (pthread_mutex_init(&userdata.plpMaskLock, NULL))
    {
        fprintf(stderr, "Failed to init lock\n");
        rval = 1;
        goto out;
    }
    fHavePLPMaskLock = 1;

    userdata.ofd = open(pOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (userdata.ofd < 0)
    {
        fprintf(stderr, "Failed to open output file %s: %s\n", pOutFilename,
            strerror(errno));
        rval = 1;
        goto out;
    }

    fileHeader.format = ENATSC3CapFormat_V2;
    fileHeader.type = ENATSC3CapType_ATSC3_J288ALP;
    fileHeader.magic = htons(ENATSC3CAP_HEADER_MAGIC);
    if (write(userdata.ofd, &fileHeader, sizeof(fileHeader)) != 
        sizeof(fileHeader))
    {
        fprintf(stderr, "Failed to write file header: %s\n", strerror(errno));
        rval = 1;
        goto out;
    }

    if (pthread_create(&plpThread, NULL, PLPStatusThread, &userdata))
    {
        fprintf(stderr, "Failed to create plp thread\n");
        rval = 1;
        goto out;
    }
    fHavePLPThread = 1;

    curl = curl_easy_init();
    if (!curl)
    {
        fprintf(stderr, "Failed to init curl\n");
        rval = 1;
        goto out;
    }

    if (userdata.nPLPIds == 0)
    {
        userdata.vPLPIds[0] = 0;
        userdata.nPLPIds = 1;
    }

    char strPLPIds[16] = "";
    int32_t off = 0;
    for (i = 0; i < userdata.nPLPIds; i++)
    {
        int ret;

        ret = snprintf(strPLPIds + off, sizeof(strPLPIds) - off, "p%d", 
            userdata.vPLPIds[i]); 
        if (ret < 0)
        {
            fprintf(stderr, "Error adding PLP %d\n", userdata.vPLPIds[i]);
            rval = 1;
            goto out;
        }
        off += ret;
        if (off >= sizeof(strPLPIds))
        {
            fprintf(stderr, "PLP ID string too long\n");
            rval = 1;
            goto out;
        }
    }

    if (snprintf(strURL, sizeof(strURL), 
        "http://%s:5004/tuner%d/ch%d%s?duration=86640&format=j288",
        userdata.pAddr, userdata.tunerNum, userdata.channel, strPLPIds) >= 
        sizeof(strURL))
    {
        fprintf(stderr, "URL too long\n");
        rval = 1;
        goto out;
    }

    printf("url: %s\n", strURL);

    curl_easy_setopt(curl, CURLOPT_URL, strURL);
    curl_easy_setopt(curl, CURLOPT_USERAGENT, "curl/7.29.0");
    curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "");
    curl_easy_setopt(curl, CURLOPT_HTTPGET, 1L);

    curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, CURL_MAX_READ_SIZE);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curlWriteFunction);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, &userdata);
    curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, strError);

    curl_easy_setopt(curl, CURLOPT_TIMEOUT, 86400);

    res = curl_easy_perform(curl);
    if (!fRunningG)
    {
        goto out;
    }

    if (res)
    {
        fprintf(stderr, "Failed to curl %s (%d: %s: %s)\n", strURL, res, 
            curl_easy_strerror(res), strError);
        rval = 1;
        goto out;
    }

out:
    fRunningG = 0;

    if (fHavePLPThread)
    {
        if (pthread_join(plpThread, NULL))
        {
            fprintf(stderr, "Failed to join plp thread\n");
        }
    }

    if (fHavePLPMaskLock)
    {
        if (pthread_mutex_destroy(&userdata.plpMaskLock))
        {
            fprintf(stderr, "Failed to destroy lock\n");
        }
    }

    if (userdata.ofd != -1)
    {
        close(userdata.ofd);
    }

    if (curl)
    {
        curl_easy_cleanup(curl);
    }

    return rval;
}
