/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <arpa/inet.h>
#include <assert.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include "enatsc3cap.h"

#define STDIN 0

int
parseArgs(int argc, char **argv, char **ppInFile)
{
    int err = 0;
    int i;

    *ppInFile = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFile = argv[i+1];
                i++;
            }
            else err = 1;
        }
     }

    if (!*ppInFile) err = 1;

    if (err)
    {
        fprintf(stderr, "Usage: %s -b infile\n", argv[0]);
    }

    return err;
}

static inline ssize_t
ReadFromFile(int fd, void *pBuffer, size_t nToRead)
{
    uint8_t *pU8Buffer = (uint8_t *)pBuffer;
    ssize_t rval = 0;

    while (nToRead)
    {
        ssize_t nb;

        nb = read(fd, pU8Buffer, nToRead);
        if (nb == 0)
        {
            break;
        }
        else if (nb < 0)
        {
            fprintf(stderr, "Read failed: %s\n", strerror(errno));
            rval = -1;
            break;
        }

        rval += nb;
        pU8Buffer += nb;
        nToRead -= nb;
    }

    return rval;
}

#define MAX_PAYLOAD (1024 * 1024)
int
ReadFileV2(int ifd)
{
    ENATSC3CapRecordHeaderV2 header;
    ENATSC3CapDemodulatorRecordV2 *pDemodRecordV2;
    ssize_t nb;
    int64_t pos = sizeof(ENATSC3CapFileHeader);
    uint8_t *pPayload = NULL;
    int32_t continuityCounter = -1;
    uint64_t totalCCErr = 0;
    uint64_t totalRFErr = 0;
    int rval = 0;

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD);
    if (!pPayload)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    while (rval == 0)
    {
        off_t startpos = pos;
        int ccerr = 0;

	nb = ReadFromFile(ifd, &header, sizeof(header));
	if (nb == 0)
	{
	    break;
	}
	else if (nb != sizeof(header))
	{
	    fprintf(stderr, "Failed to read header: %s\n",
	        strerror(errno));
	    rval = 1;
	    break;
	}
        pos += sizeof(header);

	header.length = ntohl(header.length);
	header.monowhen.sec = ENUtils_ntohll(header.monowhen.sec);
	header.monowhen.nsec = ntohl(header.monowhen.nsec);
	header.rtwhen.sec = ENUtils_ntohll(header.rtwhen.sec);
	header.rtwhen.nsec = ntohl(header.rtwhen.nsec);

        if (header.length < 0 || header.length > MAX_PAYLOAD)
        {
            fprintf(stderr, "Unhandled header length %d\n", header.length);
            rval = 1;
            goto out;
        }

	nb = ReadFromFile(ifd, pPayload, header.length);
	if (nb == 0)
	{
	    break;
	}
	else if (nb != header.length)
	{
	    fprintf(stderr, "Failed to read payload: %s\n",
	        strerror(errno));
	    rval = 1;
	    break;
	}
        pos += header.length;

        switch (header.type)
        {
            case ENATSC3CapPayloadType_ATSC1_MPEGTS:
                printf("%"PRId64" MPEGTS: (%d bytes @ %"PRId64" rt: %"PRId64".%06ds mono: %"PRId64".%06ds\n", 
                    startpos, header.length, startpos + sizeof(header), 
                    header.rtwhen.sec, header.rtwhen.nsec / 1000, 
                    header.monowhen.sec, header.monowhen.nsec / 1000);
                break;
            case ENATSC3CapPayloadType_ATSC3_J288ALP:
                printf("%"PRId64" J288-ALP: plp %d (%d bytes %"PRId64" @ rt: %"PRId64".%06ds mono: %"PRId64".%06ds)\n", 
                    startpos, header.plpId, header.length, startpos + sizeof(header),
                    header.rtwhen.sec, header.rtwhen.nsec / 1000, 
                    header.monowhen.sec, header.monowhen.nsec / 1000);
                break;
            case ENATSC3CapPayloadType_ATSC3_ALP:
                printf("%"PRId64" ALP: plp %d (%d bytes @ %"PRId64" %"PRId64".%06ds / %"PRId64".%06ds ptp: %u %.03u %.03u %.03u\n", 
                    startpos, header.plpId, header.length, startpos + sizeof(header),
                    header.rtwhen.sec, header.rtwhen.nsec / 1000, 
                    header.monowhen.sec, header.monowhen.nsec / 1000,
                    htonl(header.broadcastTime.L1D_time_sec),
                    htonl(header.broadcastTime.L1D_time_msec),
                    htonl(header.broadcastTime.L1D_time_usec),
                    htonl(header.broadcastTime.L1D_time_nsec));
                break;
            case ENATSC3CapPayloadType_ATSC3_Baseband:
                if (continuityCounter != -1)
                {
                    if (((continuityCounter + 1) & 0xff) != header.continuityCounter)
                    {
                        ccerr = 1;
                    }
                }
                continuityCounter = header.continuityCounter;
                if (ccerr)
                {
                    totalCCErr++;
                }
                if (header.fError)
                {
                    continuityCounter = header.continuityCounter;
                    totalRFErr++;
                }
                printf("%"PRId64" baseband: plp %d cc %d ccerr %d/%"PRIu64" err %d/%"PRIu64" (%d bytes @ %"PRId64" %"PRId64".%06ds / %"PRId64".%06ds ptp: %u %u %u %u\n", 
                    startpos, header.plpId, header.continuityCounter, 
                    ccerr, totalCCErr,header.fError, totalRFErr,  
                    header.length, startpos + sizeof(header),
                    header.rtwhen.sec, header.rtwhen.nsec / 1000, 
                    header.monowhen.sec, header.monowhen.nsec / 1000,
                    ntohl(header.broadcastTime.L1D_time_sec),
                    ntohl(header.broadcastTime.L1D_time_msec),
                    ntohl(header.broadcastTime.L1D_time_usec),
                    ntohl(header.broadcastTime.L1D_time_nsec));
                break;
            case ENATSC3CapPayloadType_Demod_V2:
                printf("%"PRId64" DEMOD V2: (%d bytes @ %"PRId64" %"PRId64".%06ds / %"PRId64".%06ds\n", 
                    startpos, header.length, startpos + sizeof(header),
                    header.rtwhen.sec, header.rtwhen.nsec / 1000, 
                    header.monowhen.sec, header.monowhen.nsec / 1000);
                pDemodRecordV2 = (ENATSC3CapDemodulatorRecordV2 *)pPayload;
                printf("  freq %u plppresent %"PRIx64" llspresent %"PRIx64" plpLocked %"PRIx64"\n",
                    ntohl(pDemodRecordV2->frequencyHz),
                    ENUtils_ntohll(pDemodRecordV2->atsc3.plpPresentMask),
                    ENUtils_ntohll(pDemodRecordV2->atsc3.llsPresentMask),
                    ENUtils_ntohll(pDemodRecordV2->atsc3.plpLockedMask));
                printf("  snr: %d rssi: %d errors: %u\n",
                    ntohl(pDemodRecordV2->snr),
                    ntohl(pDemodRecordV2->rssi),
                    ntohl(pDemodRecordV2->packetErrorCount));

                {
                    struct _ATSC3_L1_Detail_Subframe *pSubframe =
                        pDemodRecordV2->atsc3.l1_detail_signaling.vSubframes;
                    int32_t i;
                    uint64_t mask;

                    for (i = 0, mask = 1; i < 64; i++, mask <<= 1)
                    {
                        if (mask & ENUtils_ntohll(pDemodRecordV2->atsc3.plpPresentMask))
                        {
                            struct _ATSC3_L1_Detail_signaling_subframe_plp *pSubframePLP =
                                pSubframe->vPLPs + i;
                            printf("  %d: l1d_plp_id %d mod %d cod %d\n", i, 
                                pSubframePLP->L1D_plp_id,
                                pSubframePLP->L1D_plp_mod,
                                pSubframePLP->L1D_plp_cod);
                        }
                    }
                }
                break;
            default:
                printf("Unhandled record type %d\n", header.type);
                break;
        }

#if 0
        assert(header.length >= 0);
        if (ifd == STDIN)
        {
            int32_t nToSeek = header.length;
            char vBuffer[16384];

            while (nToSeek)
            {
                int32_t nToRead = MIN(sizeof(vBuffer), nToSeek);
                ssize_t nb;

                nb = ReadFromFile(ifd, vBuffer, nToRead);
                if (nb == 0)
                {
                    goto out;
                }
                else if (nb < 0)
                {
                    fprintf(stderr, "Failed to read input: %s\n", 
                        strerror(errno));
                    goto out;
                }

                nToSeek -= nb;
            }
        }
        else
        {
            lseek(ifd, header.length, SEEK_CUR);
        }
        pos += header.length;
#endif
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

int
main(int argc, char **argv)
{
    char *pInFilename;
    int ifd = -1;
    ENATSC3CapFileHeader fileHeader;
    ENATSC3CapFormat format;
    ssize_t nb;
    int rval = 0;

    if (parseArgs(argc, argv, &pInFilename))
    {
        rval = 1;
	goto out;
    }

    if (!strcmp(pInFilename, "-"))
    {
        ifd = STDIN;
    }
    else
    {
        ifd = open(pInFilename, O_RDONLY);
        if (ifd < 0)
        {
            fprintf(stderr, "Failed to open %s for reading (%s)\n", 
                pInFilename, strerror(errno));
            rval = 1;
            goto out;
        }
    }

    nb = ReadFromFile(ifd, &fileHeader, sizeof(fileHeader));
    if (nb != sizeof(fileHeader))
    {
        fprintf(stderr, "Failed to read file header\n");
        rval = 1;
        goto out;
    }

    if (ntohs(fileHeader.magic) != ENATSC3CAP_HEADER_MAGIC)
    {
        fprintf(stderr, "Bad magic in header\n");
        rval = 1;
        goto out;
    }
    else
    {
        if (fileHeader.format == 2)
        {
            format = ENATSC3CapFormat_V2;
            printf("File format V2, type %d\n", fileHeader.type);
        }
        else
        {
            fprintf(stderr, "Unhandled format %d\n", fileHeader.format);
            rval = 1;
            goto out;
        }
    }

    switch (format)
    {
        case ENATSC3CapFormat_V2:
            rval = ReadFileV2(ifd);
            break;
        default:
            fprintf(stderr, "Unhandled format %d\n", format);
            rval = 1;
            goto out;
    }

out:
    if (ifd >= 0)
    {
        close(ifd);
    }

    return rval;
}
