/*
 * Copyright © 2020-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <libenatsc3utils/enxml.h>

#include "atsc3esg_contentFragment.h"

static char *ensys = "esg";
static char *ensubsys = "cfrg";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static ENStatus
ContentFragmentNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGContentFragmentNode *pContentFragmentNode = 
        (ATSC3ESGContentFragmentNode *)pUserData;
    int depth;
    char *pDocumentString;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGContentFragmentNodeType_ContentFragment);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGContentFragmentNodeAttribute_id:
            pDocumentString = memmem(pContentFragmentNode->pXML,
                pContentFragmentNode->sXML, pValue, sValue);
            if (!pDocumentString)
            {
                ERRORPRINT("Failed to find id value in document\n");
                rval = ENStatus_Failure;
                goto out;
            }
            pContentFragmentNode->idXMLOffset = pDocumentString - 
                pContentFragmentNode->pXML;
            pContentFragmentNode->idXMLLen = sValue;
            VERBOSEPRINT("%*sid: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        case ATSC3ESGContentFragmentNodeAttribute_version:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pContentFragmentNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vContentFragmentAttributes[] = 
{
    {
        .id = ATSC3ESGContentFragmentNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = ContentFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGContentFragmentNodeAttribute_id,
        .sName = sizeof("id") - 1,
        .pName = "id",
        .handler = ContentFragmentNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vContentFragmentChildNodes[] =
{
};
static EnXMLParserNode serviceFragmentNodeData = 
{
    .pName = "Content",
    .pAttributes = vContentFragmentAttributes,
    .nAttributes = numof(vContentFragmentAttributes),
    .pChildNodes = vContentFragmentChildNodes,
    .nChildNodes = numof(vContentFragmentChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessContentFragmentNode(xmlTextReaderPtr reader, 
    ATSC3ESGContentFragmentNode *pContentFragmentNode)
{
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*sContent\n", 
        xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&serviceFragmentNodeData, reader, 
        pContentFragmentNode, (EnXMLParserNodeBase *)pContentFragmentNode);

    return rval;
}

ENStatus
ATSC3ESGContentFragmentParse(const char *pXML, int32_t sXML, 
    ATSC3ESGContentFragmentNode *pParsedContentFragment)
{
    xmlTextReaderPtr reader = NULL;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedContentFragment, 0, sizeof(ATSC3ESGContentFragmentNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pParsedContentFragment->pXML = (char *)pXML;
    pParsedContentFragment->sXML = sXML;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "Content"))
        {
            pParsedContentFragment->base.nodeType = 
                ATSC3ESGContentFragmentNodeType_ContentFragment;
            ProcessContentFragmentNode(reader, pParsedContentFragment);
        }
        else
        {
            ERRORPRINT("Top-level object isn't Content (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse ContentFragment XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    goto out;
}

ENStatus
ATSC3ESGContentFragmentFree(ATSC3ESGContentFragmentNode *pParsedContentFragment)
{
    ENStatus rval = ENStatus_Success;

    return rval;
}

