/*
 * Copyright © 2020-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <libenatsc3utils/enxml.h>

#include "atsc3esg_scheduleFragment.h"

static char *ensys = "esg";
static char *ensubsys = "schfrg";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static ENStatus
PresentationWindowNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentPresentationWindowNode *pPresentationWindowNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_PresentationWindow);
    pPresentationWindowNode = 
        (ATSC3ESGScheduleFragmentPresentationWindowNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGScheduleFragmentPresentationWindowNodeAttribute_startTime:
            if (EnXMLParser_antoll((char *)pValue, sValue,
                &pPresentationWindowNode->startTime))
            {
                ERRORPRINT("Failed to parse startTime %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }

            VERBOSEPRINT("%*sstartTime: %.*s\n", depth * 2,  "", sValue, 
                pValue);
            break;
        case ATSC3ESGScheduleFragmentPresentationWindowNodeAttribute_endTime:
            if (EnXMLParser_antoll((char *)pValue, sValue,
                &pPresentationWindowNode->endTime))
            {
                ERRORPRINT("Failed to parse endTime %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }

            VERBOSEPRINT("%*sendTime: %.*s\n", depth * 2,  "", sValue, 
                pValue);
            break;
        case ATSC3ESGScheduleFragmentPresentationWindowNodeAttribute_duration:
            if (EnXMLParser_antol((char *)pValue, sValue,
                &pPresentationWindowNode->duration))
            {
                ERRORPRINT("Failed to parse duration %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }

            VERBOSEPRINT("%*sduration: %.*s\n", depth * 2,  "", sValue, 
                pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vPresentationWindowAttributes[] = 
{
    {
        .id = ATSC3ESGScheduleFragmentPresentationWindowNodeAttribute_startTime,
        .sName = sizeof("startTime") - 1,
        .pName = "startTime",
        .handler = PresentationWindowNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGScheduleFragmentPresentationWindowNodeAttribute_endTime,
        .sName = sizeof("endTime") - 1,
        .pName = "endTime",
        .handler = PresentationWindowNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGScheduleFragmentPresentationWindowNodeAttribute_duration,
        .sName = sizeof("duration") - 1,
        .pName = "duration",
        .handler = PresentationWindowNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vPresentationWindowChildNodes[] =
{
};
static EnXMLParserNode presentationWindowNodeData = 
{
    .pName = "PresentationWindow",
    .pAttributes = vPresentationWindowAttributes,
    .nAttributes = numof(vPresentationWindowAttributes),
    .pChildNodes = vPresentationWindowChildNodes,
    .nChildNodes = numof(vPresentationWindowChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessPresentationWindowNode(EnXMLParserChildNode *pChildNode,
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentContentReferenceNode *pContentReferenceNode;
    ATSC3ESGScheduleFragmentPresentationWindowNode *pPresentationWindowNode;
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*sPresentationWindow\n", 
        xmlTextReaderDepth(reader) * 2, "");

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_ContentReference);
    pContentReferenceNode = 
        (ATSC3ESGScheduleFragmentContentReferenceNode *)pParentNode;
    pPresentationWindowNode = &pContentReferenceNode->presentationWindow;
    memset(pPresentationWindowNode, 0, sizeof(*pPresentationWindowNode));

    pPresentationWindowNode->base.nodeType = 
        ATSC3ESGScheduleFragmentNodeType_PresentationWindow;

    rval = EnXMLParserProcessNode(&presentationWindowNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pPresentationWindowNode);

    return rval;
}

static ENStatus
ContentReferenceNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentNode *pScheduleFragmentNode = 
        (ATSC3ESGScheduleFragmentNode *)pUserData;
    ATSC3ESGScheduleFragmentContentReferenceNode *pContentReferenceNode;
    int depth;
    char *pDocumentString;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_ContentReference);
    pContentReferenceNode = 
        (ATSC3ESGScheduleFragmentContentReferenceNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGScheduleFragmentContentReferenceNodeAttribute_idRef:
            pDocumentString = memmem(pScheduleFragmentNode->pXML,
                pScheduleFragmentNode->sXML, pValue, sValue);
            if (!pDocumentString)
            {
                ERRORPRINT("Failed to find id value in document\n");
                rval = ENStatus_Failure;
                goto out;
            }
            pContentReferenceNode->idRefXMLOffset = pDocumentString - 
                pScheduleFragmentNode->pXML;
            pContentReferenceNode->idRefXMLLen = sValue;
            VERBOSEPRINT("%*sidRef: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vContentReferenceAttributes[] = 
{
    {
        .id = ATSC3ESGScheduleFragmentContentReferenceNodeAttribute_idRef,
        .sName = sizeof("idRef") - 1,
        .pName = "idRef",
        .handler = ContentReferenceNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vContentReferenceChildNodes[] =
{
    {
        .id = ATSC3ESGScheduleFragmentContentReferenceChildNode_PresentationWindow,
        .sName = sizeof("PresentationWindow") - 1,
        .pName = "PresentationWindow",
        .handler = ProcessPresentationWindowNode,
    },
};
static EnXMLParserNode contentReferenceNodeData = 
{
    .pName = "ContentReference",
    .pAttributes = vContentReferenceAttributes,
    .nAttributes = numof(vContentReferenceAttributes),
    .pChildNodes = vContentReferenceChildNodes,
    .nChildNodes = numof(vContentReferenceChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessContentReferenceNode(EnXMLParserChildNode *pChildNode,
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentNode *pScheduleFragmentNode;
    ATSC3ESGScheduleFragmentContentReferenceNode contentReferenceNode;
    int32_t contentReferenceNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_ScheduleFragment);
    pScheduleFragmentNode = (ATSC3ESGScheduleFragmentNode *)pParentNode;
    memset(&contentReferenceNode, 0, sizeof(contentReferenceNode));

    // Allocate space for the node from the data buffer and store the offset
    // in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pScheduleFragmentNode->dataBuffer,
        sizeof(contentReferenceNode), &contentReferenceNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for contentReference node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pScheduleFragmentNode->contentReferences, contentReferenceNodeOffset))
    {
        ERRORPRINT("Failed to store contentReference node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    contentReferenceNode.base.nodeType = 
        ATSC3ESGScheduleFragmentNodeType_ContentReference;

    VERBOSEPRINT("%*sContentReference\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&contentReferenceNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&contentReferenceNode);

    memcpy(pScheduleFragmentNode->dataBuffer.pBuffer + 
        contentReferenceNodeOffset, &contentReferenceNode, 
        sizeof(contentReferenceNode));

out:
    return rval;
}

static ENStatus
ServiceReferenceNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentNode *pScheduleFragmentNode = 
        (ATSC3ESGScheduleFragmentNode *)pUserData;
    ATSC3ESGScheduleFragmentServiceReferenceNode *pServiceReferenceNode;
    int depth;
    char *pDocumentString;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_ServiceReference);
    pServiceReferenceNode = 
        (ATSC3ESGScheduleFragmentServiceReferenceNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGScheduleFragmentServiceReferenceNodeAttribute_idRef:
            pDocumentString = memmem(pScheduleFragmentNode->pXML,
                pScheduleFragmentNode->sXML, pValue, sValue);
            if (!pDocumentString)
            {
                ERRORPRINT("Failed to find id value in document\n");
                rval = ENStatus_Failure;
                goto out;
            }
            pServiceReferenceNode->idRefXMLOffset = pDocumentString - 
                pScheduleFragmentNode->pXML;
            pServiceReferenceNode->idRefXMLLen = sValue;
            VERBOSEPRINT("%*sidRef: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vServiceReferenceAttributes[] = 
{
    {
        .id = ATSC3ESGScheduleFragmentServiceReferenceNodeAttribute_idRef,
        .sName = sizeof("idRef") - 1,
        .pName = "idRef",
        .handler = ServiceReferenceNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vServiceReferenceChildNodes[] =
{
};
static EnXMLParserNode serviceReferenceNodeData = 
{
    .pName = "ServiceReference",
    .pAttributes = vServiceReferenceAttributes,
    .nAttributes = numof(vServiceReferenceAttributes),
    .pChildNodes = vServiceReferenceChildNodes,
    .nChildNodes = numof(vServiceReferenceChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessServiceReferenceNode(EnXMLParserChildNode *pChildNode,
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentNode *pScheduleFragmentNode;
    ATSC3ESGScheduleFragmentServiceReferenceNode *pServiceReferenceNode;
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*sServiceReference\n", 
        xmlTextReaderDepth(reader) * 2, "");

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_ScheduleFragment);
    pScheduleFragmentNode = (ATSC3ESGScheduleFragmentNode *)pParentNode;
    pServiceReferenceNode = &pScheduleFragmentNode->serviceReference;
    memset(pServiceReferenceNode, 0, sizeof(*pServiceReferenceNode));

    pServiceReferenceNode->base.nodeType = 
        ATSC3ESGScheduleFragmentNodeType_ServiceReference;

    rval = EnXMLParserProcessNode(&serviceReferenceNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)pServiceReferenceNode);

    return rval;
}

static ENStatus
ScheduleFragmentNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGScheduleFragmentNode *pScheduleFragmentNode = 
        (ATSC3ESGScheduleFragmentNode *)pUserData;
    int depth;
    char *pDocumentString;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGScheduleFragmentNodeType_ScheduleFragment);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGScheduleFragmentNodeAttribute_id:
            pDocumentString = memmem(pScheduleFragmentNode->pXML,
                pScheduleFragmentNode->sXML, pValue, sValue);
            if (!pDocumentString)
            {
                ERRORPRINT("Failed to find id value in document\n");
                rval = ENStatus_Failure;
                goto out;
            }
            pScheduleFragmentNode->idXMLOffset = pDocumentString - 
                pScheduleFragmentNode->pXML;
            pScheduleFragmentNode->idXMLLen = sValue;
            VERBOSEPRINT("%*sid: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        case ATSC3ESGScheduleFragmentNodeAttribute_version:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pScheduleFragmentNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vScheduleFragmentAttributes[] = 
{
    {
        .id = ATSC3ESGScheduleFragmentNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = ScheduleFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGScheduleFragmentNodeAttribute_id,
        .sName = sizeof("id") - 1,
        .pName = "id",
        .handler = ScheduleFragmentNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vScheduleFragmentChildNodes[] =
{
    {
        .id = ATSC3ESGScheduleFragmentChildNode_ServiceReference,
        .sName = sizeof("ServiceReference") - 1,
        .pName = "ServiceReference",
        .handler = ProcessServiceReferenceNode,
    },
    {
        .id = ATSC3ESGScheduleFragmentChildNode_ContentReference,
        .sName = sizeof("ContentReference") - 1,
        .pName = "ContentReference",
        .handler = ProcessContentReferenceNode,
    },
};
static EnXMLParserNode scheduleFragmentNodeData = 
{
    .pName = "Schedule",
    .pAttributes = vScheduleFragmentAttributes,
    .nAttributes = numof(vScheduleFragmentAttributes),
    .pChildNodes = vScheduleFragmentChildNodes,
    .nChildNodes = numof(vScheduleFragmentChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessScheduleFragmentNode(xmlTextReaderPtr reader, 
    ATSC3ESGScheduleFragmentNode *pScheduleFragmentNode)
{
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*sSchedule\n", 
        xmlTextReaderDepth(reader) * 2, "");

    if (EnXMLParserDataOffsetArrayInitialize(
        &pScheduleFragmentNode->contentReferences,
        &pScheduleFragmentNode->dataBuffer, 128, 128))
    {
        ERRORPRINT("Failed to allocate contentReferences array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&scheduleFragmentNodeData, reader, 
        pScheduleFragmentNode, (EnXMLParserNodeBase *)pScheduleFragmentNode);

out:
    return rval;
}

ENStatus
ATSC3ESGScheduleFragmentParse(const char *pXML, int32_t sXML, 
    ATSC3ESGScheduleFragmentNode *pParsedScheduleFragment)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedScheduleFragment, 0, sizeof(ATSC3ESGScheduleFragmentNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pParsedScheduleFragment->pXML = (char *)pXML;
    pParsedScheduleFragment->sXML = sXML;

    if (EnXMLParserDataBufferInitialize(&pParsedScheduleFragment->dataBuffer, 
        16384, 16384, ensubsys))
    {
        ERRORPRINT("Failed to initialize ScheduleFragment data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "Schedule"))
        {
            pParsedScheduleFragment->base.nodeType = 
                ATSC3ESGScheduleFragmentNodeType_ScheduleFragment;
            ProcessScheduleFragmentNode(reader, pParsedScheduleFragment);
        }
        else
        {
            ERRORPRINT("Top-level object isn't Schedule (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse ScheduleFragment XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedScheduleFragment->dataBuffer))
        {
            ERRORPRINT("Failed to free ScheduleFragment data buffer\n");
        }
    }

    goto out;
}

ENStatus
ATSC3ESGScheduleFragmentFree(
    ATSC3ESGScheduleFragmentNode *pParsedScheduleFragment)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedScheduleFragment->dataBuffer))
    {
        ERRORPRINT("Failed to free ScheduleFragment data buffer\n");
    }

    return rval;
}

