/*
 * Copyright © 2020-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>

#include <libenatsc3utils/enxml.h>

#include "atsc3esg_serviceFragment.h"

static char *ensys = "esg";
static char *ensubsys = "svcfrg";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static ENStatus
ServiceFragmentNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGServiceFragmentNode *pServiceFragmentNode = 
        (ATSC3ESGServiceFragmentNode *)pUserData;
    int depth;
    char *pDocumentString;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3ESGServiceFragmentNodeType_ServiceFragment);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGServiceFragmentNodeAttribute_id:
            pDocumentString = memmem(pServiceFragmentNode->pXML,
                pServiceFragmentNode->sXML, pValue, sValue);
            if (!pDocumentString)
            {
                ERRORPRINT("Failed to find id value in document\n");
                rval = ENStatus_Failure;
                goto out;
            }
            pServiceFragmentNode->idXMLOffset = pDocumentString - 
                pServiceFragmentNode->pXML;
            pServiceFragmentNode->idXMLLen = sValue;
            VERBOSEPRINT("%*sid: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        case ATSC3ESGServiceFragmentNodeAttribute_globalServiceID:
            pDocumentString = memmem(pServiceFragmentNode->pXML,
                pServiceFragmentNode->sXML, pValue, sValue);
            if (!pDocumentString)
            {
                ERRORPRINT("Failed to find globalServiceID value in document\n");
                rval = ENStatus_Failure;
                goto out;
            }
            pServiceFragmentNode->globalServiceIDXMLOffset = pDocumentString - 
                pServiceFragmentNode->pXML;
            pServiceFragmentNode->globalServiceIDXMLLen = sValue;
            VERBOSEPRINT("%*sglobalServiceID: %.*s\n", depth * 2,  "", sValue, 
                pValue);
            break;
        case ATSC3ESGServiceFragmentNodeAttribute_version:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pServiceFragmentNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vServiceFragmentAttributes[] = 
{
    {
        .id = ATSC3ESGServiceFragmentNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = ServiceFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGServiceFragmentNodeAttribute_id,
        .sName = sizeof("id") - 1,
        .pName = "id",
        .handler = ServiceFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGServiceFragmentNodeAttribute_globalServiceID,
        .sName = sizeof("globalServiceID") - 1,
        .pName = "globalServiceID",
        .handler = ServiceFragmentNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vServiceFragmentChildNodes[] =
{
};
static EnXMLParserNode serviceFragmentNodeData = 
{
    .pName = "Service",
    .pAttributes = vServiceFragmentAttributes,
    .nAttributes = numof(vServiceFragmentAttributes),
    .pChildNodes = vServiceFragmentChildNodes,
    .nChildNodes = numof(vServiceFragmentChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessServiceFragmentNode(xmlTextReaderPtr reader, 
    ATSC3ESGServiceFragmentNode *pServiceFragmentNode)
{
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*sService\n", 
        xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&serviceFragmentNodeData, reader, 
        pServiceFragmentNode, (EnXMLParserNodeBase *)pServiceFragmentNode);

    return rval;
}

ENStatus
ATSC3ESGServiceFragmentParse(const char *pXML, int32_t sXML, 
    ATSC3ESGServiceFragmentNode *pParsedServiceFragment)
{
    xmlTextReaderPtr reader = NULL;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedServiceFragment, 0, sizeof(ATSC3ESGServiceFragmentNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pParsedServiceFragment->pXML = (char *)pXML;
    pParsedServiceFragment->sXML = sXML;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "Service"))
        {
            pParsedServiceFragment->base.nodeType = 
                ATSC3ESGServiceFragmentNodeType_ServiceFragment;
            ProcessServiceFragmentNode(reader, pParsedServiceFragment);
        }
        else
        {
            ERRORPRINT("Top-level object isn't Service (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse ServiceFragment XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    goto out;
}

ENStatus
ATSC3ESGServiceFragmentFree(ATSC3ESGServiceFragmentNode *pParsedServiceFragment)
{
    ENStatus rval = ENStatus_Success;

    return rval;
}

