/*
 * Copyright © 2020-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <zlib.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/libenatsc3route.h>
#include <libenatsc3route/atsc3routesls_stsid.h>

#include "libenatsc3esg.h"

#include "atsc3esg_serviceFragment.h"
#include "atsc3esg_scheduleFragment.h"
#include "atsc3esg_contentFragment.h"
#include "atsc3esg_sggd.h"

static char *ensys = "esg";
static char *ensubsys = "main";

#define ESG_XML_MAX_LENGTH (1024 * 1024)

// libenatsc3esg processes ESG from a ROUTE session supplied by the 
// application. It is assumed that ESG processing is an activity which occurs
// alongside other activities such as receiving an audiovisual service. As a
// result libenatsc3esg does not attempt to manage reception of the broadcast
// below the ROUTE level (i.e. baseband, ALP, lls processing). This is left
// to the application to manage on behalf of all services which are being
// receievd.
//
// An application interested in ESG creates a ROUTE session bound to the ESG
// service, which the application discovers in the same way it discovers 
// other services (typically via the SLT). The application manages changes
// to the ROUTE session, such as adding/removing IP address/ports corresponding
// to LCT channels. libenatsc3esg deals with the output of the ROUTE session,
// namely the STSID and files delivered over the session. 
//
// In the event that the ROUTE session is destroyed (e.g. as a result of a
// a reconfiguration of the broadcast), then the ESG session must be freed 
// before the ROUTE session is destroyed. A new ESG session may be created 
// based on a new ROUTE session bound to the new ESG service configuration.

#define REP_ID_MAX_LENGTH 256

typedef struct _ESGContentFragment
{
    uint32_t receiveSessionId;
    uint32_t tsi;
    uint32_t toi;

    uint32_t fragmentTransportID;
    uint32_t fragmentVersion;

    uint32_t idXMLOffset;
    char *pId;
    int32_t idLen;

    uint32_t version;

    struct _ESGContentFragment *pNext;

    int32_t xmlLen;
    char vXML[0];
} ESGContentFragment;

typedef struct _ESGScheduleFragment
{
    uint32_t receiveSessionId;
    uint32_t tsi;
    uint32_t toi;

    uint32_t fragmentTransportID;
    uint32_t fragmentVersion;

    ATSC3ESGScheduleFragmentNode parsedScheduleFragment;

    struct _ESGScheduleFragment *pNext;

    int32_t xmlLen;
    char vXML[0];
} ESGScheduleFragment;

typedef struct _ESGServiceFragment
{
    uint32_t receiveSessionId;
    uint32_t tsi;
    uint32_t toi;

    uint32_t fragmentTransportID;
    uint32_t fragmentVersion;

    uint32_t idXMLOffset;
    char *pId;
    int32_t idLen;

    uint32_t version;

    uint32_t globalServiceIDXMLOffset;
    char *pGlobalServiceID;
    int32_t globalServiceIDLen;

    struct _ESGServiceFragment *pNext;

    int32_t xmlLen;
    char vXML[0];
} ESGServiceFragment;

typedef struct
{
    uint32_t extension_offset;
    uint16_t reserved;
    uint8_t n_o_service_guide_fragments_byte0;
    uint8_t n_o_service_guide_fragments_byte1;
    uint8_t n_o_service_guide_fragments_byte2;
    struct _serviceGuideFragmentEntry
    {
        uint32_t fragmentTransportID;
        uint32_t fragmentVersion;
        uint32_t offset;
    } __attribute__((__packed__)) vServiceGuideFragmentEntries[0];
} __attribute__((__packed__)) SGDUUnitHeader;

typedef struct
{
    uint8_t fragmentEncoding;
    union
    {
        struct _XMLFragment // encoding = 0
        {
            uint8_t fragmentType;
            uint8_t vXMLFragment[0];
        } __attribute__((__packed__)) XMLFragment;
        struct _SPDFragment // encoding = 1
        {
            uint32_t validFrom;
            uint32_t validTo;
            uint8_t vFragmentInfo[0];
        } __attribute__((__packed__)) SPDFragment;
        struct _USBDFragment // encoding = 2
        {
            uint32_t validFrom;
            uint32_t validTo;
            uint8_t vFragmentInfo[0];
        } __attribute__((__packed__)) USBDFragment;
        struct _APDFragment // encoding = 3
        {
            uint32_t validFrom;
            uint32_t validTo;
            uint8_t vFragmentInfo[0];
        } __attribute__((__packed__)) APDFragment;
    } __attribute__((__packed__));
} __attribute__((__packed__)) SGDUUnitPayload;

typedef struct _SGDFileDescription
{
    uint32_t tsi;
    uint32_t toi;
    int64_t transferLength;
    int64_t contentLength;
    int32_t fReceived;

    union
    {
        ATSC3ESGSGDDNode parsedSGDD;
    };
    int32_t fHaveParsedXML;

    struct _SGDFileDescription *pNext;
} SGDFileDescription;

typedef struct _ENATSC3ESGSessionHandle_t
{
    ENVerbosity verbosity;
    char strDebugName[16];

    int32_t fIgnoreChecksumErrors;

    // Mask of the PLPs currently being received for service reception.
    int64_t esgPLPs;

    int32_t vSLTVersion[256];
    int32_t vSLTLength[256];
    uint8_t *vpSLT[256];

    ENATSC3ROUTESessionHandle_t hROUTESession;

    struct _LCTChannelInfo
    {
        int32_t fInUse;

        uint32_t tsi;
        char strId[REP_ID_MAX_LENGTH];
        ENATSC3ROUTESessionIdType idType;


        // The files we are receiving. The number of receivers indicates the
        // number of files which can be simultaneously received on an LCT
        // channel.
    } vLCTChannelInfos[32];

    struct _FileReceiver
    {
        int32_t fInUse;

        struct _LCTChannelInfo *pLCTChannelInfo;

        uint32_t toi;

        uint8_t codepoint;
        char strURL[256];
        int64_t objectSize;
        int64_t pos;

        int64_t transferLength;
        int64_t contentLength;
        int64_t nBytesReceived;
        uint8_t *pTransferBuffer;
        int32_t sTransferBuffer;

        // When receipt of this file started. Used to determine is receipt
        // times out.
        struct timespec tAllocated;
    } vFileReceivers[4];

    // The version of the last STSID processed
    int32_t stsidVersion;
    uint32_t stsidCRC32;


    // The receiveSessionId identifies the receive session that a fragment
    // was received in. A new receive session is started whenever the STSID
    // indicates the set of SGDDs has changed. Once all SGDDs and SGDUs have
    // been processed, any fragments marked with an older receiveSessionId
    // are no longer valid and can be freed.
    uint32_t receiveSessionId;
    int32_t fReceiveSessionComplete;

    // The SGDD and SGDUs from the STSID
    SGDFileDescription *pSGDDs;
    SGDFileDescription *pSGDUs;

    // ESG items
    ESGServiceFragment *pServices;
    ESGScheduleFragment *pSchedules;
    ESGContentFragment *pContents;
    
    // Callbacks
    ENATSC3ESGSessionServiceFragmentCallback ServiceFragmentCallback;
    ENATSC3ESGSessionScheduleFragmentCallback ScheduleFragmentCallback;
    ENATSC3ESGSessionContentFragmentCallback ContentFragmentCallback;
    void *pCallbackUserData;

    pthread_mutex_t esgFragmentsLock;
} ENATSC3ESGSession, *ENATSC3ESGSessionHandle_t;

// Forward declarations
void
MediaFileAbortCallback(ENATSC3ESGSession *pSession,
    struct timespec *pPacketTimeUTC, struct _FileReceiver *pReceiver);


static ENStatus
FindLCTChannelForTSI(ENATSC3ESGSession *pSession, uint32_t tsi,
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pSession->vLCTChannelInfos); i++)
    {
        if (pSession->vLCTChannelInfos[i].fInUse &&
            pSession->vLCTChannelInfos[i].tsi == tsi)
        {
            *ppLCTChannelInfo = pSession->vLCTChannelInfos + i;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

static ENStatus
AddLCTChannelForTSI(ENATSC3ESGSession *pSession, uint32_t tsi,
    const char *pId, ENATSC3ROUTESessionIdType idType,
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    if (strlen(pId) >= sizeof(pSession->vLCTChannelInfos[0].strId))
    {
        ERRORPRINT("%s: id %s too long\n", pSession->strDebugName, pId);
        goto out;
    }

    for (i = 0; i < numof(pSession->vLCTChannelInfos); i++)
    {
        if (!pSession->vLCTChannelInfos[i].fInUse)
        {
            pSession->vLCTChannelInfos[i].fInUse = 1;
            pSession->vLCTChannelInfos[i].tsi = tsi;
            strcpy(pSession->vLCTChannelInfos[i].strId, pId);
            pSession->vLCTChannelInfos[i].idType = idType;

            *ppLCTChannelInfo = pSession->vLCTChannelInfos + i;
            rval = ENStatus_Success;
            break;
        }
    }

out:
    return rval;
}

static ENStatus
RemoveLCTChannelForTSI(ENATSC3ESGSession *pSession, uint32_t tsi)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pSession->vFileReceivers); i++)
    {
        if (pSession->vFileReceivers[i].fInUse &&
            pSession->vFileReceivers[i].pLCTChannelInfo->tsi == tsi)
        {
            struct timespec now = {.tv_sec = 0, .tv_nsec = 0};
            MediaFileAbortCallback(pSession, &now, 
                pSession->vFileReceivers + i); 
            pSession->vFileReceivers[i].fInUse = 0;
            pSession->vFileReceivers[i].pLCTChannelInfo = NULL;
        }
    }

    for (i = 0; i < numof(pSession->vLCTChannelInfos); i++)
    {
        if (pSession->vLCTChannelInfos[i].fInUse &&
            pSession->vLCTChannelInfos[i].tsi == tsi)
        {
            pSession->vLCTChannelInfos[i].fInUse = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

void
LCTChannelAddedCallback(struct timespec *pPacketTime, uint64_t packetNumber, 
    uint32_t addr, uint16_t port, uint32_t tsi, const char *pId, 
    ENATSC3ROUTESessionIdType idType, void *pCbUserData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;

    if (AddLCTChannelForTSI(pSession, tsi, pId, idType, &pLCTChannelInfo))
    {
        ERRORPRINT("%s: Failed to allocate LCT Channel info for tsi %u\n", 
            pSession->strDebugName, tsi);
    }
}

void
LCTChannelRemovedCallback(struct timespec *pPacketTime, uint64_t packetNumber, 
    uint32_t addr, uint16_t port, uint32_t tsi, void *pCbUserData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)pCbUserData;

    if (RemoveLCTChannelForTSI(pSession, tsi))
    {
        ERRORPRINT("s: Failed to free LCT Channel info for tsi %u\n", 
            pSession->strDebugName, tsi);
    }
}

ENStatus
SGDFileDescriptionAllocate(ENATSC3ESGSession *pSession, 
    SGDFileDescription **ppSGDU)
{
    SGDFileDescription *pSGDU;
    ENStatus rval = ENStatus_Success;

    pSGDU = (SGDFileDescription *)calloc(1, sizeof(SGDFileDescription));
    if (!pSGDU)
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    *ppSGDU = pSGDU;

out:
    return rval;
}
void
SGDFileDescriptionFree(SGDFileDescription *pSGDU)
{
    free(pSGDU);
}

ENStatus
ContentFragmentAllocate(ENATSC3ESGSession *pSession, char *pXML, int32_t xmlLen,
    ESGContentFragment **ppContentFragment)
{
    ESGContentFragment *pNew;
    ENStatus rval = ENStatus_Success;

    if (sizeof(ESGContentFragment) + xmlLen > ESG_XML_MAX_LENGTH)
    {
        ERRORPRINT("Content fragment too large (%d > %d)\n",
            sizeof(ESGContentFragment) + xmlLen, ESG_XML_MAX_LENGTH);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew = (ESGContentFragment *)malloc(sizeof(ESGContentFragment) + xmlLen);
    if (!pNew)
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    *ppContentFragment = pNew;

out:
    return rval;
}

void
ContentFragmentFree(ENATSC3ESGSession *pSession, 
    ESGContentFragment *pContentFragment)
{
    free(pContentFragment);
}

ENStatus
AddContentFragment(ENATSC3ESGSession *pSession, uint32_t tsi, uint32_t toi,
    uint32_t fragmentTransportID, uint32_t fragmentVersion, char *pFragment,
    int32_t fragmentLen)
{
    ATSC3ESGContentFragmentNode contentFragmentNode;
    ESGContentFragment *pIter;
    ESGContentFragment *pPrev = NULL;
    ESGContentFragment *pNew;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    for (pIter = pSession->pContents;
         pIter;
         pPrev = pIter, pIter = pIter->pNext)
    {
        if (pIter->fragmentTransportID == fragmentTransportID)
        {
            break;
        }
    }

    if (pIter)
    {
        // We already have this or a newer version of this fragment
        if (pIter->fragmentVersion >= fragmentVersion)
        {
            pIter->receiveSessionId = pSession->receiveSessionId;
            goto out;
        }

        // We need to update - remove the current version
        if (pPrev)
        {
            pPrev->pNext = pIter->pNext;
        }
        else
        {
            pSession->pContents = pSession->pContents->pNext;
        }
        ContentFragmentFree(pSession, pIter);
    }

    if (ContentFragmentAllocate(pSession, pFragment, fragmentLen, &pNew))
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew->tsi = tsi;
    pNew->toi = toi;
    pNew->receiveSessionId = pSession->receiveSessionId;
    pNew->fragmentTransportID = fragmentTransportID;
    pNew->fragmentVersion = fragmentVersion;
    pNew->xmlLen = fragmentLen;
    memcpy(pNew->vXML, pFragment, fragmentLen);

    if (ATSC3ESGContentFragmentParse(pFragment, fragmentLen,
        &contentFragmentNode))
    {
        ERRORPRINT("%s: Failed to parse content fragment\n", 
            pSession->strDebugName);
        ContentFragmentFree(pSession, pNew);
        rval = ENStatus_Failure;
        goto out;
    }

    if (!(contentFragmentNode.base.presentFlags & 
        ATSC3ESGContentFragmentNodeAttribute_id))
    {
        ERRORPRINT("Content fragment lacks id\n");
        ContentFragmentFree(pSession, pNew);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew->idXMLOffset = contentFragmentNode.idXMLOffset;
    pNew->pId = pNew->vXML + contentFragmentNode.idXMLOffset;
    pNew->idLen = contentFragmentNode.idXMLLen;

    if (!(contentFragmentNode.base.presentFlags & 
        ATSC3ESGContentFragmentNodeAttribute_version))
    {
        ERRORPRINT("Content fragment lacks version\n");
        pNew->version = 0;
    }
    else
    {
        pNew->version = contentFragmentNode.version;
    }

    if (pSession->verbosity >= ENVerbosity_ObjectTrace)
    {
        LOGPRINT_NOFILE(
            "%s: Content Fragment: transportID %d id %.*s version %d\n",
            pSession->strDebugName, fragmentTransportID,
            contentFragmentNode.idXMLLen,
            pFragment + contentFragmentNode.idXMLOffset,
            contentFragmentNode.version);
    }

    if (ATSC3ESGContentFragmentFree(&contentFragmentNode))
    {
        ERRORPRINT("%s: Failed to free content fragment\n",
            pSession->strDebugName);
    }

    pNew->pNext = pSession->pContents;
    pSession->pContents = pNew;

out:
    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }

    return rval;
}

ENStatus
ServiceFragmentAllocate(ENATSC3ESGSession *pSession, char *pXML, int32_t xmlLen,
    ESGServiceFragment **ppServiceFragment)
{
    ESGServiceFragment *pNew;
    ENStatus rval = ENStatus_Success;

    if (sizeof(ESGServiceFragment) + xmlLen > ESG_XML_MAX_LENGTH)
    {
        ERRORPRINT("Service fragment too large (%d > %d)\n",
            sizeof(ESGServiceFragment) + xmlLen, ESG_XML_MAX_LENGTH);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew = (ESGServiceFragment *)malloc(sizeof(ESGServiceFragment) + xmlLen);
    if (!pNew)
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    *ppServiceFragment = pNew;

out:
    return rval;
}

void
ServiceFragmentFree(ENATSC3ESGSession *pSession, 
    ESGServiceFragment *pServiceFragment)
{
    free(pServiceFragment);
}

ENStatus
AddServiceFragment(ENATSC3ESGSession *pSession, uint32_t tsi, uint32_t toi,
    uint32_t fragmentTransportID, uint32_t fragmentVersion, char *pFragment,
    int32_t fragmentLen)
{
    ATSC3ESGServiceFragmentNode serviceFragmentNode;
    ESGServiceFragment *pIter;
    ESGServiceFragment *pPrev = NULL;
    ESGServiceFragment *pNew;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    for (pIter = pSession->pServices;
         pIter;
         pPrev = pIter, pIter = pIter->pNext)
    {
        if (pIter->fragmentTransportID == fragmentTransportID)
        {
            break;
        }
    }

    if (pIter)
    {
        // We already have this or a newer version of this fragment
        if (pIter->fragmentVersion >= fragmentVersion)
        {
            pIter->receiveSessionId = pSession->receiveSessionId;
            goto out;
        }

        // We need to update - remove the current version
        if (pPrev)
        {
            pPrev->pNext = pIter->pNext;
        }
        else
        {
            pSession->pServices = pSession->pServices->pNext;
        }
        ServiceFragmentFree(pSession, pIter);
    }

    if (ServiceFragmentAllocate(pSession, pFragment, fragmentLen, &pNew))
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew->tsi = tsi;
    pNew->toi = toi;
    pNew->receiveSessionId = pSession->receiveSessionId;
    pNew->fragmentTransportID = fragmentTransportID;
    pNew->fragmentVersion = fragmentVersion;
    pNew->xmlLen = fragmentLen;
    memcpy(pNew->vXML, pFragment, fragmentLen);

    if (ATSC3ESGServiceFragmentParse(pFragment, fragmentLen,
        &serviceFragmentNode))
    {
        ERRORPRINT("%s: Failed to parse service fragment\n",
            pSession->strDebugName);
        ServiceFragmentFree(pSession, pNew);
        rval = ENStatus_Failure;
        goto out;
    }

    if (!(serviceFragmentNode.base.presentFlags & 
         ATSC3ESGServiceFragmentNodeAttribute_id) ||
        !(serviceFragmentNode.base.presentFlags &
         ATSC3ESGServiceFragmentNodeAttribute_globalServiceID))
    {
        ERRORPRINT("Service fragment lacks version or globalServiceID\n");
        ServiceFragmentFree(pSession, pNew);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew->idXMLOffset = serviceFragmentNode.idXMLOffset;
    pNew->pId = pNew->vXML + serviceFragmentNode.idXMLOffset;
    pNew->idLen = serviceFragmentNode.idXMLLen;
    pNew->globalServiceIDXMLOffset = 
        serviceFragmentNode.globalServiceIDXMLOffset;
    pNew->pGlobalServiceID = pNew->vXML +
        serviceFragmentNode.globalServiceIDXMLOffset;
    pNew->globalServiceIDLen = serviceFragmentNode.globalServiceIDXMLLen;

    if (!(serviceFragmentNode.base.presentFlags & 
        ATSC3ESGServiceFragmentNodeAttribute_version))
    {
        ERRORPRINT("Service fragment lacks version\n");
        pNew->version = 0;
    }
    else
    {
        pNew->version = serviceFragmentNode.version;
    }

    if (pSession->verbosity >= ENVerbosity_StreamEventTrace)
    {
        LOGPRINT_NOFILE("%s: Service Fragment: transportID: %d id %.*s globalServiceID %.*s version %d\n",
            pSession->strDebugName, fragmentTransportID,
            serviceFragmentNode.idXMLLen,
            pFragment + serviceFragmentNode.idXMLOffset,
            serviceFragmentNode.globalServiceIDXMLLen,
            pFragment + serviceFragmentNode.globalServiceIDXMLOffset,
            serviceFragmentNode.version);
    }

    if (ATSC3ESGServiceFragmentFree(&serviceFragmentNode))
    {
        ERRORPRINT("%s: Failed to free service fragment\n",
            pSession->strDebugName);
    }

    pNew->pNext = pSession->pServices;
    pSession->pServices = pNew;

out:
    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }

    return rval;
}

ENStatus
ScheduleFragmentAllocate(ENATSC3ESGSession *pSession, char *pXML, 
    int32_t xmlLen, ESGScheduleFragment **ppScheduleFragment)
{
    ESGScheduleFragment *pNew;
    ENStatus rval = ENStatus_Success;

    if (sizeof(ESGScheduleFragment) + xmlLen > ESG_XML_MAX_LENGTH)
    {
        ERRORPRINT("Schedule fragment too large (%d > %d)\n",
            sizeof(ESGScheduleFragment) + xmlLen, ESG_XML_MAX_LENGTH);
        rval = ENStatus_Failure;
        goto out;
    }

    pNew = (ESGScheduleFragment *)malloc(sizeof(ESGScheduleFragment) + xmlLen);
    if (!pNew)
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    *ppScheduleFragment = pNew;

out:
    return rval;
}

void
ScheduleFragmentFree(ENATSC3ESGSession *pSession, 
    ESGScheduleFragment *pScheduleFragment)
{
    if (ATSC3ESGScheduleFragmentFree(
        &pScheduleFragment->parsedScheduleFragment))
    {
        ERRORPRINT("s: Failed to free schedule fragment\n",
            pSession->strDebugName);
    }

    free(pScheduleFragment);
}

ENStatus
AddScheduleFragment(ENATSC3ESGSession *pSession, uint32_t tsi, uint32_t toi,
    uint32_t fragmentTransportID, uint32_t fragmentVersion, char *pFragment,
    int32_t fragmentLen)
{
    ESGScheduleFragment *pIter;
    ESGScheduleFragment *pPrev = NULL;
    ESGScheduleFragment *pNew;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    for (pIter = pSession->pSchedules;
         pIter;
         pPrev = pIter, pIter = pIter->pNext)
    {
        if (pIter->fragmentTransportID == fragmentTransportID)
        {
            break;
        }
    }

    if (pIter)
    {
        // We already have this or a newer version of this fragment
        if (pIter->fragmentVersion >= fragmentVersion)
        {
            pIter->receiveSessionId = pSession->receiveSessionId;
            goto out;
        }

        // We need to update - remove the current version
        if (pPrev)
        {
            pPrev->pNext = pIter->pNext;
        }
        else
        {
            pSession->pSchedules = pSession->pSchedules->pNext;
        }
        ScheduleFragmentFree(pSession, pIter);
    }

    if (ScheduleFragmentAllocate(pSession, pFragment, fragmentLen, &pNew))
    {
        ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pSession->verbosity >= ENVerbosity_ObjectTrace)
    {
        LOGPRINT_NOFILE("%s: Schedule Fragment: transportID %d  version %d\n",
            pSession->strDebugName, fragmentTransportID, fragmentVersion);
    }

    pNew->tsi = tsi;
    pNew->toi = toi;
    pNew->receiveSessionId = pSession->receiveSessionId;
    pNew->fragmentTransportID = fragmentTransportID;
    pNew->fragmentVersion = fragmentVersion;
    pNew->xmlLen = fragmentLen;
    memcpy(pNew->vXML, pFragment, fragmentLen);

    if (ATSC3ESGScheduleFragmentParse(pFragment, fragmentLen,
        &pNew->parsedScheduleFragment))
    {
        ERRORPRINT("%s: Failed to parse schedule fragment\n",
            pSession->strDebugName);
        free(pNew);
        rval = ENStatus_Failure;
        goto out;
    }

    if (!(pNew->parsedScheduleFragment.base.presentFlags & 
          ATSC3ESGScheduleFragmentNodeAttribute_id))
    {
        ERRORPRINT("Schedule fragment lacks version\n");
        ScheduleFragmentFree(pSession, pNew);
        rval = ENStatus_Failure;
        goto out;
    }

    if (!(pNew->parsedScheduleFragment.base.presentFlags & 
        ATSC3ESGScheduleFragmentNodeAttribute_version))
    {
        ERRORPRINT("Schedule fragment lacks version\n");
        pNew->parsedScheduleFragment.version = 0;
    }

    pNew->pNext = pSession->pSchedules;
    pSession->pSchedules = pNew;

out:
    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }

    return rval;
}

ENStatus
ParseSGDU(ENATSC3ESGSession *pSession, uint32_t tsi, uint32_t toi, 
    uint8_t *pSGDU, int32_t sgduLen)
{
    SGDUUnitHeader *pUnitHeader;
    SGDUUnitPayload *pUnitPayload;
    int32_t nFragments;
    uint8_t fragmentEncoding;
    uint8_t fragmentType;
    int32_t maxPayloadLen;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (sgduLen < sizeof(SGDUUnitHeader))
    {
        ERRORPRINT("%s: Short SGD Unit Header\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    pUnitHeader = (SGDUUnitHeader *)pSGDU;
    nFragments = (pUnitHeader->n_o_service_guide_fragments_byte0 << 16) |
        (pUnitHeader->n_o_service_guide_fragments_byte1 << 8) |
        pUnitHeader->n_o_service_guide_fragments_byte2;

    if (sgduLen < sizeof(SGDUUnitHeader) +
        nFragments * sizeof(pUnitHeader->vServiceGuideFragmentEntries[0]))
    {
        ERRORPRINT("%s: Short SGD fragment entry table\n", 
            pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    for (i = 0; i < nFragments; i++)
    {
        uint32_t offset;
        uint32_t fragmentVersion;
        uint32_t fragmentTransportID;

        fragmentTransportID = ntohl(
            pUnitHeader->vServiceGuideFragmentEntries[i].fragmentTransportID);
        fragmentVersion =
            ntohl(pUnitHeader->vServiceGuideFragmentEntries[i].fragmentVersion);
        offset = ntohl(pUnitHeader->vServiceGuideFragmentEntries[i].offset);

        if (pSession->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT_NOFILE("%s: Fragment %d\n", pSession->strDebugName, i);
            LOGPRINT_NOFILE("%s:   fragmentTransportID: %u\n", 
                pSession->strDebugName, fragmentTransportID);
            LOGPRINT_NOFILE("%s:   fragmentVersion: %u\n", 
                pSession->strDebugName, fragmentVersion);
            LOGPRINT_NOFILE("%s:   offset: %u\n", 
                pSession->strDebugName, offset);
        }

        offset += sizeof(SGDUUnitHeader) +
            sizeof(pUnitHeader->vServiceGuideFragmentEntries[0]) * nFragments;
        if (offset >= sgduLen)
        {
            ERRORPRINT("%s: Fragment data beyond data unit\n", 
                pSession->strDebugName);
            rval = ENStatus_Failure;
            goto out;
        }

        pUnitPayload = (SGDUUnitPayload *)(pSGDU + offset);
        maxPayloadLen = sgduLen - offset;

        fragmentEncoding = pUnitPayload->fragmentEncoding;
        if (fragmentEncoding == 0)
        {
            uint8_t *pXMLEnd;

            if (maxPayloadLen < 2)
            {
                ERRORPRINT("%s: Short payload\n", pSession->strDebugName);
                rval = ENStatus_Failure;
                goto out;
            }
            fragmentType = pUnitPayload->XMLFragment.fragmentType;

            // A332 5.4:
            // "For any value of i in the range 0 to
            // n_o_service_guide_fragments, when value of fragmentEncoding[i]
            // is equal to 0, the fragmentType field shall not take value of
            // 4, 5, 6, 7, 8, 9 for a SGDU."
            if (fragmentType >= 4 && fragmentType <= 9)
            {
                ERRORPRINT("%s: Invalid fragment type %d\n", 
                    pSession->strDebugName, fragmentType);
            }
            if (pSession->verbosity >= ENVerbosity_ObjectTrace)
            {
                LOGPRINT_NOFILE("%s:   XML Fragment type %d\n", 
                    pSession->strDebugName, fragmentType);
            }
            pXMLEnd = memchr(pUnitPayload->XMLFragment.vXMLFragment, 0,
                maxPayloadLen - offsetof(SGDUUnitPayload, XMLFragment) -
                offsetof(struct _XMLFragment, vXMLFragment));
            if (!pXMLEnd)
            {
                // The last string appears to not be zero-terminated.
                pXMLEnd = pSGDU + sgduLen - 1;
            }
            else
            {
                // Back up from the 0 to point to the last character of xml
                pXMLEnd--;
            }

            switch (fragmentType)
            {
                case 1: // Service
                    if (AddServiceFragment(pSession, tsi, toi, 
                        fragmentTransportID, fragmentVersion,
                        (char *)pUnitPayload->XMLFragment.vXMLFragment,
                        pXMLEnd - pUnitPayload->XMLFragment.vXMLFragment))
                    {
                        ERRORPRINT("%s: Failed to add service fragment\n",
                            pSession->strDebugName);
                    }
                    break;
                case 2: // Content
                    if (AddContentFragment(pSession, tsi, toi, 
                        fragmentTransportID, fragmentVersion,
                        (char *)pUnitPayload->XMLFragment.vXMLFragment,
                        pXMLEnd - pUnitPayload->XMLFragment.vXMLFragment))
                    {
                        ERRORPRINT("%s: Failed to add content fragment\n",
                            pSession->strDebugName);
                    }
                    break;
                case 3: // Schedule
                    if (AddScheduleFragment(pSession, tsi, toi, 
                        fragmentTransportID, fragmentVersion,
                        (char *)pUnitPayload->XMLFragment.vXMLFragment,
                        pXMLEnd - pUnitPayload->XMLFragment.vXMLFragment))
                    {
                        ERRORPRINT("%s: Failed to add service fragment\n",
                            pSession->strDebugName);
                    }
                    break;
                default:
                    LOGPRINT_NOFILE("%s: Unhandled fragment type %d\n",
                        pSession->strDebugName, fragmentType);
                    break;
            }
        }
        else
        {
            // Only fragment encoding 0 is allowed for ATSC 3.0 See A332 5.4:
            // "The fragmentEncoding[i] field shall not take value of 1 or 2
            // or 3 for any value of i in the range 0 to
            // n_o_service_guide_fragments in a SGDU."
            ERRORPRINT("%s: Invalid fragment encoding %d\n", 
                pSession->strDebugName, fragmentEncoding);
        }
    }

out:
    return rval;
}

void
PruneStaleItems(ENATSC3ESGSession *pSession)
{
    ESGScheduleFragment *pSchedule;
    ESGScheduleFragment *pPrevSchedule;
    ESGContentFragment *pContent;
    ESGContentFragment *pPrevContent;
    ESGServiceFragment *pService;
    ESGServiceFragment *pPrevService;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    pPrevSchedule = NULL;
    pSchedule = pSession->pSchedules;
    while (pSchedule)
    {
        ESGScheduleFragment *pNext = pSchedule->pNext;

        if (pSchedule->receiveSessionId < pSession->receiveSessionId)
        {
            if (pPrevSchedule)
            {
                pPrevSchedule->pNext = pNext;
            }
            else
            {
                pSession->pSchedules = pNext;
            }
            ScheduleFragmentFree(pSession, pSchedule);
        }
        else
        {
            pPrevSchedule = pSchedule;
        }

        pSchedule = pNext;
    }

    pPrevService = NULL;
    pService = pSession->pServices;
    while (pService)
    {
        ESGServiceFragment *pNext = pService->pNext;

        if (pService->receiveSessionId < pSession->receiveSessionId)
        {
            if (pPrevService)
            {
                pPrevService->pNext = pNext;
            }
            else
            {
                pSession->pServices = pNext;
            }
            ServiceFragmentFree(pSession, pService);
        }
        else
        {
            pPrevService = pService;
        }

        pService = pNext;
    }

    pPrevContent = NULL;
    pContent = pSession->pContents;
    while (pContent)
    {
        ESGContentFragment *pNext = pContent->pNext;

        if (pContent->receiveSessionId < pSession->receiveSessionId)
        {
            if (pPrevContent)
            {
                pPrevContent->pNext = pNext;
            }
            else
            {
                pSession->pContents = pNext;
            }
            ContentFragmentFree(pSession, pContent);
        }
        else
        {
            pPrevContent = pContent;
        }

        pContent = pNext;
    }

    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }
}

void
CheckReceiveSessionComplete(ENATSC3ESGSession *pSession)
{
    SGDFileDescription *pIter;

    for (pIter = pSession->pSGDDs; pIter; pIter = pIter->pNext)
    {
        if (!pIter->fReceived)
        {
            break;
        }
    }

    if (!pIter)
    {
        for (pIter = pSession->pSGDUs; pIter; pIter = pIter->pNext)
        {
            if (!pIter->fReceived)
            {
                break;
            }
        }
    }

    if (!pIter)
    {
        pSession->fReceiveSessionComplete = 1;
        LOGPRINT_NOFILE("%s: Receive session %u complete\n",
            pSession->strDebugName, pSession->receiveSessionId);
         PruneStaleItems(pSession);
    }
}

static void
CallFragmentCallbacks(ENATSC3ESGSession *pSession,
    struct timespec *pPacketTimeUTC, uint32_t tsi, uint32_t toi, 
    uint32_t receiveSessionId)
{
    ESGScheduleFragment *pSchedule;
    ESGContentFragment *pContent;
    ESGServiceFragment *pService;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    if (pSession->ServiceFragmentCallback)
    {
        for (pService = pSession->pServices;
             pService;
             pService = pService->pNext)
        {
            if (pService->tsi == tsi && pService->toi == toi)
            {
                pSession->ServiceFragmentCallback(pPacketTimeUTC, 
                    pSession->receiveSessionId, 
                    pService->idXMLOffset,
                    pService->idLen,
                    pService->version,
                    pService->globalServiceIDXMLOffset,
                    pService->globalServiceIDLen,
                    pService->vXML, pService->xmlLen, 
                    pSession->pCallbackUserData); 
            }
        }
    }

    if (pSession->ScheduleFragmentCallback)
    {
        for (pSchedule = pSession->pSchedules;
             pSchedule;
             pSchedule = pSchedule->pNext)
        {
            if (pSchedule->tsi == tsi && pSchedule->toi == toi)
            {
                ATSC3ESGScheduleFragmentNode *pParsedScheduleFragment;

                pParsedScheduleFragment = &pSchedule->parsedScheduleFragment;
                pSession->ScheduleFragmentCallback(pPacketTimeUTC, 
                    pSession->receiveSessionId, 
                    pParsedScheduleFragment->idXMLOffset,
                    pParsedScheduleFragment->idXMLLen,
                    pParsedScheduleFragment->version,
                    pParsedScheduleFragment->serviceReference.idRefXMLOffset, 
                    pParsedScheduleFragment->serviceReference.idRefXMLLen,
                    pSchedule->vXML, pSchedule->xmlLen, 
                    pSession->pCallbackUserData); 
            }
        }
    }

    if (pSession->ContentFragmentCallback)
    {
        for (pContent = pSession->pContents;
             pContent;
             pContent = pContent->pNext)
        {
            if (pContent->tsi == tsi && pContent->toi == toi)
            {
                pSession->ContentFragmentCallback(pPacketTimeUTC, 
                    pSession->receiveSessionId, 
                    pContent->idXMLOffset,
                    pContent->idLen,
                    pContent->version, 
                    pContent->vXML, pContent->xmlLen, 
                    pSession->pCallbackUserData); 
            }
        }
    }

    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }

    return;
}

ENStatus
MediaFileBeginCallback(ENATSC3ESGSession *pSession, 
    struct timespec *pPacketTimeUTC, struct _LCTChannelInfo *pLCTChannelInfo, 
    uint32_t toi, uint8_t codepoint, struct _FileReceiver *pReceiver) 
{
    int64_t transferLength;
    int64_t contentLength ;
    SGDFileDescription *pIter;
    ENStatus rval = ENStatus_Success;

    if (pSession->fReceiveSessionComplete)
    {
        CallFragmentCallbacks(pSession, pPacketTimeUTC, pLCTChannelInfo->tsi, 
            toi, pSession->receiveSessionId);
        rval = ENStatus_Failure;
        goto out;
    }

    // See if this is an SGDD we want to get
    for (pIter = pSession->pSGDDs; pIter; pIter = pIter->pNext)
    {
        if (pLCTChannelInfo->tsi == pIter->tsi &&
            toi == pIter->toi)
        {
            if (pIter->fReceived)
            {
                // We already receied this sgdd
                rval = ENStatus_Failure;
                goto out;
            }

            transferLength = pIter->transferLength;
            contentLength = pIter->contentLength;
            break;
        }
    }

    // If not an SGDD, then see if it is an SGDU
    if (!pIter)
    {
        for (pIter = pSession->pSGDUs; pIter; pIter = pIter->pNext)
        {
            if (pLCTChannelInfo->tsi == pIter->tsi &&
                toi == pIter->toi)
            {
                if (pIter->fReceived)
                {
                    // We already receied this sgdu
                    rval = ENStatus_Failure;
                    goto out;
                }

                transferLength = pIter->transferLength;
                contentLength = pIter->contentLength;
                break;
            }
        }
    }

    if (!pIter)
    {
        rval = ENStatus_Failure;
        goto out;
    }

    if (pReceiver->sTransferBuffer < transferLength)
    {
        uint8_t *pTransferBuffer;

        pTransferBuffer = (uint8_t *)realloc(
            pReceiver->pTransferBuffer, transferLength + 4096);
        if (!pTransferBuffer)
        {
            ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
            rval = ENStatus_Failure;
            goto out;
        }
        pReceiver->pTransferBuffer = pTransferBuffer;
        pReceiver->sTransferBuffer = transferLength + 4096;
    }

    pReceiver->transferLength = transferLength;
    pReceiver->contentLength = contentLength;
    pReceiver->nBytesReceived = 0;

    if (pSession->verbosity >= ENVerbosity_StreamEventTrace)
    {
        LOGPRINT_NOFILE("%s: %"PRId64".%06lu MediaFileBegin %s tsi %u toi %u codepoint %u\n",
            pSession->strDebugName, pPacketTimeUTC->tv_sec, 
            pPacketTimeUTC->tv_nsec / 1000, pReceiver->strURL, 
            pLCTChannelInfo->tsi, toi, codepoint);
    }

out:
    return rval;
}

void
MediaFileEndCallback(ENATSC3ESGSession *pSession, 
    struct timespec *pPacketTimeUTC, struct _LCTChannelInfo *pLCTChannelInfo, 
    struct _FileReceiver *pReceiver)
{
    ATSC3ESGSGDDNode parsedSGDD;
    SGDFileDescription *pIter;
    char *pXMLBuffer = NULL;

    if (pSession->verbosity >= ENVerbosity_StreamEventTrace)
    {

        LOGPRINT_NOFILE("%s: %"PRId64".%06lu MediaFileEnd %s tsi %u toi %u codepoint %u\n",
            pSession->strDebugName, pPacketTimeUTC->tv_sec,
            pPacketTimeUTC->tv_nsec / 1000, pReceiver->strURL,
            pLCTChannelInfo->tsi, pReceiver->toi, pReceiver->codepoint);
    }

    if (pReceiver->transferLength != pReceiver->nBytesReceived)
    {
        ERRORPRINT(
            "%s: Incorrect bytes received %"PRId64" != %"PRId64" for %s\n",
            pSession->strDebugName, pReceiver->transferLength, 
            pReceiver->nBytesReceived, pReceiver->strURL);
        goto out;
    }

    if (pReceiver->transferLength != pReceiver->contentLength)
    {
        int32_t xmlLength;
        int zret;

        if (pReceiver->contentLength > ESG_XML_MAX_LENGTH)
        {
            ERRORPRINT("%s: XML too large (%d)\n", pSession->strDebugName, 
                pReceiver->contentLength);
            goto out;
        }

        pXMLBuffer = (char *)malloc(pReceiver->contentLength);
        if (!pXMLBuffer)
        {
            ERRORPRINT("%s: Out of memory\n", pSession->strDebugName);
            goto out;
        }

        zret = ENUtils_GunzipUnpipe((unsigned char *)pReceiver->pTransferBuffer,
            pReceiver->transferLength, (unsigned char *)pXMLBuffer,
            pReceiver->contentLength, &xmlLength);
        if (zret != Z_OK)
        {
            ERRORPRINT("%s: Failed to unzip %s\n", pSession->strDebugName, 
                pReceiver->strURL);
            goto out;
        }
    }
    else
    {
        pXMLBuffer = (char *)pReceiver->pTransferBuffer;
    }

    // First see if this is an SGDD
    for (pIter = pSession->pSGDDs; pIter; pIter = pIter->pNext)
    {
        if (pLCTChannelInfo->tsi == pIter->tsi &&
            pReceiver->toi == pIter->toi)
        {
            if (ATSC3ESGSGDDParse(pXMLBuffer, pReceiver->contentLength,
                &parsedSGDD))
            {
                ERRORPRINT("%s: Failed to parse SGDD (%s)\n", 
                    pSession->strDebugName, pReceiver->strURL);
            }
            else
            {
                pIter->fReceived = 1;
                pIter->parsedSGDD = parsedSGDD;
                pIter->fHaveParsedXML = 1;
            }
            break;
        }
    }

    // If not an SGDD then check if it is an SGDU
    if (!pIter)
    {
        for (pIter = pSession->pSGDUs; pIter; pIter = pIter->pNext)
        {
            if (pLCTChannelInfo->tsi == pIter->tsi &&
                pReceiver->toi == pIter->toi)
            {
                if (ParseSGDU(pSession, pLCTChannelInfo->tsi, pReceiver->toi, 
                    (uint8_t *)pXMLBuffer, pReceiver->contentLength))
                {
                    ERRORPRINT("%s: Failed to parse SGDU (%s)\n", 
                        pSession->strDebugName, pReceiver->strURL);
                }
                else
                {
                    pIter->fReceived = 1;
                    CallFragmentCallbacks(pSession, pPacketTimeUTC, 
                        pLCTChannelInfo->tsi, pIter->toi, 
                        pSession->receiveSessionId);
                }
                break;
            }
        }
    }

    CheckReceiveSessionComplete(pSession);

out:
    if (pXMLBuffer && pXMLBuffer != (char *)pReceiver->pTransferBuffer)
    {
        free(pXMLBuffer);
    }

    return;
}

void
MediaFileAbortCallback(ENATSC3ESGSession *pSession, 
    struct timespec *pPacketTimeUTC, struct _FileReceiver *pReceiver)
{
    return;
}

ENStatus
MediaFileDataCallback(ENATSC3ESGSession *pSession, 
    struct timespec *pPacketTimeUTC, const uint8_t *pPayload, 
    int32_t sPayload, int64_t objectPosition,
    struct _FileReceiver *pReceiver)
{
    ENStatus rval = ENStatus_Success;

    if (objectPosition != pReceiver->nBytesReceived)
    {
        ERRORPRINT("%s: Incorrect object position %"PRId64" != %"PRId64"\n",
            pSession->strDebugName, pReceiver->nBytesReceived, 
            objectPosition);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pReceiver->nBytesReceived + sPayload > pReceiver->transferLength)
    {
        ERRORPRINT("%s: Object too large (%"PRId64" > %"PRId64")\n",
            pSession->strDebugName, pReceiver->nBytesReceived + sPayload,
            pReceiver->transferLength);
        rval = ENStatus_Failure;
        goto out;
    }

    memcpy(pReceiver->pTransferBuffer + pReceiver->nBytesReceived,
        pPayload, sPayload);
    pReceiver->nBytesReceived += sPayload;

out:
    return rval;
}

void
MigrateObjectsToReceiveSession(ENATSC3ESGSession *pSession,
    uint32_t tsi, uint32_t toi, uint32_t receiveSessionId)
{
    ESGScheduleFragment *pSchedule;
    ESGContentFragment *pContent;
    ESGServiceFragment *pService;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    for (pSchedule = pSession->pSchedules;
         pSchedule;
         pSchedule = pSchedule->pNext)
    {
        if (pSchedule->tsi == tsi && pSchedule->toi == toi)
        {
            pSchedule->receiveSessionId = receiveSessionId;
        }
    }

    for (pContent = pSession->pContents;
         pContent;
         pContent = pContent->pNext)
    {
        if (pContent->tsi == tsi && pContent->toi == toi)
        {
            pContent->receiveSessionId = receiveSessionId;
        }
    }

    for (pService = pSession->pServices;
         pService;
         pService = pService->pNext)
    {
        if (pService->tsi == tsi && pService->toi == toi)
        {
            pService->receiveSessionId = receiveSessionId;
        }
    }

    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }
}

ENStatus
ROUTESTSIDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pSTSID, int32_t sSTSID, 
    ENATSC3ROUTESLSSTSIDNodePtr pParsedSTSID, int32_t version, uint32_t crc32, 
    void *pCbData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)pCbData;
    SGDFileDescription *pOldSGDDs;
    SGDFileDescription *pOldSGDUs;
    int32_t i;
    int32_t j;
    ENStatus rval = ENStatus_Success;

    if (pSession->stsidVersion == version)
    {
        if (pSession->stsidCRC32 == crc32)
        {
            goto out;
        }
        else
        {
            ERRORPRINT("STSID version %d changed unexpectedly (%X != %X)\n",
                version, pSession->stsidCRC32, crc32);
        }
    }

    // We're starting a new receive session. Increment the id. This has the
    // side-effect of marking all objects stale. Once all SGDDs and SGDUs have
    // been received all objects still stale can be deleted.
    pSession->receiveSessionId++;
    pSession->fReceiveSessionComplete = 0;
    LOGPRINT_NOFILE("%s: Starting receive session %u\n", 
        pSession->strDebugName, pSession->receiveSessionId);

    pOldSGDDs = pSession->pSGDDs;
    pSession->pSGDDs = NULL;
    pOldSGDUs = pSession->pSGDUs;
    pSession->pSGDUs = NULL;

    if (pParsedSTSID->RSs.nOffsetsUsed)
    {
        ENATSC3ROUTESLSSTSIDRSNode *pRS;
        // XXX Do we need to handle multiple RSs for ESG?
        pRS = (ENATSC3ROUTESLSSTSIDRSNode *)EnXMLParserDataOffsetArrayGetPointer(
            &pParsedSTSID->RSs, 0);
        for (i = 0; i < pRS->LSs.nOffsetsUsed; i++)
        {
            ENATSC3ROUTESLSSTSIDLSNode *pLS;

            pLS = (ENATSC3ROUTESLSSTSIDLSNode *)EnXMLParserDataOffsetArrayGetPointer(
                &pRS->LSs, i);
            if (!(pLS->base.presentFlags & ENATSC3ROUTESLSSTSIDLSNodeAttribute_tsi))
            {
                ERRORPRINT("LS without tsi?\n");
                continue;
            }

            if (pLS->base.presentFlags & ENATSC3ROUTESLSSTSIDLSChildNode_SrcFlow &&
                pLS->srcFlow.base.presentFlags & ENATSC3ROUTESLSSTSIDSrcFlowChildNode_EFDT &&
                pLS->srcFlow.efdt.base.presentFlags & ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTInstance &&
                pLS->srcFlow.efdt.fdtInstance.base.presentFlags & ENATSC3ROUTESLSSTSIDFDTInstanceChildNode_File)
            {
                ENATSC3ROUTESLSSTSIDFileNode *pFileNode;

                for (j = 0;
                     j < pLS->srcFlow.efdt.fdtInstance.files.nOffsetsUsed;
                     j++)
                {
                    char *pContentEncoding = NULL;
                    char *pContentType = NULL;
                    int64_t contentLength = -1;
                    int64_t transferLength = -1;

                    pFileNode = (ENATSC3ROUTESLSSTSIDFileNode *)
                        EnXMLParserDataOffsetArrayGetPointer(
                            &pLS->srcFlow.efdt.fdtInstance.files, j);
                    if (!(pFileNode->base.presentFlags &
                         ENATSC3ROUTESLSSTSIDFileNodeAttribute_TOI))
                    {
                        ERRORPRINT("%s: TOI missing\n", pSession->strDebugName);
                        continue;
                    }

                    if (!(pFileNode->base.presentFlags &
                         ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentType))
                    {
                        ERRORPRINT("%s: Content-Type missing\n",
                            pSession->strDebugName);
                        continue;
                    }

                    if (!(pFileNode->base.presentFlags &
                         ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentLength))
                    {
                        ERRORPRINT("%s: content-length missing\n",
                            pSession->strDebugName);
                        continue;
                    }
                    contentLength = pFileNode->contentLength;

                    if ((pFileNode->base.presentFlags &
                         ENATSC3ROUTESLSSTSIDFileNodeAttribute_TransferLength))
                    {
                        transferLength = pFileNode->transferLength;
                    }

                    if (pFileNode->base.presentFlags &
                         ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentEncoding)
                    {
                        pContentEncoding = EnXMLParserDataBufferGetPointer(
                            &pParsedSTSID->dataBuffer,
                            pFileNode->contentEncodingDataOffset);
                    }

                    pContentType = EnXMLParserDataBufferGetPointer(
                        &pParsedSTSID->dataBuffer,
                        pFileNode->contentTypeDataOffset);

                    if (pContentEncoding && !strcmp(pContentEncoding, "gzip"))
                    {
                        if (contentLength == -1 || transferLength == -1)
                        {
                            ERRORPRINT("%s: GZIP but content-length or transfer-length missing\n",
                                pSession->strDebugName);
                            continue;
                        }
                    }

                    if (transferLength == -1)
                    {
                        transferLength = contentLength;
                    }

                    if (!strcmp(pContentType,
                        "application/vnd.oma.bcast.sgdd+xml"))
                    {
                        SGDFileDescription *pIter;
                        SGDFileDescription *pPrev;
                        SGDFileDescription *pSGDD;

                        for (pIter = pOldSGDDs, pPrev = NULL;
                             pIter;
                             pPrev = pIter, pIter = pIter->pNext)
                        {
                            if (pIter->tsi == pLS->tsi && 
                                pIter->toi == pFileNode->toi)
                            {
                                break;
                            }
                        }

                        if (pIter)
                        {
                            // There is an existing SGDD with the same TOI,
                            // reuse it.
                            pSGDD = pIter;
                            if (pPrev)
                            {
                                pPrev->pNext = pIter->pNext;
                            }
                            else
                            {
                                pOldSGDDs = pOldSGDDs->pNext;
                            }
                        }
                        else
                        {
                            // This is a new SGDD
                            if (SGDFileDescriptionAllocate(pSession, &pSGDD))
                            {
                                ERRORPRINT("%s: Out of memory\n", 
                                    pSession->strDebugName);
                                continue;
                            }
                            pSGDD->tsi = pLS->tsi;
                            pSGDD->toi = pFileNode->toi;
                            pSGDD->transferLength = pFileNode->transferLength;
                            pSGDD->contentLength = pFileNode->contentLength;
                            pSGDD->fReceived = 0;
                        }

                        pSGDD->pNext = pSession->pSGDDs;
                        pSession->pSGDDs = pSGDD;
                    }
                    else if (!strcmp(pContentType,
                        "application/vnd.oma.bcast.sgdu"))
                    {
                        SGDFileDescription *pIter;
                        SGDFileDescription *pPrev;
                        SGDFileDescription *pSGDU;

                        for (pIter = pOldSGDUs, pPrev = NULL;
                             pIter;
                             pPrev = pIter, pIter = pIter->pNext)
                        {
                            if (pIter->tsi == pLS->tsi &&
                                pIter->toi == pFileNode->toi)
                            {
                                break;
                            }
                        }

                        if (pIter)
                        {
                            // There is an existing SGDU with the same TOI,
                            // reuse it.
                            pSGDU = pIter;
                            if (pPrev)
                            {
                                pPrev->pNext = pIter->pNext;
                            }
                            else
                            {
                                pOldSGDUs = pOldSGDUs->pNext;
                            }

                            MigrateObjectsToReceiveSession(pSession, pLS->tsi,
                                pFileNode->toi, pSession->receiveSessionId);
                        }
                        else
                        {
                            // This is a new SGDU
                            if (SGDFileDescriptionAllocate(pSession, &pSGDU))
                            {
                                ERRORPRINT("%s: Out of memory\n",
                                    pSession->strDebugName);
                                continue;
                            }
                            pSGDU->tsi = pLS->tsi;
                            pSGDU->toi = pFileNode->toi;
                            pSGDU->transferLength = pFileNode->transferLength;
                            pSGDU->contentLength = pFileNode->contentLength;
                            pSGDU->fReceived = 0;
                        }

                        pSGDU->pNext = pSession->pSGDUs;
                        pSession->pSGDUs = pSGDU;
                    }
                    else
                    {
                        ERRORPRINT("%s: Unhandled content-type %s\n", 
                            pSession->strDebugName, pContentType);
                    }
                }
            }
        }
    }

    while (pOldSGDDs)
    {
        SGDFileDescription *pNext = pOldSGDDs->pNext;

        if (pOldSGDDs->fHaveParsedXML)
        {
            ATSC3ESGSGDDFree(&pOldSGDDs->parsedSGDD);
            pOldSGDDs->fHaveParsedXML = 0;
        }
        SGDFileDescriptionFree(pOldSGDDs);

        pOldSGDDs = pNext;
    }

    while (pOldSGDUs)
    {
        SGDFileDescription *pNext = pOldSGDUs->pNext;

        SGDFileDescriptionFree(pOldSGDUs);

        pOldSGDUs = pNext;
    }

    pSession->stsidVersion = version;
    pSession->stsidCRC32 = crc32;

out:
     return rval;
}

ENStatus
FindReceiver(ENATSC3ESGSession *pSession, 
    struct _LCTChannelInfo *pLCTChannelInfo, uint32_t toi,
    struct _FileReceiver **ppReceiver)
{
    struct _FileReceiver *pReceiver;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0, pReceiver = pSession->vFileReceivers; 
         i < numof(pSession->vFileReceivers); 
         i++, pReceiver++)
    {
        if (pReceiver->fInUse && 
            pReceiver->pLCTChannelInfo == pLCTChannelInfo && 
            pReceiver->toi == toi)
        {
            *ppReceiver = pReceiver;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

void
AllocateReceiver(ENATSC3ESGSession *pSession, struct timespec *pPacketTimeUTC, 
    struct _FileReceiver **ppReceiver)
{
    struct _FileReceiver *pReceiver;
    int32_t i;
    int32_t oldest = -1;
    struct timespec tOldest;

    for (i = 0, pReceiver = pSession->vFileReceivers; 
         i < numof(pSession->vFileReceivers); 
         i++, pReceiver++)
    {
        if (!pReceiver->fInUse)
        {
            clock_gettime(CLOCK_MONOTONIC, &pReceiver->tAllocated);
            pReceiver->fInUse = 1;
            *ppReceiver = pReceiver;
            break;
        }
        else
        {
            if (oldest == -1 ||
                pReceiver->tAllocated.tv_sec < tOldest.tv_sec ||
                (pReceiver->tAllocated.tv_sec == tOldest.tv_sec &&
                 pReceiver->tAllocated.tv_nsec < tOldest.tv_nsec))
            {
                tOldest = pReceiver->tAllocated;
                oldest = i;
            }
        }
    }

    // If we didn't find an inactive receiver then abort the oldest one.
    if (i == numof(pSession->vFileReceivers))
    {
        ERRORPRINT("Aborting oldest receiver (toi %"PRIu32")\n", 
            pSession->vFileReceivers[oldest].toi);
        MediaFileAbortCallback(pSession, pPacketTimeUTC,
            pSession->vFileReceivers + oldest); 
        *ppReceiver = pSession->vFileReceivers + oldest;
    }

    return;
}

void
FreeReceiver(struct _FileReceiver *pReceiver)
{
    pReceiver->fInUse = 0;
    pReceiver->pLCTChannelInfo = NULL;
}

void
ObjectDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    ENATSC3ROUTELCTHeader *pLCTHeader, int32_t formatId, int32_t frag,
    int32_t order, const char *pId, ENATSC3ROUTESessionIdType idType,
    const uint8_t *pPayload, int32_t payloadLength, int64_t objectPosition,
    int32_t fError, void *pCbUserData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;
    struct _FileReceiver *pReceiver = NULL;

    if (formatId != ENATSC3LCTChannelObjectFormatId_File)
    {
        ERRORPRINT("Unhandled formatId %d\n", formatId);
    }

    if (FindLCTChannelForTSI(pSession, pLCTHeader->transportSessionIdentifier, 
        &pLCTChannelInfo))
    {
        ERRORPRINT("%s: Failed to find LCT channel info for tsi %u\n", 
            pSession->strDebugName, pLCTHeader->transportSessionIdentifier);
        goto out;
    }

    if (FindReceiver(pSession, pLCTChannelInfo, 
        pLCTHeader->transportObjectIdentifier, &pReceiver))
    {
        char strContentType[64];

        // We only start processing a new object at offset 0.
        if (objectPosition != 0) 
        {
            goto out;
        }

        AllocateReceiver(pSession, pPacketTimeUTC, &pReceiver);

        if (ENATSC3ROUTESessionGetObjectURLAndContentType(
            pSession->hROUTESession, pLCTHeader->transportSessionIdentifier, 
            pLCTHeader->transportObjectIdentifier, pReceiver->strURL,
            sizeof(pReceiver->strURL), strContentType,
            sizeof(strContentType)))
        {
            ERRORPRINT("%s: Failed to get url for tsi %u toi %u\n", 
                pSession->strDebugName, pLCTHeader->transportSessionIdentifier, 
                pLCTHeader->transportObjectIdentifier);
            FreeReceiver(pReceiver);
            goto out;
        }

        pReceiver->pLCTChannelInfo = pLCTChannelInfo;
        pReceiver->toi = pLCTHeader->transportObjectIdentifier;
        pReceiver->codepoint = pLCTHeader->codepoint;
        pReceiver->pos = 0;
        pReceiver->objectSize = -1;

        if (MediaFileBeginCallback(pSession, pPacketTimeUTC, pLCTChannelInfo, 
            pLCTHeader->transportObjectIdentifier, pLCTHeader->codepoint, 
            pReceiver))
        {
            FreeReceiver(pReceiver);
            goto out;
        }
    }

    if (pReceiver->objectSize == -1)
    {
        if (pLCTHeader->fExtTOLPresent)
        {
            pReceiver->objectSize = pLCTHeader->tolTransferLength;
        }
        else if (pLCTHeader->fExtFTIPresent)
        {
            pReceiver->objectSize = pLCTHeader->ftiTransferLength;
        }
    }
    else
    {
        if (pLCTHeader->fExtTOLPresent)
        {
            if (pReceiver->objectSize != pLCTHeader->tolTransferLength)
            {
                ERRORPRINT("%s: %"PRIu64" TOL object size changed %"PRId64" != %"PRId64"\n",
                    pSession->strDebugName, packetNumber, 
                    pReceiver->objectSize, pLCTHeader->tolTransferLength);
                MediaFileAbortCallback(pSession, pPacketTimeUTC, pReceiver); 
                FreeReceiver(pReceiver);
                goto out;
            }
        }

        if (pLCTHeader->fExtFTIPresent)
        {
            if (pReceiver->objectSize != pLCTHeader->ftiTransferLength)
            {
                ERRORPRINT("%s: %"PRIu64" FTI object size changed %"PRId64" != %"PRId64"\n",
                    pSession->strDebugName, packetNumber, 
                    pReceiver->objectSize, pLCTHeader->ftiTransferLength);
                MediaFileAbortCallback(pSession, pPacketTimeUTC, pReceiver); 
                FreeReceiver(pReceiver);
                goto out;
            }
        }
    }

    // We currently expect data to be delivered in order.
    //
    // XXXX Strictly speaking data may arrive out of order if @order=0 in the
    // SrcFlow's Payload element.
    if (pReceiver->pos != objectPosition)
    {
        ERRORPRINT("%s: %"PRIu64" %"PRIu64".%.06lu av offset error %"PRId64" != %"PRId64" (%"PRId64" missing) (%s size %"PRId64")\n",
            pSession->strDebugName, packetNumber,
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
            pReceiver->pos, objectPosition, objectPosition - pReceiver->pos,
            pReceiver->strURL, pReceiver->objectSize);
        ERRORPRINT("%s: Out of order packet, aborting object\n",
            pSession->strDebugName);

        MediaFileAbortCallback(pSession, pPacketTimeUTC, pReceiver); 
        FreeReceiver(pReceiver);
        goto out;
    }

    if (MediaFileDataCallback(pSession, pPacketTimeUTC, pPayload, 
        payloadLength, objectPosition, pReceiver))
    {
        MediaFileAbortCallback(pSession, pPacketTimeUTC, pReceiver);
        FreeReceiver(pReceiver);
        goto out;
    }
    pReceiver->pos += payloadLength;

    // See if the object is now complete
    if (pReceiver->pos == pReceiver->objectSize)
    {
        MediaFileEndCallback(pSession, pPacketTimeUTC, pLCTChannelInfo,
            pReceiver); 
        FreeReceiver(pReceiver);
    }

out:
    return;
}

void
SessionResetCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    void *pCbUserData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)pCbUserData;
    struct _FileReceiver *pReceiver;
    int32_t i;

    for (i = 0, pReceiver = pSession->vFileReceivers;
         i < numof(pSession->vFileReceivers); 
         i++, pReceiver++)
    {
        if (pReceiver->fInUse)
        {
            MediaFileAbortCallback(pSession, pPacketTimeUTC, pReceiver);
            FreeReceiver(pReceiver);
        }
    }

    for (i = 0; i < numof(pSession->vLCTChannelInfos); i++)
    {
        if (pSession->vLCTChannelInfos[i].fInUse)
        {
            pSession->vLCTChannelInfos[i].fInUse = 0;
        }
    }
}

ENStatus
ENATSC3ESGSessionAllocate(ENATSC3ROUTESessionHandle_t hROUTESession, 
    char *pDebugName, ENATSC3ESGSessionHandle_t *phSession)
{
    int32_t fROUTESTSIDCallbackRegistered = 0;
    int32_t fROUTEObjectCallbacksRegistered = 0;
    int32_t fROUTELCTChannelCallbacksRegistered = 0;
    int32_t fHaveESGFragmentsLock = 0;
    ENATSC3ESGSession *pSession = NULL;
    ENStatus rval = ENStatus_Success;

    pSession = (ENATSC3ESGSession *)calloc(1, sizeof(ENATSC3ESGSession));
    if (!pSession)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    strncpy(pSession->strDebugName, pDebugName, 
        sizeof(pSession->strDebugName) - 1);
    pSession->strDebugName[sizeof(pSession->strDebugName) - 1] = 0;

    if (pthread_mutex_init(&pSession->esgFragmentsLock, NULL))
    {
        ERRORPRINT("%s: Failed to init mutex\n", pSession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }
    fHaveESGFragmentsLock = 1;

    pSession->hROUTESession = hROUTESession;
    pSession->verbosity = ENVerbosity_Terse;
    pSession->stsidVersion = -1;

    ENATSC3ROUTESessionSetVerbosity(pSession->hROUTESession, 
        pSession->verbosity);

    if (ENATSC3ROUTESessionRegisterSTSIDAvailableCallback(
        pSession->hROUTESession, ROUTESTSIDAvailableCallback, pSession))
    {
        ERRORPRINT("%s: Failed to register ROUTE STSDID callback\n", 
            pSession->strDebugName);
        rval = ENStatus_Failure;
        goto errout;
    }
    fROUTESTSIDCallbackRegistered = 1;

    if (ENATSC3ROUTESessionRegisterLCTChannelCallbacks(
        pSession->hROUTESession, LCTChannelAddedCallback, 
        LCTChannelRemovedCallback, pSession))
    {
        ERRORPRINT("%s: Failed to register LCT Channel callbacks\n",
            pSession->strDebugName);
        rval = 1;
        goto out;
    }
    fROUTELCTChannelCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterObjectCallbacks(pSession->hROUTESession,
        ObjectDataCallback, SessionResetCallback, pSession))
    {
        ERRORPRINT("%s: Failed to register ROUTE media file callbacks\n",
            pSession->strDebugName);
        rval = ENStatus_Failure;
        goto errout;
    }
    fROUTEObjectCallbacksRegistered = 1;

    *phSession = pSession;

out:
    return rval;

errout:
    if (pSession)
    {
        if (fROUTESTSIDCallbackRegistered)
        {
            if (ENATSC3ROUTESessionUnregisterSTSIDAvailableCallback(
                pSession->hROUTESession, ROUTESTSIDAvailableCallback, pSession))
            {
                ERRORPRINT("%s: Failed to unregister ROUTE STSDID callback\n",
                    pSession->strDebugName);
                rval = 1;
                goto out;
            }
        }

        if (fROUTEObjectCallbacksRegistered)
        {
            if (ENATSC3ROUTESessionUnregisterObjectCallbacks(
                pSession->hROUTESession, ObjectDataCallback, 
                SessionResetCallback, pSession))
            {
                ERRORPRINT(
                    "%s: Failed to unregister ROUTE media file callbacks\n",
                    pSession->strDebugName);
            }
        }

        if (fROUTELCTChannelCallbacksRegistered)
        {
            if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
                pSession->hROUTESession, LCTChannelAddedCallback, 
                LCTChannelRemovedCallback, pSession))
            {
                ERRORPRINT("%s: Failed to unregister LCT Channel callbacks\n",
                    pSession->strDebugName);
            }
        }

        if (fHaveESGFragmentsLock)
        {
            if (pthread_mutex_destroy(&pSession->esgFragmentsLock))
            {
                ERRORPRINT("%s: Failed to destroy lock\n", 
                    pSession->strDebugName);
            }
        }

        free(pSession);
    }

    goto out;
}

ENStatus
ENATSC3ESGSessionFree(ENATSC3ESGSessionHandle_t hSession)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)hSession;
    struct _FileReceiver *pReceiver;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pReceiver = pSession->vFileReceivers;
         i < numof(pSession->vFileReceivers); 
         i++, pReceiver++)
    {
        if (pReceiver->pTransferBuffer)
        {
            free(pReceiver->pTransferBuffer);
            pReceiver->pTransferBuffer = NULL;
            pReceiver->sTransferBuffer = 0;
        }
    }

    while (pSession->pServices)
    {
        ESGServiceFragment *pNext;

        pNext = pSession->pServices->pNext;
        ServiceFragmentFree(pSession, pSession->pServices);
        pSession->pServices = pNext;
    }

    while (pSession->pSchedules)
    {
        ESGScheduleFragment *pNext;

        pNext = pSession->pSchedules->pNext;
        ScheduleFragmentFree(pSession, pSession->pSchedules);
        pSession->pSchedules = pNext;
    }

    while (pSession->pContents)
    {
        ESGContentFragment *pNext;

        pNext = pSession->pContents->pNext;
        ContentFragmentFree(pSession, pSession->pContents);
        pSession->pContents = pNext;
    }

    while (pSession->pSGDDs)
    {
        SGDFileDescription *pNext;

        pNext = pSession->pSGDDs->pNext;

        if (pSession->pSGDDs->fHaveParsedXML)
        {
            ATSC3ESGSGDDFree(&pSession->pSGDDs->parsedSGDD);
            pSession->pSGDDs->fHaveParsedXML = 0;
        }

        SGDFileDescriptionFree(pSession->pSGDDs);
        pSession->pSGDDs = pNext;
    }

    while (pSession->pSGDUs)
    {
        SGDFileDescription *pNext;

        pNext = pSession->pSGDUs->pNext;
        SGDFileDescriptionFree(pSession->pSGDUs);
        pSession->pSGDUs = pNext;
    }

    if (ENATSC3ROUTESessionUnregisterSTSIDAvailableCallback(
        pSession->hROUTESession, ROUTESTSIDAvailableCallback, pSession))
    {
        ERRORPRINT("%s: Failed to unregister ROUTE STSDID callback\n",
            pSession->strDebugName);
    }

    if (ENATSC3ROUTESessionUnregisterObjectCallbacks(pSession->hROUTESession,
        ObjectDataCallback, SessionResetCallback, pSession))
    {
        ERRORPRINT("%s: Failed to unregister ROUTE media file callbacks\n",
            pSession->strDebugName);
    }

    if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
        pSession->hROUTESession, LCTChannelAddedCallback, 
        LCTChannelRemovedCallback, pSession))
    {
        ERRORPRINT("%s: Failed to unregister LCT Channel callbacks\n",
            pSession->strDebugName);
    }

    if (pthread_mutex_destroy(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to destroy lock\n", pSession->strDebugName);
    }

    free(pSession);

    return rval;
}

ENStatus
ENATSC3ESGSessionSetVerbosity(ENATSC3ESGSessionHandle_t hSession, 
    ENVerbosity verbosity)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)hSession;
    ENStatus rval = ENStatus_Success;

    pSession->verbosity = verbosity;

    return rval;
}

ENStatus
ENATSC3ESGSessionRegisterCallbacks(ENATSC3ESGSessionHandle_t hSession,
    ENATSC3ESGSessionServiceFragmentCallback ServiceFragmentCallback,
    ENATSC3ESGSessionScheduleFragmentCallback ScheduleFragmentCallback,
    ENATSC3ESGSessionContentFragmentCallback ContentFragmentCallback,
    void *pUserData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)hSession;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    pSession->ServiceFragmentCallback = ServiceFragmentCallback;
    pSession->ScheduleFragmentCallback = ScheduleFragmentCallback;
    pSession->ContentFragmentCallback = ContentFragmentCallback;
    pSession->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }

    return rval;
}

ENStatus
ENATSC3ESGSessionUnregisterCallbacks(ENATSC3ESGSessionHandle_t hSession,
    ENATSC3ESGSessionServiceFragmentCallback ServiceFragmentCallback,
    ENATSC3ESGSessionScheduleFragmentCallback ScheduleFragmentCallback,
    ENATSC3ESGSessionContentFragmentCallback ContentFragmentCallback,
    void *pUserData)
{
    ENATSC3ESGSession *pSession = (ENATSC3ESGSession *)hSession;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to lock\n", pSession->strDebugName);
    }

    if (pSession->ServiceFragmentCallback == ServiceFragmentCallback &&
        pSession->ScheduleFragmentCallback == ScheduleFragmentCallback &&
        pSession->ContentFragmentCallback == ContentFragmentCallback &&
        pSession->pCallbackUserData == pUserData)
    {
        pSession->ServiceFragmentCallback = NULL;
        pSession->ScheduleFragmentCallback = NULL;
        pSession->ContentFragmentCallback = NULL;
        pSession->pCallbackUserData = NULL;
    }
    else
    {
        ERRORPRINT("Can't match callbacks\n");
        rval = ENStatus_Failure;
    }

    if (pthread_mutex_unlock(&pSession->esgFragmentsLock))
    {
        ERRORPRINT("%s: Failed to unlock\n", pSession->strDebugName);
    }

    return rval;
}

