/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include "internetchecksum.h"

void
ENInternetChecksumInitialize(ENInternetChecksumContext *pContext)
{
    pContext->accum = 0xffff;
    pContext->fCarry = 0;
}

uint32_t
ENInternetChecksumFinalize(ENInternetChecksumContext *pContext)
{
    if (pContext->fCarry)
    {
        uint8_t t[2] = {pContext->carry, 0};
        pContext->accum += *((uint16_t *)t);
    }

    while (pContext->accum > 0xffff)
    {
        pContext->accum = (pContext->accum & 0xffff) + (pContext->accum >> 16);
    }

    return ~pContext->accum & 0xffff;
}

void
ENInternetChecksumCalc_Internal(const uint8_t *pData, int32_t sData, 
    ENInternetChecksumContext *pContext)
{
    uint32_t accum;

    accum = pContext->accum;

    if (pContext->fCarry)
    {
        uint8_t t[2] = {pContext->carry, *pData};
        accum += *((uint16_t *)t);
        pData++;
        sData--;
    }

#if defined(__x86_64__) || defined(_M_X64) || defined(i386) || defined(__i386__) || defined(__i386) || defined(_M_IX86)
    while (sData >= 16)
    {
        register uint32_t w1 = *((uint32_t *)pData);
        register uint32_t w2 = *((uint32_t *)(pData + 4));
        register uint32_t w3 = *((uint32_t *)(pData + 8));
        register uint32_t w4 = *((uint32_t *)(pData + 12));
        accum += (w1 >> 16) + (w2 >> 16) + (w3 >> 16) + (w4 >> 16) + 
                 (w1 & 0xffff) + (w2 & 0xffff) + (w3 & 0xffff) + (w4 & 0xffff);
        sData -= 16;
        pData += 16;
    }
#else
    while (sData >= 8)
    {
        accum += *((uint16_t *)pData);
        accum += *((uint16_t *)(pData + 2));
        accum += *((uint16_t *)(pData + 4));
        accum += *((uint16_t *)(pData + 6));
        sData -= 8;
        pData += 8;
    }
#endif

    while (sData > 1)
    {
        accum += *((uint16_t *)pData);
        sData -= 2;
        pData += 2;
    }

    if (sData)
    {
        pContext->fCarry = 1;
        pContext->carry = *pData;
    }
    else
    {
        pContext->fCarry = 0;
    }

    while (accum > 0xffff)
    {
        accum = (accum & 0xffff) + (accum >> 16);
    }

    pContext->accum = accum;
}


void
ENInternetChecksumCalc(const uint8_t *pData, int32_t sData, 
    ENInternetChecksumContext *pContext)
{
    // Use 32k blocks to prevent accum from overflowing.
    while (sData)
    {
        int32_t nData = MIN(sData, 32768);

        ENInternetChecksumCalc_Internal(pData, nData, pContext);
        sData -= nData;
        pData += nData;
    }
}
