/*
 * Copyright © 2019 - 2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of
 * the following permission added to Section 15 as permitted in Section 7(a):
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3lls_cdt.h"

static char *ensys = "atsc30lls";
static char *ensubsys = "cdt";

static int32_t fVerbose = 0;

#define MIN(x, y) ( ( ( x) < (y) ) ? (x) : (y) )
#define MAX(x, y) ( ( ( x) > (y) ) ? (x) : (y) )
#define numof(x) (sizeof(x) / sizeof(x[0]))

static ENStatus
CertificatesNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode =
        (ATSC3LLSCDT_CertificationDataNode *)pUserData;
    ATSC3LLSCDT_CertificatesNode *pCertificatesNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_Certificates);
    pCertificatesNode = (ATSC3LLSCDT_CertificatesNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pCertificationDataNode->dataBuffer, (char *)pValue, sValue,
        &pCertificatesNode->textDataOffset))
    {
        ERRORPRINT("Failed to store text\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sText: %.*s\n", xmlTextReaderDepth(reader) * 2, "", sValue, 
            pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vCertificatesAttributes[] = 
{
};
EnXMLParserChildNode vCertificatesChildNodes[] = 
{
};
EnXMLParserNode certificatesNodeData = 
{
    .pName = "Certificates",
    .pAttributes = vCertificatesAttributes,
    .nAttributes = numof(vCertificatesAttributes),
    .pChildNodes = vCertificatesChildNodes,
    .nChildNodes = numof(vCertificatesChildNodes),
    .textHandler = CertificatesNodeTextHandler,
};

static ENStatus
ProcessCertificatesNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode =
        (ATSC3LLSCDT_CertificationDataNode *)pUserData;
    ATSC3LLSCDT_ToBeSignedDataNode *pToBeSignedDataNode;
    ATSC3LLSCDT_CertificatesNode certificatesNode;
    int32_t certificatesNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_ToBeSignedData);
    pToBeSignedDataNode = (ATSC3LLSCDT_ToBeSignedDataNode *)pParentNode;

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pCertificationDataNode->dataBuffer, 
        sizeof(ATSC3LLSCDT_CertificatesNode), &certificatesNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for CertificatesNode\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pToBeSignedDataNode->certificates, certificatesNodeOffset))
    {
        ERRORPRINT("Failed to store certificatesNode offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sCertificates\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    certificatesNode.base.nodeType = ATSC3LLSCDTNodeType_Certificates; 
    rval = EnXMLParserProcessNode(&certificatesNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&certificatesNode);

    memcpy(pCertificationDataNode->dataBuffer.pBuffer + certificatesNodeOffset, 
        &certificatesNode, sizeof(certificatesNode));

out:
    return rval;
}

static ENStatus
NextCertNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode =
        (ATSC3LLSCDT_CertificationDataNode *)pUserData;
    ATSC3LLSCDT_NextCertNode *pNextCertNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_NextCert);
    pNextCertNode = (ATSC3LLSCDT_NextCertNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pCertificationDataNode->dataBuffer, (char *)pValue, sValue,
        &pNextCertNode->textDataOffset))
    {
        ERRORPRINT("Failed to store text\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sText: %.*s\n", xmlTextReaderDepth(reader) * 2, "", sValue, 
            pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vNextCertAttributes[] = 
{
};
EnXMLParserChildNode vNextCertChildNodes[] = 
{
};
EnXMLParserNode nextCertNodeData = 
{
    .pName = "NextCert",
    .pAttributes = vNextCertAttributes,
    .nAttributes = numof(vNextCertAttributes),
    .pChildNodes = vNextCertChildNodes,
    .nChildNodes = numof(vNextCertChildNodes),
    .textHandler = NextCertNodeTextHandler,
};

static ENStatus
ProcessNextCertNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertReplacementNode *pCertReplacementNode;
    ATSC3LLSCDT_NextCertNode *pNextCertNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CertReplacement);
    pCertReplacementNode = (ATSC3LLSCDT_CertReplacementNode *)pParentNode;

    pNextCertNode = &pCertReplacementNode->nextCertNode;
    pNextCertNode->textDataOffset = -1;

    if (fVerbose)
    {
        LOGPRINT("%*NextCert\n", xmlTextReaderDepth(reader) * 2, "");
    }

    pNextCertNode->base.nodeType = ATSC3LLSCDTNodeType_NextCert;
    rval = EnXMLParserProcessNode(&nextCertNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)pNextCertNode);

    return rval;
}

static ENStatus
CurrentCertNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode =
        (ATSC3LLSCDT_CertificationDataNode *)pUserData;
    ATSC3LLSCDT_CurrentCertNode *pCurrentCertNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CurrentCert);
    pCurrentCertNode = (ATSC3LLSCDT_CurrentCertNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pCertificationDataNode->dataBuffer, (char *)pValue, sValue,
        &pCurrentCertNode->textDataOffset))
    {
        ERRORPRINT("Failed to store text\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sText: %.*s\n", xmlTextReaderDepth(reader) * 2, "", sValue, 
            pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vCurrentCertAttributes[] = 
{
};
EnXMLParserChildNode vCurrentCertChildNodes[] = 
{
};
EnXMLParserNode currentCertNodeData = 
{
    .pName = "CurrentCert",
    .pAttributes = vCurrentCertAttributes,
    .nAttributes = numof(vCurrentCertAttributes),
    .pChildNodes = vCurrentCertChildNodes,
    .nChildNodes = numof(vCurrentCertChildNodes),
    .textHandler = CurrentCertNodeTextHandler,
};

static ENStatus
ProcessCurrentCertNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_ToBeSignedDataNode *pToBeSignedDataNode;
    ATSC3LLSCDT_CurrentCertNode *pCurrentCertNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_ToBeSignedData);
    pToBeSignedDataNode = (ATSC3LLSCDT_ToBeSignedDataNode *)pParentNode;

    pCurrentCertNode = &pToBeSignedDataNode->currentCertNode;
    pCurrentCertNode->textDataOffset = -1;

    if (fVerbose)
    {
        LOGPRINT("%*CurrentCert\n", xmlTextReaderDepth(reader) * 2, "");
    }

    pCurrentCertNode->base.nodeType = ATSC3LLSCDTNodeType_CurrentCert;
    rval = EnXMLParserProcessNode(&currentCertNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)pCurrentCertNode);

    return rval;
}

static ENStatus
CertReplacementNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertReplacementNode *pCertReplacementNode;
    int64_t seconds;
    int64_t milliseconds;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CertReplacement);
    pCertReplacementNode = (ATSC3LLSCDT_CertReplacementNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ATSC3LLSCDTCertReplacementNodeAttribute_NextCertFrom:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue,
                &seconds, &milliseconds))
            {
                ERRORPRINT("Failed to parse NextCertFrom %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            pCertReplacementNode->nextCertFromMillisecondsUTC = 
                seconds * 1000 + milliseconds;
            if (fVerbose)
            {
                LOGPRINT("%*sNextCertFrom: %.*s\n", 
                    xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSCDTCertReplacementNodeAttribute_CurrentCertUntil:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue,
                &pCertReplacementNode->currentCertUntilMillisecondsUTC, NULL))
            {
                ERRORPRINT("Failed to parse CurrentCertUntil %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            pCertReplacementNode->currentCertUntilMillisecondsUTC =
                seconds * 1000 + milliseconds;
            if (fVerbose)
            {
                LOGPRINT("%*sCurrentCertUntil: %.*s\n", 
                    xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
    }

out:
    return rval;
}

EnXMLParserAttribute vCertReplacementAttributes[] = 
{
    {
        .id = ATSC3LLSCDTCertReplacementNodeAttribute_NextCertFrom,
        .sName = sizeof("NextCertFrom") - 1,
        .pName = "NextCertFrom",
        .handler = CertReplacementNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSCDTCertReplacementNodeAttribute_CurrentCertUntil,
        .sName = sizeof("CurrentCertUntil") - 1,
        .pName = "CurrentCertUntil",
        .handler = CertReplacementNodeAttributeHandler,
    },
};
EnXMLParserChildNode vCertReplacementChildNodes[] = 
{
    {
        .id = ATSC3LLSCDTCertReplacementChildNode_NextCert,
        .sName = sizeof("NextCert") - 1,
        .pName = "NextCert",
        .handler = ProcessNextCertNode
    },
};
EnXMLParserNode certReplacementNodeData = 
{
    .pName = "CertReplacement",
    .pAttributes = vCertReplacementAttributes,
    .nAttributes = numof(vCertReplacementAttributes),
    .pChildNodes = vCertReplacementChildNodes,
    .nChildNodes = numof(vCertReplacementChildNodes),
    .textHandler = NULL,
};

static ENStatus
ToBeSignedDataNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_ToBeSignedDataNode *pToBeSignedDataNode;
    int64_t seconds;
    int64_t milliseconds;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_ToBeSignedData);
    pToBeSignedDataNode = (ATSC3LLSCDT_ToBeSignedDataNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ATSC3LLCDTToBeSignedDataNodeAttribute_OCSPRefresh:
            if (EnXMLParserDayTimeDurationToSeconds((char *)pValue, 
                sValue, &seconds, &milliseconds))
            {
                ERRORPRINT("Failed to parse OCSPRefresh %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            pToBeSignedDataNode->ocspRefreshSeconds = seconds;
            if (fVerbose)
            {
                LOGPRINT("%*sOCSPRefresh: %.*s\n", xmlTextReaderDepth
                    (reader) * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
                break;
            }
    }

out:
    return rval;
}

static ENStatus
ProcessCertReplacementNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_ToBeSignedDataNode *pToBeSignedDataNode;
    ATSC3LLSCDT_CertReplacementNode *pCertReplacementNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_ToBeSignedData);
    pToBeSignedDataNode = (ATSC3LLSCDT_ToBeSignedDataNode *)pParentNode;
    pCertReplacementNode = &pToBeSignedDataNode->certReplacementNode;

    memset(pCertReplacementNode, 0, sizeof(*pCertReplacementNode));

    if (fVerbose)
    {
        LOGPRINT("%*sDimension\n", xmlTextReaderDepth(reader) * 2, "");
    }

    pCertReplacementNode->base.nodeType = ATSC3LLSCDTNodeType_CertReplacement;
    rval = EnXMLParserProcessNode(&certReplacementNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)pCertReplacementNode);

    return rval;
}

EnXMLParserAttribute vToBeSignedDataAttributes[] = 
{
    {
        .id = ATSC3LLCDTToBeSignedDataNodeAttribute_OCSPRefresh,
        .sName = sizeof("OCSPRefresh") - 1,
        .pName = "OCSPRefresh",
        .handler = ToBeSignedDataNodeAttributeHandler,
    },
};
EnXMLParserChildNode vToBeSignedDataChildNodes[] = 
{
    {
        .id = ATSC3LLCDTToBeSignedDataChildNode_Certificates,
        .sName = sizeof("Certificates") - 1,
        .pName = "Certificates",
        .handler = ProcessCertificatesNode
    },
    {
        .id = ATSC3LLCDTToBeSignedDataChildNode_CurrentCert,
        .sName = sizeof("CurrentCert") - 1,
        .pName = "CurrentCert",
        .handler = ProcessCurrentCertNode
    },
    {
        .id = ATSC3LLCDTToBeSignedDataChildNode_CertReplacment,
        .sName = sizeof("CertReplacement") - 1,
        .pName = "CertReplacement",
        .handler = ProcessCertReplacementNode
    },
};
EnXMLParserNode toBeSignedDataNodeData = 
{
    .pName = "ToBeSignedData",
    .pAttributes = vToBeSignedDataAttributes,
    .nAttributes = numof(vToBeSignedDataAttributes),
    .pChildNodes = vToBeSignedDataChildNodes,
    .nChildNodes = numof(vToBeSignedDataChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessToBeSignedDataNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CertificationData);
    pCertificationDataNode = (ATSC3LLSCDT_CertificationDataNode *)pParentNode;

    if (EnXMLParserDataOffsetArrayInitialize(
        &pCertificationDataNode->toBeSignedDataNode.certificates,
        &pCertificationDataNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ocspResponses array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*ToBeSignedfData\n", xmlTextReaderDepth(reader) * 2, "");
    }

    pCertificationDataNode->toBeSignedDataNode.base.nodeType = 
        ATSC3LLSCDTNodeType_ToBeSignedData;
    rval = EnXMLParserProcessNode(&toBeSignedDataNodeData, reader,
        pUserData, 
        (EnXMLParserNodeBase *)&pCertificationDataNode->toBeSignedDataNode);

out:
    return rval;
}

static ENStatus
CMSSignedDataNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode =
        (ATSC3LLSCDT_CertificationDataNode *)pUserData;
    ATSC3LLSCDT_CMSSignedDataNode *pCMSSignedDataNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CMSSignedData);
    pCMSSignedDataNode = (ATSC3LLSCDT_CMSSignedDataNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pCertificationDataNode->dataBuffer, (char *)pValue, sValue,
        &pCMSSignedDataNode->textDataOffset))
    {
        ERRORPRINT("Failed to store text\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sText: %.*s\n", xmlTextReaderDepth(reader) * 2, "", sValue, 
            pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vCMSSignedDataAttributes[] = 
{
};
EnXMLParserChildNode vCMSSignedDataChildNodes[] = 
{
};
EnXMLParserNode cmsNodeData = 
{
    .pName = "CMSSignedData",
    .pAttributes = vCMSSignedDataAttributes,
    .nAttributes = numof(vCMSSignedDataAttributes),
    .pChildNodes = vCMSSignedDataChildNodes,
    .nChildNodes = numof(vCMSSignedDataChildNodes),
    .textHandler = CMSSignedDataNodeTextHandler,
};

static ENStatus
ProcessCMSSignedDataNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode;
    ATSC3LLSCDT_CMSSignedDataNode *pCMSSignedDataNode;

    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CertificationData);
    pCertificationDataNode = (ATSC3LLSCDT_CertificationDataNode *)pParentNode;

    pCMSSignedDataNode = &pCertificationDataNode->cmsSignedDataNode;
    pCMSSignedDataNode->textDataOffset = -1;

    if (fVerbose)
    {
        LOGPRINT("%*CMSSignedData\n", xmlTextReaderDepth(reader) * 2, "");
    }

    pCMSSignedDataNode->base.nodeType = ATSC3LLSCDTNodeType_CMSSignedData;
    rval = EnXMLParserProcessNode(&cmsNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)pCMSSignedDataNode);

    return rval;
}

static ENStatus
OCSPResponseNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode =
        (ATSC3LLSCDT_CertificationDataNode *)pUserData;
    ATSC3LLSCDT_OCSPResponseNode *pOCSPResponseNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_OCSPResponse);
    pOCSPResponseNode = (ATSC3LLSCDT_OCSPResponseNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pCertificationDataNode->dataBuffer, (char *)pValue, sValue,
        &pOCSPResponseNode->textDataOffset))
    {
        ERRORPRINT("Failed to store text\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sText: %.*s\n", xmlTextReaderDepth(reader) * 2, "", sValue, 
            pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vOCSPResponseAttributes[] = 
{
};
EnXMLParserChildNode vOCSPResponseChildNodes[] = 
{
};
EnXMLParserNode ocspResponseNodeData = 
{
    .pName = "OCSPResponse",
    .pAttributes = vOCSPResponseAttributes,
    .nAttributes = numof(vOCSPResponseAttributes),
    .pChildNodes = vOCSPResponseChildNodes,
    .nChildNodes = numof(vOCSPResponseChildNodes),
    .textHandler = OCSPResponseNodeTextHandler,
};

static ENStatus
ProcessOCSPResponseNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode;
    ATSC3LLSCDT_OCSPResponseNode ocspResponseNode;
    int32_t ocspResponseNodeOffset;

    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSCDTNodeType_CertificationData);
    pCertificationDataNode = (ATSC3LLSCDT_CertificationDataNode *)pParentNode;

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pCertificationDataNode->dataBuffer, 
        sizeof(ATSC3LLSCDT_OCSPResponseNode), &ocspResponseNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for OCSPResponseNode\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pCertificationDataNode->ocspResponses, ocspResponseNodeOffset))
    {
        ERRORPRINT("Failed to store ocspResponseNode offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*OCSPRespoonse\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    ocspResponseNode.base.nodeType = ATSC3LLSCDTNodeType_OCSPResponse; 
    rval = EnXMLParserProcessNode(&ocspResponseNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&ocspResponseNode);

    memcpy(pCertificationDataNode->dataBuffer.pBuffer + ocspResponseNodeOffset, 
        &ocspResponseNode, sizeof(ocspResponseNode));

out:
    return rval;
}

EnXMLParserAttribute vCertificationDataAttributes[] = 
{
};
EnXMLParserChildNode vCertificationDataChildNodes[] = 
{
    {
        .id = ATSC3LLSCDTChildNode_CMSSignedData,
        .sName = sizeof("CMSSignedData") - 1,
        .pName = "CMSSignedData",
        .handler = ProcessCMSSignedDataNode,
    },
    {
        .id = ATSC3LLSCDTChildNode_OCSPResponse,
        .sName = sizeof("OCSPResponse") - 1,
        .pName = "OCSPResponse",
        .handler = ProcessOCSPResponseNode,
    },
    {
        .id = ATSC3LLSCDTRTChildNode_ToBeSignedData,
        .sName = sizeof("ToBeSignedData") - 1,
        .pName = "ToBeSignedData",
        .handler = ProcessToBeSignedDataNode,
    },
};
EnXMLParserNode cdtNodeData = 
{
    .pName = "CertificationData",
    .pAttributes = vCertificationDataAttributes,
    .nAttributes = numof(vCertificationDataAttributes),
    .pChildNodes = vCertificationDataChildNodes,
    .nChildNodes = numof(vCertificationDataChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessCertificationDataNode(xmlTextReaderPtr reader, 
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode)
{
    ENStatus rval = ENStatus_Success;

    if (fVerbose)
    {
        LOGPRINT("RRT\n");
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &pCertificationDataNode->ocspResponses,
        &pCertificationDataNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ocspResponses array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&cdtNodeData, reader, 
        pCertificationDataNode, 
        (EnXMLParserNodeBase *)pCertificationDataNode);

out:
    return rval;
}

ENStatus
ATSC3LLSParseCDT(const char *pXML, int32_t sXML, 
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pCertificationDataNode, 0, 
        sizeof(ATSC3LLSCDT_CertificationDataNode));
    pCertificationDataNode->base.nodeType = 
        ATSC3LLSCDTNodeType_CertificationData;

    if (EnXMLParserDataBufferInitialize(&pCertificationDataNode->dataBuffer, 
        16384, 8192, ensubsys))
    {
        ERRORPRINT("Failed to initialize CDT data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "CertificationData"))
        {
            int32_t lret;

            lret = ProcessCertificationDataNode(reader, 
                pCertificationDataNode);
            if (lret)
            {
                rval = ENStatus_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't CertificationData (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse CDT XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pCertificationDataNode->dataBuffer))
        {
            ERRORPRINT("Failed to free CDT data buffer\n");
        }
    }

    goto out;
}

ENStatus
ATSC3LLSFreeParsedCDT(
    ATSC3LLSCDT_CertificationDataNode *pCertificationDataNode)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pCertificationDataNode->dataBuffer))
    {
        ERRORPRINT("Failed to free CDT string buffer\n");
    }

    return rval;
}

