/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 * All Rights Reserved.
 */

#ifndef LIBENATSC3MMTP_H
#define LIBENATSC3MMTP_H

typedef struct _MMTPSession *MMTPSessionHandle_t;

typedef struct
{
    uint16_t packet_id;
    const char *pAssetId;
    uint8_t version;
    uint32_t length;
    uint16_t service_id;
    uint16_t content_type;
    uint8_t content_version;
    uint8_t content_compression;
    uint8_t uri_length;
    const uint8_t *pURI;
    uint32_t content_length;
} ENATSC3MMTPATSC3SignalingInfo;

typedef struct
{
    uint8_t type;

    union
    {
        // For identifier_mapping_type 0: asset_id
        struct
        {
            uint16_t asset_id_scheme;
            uint16_t asset_id_length;
            const uint8_t *pAssetId; // pointer back into packet data
        } asset;

        // For identifier_mapping_type 1: a list of URLs under a common
        // packet_id mapping
        struct
        {
            uint32_t url_count;         // number of provided urls.
            const uint8_t *pRawUrlList; // Raw bytes comprilsing the url list
            struct
            {
                uint8_t url_length;
                const uint8_t *pUrl;
            } vUrls[16];                // Parsed list (up to numof(vUrls))
        } urlList;

        // For identifier_mapping_type 2: a regular expression string used
        // to match urls
        struct
        {
            uint32_t regex_length;
            const uint8_t *pRegex;
        } regex;

        // For identifier_mapping_type 3: a DASH representation@id
        struct
        {
            uint32_t representation_id_length;
            const uint8_t *pRepresentationId;
        } dashRepresentationId;
    };
} ENATSC3MMTPIdentifier_mapping;

typedef struct
{
    uint8_t location_type;
    union
    {
        // location_type == 0
        uint16_t packet_id;

        // location_type == 1
        struct
        {
            uint32_t ipv4_src_addr;
            uint32_t ipv4_dst_addr;
            uint16_t dst_port;
            uint16_t packet_id;
        } ipv4_packet_id;

        // location_type == 2
        struct
        {
            uint64_t ipv6_src_addr[2];
            uint64_t ipv6_dst_addr[2];
            uint16_t dst_port;
            uint16_t packet_id;
        } ipv6_packet_id;

        // location_type == 3
        struct
        {
            uint16_t network_id;
            uint16_t MPEG_2_transport_stream_id;
            uint16_t MPEG_2_PID;
        } mpegts_pid;

        // location_type == 4
        struct
        {
            uint64_t ipv6_src_addr[2];
            uint64_t ipv6_dst_addr[2];
            uint16_t MPEG_2_PID;
        } ipv6_pid;

        // location_type == 5
        struct
        {
            uint8_t url_length;
            const uint8_t *pUrl;
        } url;
        // location_type == 6
        struct
        {
            uint8_t url_length;
            const uint8_t *pUrl;
            uint16_t byte_offset;
            uint16_t length;
        } url_offset;

        // location_type == 8
        uint8_t message_id;

        // location_type == 9
        struct
        {
            uint16_t packet_id;
            uint8_t message_id;
        } packet_id_message_id;

        // location_type == 0x0a
        struct
        {
            uint32_t ipv4_src_addr;
            uint32_t ipv4_dst_addr;
            uint16_t dst_port;
            uint16_t packet_id;
            uint8_t message_id;
        } ipv4_packet_id_message_id;

        // location_type == 0x0b
        struct
        {
            uint64_t ipv6_src_addr[2];
            uint64_t ipv6_dst_addr[2];
            uint16_t dst_port;
            uint16_t packet_id;
            uint8_t message_id;
        } ipv6_packet_id_message_id;

        // location_type == 0x0c
        struct
        {
            uint32_t ipv4_src_addr;
            uint32_t ipv4_dst_addr;
            uint16_t dst_port;
            uint16_t MPEG_2_PID;
        } ipv4_pid;
    };
} ENATSC3MMTPLocation_info;

typedef struct
{   
    ENATSC3MMTPIdentifier_mapping Identifier_mapping;
    uint8_t asset_type[4];
    uint8_t default_asset_flag;
    uint8_t asset_clock_relation_flag;
    
    // If asset_clock_relation == 1
    uint8_t asset_clock_relation_id;
    // If asset_clock_relation == 1
    uint8_t asset_timescale_flag;
    // If asset_timescale_flag = 1 
    uint32_t asset_timescale;

    // The number of locations provided in the message
    uint8_t location_count;

    // Parsed locations (up to numof(vLocations))
    ENATSC3MMTPLocation_info vLocations[16];
    // The number of locations populated in vLocations. May be less than
    // location_count if numof(vLocations) < location_count.
    int32_t nLocations;

    uint16_t asset_descriptors_length;
    const uint8_t *pAssetDescriptors;
} ENATSC3MMTPMPTSignalingInfoAsset;

typedef struct
{
    uint16_t message_id;
    uint8_t message_version;
    uint8_t table_id;
    uint8_t table_version;
    uint8_t MP_table_mode;
    uint8_t number_of_assets;

    // The CRC32C of the MPT signaling message payload.
    uint32_t payloadCRC32C;

    uint8_t MMTP_package_id_length;
    const uint8_t *pPackageId;
    uint16_t MP_table_descriptors_length;
    const uint8_t *pTableDescriptors;

    ENATSC3MMTPMPTSignalingInfoAsset vAssets[32];
    // The number of vAssets which have been populated. May be less than
    // number_of_assets if numof(vAssets) < number_of_assets.
    int32_t nAssets;
} ENATSC3MMTPMPTSignalingInfo;

typedef void (*ENATSC3MMTPUSBDAvailableCallback)(
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const char *pUSBD, 
    int32_t sUSBD, int32_t version, void *pUserData);

typedef void (*ENATSC3MMTPHELDAvailableCallback)(
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const char *pHELD, 
    int32_t sHELD, int32_t version, void *pUserData);

typedef void (*ENATSC3MMTPVSPDAvailableCallback)(
    struct timespec *pPacketTimeUTC,  uint64_t packetNumber,
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const uint8_t *pVSPD, 
    int32_t sVSPD, int32_t version, void *pUserData);

typedef void (*ENATSC3MMTPASPDAvailableCallback)(
    struct timespec *pPacketTimeUTC,  uint64_t packetNumber,
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const uint8_t *pASPD, 
    int32_t sASPD, int32_t version, void *pUserData);

typedef void (*ENATSC3MMTPCADAvailableCallback)(
    struct timespec *pPacketTimeUTC,  uint64_t packetNumber,
    const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, const uint8_t *pCAD, 
    int32_t sCAD, int32_t version, void *pUserData);

typedef void (*ENATSC3MMTPMPTSignalingMessageCallback)(
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const ENATSC3MMTPMPTSignalingInfo *pSignalingInfo, void *pUserData);

typedef void (*ENATSC3MMTPMediaDataCallback)(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t packet_id, const char *pAssetId, 
    int32_t fragment_type, uint32_t sequence_number, 
    uint64_t mpuPresentationTime, uint32_t mpuDuration, uint64_t timeScale, 
    const uint8_t *pData, int32_t sData, void *pUserData);

typedef void (*MMTPIPAddressAddCallback)(uint32_t addr, uint16_t port, 
    void *pUserData);

typedef void (*MMTPIPAddressRemoveCallback)(uint32_t addr, uint16_t port, 
    void *pUserData);

ENStatus
ENATSC3MMTPSessionAllocate(uint32_t addr, uint16_t port, 
    MMTPIPAddressAddCallback IPAddressAddCallback,
    MMTPIPAddressRemoveCallback IPAddressRemoveCallback, 
    void *pIPAddrCallbackUserData, char *pDebugName, 
    MMTPSessionHandle_t *phMMTPSession);

ENStatus
ENATSC3MMTPSessionFree(MMTPSessionHandle_t hMMTPSession); 

ENStatus
ENATSC3MMTPSessionSetVerbosity(MMTPSessionHandle_t hMMTPSession, 
    ENVerbosity verbosity);

ENStatus
ENATSC3MMTPSessionRegisterCallbacks(
    MMTPSessionHandle_t hMMTPSession,
    ENATSC3MMTPUSBDAvailableCallback USBDAvailableCallback,
    ENATSC3MMTPHELDAvailableCallback HELDAvailableCallback,
    ENATSC3MMTPVSPDAvailableCallback VSPDAvailableCallback,
    ENATSC3MMTPASPDAvailableCallback ASPDAvailableCallback,
    ENATSC3MMTPCADAvailableCallback CADAvailableCallback,
    ENATSC3MMTPMPTSignalingMessageCallback MPTSignalingMessageCallback,
    ENATSC3MMTPMediaDataCallback MediaDataCallback,
    void *pUserData);

ENStatus
ENATSC3MMTPSessionUnregisterCallbacks(
    MMTPSessionHandle_t hMMTPSession,
    ENATSC3MMTPUSBDAvailableCallback USBDAvailableCallback,
    ENATSC3MMTPHELDAvailableCallback HELDAvailableCallback,
    ENATSC3MMTPVSPDAvailableCallback VSPDAvailableCallback,
    ENATSC3MMTPASPDAvailableCallback ASPDAvailableCallback,
    ENATSC3MMTPCADAvailableCallback CADAvailableCallback,
    ENATSC3MMTPMPTSignalingMessageCallback MPTSignalingMessageCallback,
    ENATSC3MMTPMediaDataCallback MediaDataCallback,
    void *pUserData);

ENStatus
ENATSC3MMTPSessionProcessUDPPacket(MMTPSessionHandle_t hMMTPSession,
    struct timespec *pPacketTime, uint64_t packetNumber, 
    const uint8_t *pPayload, int32_t payloadLength);

#endif
