/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3routesls_efdt.h"

static char *ensys = "routesls";
static char *ensubsys = "efdt";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static ENStatus
FileNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSEFDTNode *pEFDTNode = (ENATSC3ROUTESLSEFDTNode *)pUserData;
    ENATSC3ROUTESLSEFDTFileNode *pFileNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSEFDTNodeType_File);
    pFileNode = (ENATSC3ROUTESLSEFDTFileNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSEFDTFileNodeAttribute_contentLocation:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pEFDTNode->dataBuffer, (char *)pValue, sValue, 
                &pFileNode->contentLocationDataOffset))
            {
                ERRORPRINT("Failed to store Content-Location\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Location: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSEFDTFileNodeAttribute_toi:
            if (EnXMLParser_antoul((char *)pValue, sValue,
                &pFileNode->toi))
            {
                ERRORPRINT("Failed to parse toi\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sTOI: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3ROUTESLSEFDTFileNodeAttribute_contentLength:
            if (EnXMLParser_antoul((char *)pValue, sValue,
                &pFileNode->contentLength))
            {
                ERRORPRINT("Failed to parse Content-Length\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Length: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSEFDTFileNodeAttribute_contentType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pEFDTNode->dataBuffer, (char *)pValue, sValue, 
                &pFileNode->contentTypeDataOffset))
            {
                ERRORPRINT("Failed to store Content-Type\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Type: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", pAttribute->id, 
                pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vFileAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSEFDTFileNodeAttribute_contentLocation,
        .sName = sizeof("Content-Location") - 1,
        .pName = "Content-Location",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSEFDTFileNodeAttribute_toi,
        .sName = sizeof("TOI") - 1,
        .pName = "TOI",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSEFDTFileNodeAttribute_contentLength,
        .sName = sizeof("Content-Length") - 1,
        .pName = "Content-Length",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSEFDTFileNodeAttribute_contentType,
        .sName = sizeof("Content-Type") - 1,
        .pName = "Content-Type",
        .handler = FileNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vFileChildNodes[] =
{
};
static EnXMLParserNode fileNodeData = 
{
    .pName = "File",
    .pAttributes = vFileAttributes,
    .nAttributes = numof(vFileAttributes),
    .pChildNodes = vFileChildNodes,
    .nChildNodes = numof(vFileChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessFileNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSEFDTNode *pEFDTNode = (ENATSC3ROUTESLSEFDTNode *)pUserData;
    ENATSC3ROUTESLSEFDTFDTInstanceNode *pFDTInstanceNode;
    ENATSC3ROUTESLSEFDTFileNode fileNode;
    int32_t fileNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSEFDTNodeType_FDTInstance);
    pFDTInstanceNode = (ENATSC3ROUTESLSEFDTFDTInstanceNode *)pParentNode;

    memset(&fileNode, 0, sizeof(fileNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pEFDTNode->dataBuffer,
        sizeof(fileNode), &fileNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for file node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pFDTInstanceNode->files,
        fileNodeOffset))
    {
        ERRORPRINT("Failed to store file node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    VERBOSEPRINT("%*sFile\n", xmlTextReaderDepth(reader) * 2, "");

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingRegionTableNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    fileNode.base.nodeType = ENATSC3ROUTESLSEFDTNodeType_File;
    rval = EnXMLParserProcessNode(&fileNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)&fileNode);

    memcpy(pEFDTNode->dataBuffer.pBuffer + fileNodeOffset, &fileNode, 
        sizeof(fileNode));

out:
    return rval;
}

static ENStatus
FDTInstanceNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSEFDTFDTInstanceNode *pFDTInstanceNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSEFDTNodeType_FDTInstance);
    pFDTInstanceNode = (ENATSC3ROUTESLSEFDTFDTInstanceNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSEFDTFDTInstanceNodeAttribute_expires:
            if (EnXMLParser_antoul((char *)pValue, sValue,
                &pFDTInstanceNode->expires))
            {
                ERRORPRINT("Failed to parse Expires\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%sExpires: %.*s\n", xmlTextReaderDepth(reader) * 2, 
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSEFDTFDTInstanceNodeAttribute_efdtVersion:
            if (EnXMLParser_antoul((char *)pValue, sValue,
                &pFDTInstanceNode->efdtVersion))
            {
                ERRORPRINT("Failed to parse efdtVersion\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sefdtVersion: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", pAttribute->id, 
                pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vFDTInstanceAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSEFDTFDTInstanceNodeAttribute_expires,
        .sName = sizeof("Expires") - 1,
        .pName = "Expires",
        .handler = FDTInstanceNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSEFDTFDTInstanceNodeAttribute_efdtVersion,
        .sName = sizeof("efdtVersion") - 1,
        .pName = "efdtVersion",
        .handler = FDTInstanceNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vFDTInstanceChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSEFDTNodeType_FDTInstance,
        .sName = sizeof("File") - 1,
        .pName = "File",
        .handler = ProcessFileNode,
    },
};
static EnXMLParserNode fdtInstanceNodeData = 
{
    .pName = "FDT-Instance",
    .pAttributes = vFDTInstanceAttributes,
    .nAttributes = numof(vFDTInstanceAttributes),
    .pChildNodes = vFDTInstanceChildNodes,
    .nChildNodes = numof(vFDTInstanceChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessFDTInstanceNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSEFDTNode *pEFDTNode;
    ENATSC3ROUTESLSEFDTFDTInstanceNode *pFDTInstanceNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSEFDTNodeType_EFDT);
    pEFDTNode = (ENATSC3ROUTESLSEFDTNode *)pParentNode;
    
    pFDTInstanceNode = &pEFDTNode->fdtInstance;
    pFDTInstanceNode->base.nodeType = ENATSC3ROUTESLSEFDTNodeType_FDTInstance;

    VERBOSEPRINT("%*sFDT-Instance\n", xmlTextReaderDepth(reader) * 2, "");

    if (EnXMLParserDataOffsetArrayInitialize(&pFDTInstanceNode->files, 
        &pEFDTNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate files array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&fdtInstanceNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pFDTInstanceNode);

out:
    return rval;
}

static EnXMLParserAttribute vEFDTAttributes[] = 
{
};
static EnXMLParserChildNode vEFDTChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSEFDTNodeType_EFDT,
        .sName = sizeof("FDT-Instance") - 1,
        .pName = "FDT-Instance",
        .handler = ProcessFDTInstanceNode,
    },
};
static EnXMLParserNode efdtNodeData = 
{
    .pName = "EFDT",
    .pAttributes = vEFDTAttributes,
    .nAttributes = numof(vEFDTAttributes),
    .pChildNodes = vEFDTChildNodes,
    .nChildNodes = numof(vEFDTChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessEFDTNode(xmlTextReaderPtr reader, void *pUserData, 
    ENATSC3ROUTESLSEFDTNode *pEFDTNode)
{
    ENStatus rval = ENStatus_Success;

    pEFDTNode->base.nodeType = ENATSC3ROUTESLSEFDTNodeType_EFDT;

    VERBOSEPRINT("%*sEFDT\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&efdtNodeData, reader, pUserData, 
        &pEFDTNode->base);

    return rval;
}

ENStatus
ROUTESLSEFDTParse(const char *pXML, int32_t sXML, 
    ENATSC3ROUTESLSEFDTNode *pParsedEFDT)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedEFDT, 0, sizeof(ENATSC3ROUTESLSEFDTNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (EnXMLParserDataBufferInitialize(&pParsedEFDT->dataBuffer, 4096, 4096,
        ensubsys))
    {
        ERRORPRINT("Failed to initialize EFDT data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "EFDT"))
        {
            ProcessEFDTNode(reader, pParsedEFDT, pParsedEFDT);
        }
        else if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "FDT-Instance"))
        {
            EnXMLParserChildNode fdtInstanceChildNode =
            {
                .id = ENATSC3ROUTESLSEFDTNodeType_EFDT,
                .sName = sizeof("FDT-Instance") - 1,
                .pName = "FDT-Instance",
                .handler = ProcessFDTInstanceNode,
            };

            // Manufacture the EFDT node
            pParsedEFDT->base.nodeType = ENATSC3ROUTESLSEFDTNodeType_EFDT;
            pParsedEFDT->base.presentFlags = 
                ENATSC3ROUTESLSEFDTChildNode_FDTInstance;

            if (ProcessFDTInstanceNode(&fdtInstanceChildNode, reader, 
                pParsedEFDT, &pParsedEFDT->base))
            {
                rval = ENStatus_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't EFDT or FDT-Instance (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        LOGPRINT_NOCONSOLE("Couldn't find STSID\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedEFDT->dataBuffer))
        {
            ERRORPRINT("Failed to free USBD data buffer\n");
        }
    }

    goto out;
}

ENStatus
ROUTESLSEFDTFree(ENATSC3ROUTESLSEFDTNode *pParsedEFDT)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedEFDT->dataBuffer))
    {
        ERRORPRINT("Failed to free EFDT data buffer\n");
    }

    return rval;
}

