/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libenatsc3utils/enatsc3bitratestatistics.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include "atsc3routesls_envelope.h"
#include "atsc3routesls_efdt.h"
#include "atsc3routesls_mime.h"
#include "atsc3routesls_stsid.h"

#include "libenatsc3route.h"

char *ensys = "route";
char *ensubsys = "main";

#define MAX_PAYLOAD_SIZE 65536
#define SLS_BUFFER_SIZE 16384

#define REP_ID_MAX_LENGTH 256

#define STATS_INTERVAL_MILLISECONDS 1000

#define SLS_OBJECT_MAX_SIZE (1024 * 1024)

// A/331 7.1: SLS fragments are delivered on a dedicated LCT transport channel 
// with TSI = 0
#define SLS_TSI 0

#define EXT_TIME_SCT_HI_PRESENT_FLAG 0x8000
#define EXT_TIME_SCT_LOW_PRESENT_FLAG 0x4000
#define EXT_TIME_SCT_ERT_PRESENT_FLAG 0x2000
#define EXT_TIME_SCT_SLC_PRESENT_FLAG 0x1000

#define MAX_MEDIA_FILE_CALLBACKS 4

#define LCT_CHANNEL_MAX_PAYLOAD_DESCRIPTIONS 16

#define LiteralLength(x) (sizeof(x) - 1)

typedef struct
{
    int32_t fPresent;
    uint32_t fecPayloadId;
} ALCHeader;

typedef struct 
{
    int32_t fInUse;
    int32_t fValid;
    uint32_t addr;
    uint16_t port;
    enum LCTObjectType
    {
        LCTObjectType_Invalid = 0,
        LCTObjectType_SLS = 1,
        LCTObjectType_MediaFile = 2,
    } objectType;
    uint32_t tsi;
    char strId[REP_ID_MAX_LENGTH];
    ENATSC3ROUTESessionIdType idType;

    int64_t slsTOI;
    int64_t slsObjectSize;
    int64_t slsPos;


    // The size and crc32c of the last EFDT received
    uint32_t lastEFDTcrc32c;
    int32_t lastEFDTSize;

    // The next expected TOI for the SLS obtained from FDT-Instance
    int64_t expectedSLSTOI;

    // For some object types a buffer is allocated to aggregate the LCT/ALC
    // packets into objects so that the object can be processed in its 
    // entirety.
    uint8_t *pBuffer;
    int32_t sBuffer;

    // The template to be used in generating the Content-Location for objects
    // received on this LCT channel. Obtained from the EFDT.
    char strFileTemplate[128];
    char strContentType[64];

    // A mapping of toi to content-locations. Obtained from FDTParameters.
    struct _FDTInstanceFiles
    {
        uint32_t toi;
        char strContentLocation[128];
        char strContentType[64];
    } vFiles[16];
    int32_t nFiles;

    union
    {
        struct 
        {
            int64_t lastCompleteFDTInstanceTOI;
            int64_t lastCompleteSLSTOI;
            int32_t lastCompleteSLSVersion;
            uint32_t lastCompleteSLSCRC32C;
        } slsInfo;
    };

    int32_t nPayloadDescriptions;
    struct _LCTChannelInfoPayloadDescription
    {
        uint8_t codepoint;
        uint8_t formatId;
        uint8_t frag;
        uint8_t order;
        uint8_t srcFecPayloadId;
    } vPayloadDescriptions[LCT_CHANNEL_MAX_PAYLOAD_DESCRIPTIONS];

    // Clients can store userdata for a media file being received on this
    // LCT channel in vMediaFileCallbackUserData; 
    ENUserData vObjectCallbackUserData[MAX_MEDIA_FILE_CALLBACKS];

    // Stats
    ENATSC3BitrateStatistics stats;
} LCTChannelInfo;

typedef struct _ENATSC3ROUTESession
{
    ENATSC3ROUTESessionType type;

    ENVerbosity verbosity;
    char strDebugName[16];

    uint32_t slsAddr;
    uint32_t slsPort;

    LCTChannelInfo *pSLSLCTChannelInfo;

    struct _MPDAvailableCallbacks
    {
        int32_t fValid;
        ENATSC3ROUTESessionMPDAvailableCallbackFunction MPDAvailableCallback;
        void *pMPDAvailableCallbackUserData;
    } vMPDAvailableCallbacks[4];

    struct _HELDAvailableCallbacks
    {
        int32_t fValid;
        ENATSC3ROUTESessionHELDAvailableCallbackFunction HELDAvailableCallback;
        void *pHELDAvailableCallbackUserData;
    } vHELDAvailableCallbacks[4];

    struct _STSIDAvailableCallbacks
    {
        int32_t fValid;
        ENATSC3ROUTESessionSTSIDAvailableCallbackFunction STSIDAvailableCallback;
        void *pSTSIDAvailableCallbackUserData;
    } vSTSIDAvailableCallbacks[4];

    struct _SLSAvailableCallbacks
    {
        int32_t fValid;
        ENATSC3ROUTESessionSLSAvailableCallbackFunction SLSAvailableCallback;
        void *pSLSAvailableCallbackUserData;
    } vSLSAvailableCallbacks[4];

    struct _LCTChannelCallbacks
    {
        int32_t fValid;
        ENATSC3ROUTESessionLCTChannelAddedCallbackFunction LCTChannelAddedCallback;
        ENATSC3ROUTESessionLCTChannelRemovedCallbackFunction LCTChannelRemovedCallback;
        void *pLCTChannelCallbackUserData;
    } vLCTChannelCallbacks[4];

    struct _ObjectCallbacks
    {
        int32_t fValid;
        ENATSC3ROUTESessionObjectDataCallbackFunction ObjectDataCallback;
        ENATSC3ROUTESessionSessionResetCallbackFunction SessionResetCallback;
        void *pObjectCallbackUserData;
    } vObjectCallbacks[MAX_MEDIA_FILE_CALLBACKS];

    ENATSC3ROUTESessionIPAddressAddCallbackFunction IPAddressAddCallback;
    ENATSC3ROUTESessionIPAddressRemoveCallbackFunction IPAddressRemoveCallback;
    ENATSC3ROUTESessionIPAddressCommitCallbackFunction IPAddressCommitCallback;
    ENATSC3ROUTESessionVerifySignatureCallbackFunction VerifySignatureCallback;
    void *pCallbackUserData;

    LCTChannelInfo vLCTChannelInfos[32];

    char strShadowFilePath[128];

    // Stats
    ENATSC3BitrateStatistics stats;
    int32_t fBitrateStatisticsEnabled;
} ENATSC3ROUTESession, *ENATSC3ROUTESessionHandle_t;

static ENStatus
GenerateObjectURLFromTemplate(ENATSC3ROUTESession *pROUTESession, 
    char *pTemplate, char *pURL, int32_t sURL, uint32_t toi)
{
    int32_t nb;
    char c;
    ENStatus rval = ENStatus_Success;

    // The template from the EFDT is a string where the substring $TOI$ is
    // replaced by the toi. For example given the template "v1-$TOI$.mp4v"
    // and toi 1234, the resulting URL is "v1-1234.mp4v". The template may
    // also contain a format tag indicating the width of the toi string.
    // For example "v1-$TOI%05d$.mp4v" and toi 1234 would become 
    // "v1-01234.mp4v".
    nb = 0;
    while ((c = *(pTemplate++)) && sURL)
    {
        if (c == '$')
        {
            if (*pTemplate == '$')
            {
                // Escape sequence for '$'
                pTemplate++;
                *(pURL++) = c;
                sURL--;
            }
            else
            {
                if (!strncmp(pTemplate, "TOI", 3))
                {
                    char strFormat[128] = "%d";

                    // TOI substitution
                    pTemplate += 3;
                    if (*pTemplate == '$')
                    {
                        // Use the default "%d" format string
                        pTemplate++;
                    }
                    else if (*pTemplate == '%')
                    {
                        int32_t i;
                        char c;

                        for (i = 0; (c = *(pTemplate++)) && c != '$' && i < sizeof(strFormat); i++)
                        {
                            strFormat[i] = c;
                        }

                        if (c == 0)
                        {
                            DEBUGPRINT_NOCONSOLE(
                                "%s: Unterminated TOI format specifier\n",
                                pROUTESession->strDebugName);
                            rval = ENStatus_Failure;
                            goto out;
                        }
                        if (i == sizeof(strFormat))
                        {
                            ERRORPRINT("%s: TOI format too long\n",
                                pROUTESession->strDebugName);
                            rval = ENStatus_Failure;
                            goto out;
                        }
                    }

                    nb = snprintf(pURL, sURL, strFormat, toi);
                    if (nb >= sURL)
                    {
                        ERRORPRINT("%s: URL too long\n",
                            pROUTESession->strDebugName);
                        rval = ENStatus_Failure;
                        goto out;
                    }
                    pURL += nb;
                    sURL -= nb;
                }
                else
                {
                    // This is an undefined case per A/331 A.3.3.2.8
                }
            }
        }
        else
        {
            *(pURL++) = c;
            sURL--;
        }
    }

    if (!sURL)
    {
        ERRORPRINT("%s: URL too long\n", pROUTESession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    *pURL = 0;

out:
    return rval;
}

ENStatus
ENATSC3ROUTESessionSetShadowFilePath(ENATSC3ROUTESessionHandle_t hROUTESession, 
    char *pPath)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    ENStatus rval = ENStatus_Success;

    if (strlen(pPath) >= sizeof(pROUTESession->strShadowFilePath))
    {
        ERRORPRINT("%s: Shadow file path too long\n",
            pROUTESession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    strcpy(pROUTESession->strShadowFilePath, pPath);

out:
    return rval;
}

ENStatus
ENATSC3ROUTESessionRegisterMPDAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionMPDAvailableCallbackFunction MPDAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vMPDAvailableCallbacks); i++)
    {
        if (!pROUTESession->vMPDAvailableCallbacks[i].fValid)
        {
            struct _MPDAvailableCallbacks *pMPDAvailableCallbacks;

            pMPDAvailableCallbacks = pROUTESession->vMPDAvailableCallbacks + i;
            pMPDAvailableCallbacks->MPDAvailableCallback = MPDAvailableCallback;
            pMPDAvailableCallbacks->pMPDAvailableCallbackUserData = pUserData;
            pMPDAvailableCallbacks->fValid = 1;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionUnregisterMPDAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionMPDAvailableCallbackFunction MPDAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vMPDAvailableCallbacks); i++)
    {
        if (pROUTESession->vMPDAvailableCallbacks[i].fValid &&
            pROUTESession->vMPDAvailableCallbacks[i].MPDAvailableCallback ==
            MPDAvailableCallback &&
            pROUTESession->vMPDAvailableCallbacks[i].pMPDAvailableCallbackUserData == 
            pUserData)
        {
            pROUTESession->vMPDAvailableCallbacks[i].fValid  = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionRegisterHELDAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionHELDAvailableCallbackFunction HELDAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vHELDAvailableCallbacks); i++)
    {
        if (!pROUTESession->vHELDAvailableCallbacks[i].fValid)
        {
            struct _HELDAvailableCallbacks *pHELDAvailableCallbacks;

            pHELDAvailableCallbacks = 
                pROUTESession->vHELDAvailableCallbacks + i;
            pHELDAvailableCallbacks->HELDAvailableCallback = 
                HELDAvailableCallback;
            pHELDAvailableCallbacks->pHELDAvailableCallbackUserData = pUserData;
            pHELDAvailableCallbacks->fValid = 1;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionUnregisterHELDAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionHELDAvailableCallbackFunction HELDAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vHELDAvailableCallbacks); i++)
    {
        if (pROUTESession->vHELDAvailableCallbacks[i].fValid &&
            pROUTESession->vHELDAvailableCallbacks[i].HELDAvailableCallback == 
            HELDAvailableCallback &&
            pROUTESession->vHELDAvailableCallbacks[i].pHELDAvailableCallbackUserData == 
            pUserData)
        {
            pROUTESession->vHELDAvailableCallbacks[i].fValid = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionRegisterSTSIDAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionSTSIDAvailableCallbackFunction STSIDAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vSTSIDAvailableCallbacks); i++)
    {
        if (!pROUTESession->vSTSIDAvailableCallbacks[i].fValid)
        {
            struct _STSIDAvailableCallbacks *pSTSIDAvailableCallbacks;

            pSTSIDAvailableCallbacks = 
                pROUTESession->vSTSIDAvailableCallbacks + i;
            pSTSIDAvailableCallbacks->STSIDAvailableCallback = 
                STSIDAvailableCallback;
            pSTSIDAvailableCallbacks->pSTSIDAvailableCallbackUserData = 
                pUserData;
            pSTSIDAvailableCallbacks->fValid = 1;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionUnregisterSTSIDAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionSTSIDAvailableCallbackFunction STSIDAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vSTSIDAvailableCallbacks); i++)
    {
        if (pROUTESession->vSTSIDAvailableCallbacks[i].fValid &&
            pROUTESession->vSTSIDAvailableCallbacks[i].STSIDAvailableCallback == 
            STSIDAvailableCallback &&
            pROUTESession->vSTSIDAvailableCallbacks[i].pSTSIDAvailableCallbackUserData == 
            pUserData)
        {
            pROUTESession->vSTSIDAvailableCallbacks[i].fValid = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionRegisterSLSAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionSLSAvailableCallbackFunction SLSAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vSLSAvailableCallbacks); i++)
    {
        if (!pROUTESession->vSLSAvailableCallbacks[i].fValid)
        {
            struct _SLSAvailableCallbacks *pSLSAvailableCallbacks;

            pSLSAvailableCallbacks = pROUTESession->vSLSAvailableCallbacks + i;
            pSLSAvailableCallbacks->SLSAvailableCallback = SLSAvailableCallback;
            pSLSAvailableCallbacks->pSLSAvailableCallbackUserData = pUserData;
            pSLSAvailableCallbacks->fValid = 1;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionUnregisterSLSAvailableCallback(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionSLSAvailableCallbackFunction SLSAvailableCallback, 
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vSLSAvailableCallbacks); i++)
    {
        if (pROUTESession->vSLSAvailableCallbacks[i].fValid &&
            pROUTESession->vSLSAvailableCallbacks[i].SLSAvailableCallback == 
            SLSAvailableCallback &&
            pROUTESession->vSLSAvailableCallbacks[i].pSLSAvailableCallbackUserData == 
            pUserData)
        {
            pROUTESession->vSLSAvailableCallbacks[i].fValid = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus 
ENATSC3ROUTESessionRegisterLCTChannelCallbacks(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionLCTChannelAddedCallbackFunction LCTChannelAddedCallback,
    ENATSC3ROUTESessionLCTChannelRemovedCallbackFunction LCTChannelRemovedCallback,
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vLCTChannelCallbacks); i++)
    {
        if (!pROUTESession->vLCTChannelCallbacks[i].fValid)
        {
            struct _LCTChannelCallbacks *pLCTChannelCallbacks;

            pLCTChannelCallbacks = pROUTESession->vLCTChannelCallbacks + i;
            pLCTChannelCallbacks->LCTChannelAddedCallback = 
                LCTChannelAddedCallback;
            pLCTChannelCallbacks->LCTChannelRemovedCallback = 
                LCTChannelRemovedCallback;
            pLCTChannelCallbacks->pLCTChannelCallbackUserData = pUserData;
            pLCTChannelCallbacks->fValid = 1;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus 
ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionLCTChannelAddedCallbackFunction LCTChannelAddedCallback,
    ENATSC3ROUTESessionLCTChannelRemovedCallbackFunction LCTChannelRemovedCallback,
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    struct _LCTChannelCallbacks *pLCTChannelCallbacks;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0, pLCTChannelCallbacks = pROUTESession->vLCTChannelCallbacks; 
         i < numof(pROUTESession->vLCTChannelCallbacks); 
         i++, pLCTChannelCallbacks++)
    {
        if (pLCTChannelCallbacks->fValid &&
            pLCTChannelCallbacks->LCTChannelAddedCallback == 
            LCTChannelAddedCallback &&
            pLCTChannelCallbacks->LCTChannelRemovedCallback == 
            LCTChannelRemovedCallback &&
            pLCTChannelCallbacks->pLCTChannelCallbackUserData == 
            pUserData)
        {
            pLCTChannelCallbacks->fValid = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionRegisterObjectCallbacks(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionObjectDataCallbackFunction ObjectDataCallback,
    ENATSC3ROUTESessionSessionResetCallbackFunction SessionResetCallback,
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vObjectCallbacks); i++)
    {
        if (!pROUTESession->vObjectCallbacks[i].fValid)
        {
            pROUTESession->vObjectCallbacks[i].ObjectDataCallback = 
                ObjectDataCallback;
            pROUTESession->vObjectCallbacks[i].SessionResetCallback = 
                SessionResetCallback;
            pROUTESession->vObjectCallbacks[i].pObjectCallbackUserData = 
                pUserData;
            pROUTESession->vObjectCallbacks[i].fValid = 1;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionUnregisterObjectCallbacks(
    ENATSC3ROUTESessionHandle_t hROUTESession,
    ENATSC3ROUTESessionObjectDataCallbackFunction ObjectDataCallback,
    ENATSC3ROUTESessionSessionResetCallbackFunction SessionResetCallback,
    void *pUserData)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vObjectCallbacks); i++)
    {
        if (pROUTESession->vObjectCallbacks[i].fValid &&
            pROUTESession->vObjectCallbacks[i].ObjectDataCallback == 
                ObjectDataCallback &&
            pROUTESession->vObjectCallbacks[i].SessionResetCallback == 
                SessionResetCallback &&
            pROUTESession->vObjectCallbacks[i].pObjectCallbackUserData == 
                pUserData)
        {
            pROUTESession->vObjectCallbacks[i].fValid = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

static ENStatus
ROUTELCTChannelInfoFind(ENATSC3ROUTESession *pROUTESession, uint32_t tsi,
    LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pROUTESession->vLCTChannelInfos); i++)
    {
        if (pROUTESession->vLCTChannelInfos[i].fInUse &&
            pROUTESession->vLCTChannelInfos[i].tsi == tsi)
        {
            rval = ENStatus_Success;
            *ppLCTChannelInfo = pROUTESession->vLCTChannelInfos + i;
            break;
        }
    }

    return rval;
}

static void
ROUTECallLCTChannelAddedCallbacks(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession, uint32_t addr, 
    uint16_t port, uint32_t tsi, const char *pId, 
    ENATSC3ROUTESessionIdType idType)
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vLCTChannelCallbacks); i++)
    {
        if (pROUTESession->vLCTChannelCallbacks[i].fValid &&
            pROUTESession->vLCTChannelCallbacks[i].LCTChannelAddedCallback)
        {
            struct _LCTChannelCallbacks *pLCTChannelCallbacks;

            pLCTChannelCallbacks = pROUTESession->vLCTChannelCallbacks + i;
            pLCTChannelCallbacks->LCTChannelAddedCallback(pPacketTimeUTC,
                packetNumber, addr, port, tsi, pId, idType, 
                pLCTChannelCallbacks->pLCTChannelCallbackUserData);
        }
    }

    return;
}

static void
ROUTECallLCTChannelRemovedCallbacks(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession, uint32_t addr, 
    uint16_t port, uint32_t tsi)
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vLCTChannelCallbacks); i++)
    {
        if (pROUTESession->vLCTChannelCallbacks[i].fValid &&
            pROUTESession->vLCTChannelCallbacks[i].LCTChannelRemovedCallback)
        {
            struct _LCTChannelCallbacks *pLCTChannelCallbacks;

            pLCTChannelCallbacks = pROUTESession->vLCTChannelCallbacks + i;
            pLCTChannelCallbacks->LCTChannelRemovedCallback(pPacketTimeUTC,
                packetNumber, addr, port, tsi, 
                pLCTChannelCallbacks->pLCTChannelCallbackUserData);
        }
    }

    return;
}

static ENStatus
ROUTELCTChannelInfoAllocate(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession, uint32_t tsi, 
    char *pId, ENATSC3ROUTESessionIdType idType, enum LCTObjectType objectType, 
    uint32_t addr, uint16_t port, LCTChannelInfo **ppLCTChannelInfo)
{
    LCTChannelInfo *pLCTChannelInfo;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    *ppLCTChannelInfo = NULL;

    // If the destination address and port are not specified then they default
    // to the SLS address and port. A/331-2021 Table 7.2. (Although, A/331-2021
    // also says that the destination address and port are mandatory for 
    // ROUTE sessions not carrying SLS. Hoever some emissions do not have the
    // the destination port and address in the S-TSID.
    if (addr == 0)
    {
        addr = pROUTESession->slsAddr;
    }

    if (port == 0)
    {
        port = pROUTESession->slsPort;
    }

    if (pROUTESession->verbosity >= ENVerbosity_StreamEventTrace)
    {
        LOGPRINT_NOFILE(
            "%s: Allocate LCT channel for tsi %d id %s (%d.%d.%d.%d:%d)\n", 
            pROUTESession->strDebugName, tsi, pId, (addr >> 24) & 0xff, 
            (addr >> 16) & 0xff, (addr >> 8) & 0xff, addr & 0xff, port);
    }

    if (strlen(pId) >= sizeof(pROUTESession->vLCTChannelInfos[0].strId))
    {
        ERRORPRINT("%s: Representation ID %s too long\n", 
            pROUTESession->strDebugName, pId);
        rval = ENStatus_Failure;
        goto out;
    }

    for (i = 0, pLCTChannelInfo = pROUTESession->vLCTChannelInfos; 
         i < numof(pROUTESession->vLCTChannelInfos); 
         i++, pLCTChannelInfo++)
    {
        if (!pLCTChannelInfo->fInUse)
        {
            pLCTChannelInfo->tsi = tsi;
            pLCTChannelInfo->objectType = objectType;
            pLCTChannelInfo->expectedSLSTOI = -1;
	    pLCTChannelInfo->lastEFDTcrc32c = 0;
	    pLCTChannelInfo->lastEFDTSize = -1;
            pLCTChannelInfo->strFileTemplate[0] = 0;
            pLCTChannelInfo->strContentType[0] = 0;
            pLCTChannelInfo->nFiles = 0;
            pLCTChannelInfo->nPayloadDescriptions = 0;
            pLCTChannelInfo->pBuffer = NULL;
            pLCTChannelInfo->sBuffer = 0;
            strcpy(pLCTChannelInfo->strId, pId);
            pLCTChannelInfo->idType = idType;
            pLCTChannelInfo->addr = addr;
            pLCTChannelInfo->port = port;
            ENATSC3BitrateStatisticsInitialize(&pLCTChannelInfo->stats, 2000); 
            snprintf(pLCTChannelInfo->stats.strName, 
                sizeof(pLCTChannelInfo->stats.strName), "LCT tsi %d", tsi);
            pLCTChannelInfo->stats.strName[
                sizeof(pLCTChannelInfo->stats.strName) -1 ] = 0;

            pLCTChannelInfo->fInUse = 1;
            pLCTChannelInfo->fValid = 1;
            *ppLCTChannelInfo = pLCTChannelInfo;

            if (addr)
            {
                if (pROUTESession->IPAddressAddCallback)
                {
                    pROUTESession->IPAddressAddCallback(addr, port, 
                        pROUTESession->pCallbackUserData);
                }

                ROUTECallLCTChannelAddedCallbacks(pPacketTimeUTC, packetNumber, 
                    pROUTESession, addr, port, tsi, pId, idType);
            }

            rval = ENStatus_Success;
            break;
        }
    }

    if (i == numof(pROUTESession->vLCTChannelInfos))
    {
        ERRORPRINT("%s: All LCT Channels in use\n", 
            pROUTESession->strDebugName);
        for (i = 0; i < numof(pROUTESession->vLCTChannelInfos); i++)
        {
            ERRORPRINT("%s:   %d: id %s tsi %u\n",
                pROUTESession->strDebugName, i,
                pROUTESession->vLCTChannelInfos[i].strId,
                pROUTESession->vLCTChannelInfos[i].tsi);
        }
    }

out:
    return rval;
}

static ENStatus
ROUTELCTChannelInfoReturn(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession, 
    LCTChannelInfo *pLCTChannelInfo)
{
    ENStatus rval = ENStatus_Success;

    if (pLCTChannelInfo->addr)
    {
        if (pROUTESession->IPAddressRemoveCallback)
        {
            pROUTESession->IPAddressRemoveCallback(pLCTChannelInfo->addr, 
                pLCTChannelInfo->port, pROUTESession->pCallbackUserData);
        } 

        // Don't call the callback for tsi 0, since tsi 0 is implicitly managed
        // by the ROUTE layer.
        if (pLCTChannelInfo->tsi)
        {
            ROUTECallLCTChannelRemovedCallbacks(pPacketTimeUTC, packetNumber, 
                pROUTESession, pLCTChannelInfo->addr, pLCTChannelInfo->port, 
                pLCTChannelInfo->tsi);
        }
    }

    if (pLCTChannelInfo->pBuffer)
    {
        free(pLCTChannelInfo->pBuffer);
        pLCTChannelInfo->pBuffer = NULL;
    }
    pLCTChannelInfo->fInUse = 0;
    pLCTChannelInfo->fValid = 0;

    return rval;
}

static void
ROUTELCTChannelInfoChangeAddress(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession, uint32_t addr, 
    uint16_t port, LCTChannelInfo *pLCTChannelInfo)
{
    if (pLCTChannelInfo->addr)
    {
        if (pROUTESession->IPAddressRemoveCallback)
        {
            pROUTESession->IPAddressRemoveCallback(pLCTChannelInfo->addr, 
                pLCTChannelInfo->port, pROUTESession->pCallbackUserData);
        }

        ROUTECallLCTChannelRemovedCallbacks(pPacketTimeUTC, packetNumber, 
            pROUTESession, pLCTChannelInfo->addr, pLCTChannelInfo->port, 
            pLCTChannelInfo->tsi);
    }

    pLCTChannelInfo->addr = addr;
    pLCTChannelInfo->port = port;

    if (pLCTChannelInfo->addr)
    {
        if (pROUTESession->IPAddressAddCallback)
        {
            pROUTESession->IPAddressAddCallback(pLCTChannelInfo->addr, 
                pLCTChannelInfo->port, pROUTESession->pCallbackUserData);
        }

        ROUTECallLCTChannelAddedCallbacks(pPacketTimeUTC, packetNumber, 
            pROUTESession, pLCTChannelInfo->addr, pLCTChannelInfo->port, 
            pLCTChannelInfo->tsi, pLCTChannelInfo->strId, 
            pLCTChannelInfo->idType);
    }
}

static void
ROUTELCTChannelInfoMarkAllInvalid(ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vLCTChannelInfos); i++)
    {
        pROUTESession->vLCTChannelInfos[i].fValid = 0;
    }
}

static void
ROUTELCTChannelInfoReturnAllInvalidNonSLS(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vLCTChannelInfos); i++)
    {
        if (pROUTESession->vLCTChannelInfos[i].fInUse &&
            !pROUTESession->vLCTChannelInfos[i].fValid &&
            pROUTESession->vLCTChannelInfos[i].objectType != LCTObjectType_SLS)
        {
            if (ROUTELCTChannelInfoReturn(pPacketTimeUTC, packetNumber, 
                pROUTESession, pROUTESession->vLCTChannelInfos + i))
            {
                ERRORPRINT("%s: Failed to return LCT Channel\n",
                    pROUTESession->strDebugName);
            }
        }
    }
}

static void
ROUTELCTChannelInfoReturnAll(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vLCTChannelInfos); i++)
    {
        if (pROUTESession->vLCTChannelInfos[i].fInUse)
        {
            if (ROUTELCTChannelInfoReturn(pPacketTimeUTC, packetNumber, 
                pROUTESession, pROUTESession->vLCTChannelInfos + i))
            {
                ERRORPRINT("%s: Failed to return LCT Channel\n",
                    pROUTESession->strDebugName);
            }
        }
    }
}

static ENStatus
ROUTECallMPDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pDocument, int32_t sDocument, 
    int32_t version, uint32_t crc32c, ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < numof(pROUTESession->vMPDAvailableCallbacks); i++)
    {
        if (pROUTESession->vMPDAvailableCallbacks[i].fValid &&
            pROUTESession->vMPDAvailableCallbacks[i].MPDAvailableCallback)
        {
            struct _MPDAvailableCallbacks *pMPDAvailableCallbacks;

            pMPDAvailableCallbacks = pROUTESession->vMPDAvailableCallbacks + i;
            if (pMPDAvailableCallbacks->MPDAvailableCallback(pPacketTimeUTC, 
                packetNumber, pDocument, sDocument, version, crc32c, 
                pMPDAvailableCallbacks->pMPDAvailableCallbackUserData))
            {
                rval = ENStatus_Failure;
            }
        }
    }

    return rval;
}

static void
ROUTECallSLSAvailableCallbacks(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint8_t plpId, char *pURL, 
    ENATSC3ROUTELCTHeader *pLCTHeader, LCTChannelInfo *pLCTChannelInfo, 
    int32_t version, uint32_t crc32c, ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vSLSAvailableCallbacks); i++)
    {
        if (pROUTESession->vSLSAvailableCallbacks[i].fValid &&
            pROUTESession->vSLSAvailableCallbacks[i].SLSAvailableCallback)
        {
            struct _SLSAvailableCallbacks *pSLSAvailableCallbacks;

            pSLSAvailableCallbacks = pROUTESession->vSLSAvailableCallbacks + i;
            pSLSAvailableCallbacks->SLSAvailableCallback(pPacketTimeUTC,
                packetNumber, plpId, pLCTHeader->transportObjectIdentifier, 
                pURL, pLCTChannelInfo->pBuffer, pLCTChannelInfo->slsObjectSize, 
                pLCTHeader->transportObjectIdentifier & 0x80000000,
                version, crc32c, 
                pSLSAvailableCallbacks->pSLSAvailableCallbackUserData);
        }
    }
}

static ENStatus
ROUTECallHELDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pDocument, int32_t sDocument, 
    int32_t version, uint32_t crc32c, ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < numof(pROUTESession->vHELDAvailableCallbacks); i++)
    {
        if (pROUTESession->vHELDAvailableCallbacks[i].fValid &&
            pROUTESession->vHELDAvailableCallbacks[i].HELDAvailableCallback)
        {
            struct _HELDAvailableCallbacks *pHELDAvailableCallbacks;

            pHELDAvailableCallbacks = 
                pROUTESession->vHELDAvailableCallbacks + i;
            if (pHELDAvailableCallbacks->HELDAvailableCallback(pPacketTimeUTC, 
                packetNumber, pDocument, sDocument, version, crc32c, 
                pHELDAvailableCallbacks->pHELDAvailableCallbackUserData))
            {
                rval = ENStatus_Failure;
            }
        }
    }

    return rval;
}

static ENStatus
ROUTECallSTSIDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pDocument, int32_t sDocument, 
    ENATSC3ROUTESLSSTSIDNode *pParsedSTSID, int32_t version, uint32_t crc32c, 
    ENATSC3ROUTESession *pROUTESession)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < numof(pROUTESession->vSTSIDAvailableCallbacks); i++)
    {
        if (pROUTESession->vSTSIDAvailableCallbacks[i].fValid &&
            pROUTESession->vSTSIDAvailableCallbacks[i].STSIDAvailableCallback)
        {
            struct _STSIDAvailableCallbacks *pSTSIDAvailableCallbacks;

            pSTSIDAvailableCallbacks = 
                pROUTESession->vSTSIDAvailableCallbacks + i;
            if (pSTSIDAvailableCallbacks->STSIDAvailableCallback(
                pPacketTimeUTC, packetNumber, pDocument, sDocument, 
                pParsedSTSID, version, crc32c,  
                pSTSIDAvailableCallbacks->pSTSIDAvailableCallbackUserData))
            {
                rval = ENStatus_Failure;
            }
        }
    }

    return rval;
}

static inline ENStatus
ROUTEGetValuesFromSTSIDPayloadNode(
    ENATSC3ROUTESLSSTSIDPayloadNode *pPayloadNode, uint8_t *pCodePoint, 
    uint8_t *pFormatId, uint8_t *pFrag, uint8_t *pOrder, 
    uint8_t *pSrcFecPayloadId)
{
    ENStatus rval = ENStatus_Success;

    if (!(pPayloadNode->base.presentFlags & 
          ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_codePoint) ||
        !(pPayloadNode->base.presentFlags & pPayloadNode->base.presentFlags))
    {
        ERRORPRINT("Payload description missing codepoint or formatId\n");
        rval = ENStatus_Failure;
        goto out;
    }

    *pCodePoint = pPayloadNode->codePoint;
    *pFormatId = pPayloadNode->formatId;

    if (pPayloadNode->base.presentFlags & 
        ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_frag)
    { 
        *pFrag = pPayloadNode->frag;
    }
    else
    {
        *pFrag = 0;
    }

    if (pPayloadNode->base.presentFlags & 
        ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_order)
    { 
        *pOrder = pPayloadNode->order;
    }
    else
    {
        *pOrder = 0;
    }

    if (pPayloadNode->base.presentFlags & 
        ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_srcFecPayloadID)
    { 
        *pSrcFecPayloadId = pPayloadNode->srcFecPayloadID;
    }
    else
    {
        *pSrcFecPayloadId = 0;
    }

out:
    return rval;
}

// This is a fairly loose check for the content-type. Strictly speaking there
// should be no spaces in the content-type value, but in practice white space
// tends to appear after and sometimes before the semicolon separating the
// content-type from its arguments.
#define IsContentType(String, Literal)                          \
  (!strncasecmp(String, Literal, sizeof(Literal) - 1) &&        \
   (String[sizeof(Literal) - 1] == 0 ||                         \
    String[sizeof(Literal) - 1] == ' ' ||                       \
    String[sizeof(Literal) - 1] == ';'))

static void
ROUTEProcessParsedSTSID(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode, ENATSC3ROUTESession *pROUTESession)
{
    char strSegment0URL[REP_ID_MAX_LENGTH];
    ENATSC3ROUTESessionIdType idType = ENATSC3ROUTESessionIdType_Unknown;
    int32_t i;
    int32_t j;
    int32_t k;
    int32_t l;

    // Mark all of the LCT channels associated with the route session as 
    // invalid. As we process the S-TSID we will mark sessions it contains 
    // as valid. After processing the S-TSID any LCT channels which are 
    // still marked as invalid will be freed.
    ROUTELCTChannelInfoMarkAllInvalid(pROUTESession);

    for (i = 0; i < pSTSIDNode->RSs.nOffsetsUsed; i++)
    {
        ENATSC3ROUTESLSSTSIDRSNode *pRS;
        ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
        char *pFileTemplate = NULL;
        uint32_t addr = 0;
        uint16_t port = 0;
        int32_t nPayloadDescriptions = 0;

        pRS = 
            (ENATSC3ROUTESLSSTSIDRSNode *)EnXMLParserDataOffsetArrayGetPointer(
                &pSTSIDNode->RSs, i);
        if (pRS->base.presentFlags & 
            ENATSC3ROUTESLSSTSIDRSNodeAttribute_dIpAddr) 
        {
            char *pAddr;
            struct in_addr in_addr;
            pAddr = EnXMLParserDataBufferGetPointer(&pSTSIDNode->dataBuffer, 
                pRS->dIpAddrDataOffset);
            if (inet_aton(pAddr, &in_addr))
            {
                addr = htonl(in_addr.s_addr);
            }
            else
            {
                ERRORPRINT("%s: Failed to parse IP addr %s\n", 
                    pROUTESession->strDebugName, pAddr);
            }
        }
        else
        {
            // If the destination address is not in the S-TSID default it to 
            // the ROUTE session carrying SLS. A221-2021 Table 7.2.
            addr = pROUTESession->slsAddr;
        }

        if (pRS->base.presentFlags & ENATSC3ROUTESLSSTSIDRSNodeAttribute_dPort)
        {
            port = pRS->dPort;
        }
        else
        {
            // If the destination port is not in the S-TSID default it to the 
            // ROUTE session carrying SLS. A221-2021 Table 7.2.
            port = pROUTESession->slsPort;
        }

        for (j = 0; j < pRS->LSs.nOffsetsUsed; j++)
        {
            ENATSC3ROUTESLSSTSIDLSNode *pLS;
            LCTChannelInfo *pLCTChannelInfo = NULL;
            char *pContentInfoContentType = "";
            char *pId = NULL;

            pLS = (ENATSC3ROUTESLSSTSIDLSNode *)
                EnXMLParserDataOffsetArrayGetPointer(&pRS->LSs, j);

            // Make sure we have:
            // - a SrcFlow and tsi
            if (!(pLS->base.presentFlags & 
                  ENATSC3ROUTESLSSTSIDLSChildNode_SrcFlow) ||
                !(pLS->base.presentFlags & 
                  ENATSC3ROUTESLSSTSIDLSNodeAttribute_tsi))
            {
                ERRORPRINT("%s: WARN STSID LS missing elements: SrcFlow: %d tsi: %d\n", 
                    pROUTESession->strDebugName, 
                    pLS->base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDLSChildNode_SrcFlow,
                    pLS->base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDLSNodeAttribute_tsi);
                continue;
            }

            // tsi == 0 is reserved for signaling. It cannot appear in a 
            // SrcFlow.
            if (pLS->tsi == 0)
            {
                ERRORPRINT("%s Invalid tsi 0 for SrcFlow\n",
                    pROUTESession->strDebugName);
                continue;
            }

            // Check if the SrcFlow contains an EFDT
            // XXX The EFDT may be delivered instead as a FLUTE FDT-Instance
            if (!(pLS->srcFlow.base.presentFlags & 
                ENATSC3ROUTESLSSTSIDSrcFlowChildNode_EFDT))
            {
                ERRORPRINT("%s: WARN STSID SrcFlow missing EFDT\n", 
                    pROUTESession->strDebugName);
                ERRORPRINT("%s: FIXME - Support EFDT in TOI=0 FDT-Instance\n", 
                    pROUTESession->strDebugName);
                continue;
            }

            if (pROUTESession->type == ENATSC3ROUTESessionType_ATSC3_DASH)
            {
                // Check if:
                // - The SrcFlow has a ContentInfo
                // - The ContentInfo has a MediaInfo
                // - The MediaInfo has a repId
                if (!(pLS->srcFlow.base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDSrcFlowChildNode_ContentInfo) || 
                    !(pLS->srcFlow.contentInfo.base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDContentInfoChildNode_MediaInfo) ||
                    !(pLS->srcFlow.contentInfo.mediaInfo.base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_repId))
                {
                    if (pLS->base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDLSChildNode_SrcFlow)
                    {
                        ERRORPRINT("%s: WARN STSID: no repId to map LCT Channel %d to DASH representation: SrcFlow:ContentInfo: %d, ContentInfo:MediaInfo: %d\n",
                            pROUTESession->strDebugName, pLS->tsi, 
                            pLS->srcFlow.base.presentFlags & 
                                ENATSC3ROUTESLSSTSIDSrcFlowChildNode_ContentInfo,
                            pLS->srcFlow.contentInfo.base.presentFlags & 
                                ENATSC3ROUTESLSSTSIDContentInfoChildNode_MediaInfo);
                    }
                    if (pLS->srcFlow.contentInfo.base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDContentInfoChildNode_MediaInfo)
                    {
                        ERRORPRINT("%s:     MediaInfo:repId: %d\n",
                            pROUTESession->strDebugName, 
                            pLS->srcFlow.contentInfo.mediaInfo.base.presentFlags & 
                                ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_repId);
                    }
                }
                else
                {
                    pId = EnXMLParserDataBufferGetPointer(
                        &pSTSIDNode->dataBuffer, 
                        pLS->srcFlow.contentInfo.mediaInfo.repIdDataOffset);
                    idType = ENATSC3ROUTESessionIdType_RepId;
                    if (pROUTESession->verbosity >= ENVerbosity_Terse)
                    {
                        LOGPRINT("%s: Using repId %s for tsi %d\n", 
                            pROUTESession->strDebugName, pId, pLS->tsi);
                    }
                }
            }

            if ((pLS->srcFlow.base.presentFlags & 
                 ENATSC3ROUTESLSSTSIDSrcFlowChildNode_ContentInfo) &&
                (pLS->srcFlow.contentInfo.base.presentFlags & 
                 ENATSC3ROUTESLSSTSIDContentInfoChildNode_MediaInfo) &&
                (pLS->srcFlow.contentInfo.mediaInfo.base.presentFlags &
                 ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_contentType))
            {
                pContentInfoContentType = EnXMLParserDataBufferGetPointer(
                    &pSTSIDNode->dataBuffer, 
                    pLS->srcFlow.contentInfo.mediaInfo.contentTypeDataOffset);
            }

            pEFDTNode = &pLS->srcFlow.efdt;

            // Make sure we have a FDT-Instance
            if (!(pEFDTNode->base.presentFlags &
                ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTInstance))
            {
                // Hmmm...the Sinclair CES broadcast seems to have FileTemplate
                // and FDTParameters in place of FDT-Instance
                if (!(pEFDTNode->base.presentFlags & 
                    ENATSC3ROUTESLSSTSIDEFDTChildNode_FileTemplate))
                {
                    ERRORPRINT("%s: Sinclair CES 2019 Hack: EFDT doesn't have fileTemplate\n",
                        pROUTESession->strDebugName);
                    continue;
                }
                pFileTemplate = EnXMLParserDataBufferGetPointer(
                    &pSTSIDNode->dataBuffer,
                    pEFDTNode->fileTemplate.fileTemplateDataOffset);
            }
            else
            {
                if (!(pEFDTNode->fdtInstance.base.presentFlags & 
                    ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_fileTemplate))
                {
                    pFileTemplate = "";
                }
                else
                {
                    pFileTemplate = EnXMLParserDataBufferGetPointer(
                        &pSTSIDNode->dataBuffer, 
                        pEFDTNode->fdtInstance.fileTemplateDataOffset);
                }
            }

            if (pROUTESession->type == ENATSC3ROUTESessionType_ATSC3_DASH)
            {
                if (!pId)
                {
                    if (GenerateObjectURLFromTemplate(pROUTESession, 
                        pFileTemplate, strSegment0URL, sizeof(strSegment0URL), 
                        0))
                    {
                        ERRORPRINT("%s: Segment 0 URL too long\n",
                            pROUTESession->strDebugName);
                        continue;
                    }
                    pId = strSegment0URL;
                    idType = ENATSC3ROUTESessionIdType_Segment0URL;
                    ERRORPRINT("%s: WARN: Using Segment 0 URL %s in place of DASH Representation Id to map LCT channel tsi %d\n", 
                        pROUTESession->strDebugName, pId, pLS->tsi);
                }
            }
            else
            {
                pId = "";
            }
 
            // Check if the SrcFlow contains a Payload
            if (pLS->srcFlow.base.presentFlags & 
                ENATSC3ROUTESLSSTSIDSrcFlowChildNode_Payload)
            {
                nPayloadDescriptions = pLS->srcFlow.payloads.nOffsetsUsed;
            }
            else
            {
                nPayloadDescriptions = 0;
            }

            // See if the LCT channel already exists. If it does, and none of
            // its properties have changed, then we leave it alone. However if
            // a characteristic has changed then we treat it as a new channel
            // 
            // Note if the address and port change we don't require a new 
            // channel. We just need to ensure we listen on the new address
            // and port.
            // (releasing the current channel for the tsi).
            if (!ROUTELCTChannelInfoFind(pROUTESession, pLS->tsi,
                &pLCTChannelInfo))
            {
                int32_t fChanged;

                fChanged =  
                    pLCTChannelInfo->objectType != LCTObjectType_MediaFile ||
                    pLCTChannelInfo->idType != idType ||
                    strcmp(pLCTChannelInfo->strId, pId) ||
                    strcmp(pLCTChannelInfo->strFileTemplate, pFileTemplate) ||
                    strcmp(pLCTChannelInfo->strContentType, pContentInfoContentType) ||
                    pLCTChannelInfo->nPayloadDescriptions != nPayloadDescriptions;
                if (!fChanged)
                {
                    for (k = 0; k < nPayloadDescriptions; k++)
                    {
                        ENATSC3ROUTESLSSTSIDPayloadNode *pPayloadNode;
                        uint8_t codepoint;
                        uint8_t formatId;
                        uint8_t frag = 0;
                        uint8_t order = 1;
                        uint8_t srcFecPayloadId = 0;

                        pPayloadNode = (ENATSC3ROUTESLSSTSIDPayloadNode *)EnXMLParserDataOffsetArrayGetPointer(&pLS->srcFlow.payloads, k);

                        if (ROUTEGetValuesFromSTSIDPayloadNode(pPayloadNode, 
                            &codepoint, &formatId, &frag, &order, 
                            &srcFecPayloadId))
                        {
                            ERRORPRINT("Payload description missing codepoint or formatId\n");
                            continue;
                        }

                        for (l = 0; l < nPayloadDescriptions; l++)
                        {
                            if (pLCTChannelInfo->vPayloadDescriptions[l].codepoint == codepoint && 
                                pLCTChannelInfo->vPayloadDescriptions[l].order == order && 
                                pLCTChannelInfo->vPayloadDescriptions[l].formatId == formatId && 
                                pLCTChannelInfo->vPayloadDescriptions[l].frag == frag && 
                                pLCTChannelInfo->vPayloadDescriptions[l].srcFecPayloadId == srcFecPayloadId)
                            {
                                break;
                            }
                        }

                        if (i == nPayloadDescriptions)
                        {
                            fChanged = 1;
                            break;
                        }
                    }
                }
                if (fChanged)
                {
                    LOGPRINT("%s: LCT Channel tsi %d changed (%s, %s, %s, %d, %d, %d) -> (%s, %s, %s, %d, %d, %d)\n",
                        pROUTESession->strDebugName,
                        pLS->tsi,
                        pLCTChannelInfo->strId,
                        pLCTChannelInfo->strFileTemplate,
                        pLCTChannelInfo->strContentType, 
                        pLCTChannelInfo->objectType,
                        pLCTChannelInfo->idType,
                        pLCTChannelInfo->nPayloadDescriptions,
                        pId,
                        pFileTemplate,
                        pContentInfoContentType, 
                        LCTObjectType_MediaFile,
                        idType,
                        nPayloadDescriptions);
                    if (ROUTELCTChannelInfoReturn(pPacketTimeUTC, packetNumber, 
                        pROUTESession, pLCTChannelInfo))
                    {
                        ERRORPRINT("%s: Failed to release LCT channel for %d\n",
                            pROUTESession->strDebugName, pLS->tsi);
                    }
                    pLCTChannelInfo = NULL;
                }
                else
                {
                    if (pLCTChannelInfo->addr != addr || 
                        pLCTChannelInfo->port != port)
                    {
                        ROUTELCTChannelInfoChangeAddress(pPacketTimeUTC, 
                            packetNumber, pROUTESession, addr, port, 
                            pLCTChannelInfo);
                    }
                    pLCTChannelInfo->fValid = 1;
                }
            }

            if (!pLCTChannelInfo)
            {
                if (ROUTELCTChannelInfoAllocate(pPacketTimeUTC, packetNumber, 
                    pROUTESession, pLS->tsi, pId, idType, 
                    LCTObjectType_MediaFile, addr, port, &pLCTChannelInfo))
                {
                    ERRORPRINT("%s: Failed to allocate LCTChannelInfo\n",
                        pROUTESession->strDebugName);
                    continue;
                }

                if (strlen(pFileTemplate) >= 
                    sizeof(pLCTChannelInfo->strFileTemplate))
                {
                    ERRORPRINT("%s: File template too long\n",
                        pROUTESession->strDebugName);
                    if (ROUTELCTChannelInfoReturn(pPacketTimeUTC, packetNumber, 
                        pROUTESession, pLCTChannelInfo))
                    {
                        ERRORPRINT("%s: Failed to return LCTChannelInfo\n",
                            pROUTESession->strDebugName);
                    }
                    continue;
                }
                strcpy(pLCTChannelInfo->strFileTemplate, pFileTemplate); 

                if (strlen(pContentInfoContentType) >= 
                    sizeof(pLCTChannelInfo->strContentType))
                {
                    ERRORPRINT("%s: Content-Type too long\n", 
                        pROUTESession->strDebugName);
                    if (ROUTELCTChannelInfoReturn(pPacketTimeUTC, packetNumber, 
                        pROUTESession, pLCTChannelInfo))
                    {
                        ERRORPRINT("%s: Failed to return LCTChannelInfo\n",
                            pROUTESession->strDebugName);
                    }
                    continue;
                }
                strcpy(pLCTChannelInfo->strContentType, 
                    pContentInfoContentType); 

                if (pLS->srcFlow.base.presentFlags & 
                    ENATSC3ROUTESLSSTSIDSrcFlowChildNode_Payload)
                {
                    for (k = 0; k < pLS->srcFlow.payloads.nOffsetsUsed; k++)
                    {
                        ENATSC3ROUTESLSSTSIDPayloadNode *pPayloadNode;
                        struct _LCTChannelInfoPayloadDescription *pPayloadDescription;

                        if (pLCTChannelInfo->nPayloadDescriptions ==
                            numof(pLCTChannelInfo->vPayloadDescriptions))
                        {
                            ERRORPRINT("Too many payload descriotions\n");
                            break;
                        }

                        pPayloadDescription = 
                            pLCTChannelInfo->vPayloadDescriptions +
                            pLCTChannelInfo->nPayloadDescriptions;

                        pPayloadNode = (ENATSC3ROUTESLSSTSIDPayloadNode *)EnXMLParserDataOffsetArrayGetPointer(&pLS->srcFlow.payloads, k);

                        if (ROUTEGetValuesFromSTSIDPayloadNode(pPayloadNode, 
                            &pPayloadDescription->codepoint, 
                            &pPayloadDescription->formatId, 
                            &pPayloadDescription->frag, 
                            &pPayloadDescription->order, 
                            &pPayloadDescription->srcFecPayloadId))
                        {
                            ERRORPRINT("Payload description missing codepoint or formatId\n");
                            continue;
                        }
                        pLCTChannelInfo->nPayloadDescriptions++;
                    }       
                }
            }

            if (pEFDTNode->base.presentFlags &
                (ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTParameters |
                ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTInstance))
            {
                EnXMLParserDataOffsetArray *pFileArray = NULL;

                if (pEFDTNode->base.presentFlags & 
                    ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTParameters)
                {
                    pFileArray = &pEFDTNode->fdtParameters.files;
                }
                else if (pEFDTNode->base.presentFlags &
                    ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTInstance)
                {
                    pFileArray = &pEFDTNode->fdtInstance.files;
                }

                pLCTChannelInfo->nFiles = 0;
                pLCTChannelInfo->vFiles[0].strContentLocation[0] = 0;
                for (k = 0; k < pFileArray->nOffsetsUsed; k++)
                {
                    char *pContentLocation = NULL;
                    ENATSC3ROUTESLSSTSIDFileNode *pFileNode;
                    char *pContentType = pContentInfoContentType;
                    struct _FDTInstanceFiles *pFile;

                    pFileNode = EnXMLParserDataOffsetArrayGetPointer(
                        pFileArray, k);

                    if (!(pFileNode->base.presentFlags &
                          ENATSC3ROUTESLSSTSIDFileNodeAttribute_TOI) ||
                        !(pFileNode->base.presentFlags &
                          ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentLocation))
                    {
                        ERRORPRINT("%s: Incomplete fdt:File node\n",
                            pROUTESession->strDebugName);
                        continue;
                    }

                    pContentLocation = EnXMLParserDataBufferGetPointer(
                        &pSTSIDNode->dataBuffer,
                        pFileNode->contentLocationDataOffset);

                    if (pFileNode->base.presentFlags &
                        ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentType)
                    {
                        pContentType = EnXMLParserDataBufferGetPointer(
                            &pSTSIDNode->dataBuffer, 
                            pFileNode->contentTypeDataOffset);
                    }

                    if (strlen(pContentLocation) > 
                        sizeof(pLCTChannelInfo->vFiles[0].strContentLocation))
                    {
                        ERRORPRINT("%s: Content-Location too long\n",
                            pROUTESession->strDebugName);
                        continue;
                    }
                    if (strlen(pContentType) > 
                        sizeof(pLCTChannelInfo->vFiles[0].strContentType))
                    {
                        ERRORPRINT("%s: Content-Type too long\n", 
                            pROUTESession->strDebugName);
                        continue;
                    }

                    if (pLCTChannelInfo->nFiles ==  
                        numof(pLCTChannelInfo->vFiles))
                    {
                        ERRORPRINT("%s: Too many fdt:Files\n",
                            pROUTESession->strDebugName);
                        continue;
                    }

                    pFile = pLCTChannelInfo->vFiles + pLCTChannelInfo->nFiles;

                    strcpy(pFile->strContentLocation, pContentLocation);
                    strcpy(pFile->strContentType, pContentType);
                    pFile->toi = pFileNode->toi;
                    pLCTChannelInfo->nFiles++;
                }
            }
        }
    }

    if (pROUTESession->IPAddressCommitCallback)
    {
        pROUTESession->IPAddressCommitCallback(
            pROUTESession->pCallbackUserData);
    }

    ROUTELCTChannelInfoReturnAllInvalidNonSLS(pPacketTimeUTC, packetNumber, 
        pROUTESession);

    return;
}

typedef struct
{
    struct timespec *pPacketTimeUTC;
    ENATSC3ROUTESession *pROUTESession;
    uint64_t packetNumber;
    uint32_t version;
    uint32_t crc32c;
    ENATSC3ROUTELCTHeader *pLCTHeader;
    LCTChannelInfo *pLCTChannelInfo;
    ENATSC3ROUTESLSEnvelopeNode envelopeNode;
    int32_t fHaveEnvelopeNode;
    struct
    {
        const char *pMIME;
        int32_t sMIME;
    } signedData;
} ROUTEMIMECallbackUserData;


static int32_t
ROUTEMIMEGetRelatedDocumentVersion(
    ROUTEMIMECallbackUserData *pCallbackUserData, char *pContentLocation)
{
    int32_t i;
    int32_t version = -1;

    if (!pCallbackUserData->fHaveEnvelopeNode)
    {
        ERRORPRINT("%s: Received MIME part %s before envelope\n", 
            pCallbackUserData->pROUTESession->strDebugName, pContentLocation);
        goto out;
    }

    for (i = 0; 
         i < pCallbackUserData->envelopeNode.items.nOffsetsUsed; 
         i++)
    {
        ENATSC3ROUTESLSEnvelopeItemNode *pItemNode;

        pItemNode = EnXMLParserDataOffsetArrayGetPointer(
            &pCallbackUserData->envelopeNode.items, i);
        if ((pItemNode->base.presentFlags & 
             ENATSC3ROUTESLSEnvelopeItemNodeAttribute_metadataURI))
        {
            char *pMetadataURI;

            pMetadataURI = EnXMLParserDataBufferGetPointer(
                &pCallbackUserData->envelopeNode.dataBuffer,
                pItemNode->metadataURIDataOffset);
            if (!strcmp(pMetadataURI, pContentLocation))
            {
                if (!(pItemNode->base.presentFlags &
                     ENATSC3ROUTESLSEnvelopeItemNodeAttribute_version))
                {
                    ERRORPRINT("%s: No version for %s\n", 
                        pCallbackUserData->pROUTESession->strDebugName, 
                        pContentLocation);
                }
                else
                {
                    version = pItemNode->version;
                }
                break;
            }
        }
    }

out:
    return version;
}

static int32_t
ROUTEMIMERelatedDocumentAvailableCallback(const char *pDocument, 
    int32_t sDocument, const char *pPart, int32_t sPart, char *pContentType, 
    char *pContentLocation, void *pUserData)
{
    ROUTEMIMECallbackUserData *pCallbackUserData;
    int rval = 0;

    pCallbackUserData = (ROUTEMIMECallbackUserData *)pUserData;

    if (pCallbackUserData->pROUTESession->verbosity >= 
        ENVerbosity_StreamEventTrace)
    {
        LOGPRINT("%s: Got MIME multipart/related document %s\n", 
            pCallbackUserData->pROUTESession->strDebugName, pContentType);
    }

    if (IsContentType(pContentType, "application/mbms-envelope+xml"))
    {
        if (ENATSC3ROUTESLSEnvelopeParse(pDocument, sDocument, 
            &pCallbackUserData->envelopeNode))
        {
            ERRORPRINT("%s: Failed to parse envelope part\n", 
                pCallbackUserData->pROUTESession->strDebugName);
            rval = 1;
        }
        else
        {
            pCallbackUserData->fHaveEnvelopeNode = 1;
        }
    }
    else if (IsContentType(pContentType, "application/s-tsid") ||
        IsContentType(pContentType, "application/route-s-tsid+xml"))
    {
        ENATSC3ROUTESLSSTSIDNode stsidNode;
        int32_t version;
        uint32_t crc32c;

        version = ROUTEMIMEGetRelatedDocumentVersion(pCallbackUserData, 
            pContentLocation);
        if (version == -1)
        {
            ERRORPRINT("%s: WARN: Failed to find version for %s\n",
                pCallbackUserData->pROUTESession->strDebugName, 
                pContentLocation); 
        }

        crc32c = ENUtils_crc32c(0, (uint8_t *)pDocument, 
	    sDocument);

        if (ENATSC3ROUTESLSSTSIDParse(pDocument, sDocument, &stsidNode))
        {
            ERRORPRINT("%s: Failed to parse s-tsid part\n",
                pCallbackUserData->pROUTESession->strDebugName);
            rval = 1;
        }
        else
        {
            ENStatus lret;

            ROUTEProcessParsedSTSID(pCallbackUserData->pPacketTimeUTC, 
                pCallbackUserData->packetNumber, &stsidNode, 
                pCallbackUserData->pROUTESession);

            lret = ROUTECallSTSIDAvailableCallback(
                pCallbackUserData->pPacketTimeUTC, 
                pCallbackUserData->packetNumber, pDocument, sDocument, 
                &stsidNode, version, crc32c, pCallbackUserData->pROUTESession);

            if (ENATSC3ROUTESLSSTSIDFree(&stsidNode))
            {
                ERRORPRINT("%s: Failed to free stsid\n",
                    pCallbackUserData->pROUTESession->strDebugName);
            }

            if (lret)
            {
                LOGPRINT("STSID rejected\n");
                rval = 1;
            }
        }
    }
    else if (IsContentType(pContentType, "application/dash+xml"))
    {
        int32_t version;
        uint32_t crc32c;

        version = ROUTEMIMEGetRelatedDocumentVersion(pCallbackUserData, 
            pContentLocation);
        if (version == -1)
        {
            ERRORPRINT("%s: WARN: Failed to find version for %s\n",
                pCallbackUserData->pROUTESession->strDebugName, 
                pContentLocation); 
        }

        crc32c = ENUtils_crc32c(0, (uint8_t *)pDocument, 
	    sDocument);

        if (ROUTECallMPDAvailableCallback(pCallbackUserData->pPacketTimeUTC, 
            pCallbackUserData->packetNumber, pDocument, sDocument, version, 
            crc32c, pCallbackUserData->pROUTESession))
        {
            LOGPRINT("MPD rejected\n");
            rval = 1;
        }
    }
    else if (IsContentType(pContentType, "application/atsc-held+xml"))
    {
        int32_t version;
        uint32_t crc32c;

        version = ROUTEMIMEGetRelatedDocumentVersion(pCallbackUserData, 
            pContentLocation);
        if (version == -1)
        {
            ERRORPRINT("%s: WARN: Failed to find version for %s\n",
                pCallbackUserData->pROUTESession->strDebugName, 
                pContentLocation); 
        }

        crc32c = ENUtils_crc32c(0, (uint8_t *)pDocument, 
	    sDocument);

        if (ROUTECallHELDAvailableCallback(pCallbackUserData->pPacketTimeUTC, 
            pCallbackUserData->packetNumber, pDocument, sDocument, version, 
            crc32c, pCallbackUserData->pROUTESession))
        {
            LOGPRINT("HELD rejected\n");
            rval = 1;
        }
    }

    return rval;
}

static int32_t
ROUTEMIMESignedDocumentAvailableCallback(const char *pDocument, 
    int32_t sDocument, const char *pPart, int32_t sPart, char *pContentType, 
    char *pContentLocation, void *pUserData)
{
    ROUTEMIMECallbackUserData *pCallbackUserData;
    int rval = 0;

    pCallbackUserData = (ROUTEMIMECallbackUserData *)pUserData;

    if (pCallbackUserData->pROUTESession->verbosity >= 
        ENVerbosity_StreamEventTrace)
    {
        LOGPRINT("%s: Got MIME multipart/signed document %s\n", 
            pCallbackUserData->pROUTESession->strDebugName, pContentType);
    }

    if (IsContentType(pContentType, "multipart/related"))
    {
        // Just note where the signed content is. If the signature verifies 
        // then we'll process it.
        pCallbackUserData->signedData.pMIME = pPart;
        pCallbackUserData->signedData.sMIME = sPart;
    }
    else if (IsContentType(pContentType, "application/pkcs7-signature"))
    {
        ROUTEMIMECallbackUserData mimeCallbackUserData;
        ENATSC3ROUTESLSMultipartMIMEHeaderInfo_t mimeHeaderInfo;
        const char *pParts;
        int32_t partsLen;
        const char *pBoundary = NULL;
        int32_t boundaryLen;
        int32_t i;

        if (!pCallbackUserData->pROUTESession->VerifySignatureCallback)
        {
            ERRORPRINT(
                "%s: No VerifySignature callback, rejecting signed SLS\n",
                pCallbackUserData->pROUTESession->strDebugName);
            rval = ENStatus_Failure;
            goto out;
        }

        if (!pCallbackUserData->signedData.pMIME)
        {
            ERRORPRINT("S/MIME parse failure - no signed data\n");
            rval = ENStatus_Failure;
            goto out;
        }

        if (pCallbackUserData->pROUTESession->VerifySignatureCallback(
            (const uint8_t *)pPart, sPart, 
            (const uint8_t *)pCallbackUserData->signedData.pMIME, 
            pCallbackUserData->signedData.sMIME,
            pCallbackUserData->pROUTESession->pCallbackUserData))
        {
            ERRORPRINT("%s: Failed to verify signature, rejecting SLS\n",
                pCallbackUserData->pROUTESession->strDebugName);
            rval = ENStatus_Failure;
            goto out;
        }

        if (ENATSC3ROUTESLSMultipartMIMEHeaderParse(
            pCallbackUserData->signedData.pMIME, 
            pCallbackUserData->signedData.sMIME,
            &mimeHeaderInfo, &pParts, &partsLen))
        {
            ERRORPRINT(
                "%s: %"PRIu64" Failed to parse ROUTE MIME document headers\n",
                pCallbackUserData->pROUTESession->strDebugName, 
                pCallbackUserData->packetNumber);
        }

        if (!mimeHeaderInfo.pContentType)
        {
            ERRORPRINT("%s: Couldn't parse mime content-type\n",
                pCallbackUserData->pROUTESession->strDebugName);
            rval = ENStatus_Failure;
            goto out;
        }

        for (i = 0; i < mimeHeaderInfo.nParameters; i++)
        {
            if (mimeHeaderInfo.vParameters[i].nameLen == sizeof("boundary") -1 &&
                !strncasecmp(mimeHeaderInfo.vParameters[i].pName, "boundary", 
                    sizeof("boundary") - 1))
            {
                pBoundary = mimeHeaderInfo.vParameters[i].pValue;
                boundaryLen = mimeHeaderInfo.vParameters[i].valueLen;
                break;
            }
        }

        if (!pBoundary)
        {
            ERRORPRINT(
                "%s: %"PRIu64" Failed to parse ROUTE MIME document headers\n",
                    pCallbackUserData->pROUTESession->strDebugName, 
                    pCallbackUserData->packetNumber);
            rval = ENStatus_Failure;
            goto out;
        }

        mimeCallbackUserData.pPacketTimeUTC = pCallbackUserData->pPacketTimeUTC;
        mimeCallbackUserData.pROUTESession = pCallbackUserData->pROUTESession;
        mimeCallbackUserData.packetNumber = pCallbackUserData->packetNumber;
        mimeCallbackUserData.version = pCallbackUserData->version;
        mimeCallbackUserData.pLCTHeader = pCallbackUserData->pLCTHeader;
        mimeCallbackUserData.pLCTChannelInfo = pCallbackUserData->pLCTChannelInfo;
        mimeCallbackUserData.fHaveEnvelopeNode = 0;
        if (!strncasecmp(mimeHeaderInfo.pContentType, "multipart/related", 
            sizeof("multipart/related") - 1))
        {
            if (ENATSC3ROUTESLSMultipartMIMEPartsParse(pParts, partsLen, 
                pBoundary, boundaryLen, 
                ROUTEMIMERelatedDocumentAvailableCallback, 
                &mimeCallbackUserData))
            {
                ERRORPRINT(
                    "%s: %"PRIu64" Failed to parse ROUTE MIME related document parts\n",
                    pCallbackUserData->pROUTESession->strDebugName, 
                    pCallbackUserData->packetNumber);
                pCallbackUserData->pLCTChannelInfo->slsInfo.lastCompleteSLSVersion = -1;
                pCallbackUserData->pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = 0;
            }
            else
            {
                if (pCallbackUserData->pLCTChannelInfo->expectedSLSTOI != -1)
                {
                    pCallbackUserData->pLCTChannelInfo->slsInfo.lastCompleteSLSTOI = 
                        pCallbackUserData->pLCTHeader->transportObjectIdentifier;
                    pCallbackUserData->pLCTChannelInfo->slsInfo.lastCompleteSLSVersion = pCallbackUserData->version;
                    pCallbackUserData->pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = pCallbackUserData->crc32c;
                }
            }

            if (mimeCallbackUserData.fHaveEnvelopeNode)
            {
                if (ENATSC3ROUTESLSEnvelopeFree(
                    &mimeCallbackUserData.envelopeNode))
                {
                    ERRORPRINT("%s: Failed to free parsed envelope\n",
                        pCallbackUserData->pROUTESession->strDebugName);
                }
            }
        }
        else
        {
            ERRORPRINT("%s: Unhandled mime content-type %.*s\n", 
                pCallbackUserData->pROUTESession->strDebugName, 
                mimeHeaderInfo.contentTypeLen, mimeHeaderInfo.pContentType);
        }
    }

out:
    return rval;
}

static ENStatus
ROUTEHandleEFDT(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    ENATSC3ROUTESession *pROUTESession, ENATSC3ROUTELCTHeader *pLCTHeader, 
    ALCHeader *pALCHeader, LCTChannelInfo *pLCTChannelInfo)
{
    ENATSC3ROUTESLSEFDTNode efdtNode;
    uint32_t crc32c;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    // Note at present we really don't use the FDT Instance on tsi=0 (i.e.
    // the SLS LCT channel). We are parsing it purely for verification 
    // purposes. The expected TOI for SLS is as described in A/331 Annex C.
    //
    // Also note that in some cases a FLUTE FDT-Instance object is present 
    // instead of an EFDT. RouteHandleEFDT will detect this case.

    if (pROUTESession->strShadowFilePath[0])
    {
        char strFilename[256];
        int ofd = -1;

        if (snprintf(strFilename, sizeof(strFilename), 
            "%s/fdtInstance-%"PRIu64".xml", pROUTESession->strShadowFilePath, 
            packetNumber) >= sizeof(strFilename))
        {
            ERRORPRINT("%s: Filename too long\n", pROUTESession->strDebugName);
        }
        else
        {
            ofd = open(strFilename, O_RDWR | O_CREAT, 0666);
            if (ofd < 0)
            {
                ERRORPRINT("%s: Failed to open %s: %s\n", strFilename, 
                    pROUTESession->strDebugName, strerror(errno));
            }
            else
            {
                if (write(ofd, pLCTChannelInfo->pBuffer, 
                    pLCTChannelInfo->slsObjectSize) != 
                    pLCTChannelInfo->slsObjectSize)
                {
                    ERRORPRINT("%s: Write failed; %s\n", 
                        pROUTESession->strDebugName, strerror(errno));
                }
            }
        }

        if (ofd >= 0)
        {
            close(ofd);
        }
    }

    crc32c = ENUtils_crc32c(0, pLCTChannelInfo->pBuffer, 
	pLCTChannelInfo->slsObjectSize);
    if (pLCTChannelInfo->lastEFDTcrc32c && 
	pLCTChannelInfo->lastEFDTcrc32c == crc32c && 
	pLCTChannelInfo->lastEFDTSize == pLCTChannelInfo->slsObjectSize && 
	pLCTChannelInfo->expectedSLSTOI != -1)
    {
	goto out;
    }

    if (ROUTESLSEFDTParse((char *)pLCTChannelInfo->pBuffer, 
        pLCTChannelInfo->slsObjectSize, &efdtNode))
    {
        LOGPRINT_NOCONSOLE("%s: Failed to parse fdt instance {%.*s}\n", 
            pROUTESession->strDebugName, pLCTChannelInfo->slsObjectSize, 
            pLCTChannelInfo->pBuffer);
        goto out;
    }

    pLCTChannelInfo->lastEFDTcrc32c = crc32c;
    pLCTChannelInfo->lastEFDTSize = pLCTChannelInfo->slsObjectSize;
    pLCTChannelInfo->expectedSLSTOI = -1;

    if (efdtNode.fdtInstance.files.nOffsetsUsed != 1)
    {
        ERRORPRINT(
            "%s: WARN: FDT-Instance for SLS has 0 or more than 1 files: %d\n",
            pROUTESession->strDebugName, 
            efdtNode.fdtInstance.files.nOffsetsUsed);
    }

    for (i = 0; i < efdtNode.fdtInstance.files.nOffsetsUsed; i++)
    {
        ENATSC3ROUTESLSEFDTFileNode *pFileNode;

        pFileNode = (ENATSC3ROUTESLSEFDTFileNode *)
            EnXMLParserDataOffsetArrayGetPointer(&efdtNode.fdtInstance.files, 
            0);

        if (!(pFileNode->base.presentFlags & 
              ENATSC3ROUTESLSEFDTFileNodeAttribute_toi) || 
            !(pFileNode->base.presentFlags & 
              ENATSC3ROUTESLSEFDTFileNodeAttribute_contentLocation) || 
            !(pFileNode->base.presentFlags & 
              ENATSC3ROUTESLSEFDTFileNodeAttribute_contentType))
        {
            ERRORPRINT("%s: WARN: FDT-Instance for SLS missing elements (presentFlags 0x%x\n", 
                pROUTESession->strDebugName, pFileNode->base.presentFlags);
            continue;
        }

        pLCTChannelInfo->expectedSLSTOI = pFileNode->toi;
    }

    if (ROUTESLSEFDTFree(&efdtNode))
    {
        ERRORPRINT("%s: Failed to free fdt instance\n", 
            pROUTESession->strDebugName);
    }

    pLCTChannelInfo->slsInfo.lastCompleteFDTInstanceTOI = 
        pLCTHeader->transportObjectIdentifier;

out:
    return rval;
}

static ENStatus
ROUTEHandleSLSObject(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint8_t plpId, char *pURL, ENATSC3ROUTESession *pROUTESession, 
    ENATSC3ROUTELCTHeader *pLCTHeader, ALCHeader *pALCHeader, 
    LCTChannelInfo *pLCTChannelInfo)
{
    ROUTEMIMECallbackUserData mimeCallbackUserData; 
    uint32_t version;
    uint32_t crc32c;
    char *pXML;
    int32_t sXML;
    ENStatus rval = ENStatus_Success;

    version = pLCTHeader->transportObjectIdentifier & 0xff;
    crc32c = ENUtils_crc32c(0, pLCTChannelInfo->pBuffer, 
        pLCTChannelInfo->slsObjectSize);

    ROUTECallSLSAvailableCallbacks(pPacketTimeUTC, packetNumber, plpId,
        pURL, pLCTHeader, pLCTChannelInfo, version, crc32c, pROUTESession);

    if (pLCTChannelInfo->slsInfo.lastCompleteSLSVersion == version && 
        pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C == crc32c)
    {
        goto out;
    }

    if (pROUTESession->strShadowFilePath[0])
    {
        char strFilename[256];
        int ofd = -1;

        if (snprintf(strFilename, sizeof(strFilename), "%s/sls-%d.xml", 
            pROUTESession->strShadowFilePath, version) >= 
            sizeof(strFilename))
        {
            ERRORPRINT("%s: Filename too long\n", pROUTESession->strDebugName);
        }
        else
        {
            ofd = open(strFilename, O_RDWR | O_CREAT, 0666);
            if (ofd < 0)
            {
                ERRORPRINT("%s: Failed to open %s: %s\n", strFilename, 
                    pROUTESession->strDebugName, strerror(errno));
            }
            else
            {
                if (write(ofd, pLCTChannelInfo->pBuffer, 
                    pLCTChannelInfo->slsObjectSize) != 
                    pLCTChannelInfo->slsObjectSize)
                {
                    ERRORPRINT("%s: Write failed; %s\n", 
                        pROUTESession->strDebugName, strerror(errno));
                }
            }
        }

        if (ofd >= 0)
        {
            close(ofd);
        }
    }

    if (pLCTHeader->transportObjectIdentifier != 
        pLCTChannelInfo->expectedSLSTOI &&
        pLCTChannelInfo->expectedSLSTOI != -1)
    {
        ERRORPRINT(
            "%s: WARN: SLS TOI %u doesn't match FDT-Instance TOI %u (url %s)\n",
            pROUTESession->strDebugName, pLCTHeader->transportObjectIdentifier,
            pLCTChannelInfo->expectedSLSTOI, pURL);
    }

    pXML = (char *)pLCTChannelInfo->pBuffer;
    sXML = pLCTChannelInfo->slsObjectSize;

    ENATSC3ROUTESLSMultipartMIMEHeaderInfo_t mimeHeaderInfo;
    const char *pParts;
    int32_t partsLen;
    const char *pBoundary = NULL;
    int32_t boundaryLen;
    int32_t i;

    if (ENATSC3ROUTESLSMultipartMIMEHeaderParse(pXML, sXML, &mimeHeaderInfo, 
        &pParts, &partsLen))
    {
        ERRORPRINT(
            "%s: %"PRIu64" Failed to parse ROUTE MIME document headers\n",
            pROUTESession->strDebugName, packetNumber);
        rval = ENStatus_Failure;
        goto out;
    }

    if (!mimeHeaderInfo.pContentType)
    {
        ERRORPRINT("%s: Couldn't parse mime content-type\n", 
            pROUTESession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    for (i = 0; i < mimeHeaderInfo.nParameters; i++)
    {
        if (mimeHeaderInfo.vParameters[i].nameLen == sizeof("boundary") -1 &&
            !strncasecmp(mimeHeaderInfo.vParameters[i].pName, "boundary", 
                sizeof("boundary") - 1))
        {
            pBoundary = mimeHeaderInfo.vParameters[i].pValue;
            boundaryLen = mimeHeaderInfo.vParameters[i].valueLen;
            break;
        }
    }

    if (!pBoundary)
    {
        ERRORPRINT(
            "%s: %"PRIu64" Failed to parse ROUTE MIME document headers\n",
                pROUTESession->strDebugName, packetNumber);
        rval = ENStatus_Failure;
        goto out;
    }

    mimeCallbackUserData.pPacketTimeUTC = pPacketTimeUTC;
    mimeCallbackUserData.pROUTESession = pROUTESession;
    mimeCallbackUserData.packetNumber = packetNumber;
    mimeCallbackUserData.version = version;
    mimeCallbackUserData.crc32c = crc32c;
    mimeCallbackUserData.pLCTHeader = pLCTHeader;
    mimeCallbackUserData.pLCTChannelInfo = pLCTChannelInfo;
    mimeCallbackUserData.fHaveEnvelopeNode = 0;
    mimeCallbackUserData.signedData.pMIME = NULL;
    mimeCallbackUserData.signedData.sMIME = 0;

    if (!strncasecmp(mimeHeaderInfo.pContentType, "multipart/related", 
        sizeof("multipart/related") - 1))
    {
        if (ENATSC3ROUTESLSMultipartMIMEPartsParse(pParts, partsLen, pBoundary, 
            boundaryLen, ROUTEMIMERelatedDocumentAvailableCallback, 
            &mimeCallbackUserData))
        {
            ERRORPRINT("%s: %"PRIu64" Failed to parse ROUTE MIME related document parts\n",
                pROUTESession->strDebugName, packetNumber);
            pLCTChannelInfo->slsInfo.lastCompleteSLSVersion = -1;
            pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = 0;
        }
        else
        {
            if (pLCTChannelInfo->expectedSLSTOI != -1)
            {
                pLCTChannelInfo->slsInfo.lastCompleteSLSTOI = 
                    pLCTHeader->transportObjectIdentifier;
                pLCTChannelInfo->slsInfo.lastCompleteSLSVersion = version;
                pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = crc32c;
            }
        }
    }
    else if (!strncasecmp(mimeHeaderInfo.pContentType, "multipart/signed", 
        sizeof("multipart/signed") - 1))
    {
        if (ENATSC3ROUTESLSMultipartMIMEPartsParse(pParts, partsLen, pBoundary, 
            boundaryLen, ROUTEMIMESignedDocumentAvailableCallback, 
            &mimeCallbackUserData))
        {
            ERRORPRINT("%s: %"PRIu64" Failed to parse ROUTE MIME signed document parts\n",
                pROUTESession->strDebugName, packetNumber);
            pLCTChannelInfo->slsInfo.lastCompleteSLSVersion = -1;
            pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = 0;
        }
        else
        {
            if (pLCTChannelInfo->expectedSLSTOI != -1)
            {
                pLCTChannelInfo->slsInfo.lastCompleteSLSTOI = 
                    pLCTHeader->transportObjectIdentifier;
                pLCTChannelInfo->slsInfo.lastCompleteSLSVersion = version;
                pLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = crc32c;
            }
        }
    }
    else
    {
        ERRORPRINT("%s: Unhandled mime content-type %.*s\n", 
            pROUTESession->strDebugName, mimeHeaderInfo.contentTypeLen, 
            mimeHeaderInfo.pContentType);
    }

    if (mimeCallbackUserData.fHaveEnvelopeNode)
    {
        if (ENATSC3ROUTESLSEnvelopeFree(&mimeCallbackUserData.envelopeNode))
        {
            ERRORPRINT("%s: Failed to free parsed envelope\n", 
                pROUTESession->strDebugName);
        }
    }

out:
    return rval;
}

static ENStatus
GetObjectURLAndContentTypeFromLCTChannelInfo(ENATSC3ROUTESession *pROUTESession,
    LCTChannelInfo *pLCTChannelInfo, uint32_t toi, char *pURL, int32_t sURL, 
    char *pContentType, int32_t sContentType) 
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    // First check to see if there is an explicit toi-to-filename mapping
    for (i = 0; i < pLCTChannelInfo->nFiles; i++)
    {
        if (pLCTChannelInfo->vFiles[i].toi == toi)
        {
            if (pURL)
            {
                if (snprintf(pURL, sURL, "%s", 
                    pLCTChannelInfo->vFiles[i].strContentLocation) >= sURL)
                {
                    ERRORPRINT("%s: URL too long\n", 
                        pROUTESession->strDebugName);
                    rval = ENStatus_Failure;
                    goto out;
                }
            }

            if (pContentType)
            {
                if (snprintf(pContentType, sContentType, "%s", 
                    pLCTChannelInfo->vFiles[i].strContentType) >= 
                    sContentType)
                {
                    ERRORPRINT("%s: ContentType too long\n", 
                        pROUTESession->strDebugName);
                    rval = ENStatus_Failure;
                    goto out;
                }
            }

            goto out;
        }
    }

    // If there is no explicit mapping then use the FileTemplate from the
    // EFDT
    if (pURL)
    {
        if (GenerateObjectURLFromTemplate(pROUTESession, 
            pLCTChannelInfo->strFileTemplate, pURL, sURL, toi))
        {
            rval = ENStatus_Failure;
            goto out;
        }
    }

    if (pContentType)
    {
        if (snprintf(pContentType, sContentType, "%s", 
            pLCTChannelInfo->strContentType) >= sContentType)
        {
            ERRORPRINT("%s: ContentType too long\n",
                pROUTESession->strDebugName);
            rval = ENStatus_Failure;
            goto out;
        }
    }

out:
    return rval;
}

ENStatus
ENATSC3ROUTESessionGetObjectURLAndContentType(
    ENATSC3ROUTESessionHandle_t hROUTESession, uint32_t tsi, uint32_t toi, 
    char *pURL, int32_t sURL, char *pContentType, int32_t sContentType) 
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    LCTChannelInfo *pLCTChannelInfo = NULL;
    ENStatus rval = ENStatus_Success;

    if (ROUTELCTChannelInfoFind(pROUTESession, tsi, &pLCTChannelInfo))
    {
        ERRORPRINT("%s: Failed to find LCT Channel for tsi %u\n", 
            pROUTESession->strDebugName, tsi);
        rval = ENStatus_Failure;
        goto out;
    }

    rval = GetObjectURLAndContentTypeFromLCTChannelInfo(pROUTESession, 
        pLCTChannelInfo, toi, pURL, sURL, pContentType, sContentType);
    if (rval)
    {
        ERRORPRINT("%s: Failed to get URL/Content-Type for LCT Channel\n",
            pROUTESession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

out:
    return rval;
}

ENStatus
ENATSC3ROUTESessionGetEntityModeObjectURLAndContentType(
    ENATSC3ROUTESessionHandle_t hROUTESession, uint32_t tsi, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t *pPayloadOffset, 
    char *pURL, int32_t sURL, char *pContentType, int32_t sContentType)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    LCTChannelInfo *pLCTChannelInfo = NULL;
    int32_t payloadOffset = 0;
    int32_t fHaveContentType = 0;
    ENStatus rval = ENStatus_Success;
    
    pURL[0] = 0;

    if (ROUTELCTChannelInfoFind(pROUTESession, tsi, &pLCTChannelInfo))
    {
        ERRORPRINT("%s: Failed to find LCT Channel for tsi %u\n",
            pROUTESession->strDebugName, tsi);
        rval = ENStatus_Failure;
        goto out;
    }

    // Iterate over the headers
    while (1)
    {
        int32_t headerStart = payloadOffset;

        // Search for end of header line
        while (payloadOffset + 1 < payloadLength &&
            (pPayload[payloadOffset] != '\r' ||
             pPayload[payloadOffset + 1] != '\n'))
        {
            payloadOffset++;
        }

        if (payloadOffset + 1 == payloadLength)
        {
            fprintf(stderr, "Unexpected end of headers\n");
            break;
        }

        if (payloadOffset == headerStart)
        {
            payloadOffset += 2;
            break;
        }

        if (payloadOffset - headerStart > LiteralLength("Content-Location:") &&
            !strncmp((char *)pPayload, "Content-Location:",
                LiteralLength("Content-Location:")))
        {
            headerStart += LiteralLength("Content-Location:");
            while (isspace(pPayload[headerStart]))
            {
                headerStart++;
            }
            if (payloadOffset - headerStart >= sURL)
            {
                rval = ENStatus_Failure;
                goto out;
            }
            strncpy(pURL, (char *)pPayload + headerStart,
                payloadOffset - headerStart);
            pURL[payloadOffset - headerStart] = 0;
        }
        else if (payloadOffset - headerStart > LiteralLength("Content-Type:") &&
            !strncmp((char *)pPayload, "Content-Type:",
                LiteralLength("Content-Type:")))
        {
            headerStart += LiteralLength("Content-Type:");
            while (isspace(pPayload[headerStart]))
            {
                headerStart++;
            }
            if (payloadOffset - headerStart >= sContentType)
            {
                rval = ENStatus_Failure;
                goto out;
            }
            if (pContentType)
            {
                strncpy(pContentType, (char *)pPayload + headerStart,
                    payloadOffset - headerStart);
                pContentType[payloadOffset - headerStart] = 0;
                fHaveContentType = 0;
            }
        }

        payloadOffset += 2;
    }

    if (pContentType && !fHaveContentType)
    {
        strncpy(pContentType, pLCTChannelInfo->strContentType, 
            sContentType);
        if (pContentType[sContentType - 1])
        {
            ERRORPRINT("Content-Type too long\n");
            rval = ENStatus_Failure;
            goto out;
        }
    }

    *pPayloadOffset = payloadOffset;

out:
    return rval;
}

static ENStatus
ROUTEHandleSLSPayload(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint8_t plpId, const uint8_t *pPayload, int32_t payloadLength, 
    ENATSC3ROUTESession *pROUTESession, ENATSC3ROUTELCTHeader *pLCTHeader, 
    ALCHeader *pALCHeader, LCTChannelInfo *pLCTChannelInfo)
{
    ENStatus rval = ENStatus_Success;

    if (pLCTChannelInfo->slsTOI == -1)
    {
        // We only start processing a new object at offset 0. 
        if (pALCHeader->fecPayloadId != 0)
        {
            goto out;
        }

        pLCTChannelInfo->slsTOI = pLCTHeader->transportObjectIdentifier;
        pLCTChannelInfo->slsPos = 0;
        pLCTChannelInfo->slsObjectSize = -1;
    }
    else if (pLCTHeader->transportObjectIdentifier != 
        pLCTChannelInfo->slsTOI)
    {
        // If we get here then the object Id as changed prematurely. This
        // would suggest that we didn't receive the end of the previous
        // object.
        DEBUGPRINT_NOCONSOLE(
            "%s: %"PRIu64" Premature end of FDT-Instance SLS object toi %u\n", 
            pROUTESession->strDebugName, packetNumber, 
            pLCTChannelInfo->slsTOI);
        if (pALCHeader->fecPayloadId == 0)
        {
            // We're at the beginning of a new object. Start processing it.
            pLCTChannelInfo->slsTOI = 
                pLCTHeader->transportObjectIdentifier;
            pLCTChannelInfo->slsPos = 0;
            pLCTChannelInfo->slsObjectSize = -1;
        }
        else
        {
            // Start searching for a new object
            pLCTChannelInfo->slsTOI = -1;
            goto out;
        }
    }

    if (pLCTChannelInfo->slsObjectSize == -1)
    {
        if (pLCTHeader->fExtTOLPresent)
        {
            pLCTChannelInfo->slsObjectSize = pLCTHeader->tolTransferLength;
        }
        else if (pLCTHeader->fExtFTIPresent)
        {
            pLCTChannelInfo->slsObjectSize = pLCTHeader->ftiTransferLength;
        }

        if (pLCTChannelInfo->slsObjectSize > pLCTChannelInfo->sBuffer)
        {
            ERRORPRINT("%s: %"PRIu64" SLS object larger than buffer (%"PRId64" > %d)\n",
                pROUTESession->strDebugName, packetNumber, 
                pLCTChannelInfo->slsObjectSize, pLCTChannelInfo->sBuffer);
            pLCTChannelInfo->slsTOI = -1;
            goto out;
        }
    }
    else
    {
        if (pLCTHeader->fExtTOLPresent)
        {
            if (pLCTChannelInfo->slsObjectSize != 
                pLCTHeader->tolTransferLength)
            {
                DEBUGPRINT_NOCONSOLE(
                    "%s: %"PRIu64" SLS TOL object size changed %"PRId64" != %"PRId64"\n", 
                    pROUTESession->strDebugName, packetNumber, 
                    pLCTChannelInfo->slsObjectSize,
                    pLCTHeader->tolTransferLength);
                pLCTChannelInfo->slsTOI = -1;
                goto out;
            }
        }
        else if (pLCTHeader->fExtFTIPresent)
        {
            if (pLCTChannelInfo->slsObjectSize != pLCTHeader->ftiTransferLength)
            {
                DEBUGPRINT_NOCONSOLE(
                    "%s: %"PRIu64" FTI object size changed %"PRId64" != %"PRId64"\n", 
                    pROUTESession->strDebugName, packetNumber, 
                    pLCTChannelInfo->slsObjectSize,
                    pLCTHeader->ftiTransferLength);
                pLCTChannelInfo->slsTOI = -1;
                goto out;
            }
        }
    }

    // We currently expect data to be delivered in order.
    //
    // XXXX Strictly speaking data may arrive out of order if @order=0 in the 
    // SrcFlow's Payload element.
    if (pLCTChannelInfo->slsPos != pALCHeader->fecPayloadId)
    {
        DEBUGPRINT_NOCONSOLE(
            "%s: %"PRIu64" SLS offset error %"PRId64" != %d\n", 
            pROUTESession->strDebugName, packetNumber, pLCTChannelInfo->slsPos, 
            pALCHeader->fecPayloadId);
        pLCTChannelInfo->slsTOI = -1;
        goto out;
    }

    if ((pLCTChannelInfo->slsObjectSize != -1 && 
         pLCTChannelInfo->slsPos + payloadLength > 
         pLCTChannelInfo->slsObjectSize) ||
        (pLCTChannelInfo->slsObjectSize == -1 &&
         pLCTChannelInfo->slsPos + payloadLength > pLCTChannelInfo->sBuffer))
    {
        ERRORPRINT("%s: %"PRIu64" SLS fragment buffer overflow (%"PRId64" > %d)\n", 
            pROUTESession->strDebugName, packetNumber, 
            pLCTChannelInfo->slsPos + payloadLength, pLCTChannelInfo->sBuffer);
        pLCTChannelInfo->slsTOI = -1;
        goto out;
    }

    memcpy(pLCTChannelInfo->pBuffer + pLCTChannelInfo->slsPos, pPayload, 
        payloadLength);
    pLCTChannelInfo->slsPos += payloadLength;

    // See if the object is now complete
    if (pLCTChannelInfo->slsPos == pLCTChannelInfo->slsObjectSize)
    {
        uint32_t fGzipped;

        fGzipped = pLCTHeader->transportObjectIdentifier & 0x80000000;

        if (fGzipped)
        {
            // Hmmm....seems there are packets flagged as gzipped but they 
            // really aren't. Let's check for the gzip signature.
            if (pLCTChannelInfo->pBuffer[0] == 0x1f &&
                pLCTChannelInfo->pBuffer[1] == 0x8b)
            {
                ERRORPRINT("%s; Unhandled gzip'd SLS fragment", 
                    pROUTESession->strDebugName);
                goto out;
            }
            else
            {
                ERRORPRINT("%s: WARN: TOI indicates gzip but payload doesn't have gzip header\n",
                    pROUTESession->strDebugName);
            }
        }

        if (pLCTHeader->transportObjectIdentifier == 0) 
        {
            // If toi=0 then this is an EFDT which will provide the toi for
            // the MIME SLS. 
            rval = ROUTEHandleEFDT(pPacketTimeUTC, packetNumber, pROUTESession, 
                pLCTHeader, pALCHeader, pLCTChannelInfo);
        }
        else
        {
            char strURL[256];

            if (GetObjectURLAndContentTypeFromLCTChannelInfo(pROUTESession, 
                pLCTChannelInfo, pLCTHeader->transportObjectIdentifier, strURL, 
                sizeof(strURL), NULL, 0))
            {
                ERRORPRINT("%s: Failed to get URL for SLS object\n",
                    pROUTESession->strDebugName);
            }
            else
            {
                rval = ROUTEHandleSLSObject(pPacketTimeUTC, packetNumber, 
                    plpId, strURL, pROUTESession, pLCTHeader, pALCHeader, 
                    pLCTChannelInfo);
            }
        }

        pLCTChannelInfo->slsTOI = -1;
    }

out:
    return rval;
}

static struct
{
    int32_t formatId;
    int32_t frag;
    int32_t order;
} vCodepointMapG[] = 
{
    {
        // codepoint = 0: ATSC Reserved
        .formatId = -1,
        .frag = -1,
        .order = -1,
    },
    {
        // codepoint = 1 : NRT - File Mode
        .formatId = 1, // File Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 2 : NRT - Entity Mode
        .formatId = 2, // Entity Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 3 : NRT - Unsigned Package Mode
        .formatId = 3, // Unsigned Package Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 4 : NRT - Signed Package Mode
        .formatId = 4, // Signed Package Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 5 : New IS, timeline changed
        .formatId = 1, // File Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 6 : New IS: timeline continued
        .formatId = 1, // File Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 7: Redundant IS
        .formatId = 1, // File Mode
        .frag = 0,
        .order = 1,
    },
    {
        // codepoint = 8: Media Sregment, File Mode
        .formatId = 1, // File Mode
        .frag = 1,
        .order = 1,
    },
    {
        // codepoint = 9 : Media Segment, Entity Mode
        .formatId = 2, // File Mode
        .frag = 1,
        .order = 1,
    },
};

static void
ROUTECallObjectDataCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTELCTHeader *pLCTHeader, 
    LCTChannelInfo *pLCTChannelInfo, const uint8_t *pPayload, int32_t sPayload, 
    int64_t objectPosition, int32_t fError, ENATSC3ROUTESession *pROUTESession)
{
    int32_t formatId = -1;
    int32_t frag = -1;
    int32_t order = -1;
    int32_t i;

    if (pLCTHeader->codepoint < 10)
    {
        formatId = vCodepointMapG[pLCTHeader->codepoint].formatId;
        frag = vCodepointMapG[pLCTHeader->codepoint].frag;
        order = vCodepointMapG[pLCTHeader->codepoint].order;
    }
    else if (pLCTHeader->codepoint >= 128)
    {
        struct _LCTChannelInfoPayloadDescription *pPayloadDescription;

        for (i = 0, pPayloadDescription = pLCTChannelInfo->vPayloadDescriptions;
             i < pLCTChannelInfo->nPayloadDescriptions; 
             i++, pPayloadDescription++)
        {
            if (pPayloadDescription->codepoint == pLCTHeader->codepoint)
            {
                formatId = pPayloadDescription->formatId;
                frag = pPayloadDescription->frag;
                order = pPayloadDescription->order;
                break;
            }
        }
    }

    for (i = 0; i < numof(pROUTESession->vObjectCallbacks); i++)
    {
        if (pROUTESession->vObjectCallbacks[i].fValid &&
            pROUTESession->vObjectCallbacks[i].ObjectDataCallback)
        {
            struct _ObjectCallbacks *pObjectCallbacks;

            pObjectCallbacks = pROUTESession->vObjectCallbacks + i;
            pObjectCallbacks->ObjectDataCallback(pPacketTimeUTC, packetNumber,
                pLCTHeader, formatId, frag, order, pLCTChannelInfo->strId, 
                pLCTChannelInfo->idType, pPayload, sPayload, objectPosition, 
                fError, pObjectCallbacks->pObjectCallbackUserData);
        }
    }
}

static void
ROUTECallSessionResetCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3ROUTESession *pROUTESession) 
{
    int32_t i;

    for (i = 0; i < numof(pROUTESession->vObjectCallbacks); i++)
    {
        if (pROUTESession->vObjectCallbacks[i].fValid &&
            pROUTESession->vObjectCallbacks[i].SessionResetCallback)
        {
            struct _ObjectCallbacks *pObjectCallbacks;

            pObjectCallbacks = pROUTESession->vObjectCallbacks + i;
            pObjectCallbacks->SessionResetCallback(pPacketTimeUTC, 
                packetNumber, pObjectCallbacks->pObjectCallbackUserData);
        }
    }
}

#define LiteralLength(x) (sizeof(x) - 1)
static ENStatus
ROUTEHandleMediaFilePayload(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint8_t plpId, const uint8_t *pPayload, 
    int32_t payloadLength, ENATSC3ROUTESession *pROUTESession, 
    ENATSC3ROUTELCTHeader *pLCTHeader, uint64_t objectPosition, 
    LCTChannelInfo *pLCTChannelInfo)
{
    ENStatus rval = ENStatus_Success;

    ROUTECallObjectDataCallback(pPacketTimeUTC, packetNumber, pLCTHeader,
        pLCTChannelInfo, pPayload, payloadLength, objectPosition, 0, 
        pROUTESession);

    if (pROUTESession->strShadowFilePath[0])
    {
        char strOutFilename[256];
        char strURL[128];

        if (GetObjectURLAndContentTypeFromLCTChannelInfo(pROUTESession,
            pLCTChannelInfo, pLCTHeader->transportObjectIdentifier, 
            strURL, sizeof(strURL), NULL, 0))
        {
            ERRORPRINT("%s: URL too long\n", pROUTESession->strDebugName);
        }
        else
        {
            if (snprintf(strOutFilename, sizeof(strOutFilename), "%s/%s", 
                pROUTESession->strShadowFilePath, strURL) >= 
                sizeof(strOutFilename))
            {
                ERRORPRINT("%s: Filename too long\n", 
                    pROUTESession->strDebugName);
            }
            else
            {
                int ofd;

                ofd = open(strOutFilename, O_WRONLY);
                if (ofd <= 0)
                {
                    ofd = open(strOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 
                        0666);
                }

                if (ofd < 0)
                {
                    ERRORPRINT("%s: Failed to open %s: %s\n", 
                        pROUTESession->strDebugName, strOutFilename, 
                        strerror(errno));
                }
                else
                {
                    if (lseek(ofd, objectPosition, SEEK_SET) != objectPosition)
                    {
                        ERRORPRINT("%s: Failed to seek %s to %"PRIu64"\n",
                            pROUTESession->strDebugName, strOutFilename, 
                            objectPosition);
                    }
                    if (write(ofd, pPayload, payloadLength) != payloadLength) 
                    {
                        ERRORPRINT("%s: Write failed: %s\n", 
                            pROUTESession->strDebugName, strerror(errno));
                    }

                    close(ofd);
                }
            }
        }
    }

    return rval;
}

ENStatus
ENATSC3ROUTESessionProcessUDPPacket(ENATSC3ROUTESessionHandle_t hROUTESession, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, uint8_t plpId, 
    const uint8_t *pPayload, int32_t payloadLength) 
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    ENATSC3ROUTELCTHeader lctHeader;
    ALCHeader alcHeader;
    int32_t extensionHeaderOffset;
    int32_t offset;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (pROUTESession->fBitrateStatisticsEnabled)
    {
        ENATSC3BitrateStatisticsUpdate(&pROUTESession->stats, pPacketTimeUTC, 
            payloadLength);
    }

    if (payloadLength < 16)
    {
        ERRORPRINT("%s: %"PRIu64" Packet too small for LCT\n", 
            pROUTESession->strDebugName, packetNumber);
        rval = ENStatus_Failure;
        goto out;
    }
    
    lctHeader.version = pPayload[0] >> 4;
    lctHeader.congestionControlFlag = (pPayload[0] >> 2) && 0x3;
    lctHeader.protocolSpecificIndication = pPayload[0] & 0x3;
    lctHeader.transportSessionIdentifierFlag = (pPayload[1] >> 7) & 0x1;
    lctHeader.transportObjectIdentifierFlag = (pPayload[1] >> 5) & 0x3;
    lctHeader.halfWordFlag = (pPayload[1] >> 4) & 0x1;
    lctHeader.reserved = (pPayload[1] >> 3) & 0x3;
    lctHeader.closeSessionFlag = (pPayload[1] >> 1) & 0x1;
    lctHeader.closeObjectFlag = (pPayload[1] >> 0) & 0x1;
    lctHeader.headerLength = pPayload[2];
    lctHeader.codepoint = pPayload[3];

    // See A/331 A.3.6
    if (lctHeader.version != 1)
    {
        DEBUGPRINT_NOCONSOLE("%s: Invalid LCT header version %d != 1\n", 
            pROUTESession->strDebugName, lctHeader.version);
        rval = ENStatus_Failure;
        goto out;
    }

    if (lctHeader.protocolSpecificIndication != 2)
    {
        DEBUGPRINT_NOCONSOLE("%s: Invalid LCT header PSI %d != 2\n", 
            pROUTESession->strDebugName, lctHeader.protocolSpecificIndication);
        rval = ENStatus_Failure;
        goto out;
    }

    if (lctHeader.transportSessionIdentifierFlag != 1)
    {
        DEBUGPRINT_NOCONSOLE("%s: Invalid LCT header TSI flag %d != 1\n", 
            pROUTESession->strDebugName, 
            lctHeader.transportSessionIdentifierFlag);
        rval = ENStatus_Failure;
        goto out;
    }

    if (lctHeader.transportObjectIdentifierFlag != 1)
    {
        DEBUGPRINT_NOCONSOLE("%s: Invalid LCT header TOI flag  %d != 1\n", 
            pROUTESession->strDebugName, 
            lctHeader.transportObjectIdentifierFlag);
        rval = ENStatus_Failure;
        goto out;
    }

    if (lctHeader.halfWordFlag != 0)
    {
        DEBUGPRINT_NOCONSOLE("%s: Invalid LCT header half-word flag %d != 0\n", 
            pROUTESession->strDebugName, lctHeader.halfWordFlag);
        rval = ENStatus_Failure;
        goto out;
    }

    if (lctHeader.reserved != 0)
    {
        DEBUGPRINT_NOCONSOLE("%s: Invalid LCT header reserved field %d != 0\n", 
            pROUTESession->strDebugName, lctHeader.reserved);
        rval = ENStatus_Failure;
        goto out;
    }

    lctHeader.congestionControlInformation = 
        (pPayload[4] << 24) | (pPayload[5] << 16) |
        (pPayload[6] << 8) | (pPayload[7] << 0);
    lctHeader.transportSessionIdentifier = 
        (pPayload[8] << 24) | (pPayload[9] << 16) |
        (pPayload[10] << 8) | (pPayload[11] << 0);
    lctHeader.transportObjectIdentifier = 
        (pPayload[12] << 24) | (pPayload[13] << 16) |
        (pPayload[14] << 8) | (pPayload[15] << 0);

    if (pROUTESession->verbosity >= ENVerbosity_PacketTrace)
    {
        LOGPRINT_NOFILE("%s: %"PRIu64": lct tsi: %"PRId32" toi: %"PRIu32"\n", 
            pROUTESession->strDebugName, packetNumber, 
            lctHeader.transportSessionIdentifier, 
            lctHeader.transportObjectIdentifier);
    }

    if (payloadLength < lctHeader.headerLength * 4)
    {
        DEBUGPRINT_NOCONSOLE(
            "%s: %"PRIu64" Packet too small for LCT extension headers\n", 
            pROUTESession->strDebugName, packetNumber);
        rval = ENStatus_Failure;
        goto out;
    }

    // Process extension headers.
    lctHeader.fExtTOLPresent = 0;
    lctHeader.tolTransferLength = 0;
    lctHeader.fExtFTIPresent = 0;
    lctHeader.ftiTransferLength = 0;
    lctHeader.fExtFDTPresent = 0;
    extensionHeaderOffset = 16;
    while (extensionHeaderOffset < lctHeader.headerLength * 4)
    {
        uint8_t het;
        uint8_t hel;

        het = pPayload[extensionHeaderOffset];

        if (het <= 127)
        {
            hel = pPayload[extensionHeaderOffset + 1];
            if (hel == 0)
            {
                DEBUGPRINT_NOCONSOLE(
                    "%s: %"PRIu64" zero extension header length het %d\n",
                    pROUTESession->strDebugName, packetNumber, het);
                rval = ENStatus_Failure;
                goto out;
            }
        }
        else 
        {
            hel = 1;
        }

        if (payloadLength < extensionHeaderOffset + 4 * hel)
        {
            DEBUGPRINT_NOCONSOLE(
                "%s: %"PRIu64" Packet too small for header extension %d\n", 
                pROUTESession->strDebugName, packetNumber, het);
            rval = ENStatus_Failure;
            goto out;
        }

        switch (het)
        {
            case 2: // EXT_TIME
                lctHeader.fExtTimePresent = 1;
                lctHeader.timeUseBits = 
                    ((uint16_t)pPayload[extensionHeaderOffset + 2] << 8) | 
                    ((uint16_t)pPayload[extensionHeaderOffset + 3] << 0);
                offset = extensionHeaderOffset + 4;
                if (lctHeader.fExtTimePresent & EXT_TIME_SCT_HI_PRESENT_FLAG)
                {
                    lctHeader.systemClockTimeHi = 
                        ((uint32_t)pPayload[offset + 0] << 24) |
                        ((uint32_t)pPayload[offset + 1] << 16) |
                        ((uint32_t)pPayload[offset + 2] << 8) |
                        ((uint32_t)pPayload[offset + 3] << 0);
                    offset += 4;
                }
                if (lctHeader.fExtTimePresent & EXT_TIME_SCT_LOW_PRESENT_FLAG)
                {
                    lctHeader.systemClockTimeLow = 
                        ((uint32_t)pPayload[offset + 0] << 24) |
                        ((uint32_t)pPayload[offset + 1] << 16) |
                        ((uint32_t)pPayload[offset + 2] << 8) |
                        ((uint32_t)pPayload[offset + 3] << 0);
                    offset += 4;
                }
                if (lctHeader.fExtTimePresent & EXT_TIME_SCT_ERT_PRESENT_FLAG)
                {
                    lctHeader.expectedResidualTime = 
                        ((uint32_t)pPayload[offset + 0] << 24) |
                        ((uint32_t)pPayload[offset + 1] << 16) |
                        ((uint32_t)pPayload[offset + 2] << 8) |
                        ((uint32_t)pPayload[offset + 3] << 0);
                    offset += 4;
                }
                if (lctHeader.fExtTimePresent & EXT_TIME_SCT_SLC_PRESENT_FLAG)
                {
                    lctHeader.sessionLastChangedTime = 
                        ((uint32_t)pPayload[offset + 0] << 24) |
                        ((uint32_t)pPayload[offset + 1] << 16) |
                        ((uint32_t)pPayload[offset + 2] << 8) |
                        ((uint32_t)pPayload[offset + 3] << 0);
                    offset += 4;
                }
                extensionHeaderOffset += 4 * hel;
                break;
            case 64:  // EXT_FTI
                lctHeader.fExtFTIPresent = 1;
                lctHeader.ftiTransferLength =
                    ((uint64_t)pPayload[extensionHeaderOffset + 2] << 40) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 3] << 32) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 4] << 24) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 5] << 16) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 6] << 8) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 7] << 0);
                lctHeader.ftiEncodingSymbolLength = 
                    (pPayload[extensionHeaderOffset + 10] << 8) |
                    (pPayload[extensionHeaderOffset + 11] << 0);
                lctHeader.ftiMaxSourceBlockLength =
                    (pPayload[extensionHeaderOffset + 12] << 24) |
                    (pPayload[extensionHeaderOffset + 13] << 16) |
                    (pPayload[extensionHeaderOffset + 14] << 8) |
                    (pPayload[extensionHeaderOffset + 15] << 0);
                extensionHeaderOffset += 4 * hel;
                break;
            case 66: // EXT_ROUTE_PRESENTATION_TIME
                lctHeader.fExtRoutePresentationTimePresent = 1;
                lctHeader.routePresentationTime = 
                    ((uint64_t)pPayload[extensionHeaderOffset + 2] << 40) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 3] << 32) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 4] << 24) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 5] << 16) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 6] << 8) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 7] << 0);
                extensionHeaderOffset += 4 * hel;
                break;
            case 67:  // 48-bit EXT_TOL
                lctHeader.fExtTOLPresent = 1;
                lctHeader.tolTransferLength = 
                    ((uint64_t)pPayload[extensionHeaderOffset + 2] << 40) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 3] << 32) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 4] << 24) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 5] << 16) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 6] << 8) |
                    ((uint64_t)pPayload[extensionHeaderOffset + 7] << 0);
                extensionHeaderOffset += 4 * hel;
                break;
            case 192: // FDT_INSTANCE
                lctHeader.fExtFDTPresent = 1;
                lctHeader.fdtInstanceVersion = 
                    pPayload[extensionHeaderOffset + 1] >> 4;
                lctHeader.fdtInstanceId =
                    ((pPayload[extensionHeaderOffset + 1] & 0xf) << 16) |
                    (pPayload[extensionHeaderOffset + 2] << 8) |
                    (pPayload[extensionHeaderOffset + 3] << 0);
                extensionHeaderOffset += 4;
                break;
            case 194: // 24-bit EXT_TOL
                lctHeader.fExtTOLPresent = 1;
                lctHeader.tolTransferLength = 
                    (pPayload[extensionHeaderOffset + 1] << 16) |
                    (pPayload[extensionHeaderOffset + 2] << 8) |
                    (pPayload[extensionHeaderOffset + 3] << 0);
                extensionHeaderOffset += 4;
                break;
            default:
                DEBUGPRINT_NOCONSOLE(
                    "%s: %"PRIu64" Unhandled header extension %d len %d\n", 
                    pROUTESession->strDebugName, packetNumber, het, hel * 4);
                if (het <= 127)
                {
                    extensionHeaderOffset += 4 * hel;
                }
                else
                {
                    extensionHeaderOffset += 4;
                }

                // If the extension length is zero then we are going to bail
                // to prevent spinning
                if (hel * 4 == 0)
                {
                    extensionHeaderOffset = lctHeader.headerLength * 4;
                }
                break;
        }
    }

    // It is possible for an LCT packet to have no payload (in which case
    // fecPayloadId will not be present). The no-payload case generally
    // occurs when information in the headers, but no payload, needs to be
    // conveyed.
    if (payloadLength >= lctHeader.headerLength * 4)
    {
        alcHeader.fPresent = 1;
        alcHeader.fecPayloadId = 
            (pPayload[lctHeader.headerLength * 4] << 24) | 
            (pPayload[lctHeader.headerLength * 4 + 1] << 16) | 
            (pPayload[lctHeader.headerLength * 4 + 2] << 8) | 
            (pPayload[lctHeader.headerLength * 4 + 3] << 0);

        if (pROUTESession->verbosity >= ENVerbosity_PacketTrace)
        {
            LOGPRINT_NOFILE("%s: %"PRIu64": alc: FEC Payload Id: %"PRId32"\n", 
                pROUTESession->strDebugName, packetNumber, 
                alcHeader.fecPayloadId);
        }
    }
    else
    {
        alcHeader.fPresent = 0;
        alcHeader.fecPayloadId = 0;
        DEBUGPRINT_NOCONSOLE("%s: %"PRId64" LCT packet with no payload\n", 
            pROUTESession->strDebugName, packetNumber);
    }

    for (i = 0; i < numof(pROUTESession->vLCTChannelInfos); i++)
    {
        if (pROUTESession->vLCTChannelInfos[i].fInUse &&
            pROUTESession->vLCTChannelInfos[i].tsi == 
            lctHeader.transportSessionIdentifier)
        {
            if (pROUTESession->fBitrateStatisticsEnabled)
            {
                ENATSC3BitrateStatisticsUpdate(
                    &pROUTESession->vLCTChannelInfos[i].stats, pPacketTimeUTC, 
                    payloadLength); 
            }

            switch (pROUTESession->vLCTChannelInfos[i].objectType)
            {
                case LCTObjectType_SLS:
                    if (ROUTEHandleSLSPayload(pPacketTimeUTC, packetNumber,
                        plpId, 
                        pPayload + lctHeader.headerLength * 4 + 4,
                        payloadLength - lctHeader.headerLength * 4 - 4, 
                        pROUTESession, &lctHeader, &alcHeader,
                        pROUTESession->vLCTChannelInfos + i))
                    {
                        DEBUGPRINT_NOCONSOLE(
                            "%s: %"PRIu64" Failed to process SLS packet\n", 
                            pROUTESession->strDebugName, packetNumber);
                        rval = ENStatus_Failure;
                        goto out;
                    }
                    break;
                case LCTObjectType_MediaFile:
                    if (ROUTEHandleMediaFilePayload(pPacketTimeUTC, 
                        packetNumber, plpId,  
                        pPayload + lctHeader.headerLength * 4 + 4,
                        payloadLength - lctHeader.headerLength * 4 - 4, 
                        pROUTESession, &lctHeader, alcHeader.fecPayloadId,
                        pROUTESession->vLCTChannelInfos + i))
                    {
                        DEBUGPRINT_NOCONSOLE(
                            "%s: %"PRIu64" Failed to process media file packet\n", 
                            pROUTESession->strDebugName, packetNumber);
                        rval = ENStatus_Failure;
                        goto out;
                    }
                    break;
                default:
                    DEBUGPRINT_NOCONSOLE(
                        "%s: %"PRIu64" Unhandled object type %d\n",
                        pROUTESession->strDebugName, packetNumber,
                        pROUTESession->vLCTChannelInfos[i].objectType);
            }
            break;
        }
    }

    if (i == numof(pROUTESession->vLCTChannelInfos))
    {
        if (pROUTESession->verbosity > ENVerbosity_Terse)
        {
            DEBUGPRINT_NOCONSOLE("%s: No LCT channel for tsi %d\n", 
                pROUTESession->strDebugName, 
                lctHeader.transportSessionIdentifier); 
        }
    }

out:
    return rval;
}

ENStatus
ENATSC3ROUTESessionAllocate(ENATSC3ROUTESessionType routeSessionType, 
   uint32_t slsAddr, int32_t slsPort, 
   ENATSC3ROUTESessionIPAddressAddCallbackFunction IPAddressAddCallback, 
   ENATSC3ROUTESessionIPAddressRemoveCallbackFunction IPAddressRemoveCallback,
   ENATSC3ROUTESessionIPAddressCommitCallbackFunction IPAddressCommitCallback,
   ENATSC3ROUTESessionVerifySignatureCallbackFunction VerifySignatureCallback,
   void *pCallbackUserData, char *pDebugName, 
   ENATSC3ROUTESessionHandle_t *phROUTESession)
{
    ENATSC3ROUTESession *pROUTESession = NULL;
    struct timespec timezero = {.tv_sec = 0, .tv_nsec = 0};
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pROUTESession = (ENATSC3ROUTESession *)calloc(sizeof(ENATSC3ROUTESession), 
        1);
    if (!pROUTESession)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    strncpy(pROUTESession->strDebugName, pDebugName, 
        sizeof(pROUTESession->strDebugName) - 1);
    pROUTESession->strDebugName[sizeof(pROUTESession->strDebugName) - 1] = 0;

    pROUTESession->type = routeSessionType;
    pROUTESession->IPAddressAddCallback = IPAddressAddCallback;
    pROUTESession->IPAddressRemoveCallback = IPAddressRemoveCallback;
    pROUTESession->IPAddressCommitCallback = IPAddressCommitCallback;
    pROUTESession->VerifySignatureCallback = VerifySignatureCallback;
    pROUTESession->pCallbackUserData = pCallbackUserData;

    pROUTESession->slsAddr = slsAddr;
    pROUTESession->slsPort = slsPort;

    for (i = 0; i < numof(pROUTESession->vObjectCallbacks); i++)
    {
        pROUTESession->vObjectCallbacks[i].fValid = 0;
    }

    for (i = 0; i < numof(pROUTESession->vMPDAvailableCallbacks); i++)
    {
        pROUTESession->vMPDAvailableCallbacks[i].fValid = 0;
    }

    for (i = 0; i < numof(pROUTESession->vSLSAvailableCallbacks); i++)
    {
        pROUTESession->vSLSAvailableCallbacks[i].fValid = 0;
    }

    for (i = 0; i < numof(pROUTESession->vHELDAvailableCallbacks); i++)
    {
        pROUTESession->vHELDAvailableCallbacks[i].fValid = 0;
    }

    for (i = 0; i < numof(pROUTESession->vSTSIDAvailableCallbacks); i++)
    {
        pROUTESession->vSTSIDAvailableCallbacks[i].fValid = 0;
    }

    if (ROUTELCTChannelInfoAllocate(&timezero, 0, pROUTESession, 0, "", 
        ENATSC3ROUTESessionIdType_Unknown, LCTObjectType_SLS, slsAddr, slsPort, 
        &pROUTESession->pSLSLCTChannelInfo))
    {
        ERRORPRINT("%s: Failed to allocate LCT channel for SLS\n", 
            pROUTESession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    if (pROUTESession->IPAddressCommitCallback)
    {
        pROUTESession->IPAddressCommitCallback(
            pROUTESession->pCallbackUserData);
    }

    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteFDTInstanceTOI = -1;
    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteSLSVersion = -1;
    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = 0;
    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteSLSTOI = -1;
    pROUTESession->pSLSLCTChannelInfo->sBuffer = SLS_BUFFER_SIZE;
    pROUTESession->pSLSLCTChannelInfo->pBuffer = 
        (uint8_t *)malloc(pROUTESession->pSLSLCTChannelInfo->sBuffer);
    if (!pROUTESession->pSLSLCTChannelInfo->pBuffer)
    {
        ERRORPRINT("%s: Out of memory\n", pROUTESession->strDebugName);
        rval = ENStatus_Failure;
        goto out;
    }

    ENATSC3BitrateStatisticsInitialize(&pROUTESession->stats, 2000);
    snprintf(pROUTESession->stats.strName, 
        sizeof(pROUTESession->stats.strName), "ROUTE %d.%d.%d.%d:%d", 
        (slsAddr >> 24) & 0xff, (slsAddr >> 16) & 0xff, 
        (slsAddr >> 8) & 0xff, slsAddr & 0xff, slsPort); 
    pROUTESession->stats.strName[sizeof(pROUTESession->stats.strName) - 1] = 0;

    *phROUTESession = pROUTESession;

out:
    return rval;

errout:
    if (pROUTESession)
    {
        if (pROUTESession->pSLSLCTChannelInfo)
        {
            struct timespec timezero = {.tv_sec = 0, .tv_nsec = 0};
            if (ROUTELCTChannelInfoReturn(&timezero, 0, pROUTESession, 
                pROUTESession->pSLSLCTChannelInfo))
            {
                ERRORPRINT("%s: Failed to return SLS LCT channel\n",
                    pROUTESession->strDebugName);
            }
        }

        if (pROUTESession->IPAddressCommitCallback)
        {
            pROUTESession->IPAddressCommitCallback(
                pROUTESession->pCallbackUserData);
        }

        if (pROUTESession->pSLSLCTChannelInfo->pBuffer)
        {
            free(pROUTESession->pSLSLCTChannelInfo->pBuffer);
        }

        free(pROUTESession);
    }

    goto out;
}

ENStatus
ENATSC3ROUTESessionFree(ENATSC3ROUTESessionHandle_t hROUTESession)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    struct timespec timezero = {.tv_sec = 0, .tv_nsec = 0};
    ENStatus rval = ENStatus_Success;

    ROUTELCTChannelInfoReturnAll(&timezero, 0, pROUTESession); 

    if (pROUTESession->IPAddressCommitCallback)
    {
        pROUTESession->IPAddressCommitCallback(
            pROUTESession->pCallbackUserData);
    }

    free(pROUTESession);

    return rval;
}

void
ENATSC3ROUTESessionSetBitrateStatisticsEnabled(
    ENATSC3ROUTESessionHandle_t hROUTESession, int32_t fEnabled)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;

    pROUTESession->fBitrateStatisticsEnabled = fEnabled;
}

ENStatus 
ENATSC3ROUTESessionSetVerbosity(ENATSC3ROUTESessionHandle_t hROUTESession,
    ENVerbosity verbosity)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    ENStatus rval = ENStatus_Success;

    pROUTESession->verbosity = verbosity;

    return rval;
}

ENStatus
ENATSC3ROUTESessionReset(ENATSC3ROUTESessionHandle_t hROUTESession, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber)
{
    ENATSC3ROUTESession *pROUTESession = (ENATSC3ROUTESession *)hROUTESession;
    LCTChannelInfo *pLCTChannelInfo;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    LOGPRINT("%s: Resetting ROUTE session\n", pROUTESession->strDebugName);

    ROUTECallSessionResetCallback(pPacketTimeUTC, packetNumber, pROUTESession);

    // Reset all LCT channels except for the signaling channel
    for (i = 0, pLCTChannelInfo = pROUTESession->vLCTChannelInfos; 
         i < numof(pROUTESession->vLCTChannelInfos); 
         i++, pLCTChannelInfo++)
    {
        if (pROUTESession->vLCTChannelInfos[i].fInUse)
        {
            // The LCT Channel for tsi 0 persists across resets. This channel
            // provides signaling information and we can continue to acquire
            // any signaling that is in-progress.
            if (pLCTChannelInfo->tsi != 0)
            {
                struct timespec timezero = {.tv_sec = 0, .tv_nsec = 0};
                if (ROUTELCTChannelInfoReturn(&timezero, 0, pROUTESession, 
                    pLCTChannelInfo))
                {
                    ERRORPRINT("%s: WARN: Failed to return LCT channel\n",
                        pROUTESession->strDebugName);
                }

                // But reset the expected TOI for SLS
                pLCTChannelInfo->expectedSLSTOI = -1;
		pLCTChannelInfo->lastEFDTcrc32c = 0;
		pLCTChannelInfo->lastEFDTSize = -1;
            }
        }
    }

    if (pROUTESession->IPAddressCommitCallback)
    {
        pROUTESession->IPAddressCommitCallback(
            pROUTESession->pCallbackUserData);
    }

    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteFDTInstanceTOI = -1;
    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteSLSVersion = -1;
    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteSLSCRC32C = 0;
    pROUTESession->pSLSLCTChannelInfo->slsInfo.lastCompleteSLSTOI = -1;

    return rval;
}
