/*
 * Copyright © 2019 - 2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of
 * the following permission added to Section 15 as permitted in Section 7(a):
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include "config.h"
  
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <arpa/inet.h>
#include <pthread.h>

#include <libxml2/libxml/xmlreader.h>

#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/cms.h>
#include <openssl/ocsp.h>

DEFINE_STACK_OF(OCSP_RESPONSE)
DEFINE_STACK_OF(OCSP_BASICRESP)

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3utils/enxml.h>
#include <libenatsc3lls/atsc3lls_cdt.h>
#include <libenatsc3lls/libenatsc3lls.h>

#include <enatsc3-common/enatsc3.h>

#include "libenatsc3ssp.h"

static char *ensys = "ssp";
static char *ensubsys = "main";

typedef struct _ENATSC3SSP
{
    // Contains CurrentCert from the CDT
    STACK_OF(X509) *pCurrentSMTSigningCertStack;
    // Contains NextCert from the CDT
    STACK_OF(X509) *pNextSMTSigningCertStack;
    // Contains all certificates from the CDT
    STACK_OF(X509) *pCertificates;
    // Contains all certificates from the CDT
    X509_STORE *pStore;

    STACK_OF(OCSP_RESPONSE) *pOCSPResponses;

    // The last signingTime from a CDT. 
    int64_t tLastCDTSigningTime;

    // The last signingTime from an SMT
    int64_t tLastSMTSigningTime;

    // Callback to request app to load any certificates it has in
    // secure storage and that are required by this SSP.
    ENATSC3SSPLoadCertificatesCallback LoadCertificatesCallback;
    void *pCallbackUserData;

    int32_t nReaders;
    pthread_mutex_t lock;
    pthread_cond_t cond;
} ENATSC3SSP, *ENATSC3SSPHandle_t;

ENStatus
ENATSC3SSPCreate(ENATSC3SSPLoadCertificatesCallback LoadCertificatesCallback, 
    void *pCallbackUserData, ENATSC3SSPHandle_t *phSSP)
{
    ENATSC3SSP *pSSP = NULL;
    int32_t fHaveLock = 0;
    int32_t fHaveCond = 0;
    int32_t fHaveCondattr = 0;
    pthread_condattr_t condattr;
    ENStatus rval = ENStatus_Success;

    pSSP = (ENATSC3SSP *)calloc(sizeof(ENATSC3SSP), 1);
    if (!pSSP)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (pthread_mutex_init(&pSSP->lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    if (pthread_condattr_init(&condattr))
    {
        ERRORPRINT("Failed to init condattr\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveCondattr = 1;

    if (pthread_condattr_setclock(&condattr, CLOCK_MONOTONIC))
    {
        ERRORPRINT("Failed to set condattr clock\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (pthread_cond_init(&pSSP->cond, &condattr))
    {
        ERRORPRINT("Failed to init cond\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveCond = 1;

    pSSP->LoadCertificatesCallback = LoadCertificatesCallback;
    pSSP->pCallbackUserData = pCallbackUserData;

    *phSSP = pSSP;

out:
    if (fHaveCondattr)
    {
        if (pthread_condattr_destroy(&condattr))
        {
            ERRORPRINT("Failed to destroy condattr\n");
        }
    }

    return rval;

errout:
    if (pSSP)
    {
        if (fHaveCond)
        {
            if (pthread_cond_destroy(&pSSP->cond))
            {
                ERRORPRINT("Failed to destory cond\n");
            }
        }

        if (fHaveLock)
        {
            if (pthread_mutex_destroy(&pSSP->lock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        free(pSSP);
    }

    goto out;
}

ENStatus
ENATSC3SSPDestroy(ENATSC3SSPHandle_t hSSP)
{
    ENATSC3SSP *pSSP = (ENATSC3SSP *)hSSP;
    ENStatus rval = ENStatus_Success;

    if (pSSP->pOCSPResponses)
    {
        sk_OCSP_RESPONSE_pop_free(pSSP->pOCSPResponses, OCSP_RESPONSE_free);
    }

    if (pSSP->pCurrentSMTSigningCertStack)
    {
        sk_X509_free(pSSP->pCurrentSMTSigningCertStack);
    }

    if (pSSP->pNextSMTSigningCertStack)
    {
        sk_X509_free(pSSP->pNextSMTSigningCertStack);
    }

    if (pSSP->pCertificates)
    {
        sk_X509_pop_free(pSSP->pCertificates, X509_free);
    }

    if (pSSP->pStore)
    {
        X509_STORE_free(pSSP->pStore);
    }

    if (pthread_cond_destroy(&pSSP->cond))
    {
        ERRORPRINT("Failed to destory cond\n");
    }

    if (pthread_mutex_destroy(&pSSP->lock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    free(pSSP);

    return rval;
}

static ENStatus
ENATSC3SSPCertificateStoreRequestAccess(ENATSC3SSP *pSSP)
{
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSSP->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSSP->nReaders++;

    if (pthread_mutex_unlock(&pSSP->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

static ENStatus
ENATSC3SSPCertificateStoreRelease(ENATSC3SSP *pSSP)
{
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSSP->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (!(--pSSP->nReaders))
    {
        if (pthread_cond_signal(&pSSP->cond))
        {
            ERRORPRINT("Failed to signal\n");
        }
    }

    if (pthread_mutex_unlock(&pSSP->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

static ENStatus
ENATSC3SSPCertificateStoreReplaceCertificates(ENATSC3SSP *pSSP,
    STACK_OF(X509) *pCertificates, STACK_OF(OCSP_RESPONSE) *pOCSPResponses,
    STACK_OF(X509) *pCurrentSMTSigningCert, 
    STACK_OF(X509) *pNextSMTSigningCert, X509_STORE *pStore, 
    int64_t signingTime, const struct timespec *pAbsTime)
{
    int ptret;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSSP->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    while(pSSP->nReaders)
    {
        ptret = pthread_cond_timedwait(&pSSP->cond, &pSSP->lock, pAbsTime);
        if (ptret)
        {
            if (ptret == ETIMEDOUT)
            {
                ERRORPRINT("Timeout gaining access to store\n");
                rval = ENStatus_Retry;
            }
            else
            {
                ERRORPRINT("timedwait failed: %s\n", strerror(errno));
                rval = ENStatus_Failure;
            }
            break;
        }
    }

    if (!rval)
    {
        if (signingTime < pSSP->tLastCDTSigningTime)
        {
            ERRORPRINT(
                "CDT signing time went backwards (%"PRId64" < %"PRId64")\n",
                signingTime, pSSP->tLastCDTSigningTime);
            rval = ENStatus_Failure;
        }
    }

    if (!rval)
    {
        if (pSSP->pCertificates)
        {
            sk_X509_pop_free(pSSP->pCertificates, X509_free);
        }
        pSSP->pCertificates = pCertificates;

        if (pSSP->pOCSPResponses)
        {
            sk_OCSP_RESPONSE_pop_free(pSSP->pOCSPResponses, 
                OCSP_RESPONSE_free);
        }
        pSSP->pOCSPResponses = pOCSPResponses;
        pSSP->pCurrentSMTSigningCertStack = pCurrentSMTSigningCert;
        pSSP->pNextSMTSigningCertStack = pNextSMTSigningCert;
        pSSP->pStore = pStore;
    }
    
    if (pthread_mutex_unlock(&pSSP->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

static ENStatus
ENATSC3SSPVerifySignature_Internal(STACK_OF(X509) *pStack, X509_STORE *pStore, 
    CMS_ContentInfo *pCMS, const uint8_t *pContent, int32_t contentLength, 
    int32_t fDontVerifySigningCertificate, int64_t *pSigningTime)
{
    BIO *contentMemBIO = NULL;
    int flags = CMS_DETACHED | CMS_BINARY;
    int32_t fHaveSigningTime = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (fDontVerifySigningCertificate)
    {
        ERRORPRINT("WARN: Not verifying signer certificate chain\n");
        flags |= CMS_NO_SIGNER_CERT_VERIFY;
    }

    STACK_OF(CMS_SignerInfo) *pCMSSignerInfos;

    pCMSSignerInfos = CMS_get0_SignerInfos(pCMS);
    if (pCMSSignerInfos)
    {
        for (i = 0; i < sk_CMS_SignerInfo_num(pCMSSignerInfos); i++)
        {
            CMS_SignerInfo *pCMSSignerInfo;
            int loc;
            int32_t j;

            pCMSSignerInfo = sk_CMS_SignerInfo_value(pCMSSignerInfos, i);

#if 0
            // The following could be used to determine which certificate was
            // used to sign, particularly for the CDT. For the SMT CurrentCert
            // and NextCert are called out in the CDT. However at the moment
            // we're letting CMS_verify examine all certs (in the case of the
            // CDT).
            ASN1_OCTET_STRING *pKeyId;
            X509_NAME *pIssuer;
            ASN1_INTEGER *pSerial;
            if (CMS_SignerInfo_get0_signer_id(pCMSSignerInfo, 
                &pKeyId, &pIssuer, &pSerial) != 1)
            {
                ERRORPRINT("Failed to get SignerInfo ids\n");
                rval = ENStatus_Failure;
                goto out;
            }
#endif

            loc = CMS_signed_get_attr_by_NID(pCMSSignerInfo, 
                NID_pkcs9_signingTime, -1);
            if (loc != -1)
            {
                X509_ATTRIBUTE *pAttr;

                pAttr = CMS_signed_get_attr(pCMSSignerInfo, loc);
                for (j = 0; j < X509_ATTRIBUTE_count(pAttr); j++)
                {
                    ASN1_TYPE *pType;

                    pType = X509_ATTRIBUTE_get0_type(pAttr, j);
                    if (pType->type == V_ASN1_UTCTIME)
                    {
                        struct tm signingTime_tm;

                        if (ASN1_TIME_to_tm(pType->value.asn1_string, 
                            &signingTime_tm) != 1)
                        {
                            ERRORPRINT("Failed to parse signingTime\n");
                            rval = ENStatus_Failure;
                            goto out;
                        }
                        *pSigningTime = timegm(&signingTime_tm);
                        if (*pSigningTime > time(NULL))
                        {
                            ERRORPRINT("Invalid signingTime %"PRId64"\n",
                                *pSigningTime);
                            rval = ENStatus_Failure;
                            goto out;
                        }
                        fHaveSigningTime = 1;
                    }
                }
            }
        }
    }

    if (!fHaveSigningTime)
    {
        ERRORPRINT("No signing time in CMS\n");
        rval = ENStatus_Failure;
        goto out;
    }

    contentMemBIO = BIO_new_mem_buf(pContent, contentLength);
    if (!contentMemBIO)
    {
        ERRORPRINT("Failed to create content memory bio\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (CMS_verify(pCMS, pStack, pStore, contentMemBIO, NULL, flags) != 1)
    {
        unsigned long err;
        char strError[1024];

        ERRORPRINT("CMS verification failed\n");
        while ((err = ERR_get_error()))
        {
            ERR_error_string_n(err, strError, sizeof(strError));
            strError[sizeof(strError) -1] = 0;
            ERRORPRINT("%s\n", strError);
        }
        rval = ENStatus_Failure;
        goto out;
    }

out:
    if (contentMemBIO)
    {
        BIO_free(contentMemBIO);
    }

    return rval;
}

static CMS_ContentInfo *
ReadCMSFromPKCS7(BIO *bio)
{
    BIO *b64BIO;
    CMS_ContentInfo *pCMS = NULL;

    b64BIO = BIO_new(BIO_f_base64());
    if (!b64BIO)
    {
        ASN1err(ASN1_F_B64_READ_ASN1, ERR_R_MALLOC_FAILURE);
        goto out;
    }

    bio = BIO_push(b64BIO, bio);

    pCMS = (CMS_ContentInfo *)ASN1_item_d2i_bio(ASN1_ITEM_rptr(
        CMS_ContentInfo), bio, NULL);
    if (!pCMS)
    {
        ASN1err(ASN1_F_B64_READ_ASN1, ASN1_R_DECODE_ERROR);
    }

    (void)BIO_flush(bio);
    BIO_pop(bio);
    BIO_free(b64BIO);

out:
    return pCMS;
}

ENStatus
ENATSC3SSPVerifySignature(ENATSC3SSPHandle_t hSSP,
    const uint8_t *pSignature, int32_t signatureLength, 
    const uint8_t *pSignedData, int32_t signedDataLength,
    int32_t fDontVerifySigningCertificate, int64_t *pSigningTime)
{
    ENATSC3SSP *pSSP = (ENATSC3SSP *)hSSP;
    BIO *signatureMemBIO = NULL;
    CMS_ContentInfo *pCMS = NULL;
    ENStatus rval = ENStatus_Success;

    while (signatureLength > 4)
    {
        if (pSignature[0] == '\r' && pSignature[1] == '\n' && 
            pSignature[2] == '\r' && pSignature[3] == '\n')
        {
            pSignature += 4;
            signatureLength -= 4;
            break;
        }
        pSignature++;
        signatureLength--;
    }

    signatureMemBIO = BIO_new_mem_buf(pSignature, signatureLength);
    if (!signatureMemBIO)
    {
        ERRORPRINT("Failed to create signature membio\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pCMS = ReadCMSFromPKCS7(signatureMemBIO); 

    BIO_free(signatureMemBIO);

    if (!pCMS)
    {
        ERRORPRINT("Failed to read CMS\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENATSC3SSPCertificateStoreRequestAccess(pSSP))
    {
        ERRORPRINT("Failed to get SSP access\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (!pSSP->pCurrentSMTSigningCertStack && !pSSP->pNextSMTSigningCertStack)
    {
        ERRORPRINT("No signing certificate stack\n");
        rval = ENStatus_NotReady;
        goto releaseout;
    }

    if (ENATSC3SSPVerifySignature_Internal(pSSP->pCurrentSMTSigningCertStack, 
        NULL, pCMS, pSignedData, signedDataLength, 
        fDontVerifySigningCertificate, pSigningTime))
    {
        ERRORPRINT(
            "Failed to verify signed multitable signature with CurrentCert\n");
        if (pSSP->pNextSMTSigningCertStack) 
        {
            if (ENATSC3SSPVerifySignature_Internal(
                pSSP->pNextSMTSigningCertStack, NULL, pCMS,
                pSignedData, signedDataLength, 
                fDontVerifySigningCertificate, pSigningTime))
            {
                ERRORPRINT("Failed to verify signed multitable signature with NextCert\n");
                rval = ENStatus_Failure;
            }
        }
        else
        {
            ERRORPRINT("No next cert\n");
            rval = ENStatus_Failure;
        }
    }

releaseout:
    if (ENATSC3SSPCertificateStoreRelease(pSSP))
    {
        ERRORPRINT("WARN: Failed to release SSP\n");
    }

out:
    if (pCMS)
    {
        CMS_ContentInfo_free(pCMS);
    }

    return rval;
}

ENStatus
ENATSC3SSPVerifySignedMultitableSignature(ENATSC3SSPHandle_t hSSP, 
    const uint8_t *pSignedMultitable, int32_t signedMultitableLen,
    int32_t fDontVerifySigningCertificate)
{
    ENATSC3SSP *pSSP = (ENATSC3SSP *)hSSP;
    int32_t payloadCount;
    struct _LLSPayloadHeader
    {
        uint8_t payloadId;
        uint8_t payloadVersion;
        uint16_t payloadLength;
    };
    const uint8_t *pIter;
    int32_t iterLen;
    const uint8_t *pSignedData;
    int32_t signedDataLength;
    uint16_t signatureLength;
    const uint8_t *pSignature;
    int64_t signingTime;
    BIO *signatureMemBIO;
    CMS_ContentInfo *pCMS = NULL;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (signedMultitableLen < 1)
    {
        ERRORPRINT("SignedMultitable too short\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pSignedData = pSignedMultitable;
    pIter = pSignedMultitable;
    iterLen = signedMultitableLen;

    payloadCount = *(pIter++);
    iterLen--;

    for (i = 0;
         i < payloadCount && iterLen > sizeof(struct _LLSPayloadHeader);
         i++)
    {
        uint16_t payloadLength;

        payloadLength = ntohs(
            ((struct _LLSPayloadHeader*)pIter)->payloadLength);
        pIter += sizeof(struct _LLSPayloadHeader) + payloadLength;
        iterLen -= sizeof(struct _LLSPayloadHeader) + payloadLength;
    }

    signedDataLength = pIter - pSignedMultitable;

    if (iterLen < 2)
    {
        ERRORPRINT("SignedMultitable too short\n");
        rval = ENStatus_Failure;
        goto out;
    }
    signatureLength = (pIter[0] << 8) | pIter[1];
    pIter += 2;
    iterLen -= 2;

    pSignature = pIter;
    if (iterLen < signatureLength)
    {
        ERRORPRINT("SignedMultitable too short\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENATSC3SSPCertificateStoreRequestAccess(pSSP))
    {
        ERRORPRINT("Failed to get SSP access\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (!pSSP->pCurrentSMTSigningCertStack && !pSSP->pNextSMTSigningCertStack)
    {
        ERRORPRINT("No signing certificate stack\n");
        rval = ENStatus_NotReady;
        goto releaseout;
    }

    signatureMemBIO = BIO_new_mem_buf(pSignature, signatureLength);
    if (!signatureMemBIO)
    {
        ERRORPRINT("Failed to create signature membio\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pCMS = d2i_CMS_bio(signatureMemBIO, NULL);

    BIO_free(signatureMemBIO);

    if (!pCMS)
    {
        ERRORPRINT("Failed to read CMS\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENATSC3SSPVerifySignature_Internal(pSSP->pCurrentSMTSigningCertStack, 
        NULL, pCMS, pSignedData, signedDataLength, 
        fDontVerifySigningCertificate, &signingTime))
    {
        ERRORPRINT(
            "Failed to verify signed multitable signature with CurrentCert\n");
        if (pSSP->pNextSMTSigningCertStack) 
        {
            if (ENATSC3SSPVerifySignature_Internal(
                pSSP->pNextSMTSigningCertStack, NULL, pCMS, pSignedData,
                signedDataLength, fDontVerifySigningCertificate, &signingTime))
            {
                ERRORPRINT("Failed to verify signed multitable signature with NextCert\n");
                rval = ENStatus_Failure;
            }
        }
        else
        {
            ERRORPRINT("No next cert\n");
            rval = ENStatus_Failure;
        }
    }
    
    if (rval == ENStatus_Success)
    {
        if (signingTime < pSSP->tLastSMTSigningTime)
        {
            ERRORPRINT(
                "SMT signing time went backwards (%"PRId64" < %"PRId64")\n",
                signingTime, pSSP->tLastSMTSigningTime);
            rval = ENStatus_Failure;
        }
    }

    if (rval == ENStatus_Success)
    {
        pSSP->tLastSMTSigningTime = signingTime;
    }

releaseout:
    if (ENATSC3SSPCertificateStoreRelease(pSSP))
    {
        ERRORPRINT("WARN: Failed to release SSP\n");
    }

out:
    if (pCMS)
    {
        CMS_ContentInfo_free(pCMS);
    }

    return rval;
}

static ENStatus
ENATSC3SSPExtractCertificationDataTableCertificates(
    ATSC3LLSCDT_CertificationDataNode *pParsedCDT, 
    STACK_OF(X509) *pCertificatesStack,
    STACK_OF(X509) *pCurrentSMTSigningCertStack, 
    STACK_OF(X509) *pNextSMTSigningCertStack,
    X509_STORE *pStore)
{
    int32_t i;
    uint8_t vDecoded[16384];
    uint8_t vDecodedCurrentCertSubjectKeyIdentifier[4096];
    int32_t currentCertSubjectKeyIdentifiertDecodedLen;
    uint8_t vDecodedNextCertSubjectKeyIdentifier[4096];
    int32_t nextCertSubjectKeyIdentifiertDecodedLen;
    int32_t fHaveNextCertSubjectKeyIdentifier = 0;
    char *pText;
    ENStatus rval = ENStatus_Success;

    if (!(pParsedCDT->base.presentFlags & 
        ATSC3LLSCDTRTChildNode_ToBeSignedData))
    {
        ERRORPRINT("ToBeSignedData missing\n");
        rval = ENStatus_Failure;
        goto out;
    }

    // Get CurrentCert
    if (!(pParsedCDT->toBeSignedDataNode.base.presentFlags & 
        ATSC3LLCDTToBeSignedDataChildNode_CurrentCert) || 
        !pParsedCDT->toBeSignedDataNode.currentCertNode.base.textPresent)
    {
        ERRORPRINT("CurrentCert missing or empty\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pText = (char *)EnXMLParserDataBufferGetPointer(
        &pParsedCDT->dataBuffer, 
        pParsedCDT->toBeSignedDataNode.currentCertNode.textDataOffset);
    if (ENUtils_Base64DecodedSize(strlen(pText)) > 
        sizeof(vDecodedCurrentCertSubjectKeyIdentifier))
    {
        ERRORPRINT("SubjectKeyIdentifier too long (%ld > %ld)\n",
            ENUtils_Base64DecodedSize(strlen(pText)), 
            sizeof(vDecodedCurrentCertSubjectKeyIdentifier));
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENUtils_Base64Decode(pText, strlen(pText), 
        vDecodedCurrentCertSubjectKeyIdentifier, 
        &currentCertSubjectKeyIdentifiertDecodedLen))
    {
        ERRORPRINT("Failed to decode CurrentCert\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if ((pParsedCDT->toBeSignedDataNode.base.presentFlags & 
         ATSC3LLCDTToBeSignedDataChildNode_CertReplacment) &&
        (pParsedCDT->toBeSignedDataNode.certReplacementNode.base.presentFlags &
         ATSC3LLSCDTCertReplacementChildNode_NextCert))
    {
        ATSC3LLSCDT_NextCertNode *pNextCertNode;

        pNextCertNode = 
            &pParsedCDT->toBeSignedDataNode.certReplacementNode.nextCertNode;
        if (pNextCertNode->base.textPresent)
        {
            pText = (char *)EnXMLParserDataBufferGetPointer(
                &pParsedCDT->dataBuffer, pNextCertNode->textDataOffset);
            if (ENUtils_Base64DecodedSize(strlen(pText)) > 
                sizeof(vDecodedNextCertSubjectKeyIdentifier))
            {
                ERRORPRINT("SubjectKeyIdentifier too long (%ld > %ld)\n",
                    ENUtils_Base64DecodedSize(strlen(pText)), 
                    sizeof(vDecodedNextCertSubjectKeyIdentifier));
                rval = ENStatus_Failure;
                goto out;
            }

            if (ENUtils_Base64Decode(pText, strlen(pText), 
                vDecodedNextCertSubjectKeyIdentifier, &
                nextCertSubjectKeyIdentifiertDecodedLen))
            {
                ERRORPRINT("Failed to decode NextCert\n");
                rval = ENStatus_Failure;
                goto out;
            }
            fHaveNextCertSubjectKeyIdentifier = 1;
        }
    }

    for (i = 0; 
         i < pParsedCDT->toBeSignedDataNode.certificates.nOffsetsUsed; 
         i++)
    {
        ATSC3LLSCDT_CertificatesNode *pCertificatesNode;
        int32_t decodedLen;

        pCertificatesNode = (ATSC3LLSCDT_CertificatesNode *)
            EnXMLParserDataOffsetArrayGetPointer(
                &pParsedCDT->toBeSignedDataNode.certificates, i);
        if (pCertificatesNode->base.textPresent)
        {
            pText = (char *)EnXMLParserDataBufferGetPointer(
                &pParsedCDT->dataBuffer, pCertificatesNode->textDataOffset);
            if (ENUtils_Base64DecodedSize(strlen(pText)) > sizeof(vDecoded))
            {
                ERRORPRINT("Certificate too long (%ld > %ld)\n",
                    ENUtils_Base64DecodedSize(strlen(pText)),
                    sizeof(vDecoded));
                rval = ENStatus_Failure;
                goto out;
            }
            if (ENUtils_Base64Decode(pText, strlen(pText), vDecoded, 
                &decodedLen))
            {
                ERRORPRINT("Failed to base64 decode certificate\n");
                rval = ENStatus_Failure;
                goto out;
            }
            else
            {
                X509 *pX509 = NULL;
                const unsigned char *pDERCertificate;

                pDERCertificate = (const unsigned char *)vDecoded;
                pX509 = d2i_X509(NULL, &pDERCertificate, decodedLen);
                if (!pX509)
                {
                    ERRORPRINT("Failed to decode X509 cert\n");
                    rval = ENStatus_Failure;
                    goto out;
                }

                ASN1_OCTET_STRING *pSubjectKeyIdentifier_der;
                pSubjectKeyIdentifier_der = X509_get_ext_d2i(pX509, 
                    NID_subject_key_identifier, NULL, NULL); 
                if (pSubjectKeyIdentifier_der)
                {
                    size_t len;
                    const uint8_t *pData;
                    len = ASN1_STRING_length(pSubjectKeyIdentifier_der);
                    pData = ASN1_STRING_get0_data(pSubjectKeyIdentifier_der);

                    if (len == currentCertSubjectKeyIdentifiertDecodedLen)
                    {
                        if (!memcmp(pData, 
                            vDecodedCurrentCertSubjectKeyIdentifier, len))
                        {
                            if (!sk_X509_push(pCurrentSMTSigningCertStack, 
                                pX509))
                            {
                                ERRORPRINT("Failed to add X509 cert\n");
                                X509_free(pX509);
                                rval = ENStatus_Failure;
                                goto out;
                            }
                        }
                    }

                    if (fHaveNextCertSubjectKeyIdentifier)
                    {
                        if (len == nextCertSubjectKeyIdentifiertDecodedLen)
                        {
                            if (!memcmp(pData,
                                vDecodedNextCertSubjectKeyIdentifier, len))
                            {
                                if (!sk_X509_push(pNextSMTSigningCertStack, 
                                    pX509))
                                {
                                    ERRORPRINT("Failed to add X509 cert\n");
                                    X509_free(pX509);
                                    rval = ENStatus_Failure;
                                    goto out;
                                }
                            }
                        }
                    }

                    ASN1_STRING_free(pSubjectKeyIdentifier_der);
                }

                if (X509_STORE_add_cert(pStore, pX509) != 1)
                {
                    ERRORPRINT("Failed to add X509 cert\n");
                    X509_free(pX509);
                    rval = ENStatus_Failure;
                    goto out;
                }

                if (!sk_X509_push(pCertificatesStack, pX509))
                {
                    ERRORPRINT("Failed to add X509 cert\n");
                    X509_free(pX509);
                    rval = ENStatus_Failure;
                    goto out;
                }
            }
        }
    }

out:
    return rval;
}

static ENStatus
ParseOCSPResponse(const uint8_t *pOCSPASN1, int32_t ocspASN1Length, 
    OCSP_RESPONSE **ppOCSPResponse) 
{
    BIO *ocspMemBIO = NULL;
    ENStatus rval = ENStatus_Success;

    ocspMemBIO = BIO_new_mem_buf(pOCSPASN1, ocspASN1Length);
    if (!ocspMemBIO)
    {
        ERRORPRINT("Failed to create membio\n");
        rval = ENStatus_Failure;
        goto out;
    }

    *ppOCSPResponse = d2i_OCSP_RESPONSE_bio(ocspMemBIO, NULL);

    BIO_free(ocspMemBIO);

    if (!*ppOCSPResponse)
    {
        ERRORPRINT("Failed to read OCSP response\n");
        rval = ENStatus_Failure;
        goto out;
    }

out:
   return rval;
}

static ENStatus
ENATSC3SSPExtractCertificationDataTableOCSPResponses(
    ATSC3LLSCDT_CertificationDataNode *pParsedCDT, 
    STACK_OF(OCSP_RESPONSE) *pOCSPResponses)
{
    int32_t i;
    uint8_t vDecoded[16384];
    int32_t decodedLen;
    ENStatus rval = ENStatus_Success;

    if (pParsedCDT->base.presentFlags & ATSC3LLSCDTChildNode_OCSPResponse)
    {
        for (i = 0; 
             i < pParsedCDT->ocspResponses.nOffsetsUsed; 
             i++)
        {
            ATSC3LLSCDT_OCSPResponseNode *pOCSPResponseNode;
            char *pText;

            pOCSPResponseNode = (ATSC3LLSCDT_OCSPResponseNode *)
                EnXMLParserDataOffsetArrayGetPointer(
                    &pParsedCDT->ocspResponses, i);
            if (pOCSPResponseNode->base.textPresent)
            {
                pText = (char *)EnXMLParserDataBufferGetPointer(
                    &pParsedCDT->dataBuffer, pOCSPResponseNode->textDataOffset);
                if (ENUtils_Base64DecodedSize(strlen(pText)) > sizeof(vDecoded))
                {
                    ERRORPRINT("OCSP response too long (%ld > %ld)\n",
                        ENUtils_Base64DecodedSize(strlen(pText)),
                        sizeof(vDecoded));
                    rval = ENStatus_Failure;
                    goto errout;
                }
                if (ENUtils_Base64Decode(pText, strlen(pText),
                    vDecoded, &decodedLen))
                {
                    ERRORPRINT("Failed to base64 decode OCSP response`\n");
                }
                else
                {
                    OCSP_RESPONSE *pOCSPResponse;

                    if (ParseOCSPResponse(vDecoded, decodedLen, 
                        &pOCSPResponse))
                    {
                        ERRORPRINT("Failed to parse OCSP response\n");
                        rval = ENStatus_Failure;
                        goto errout;
                    }

                    if (!sk_OCSP_RESPONSE_push(pOCSPResponses, pOCSPResponse))
                    {
                        ERRORPRINT("Failed to add OCSP response\n");
                        OCSP_RESPONSE_free(pOCSPResponse);
                        rval = ENStatus_Failure;
                        goto errout;
                    }
                }
            }
        }
    }

out:
    return rval;

errout:
    goto out;
}

static ENStatus
ENATSC3SSPVerifyCertificationDataSignature(ENATSC3SSP *pSSP, char *pXML, 
    int32_t xmlLength, 
    ATSC3LLSCDT_CertificationDataNode *pParsedCertificationData,
    STACK_OF(X509) *pStack, X509_STORE *pChainStore,
    int32_t fDontVerifySigningCertificate, int64_t *pSigningTime)
{
    uint8_t vSignature[4096];
    int32_t signatureLength;
    uint8_t *pSignedData;
    uint8_t *pSignedDataEnd;
    char *pCMSSignedData;
    char *pXMLEnd;
    BIO *signatureMemBIO;
    CMS_ContentInfo *pCMS = NULL;
    ENStatus rval = ENStatus_Success;

    pXMLEnd = pXML + xmlLength;

    if (!(pParsedCertificationData->base.presentFlags &
           ATSC3LLSCDTChildNode_CMSSignedData) ||
        !pParsedCertificationData->cmsSignedDataNode.base.textPresent)
    {
        ERRORPRINT("CertificationData missing or empty CMSSignedData\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pCMSSignedData = (char *)EnXMLParserDataBufferGetPointer(
        &pParsedCertificationData->dataBuffer,
        pParsedCertificationData->cmsSignedDataNode.textDataOffset);

    if (ENUtils_Base64DecodedSize(strlen(pCMSSignedData)) > sizeof(vSignature))
    {
        ERRORPRINT("Signature too long (%ld > %ld)\n",
            ENUtils_Base64DecodedSize(strlen(pCMSSignedData)),
            sizeof(vSignature));
        rval = ENStatus_Failure;
        goto out;
    }
    if (ENUtils_Base64Decode(pCMSSignedData, strlen(pCMSSignedData),
        vSignature, &signatureLength))
    {
        ERRORPRINT("Failed to base64 decode\n");
        rval = ENStatus_Failure;
        goto out;
    }

    // Find the boundaries of ToBeSignedData. Per A/360 the signed data spans
    // the ToBoSignedData element from the opening '<' to the closing '>'.
    // Note that per XML spec whitespace is not allowed between the opening '<'
    // and a tag name, however whitespace is allowed before the closing '>'.
    //
    // It would be nice if libxml2  provided the starting and ending document
    // location for each node, but it doesn't at present.
    pSignedData = memmem(pXML, xmlLength, "<ToBeSignedData",
       sizeof("<ToBeSignedData") - 1);
    if (!pSignedData)
    {
        ERRORPRINT("Failed to find ToBeSignedData\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pSignedDataEnd = memmem(pSignedData + sizeof("<ToBeSignedData") - 1,
        (uint8_t *)pXMLEnd - pSignedData - (sizeof("<ToBeSignedData") - 1),
        "</ToBeSignedData", sizeof("</ToBeSignedData") -1);
    if (!pSignedDataEnd)
    {
        ERRORPRINT("Failed to find ToBeSignedData end element\n");
        rval = ENStatus_Failure;
        goto out;
    }
    pSignedDataEnd += sizeof("</ToBeSignedData") - 1;
    while (isspace(*pSignedDataEnd) && pSignedDataEnd < (uint8_t *)pXMLEnd)
    {
        pSignedDataEnd++;
    }
    if (pSignedDataEnd == (uint8_t *)pXMLEnd || *pSignedDataEnd != '>')
    {
        ERRORPRINT("Malformed ToBeSignedData end element\n");
        rval = ENStatus_Failure;
        goto out;
    }
    pSignedDataEnd++;

    signatureMemBIO = BIO_new_mem_buf(vSignature, signatureLength);
    if (!signatureMemBIO)
    {
        ERRORPRINT("Failed to create signature membio\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pCMS = d2i_CMS_bio(signatureMemBIO, NULL);

    BIO_free(signatureMemBIO);

    if (!pCMS)
    {
        ERRORPRINT("Failed to read CMS\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (ENATSC3SSPVerifySignature_Internal(pStack, pChainStore,
        pCMS, pSignedData, pSignedDataEnd - pSignedData,
        fDontVerifySigningCertificate, pSigningTime))
    {
        ERRORPRINT("Failed to verify CertificationData signature\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (*pSigningTime < pSSP->tLastCDTSigningTime)
    {
        ERRORPRINT("CDT signing time went backwards (%"PRId64" < %"PRId64")\n",
            *pSigningTime, pSSP->tLastCDTSigningTime);
        rval = ENStatus_Failure;
        goto out;
    }

out:
    if (pCMS)
    {
        CMS_ContentInfo_free(pCMS);
    }

    return rval;
}

static ENStatus
ENATSC3SSPVerifyOCSPResponses(STACK_OF(OCSP_RESPONSE) *pOCSPResponseStack,
    STACK_OF(X509) *pCertificates, X509_STORE *pStore, 
    int64_t ocspRefreshSeconds, int32_t fDontVerifyCertificateChain, 
    int32_t fIgnoreOCSPResponseExpired)
{
    int opensslret;
    int32_t i;
    int64_t tNowUTC;
    STACK_OF(OCSP_BASICRESP) *pBasicResponseStack = NULL;
    BIO *out = NULL;
    ENStatus rval = ENStatus_Success;

    out = BIO_new_fp(stdout, BIO_NOCLOSE);

    tNowUTC = time(NULL);

    pBasicResponseStack = sk_OCSP_BASICRESP_new_null();

    for (i = 0; i < sk_OCSP_RESPONSE_num(pOCSPResponseStack); i++)
    {
        OCSP_RESPONSE *pOCSPResponse;
        OCSP_BASICRESP *pOCSPBasicResponse;
        const ASN1_GENERALIZEDTIME *pProducedAtASN1;
        struct tm tProducedAtTime_tm;
        int64_t tProducedAtTimeUTC;

        pOCSPResponse = sk_OCSP_RESPONSE_value(pOCSPResponseStack, i);
        pOCSPBasicResponse = OCSP_response_get1_basic(pOCSPResponse);

        pProducedAtASN1 = OCSP_resp_get0_produced_at(pOCSPBasicResponse);
        if (ASN1_TIME_to_tm(pProducedAtASN1, &tProducedAtTime_tm) != 1)
        {
            ERRORPRINT("Failed to parse producedAtTime\n");
            rval = ENStatus_Failure;
            OCSP_BASICRESP_free(pOCSPBasicResponse);
            goto out;
        }
        tProducedAtTimeUTC = timegm(&tProducedAtTime_tm);
        if (tProducedAtTimeUTC == -1)
        {
            ERRORPRINT("Failed to convert producedAtTime to UTC\n");
            rval = ENStatus_Failure;
            OCSP_BASICRESP_free(pOCSPBasicResponse);
            goto out;
        }

        if (tProducedAtTimeUTC + ocspRefreshSeconds < tNowUTC)
        {
            ERRORPRINT(
                "OCSPResponse expired (%"PRId64" + %"PRId64" < %"PRId64")%s\n",
                tProducedAtTimeUTC, ocspRefreshSeconds, tNowUTC,
                fIgnoreOCSPResponseExpired ? " (ignored)" : "");
            if (!fIgnoreOCSPResponseExpired)
            {
                rval = ENStatus_Failure;
                OCSP_BASICRESP_free(pOCSPBasicResponse);
                goto out;
            }
        }

        opensslret = OCSP_basic_verify(pOCSPBasicResponse, pCertificates, 
            pStore, fDontVerifyCertificateChain ? OCSP_NOVERIFY : 0);
        if (opensslret != 1)
        {
            ERRORPRINT("Failed to verify OCSP: %d\n", opensslret);
            rval = ENStatus_Failure;
            OCSP_BASICRESP_free(pOCSPBasicResponse);
            goto out;
        }

        sk_OCSP_BASICRESP_push(pBasicResponseStack, pOCSPBasicResponse);
    }

    for (i = 0; i < sk_X509_num(pCertificates); i++)
    {
        X509 *pCertificate;
        const ASN1_INTEGER *pCertId;
        int32_t j;
        int status = -1;

        pCertificate = sk_X509_value(pCertificates, i);

        pCertId = X509_get0_serialNumber(pCertificate);
        if (!pCertId)
        {
            ERRORPRINT("Failed to ger serial number for cert %d\n", i);
            rval = ENStatus_Failure;
            goto out;
        }

        for (j = 0; j < sk_OCSP_BASICRESP_num(pBasicResponseStack); j++)
        {
            OCSP_BASICRESP *pOCSPBasicResponse;
            int32_t k;

            pOCSPBasicResponse = sk_OCSP_BASICRESP_value(pBasicResponseStack, 
                j);
            if (!pOCSPBasicResponse)
            {
                ERRORPRINT("Failed to get OCSP basic response\n");
                rval = ENStatus_Failure;
                goto out;
            }

            for (k = 0; k < OCSP_resp_count(pOCSPBasicResponse); k++)
            {
                OCSP_SINGLERESP *pOCSPSingleResponse;
                const OCSP_CERTID *pRespCertId;
                ASN1_INTEGER *pRespCertIdSerial;

                pOCSPSingleResponse = OCSP_resp_get0(pOCSPBasicResponse, k);
                if (!pOCSPSingleResponse)
                {
                    ERRORPRINT("Failed to get single response\n");
                    rval = ENStatus_Failure;
                    OCSP_BASICRESP_free(pOCSPBasicResponse);
                    goto out;
                }

                pRespCertId = OCSP_SINGLERESP_get0_id(pOCSPSingleResponse);
                if (OCSP_id_get0_info(NULL, NULL, NULL, &pRespCertIdSerial, 
                    (OCSP_CERTID *)pRespCertId) != 1)
                {
                    ERRORPRINT("Failed to get single response cert serial\n");
                    rval = ENStatus_Failure;
                    OCSP_BASICRESP_free(pOCSPBasicResponse);
                    goto out;
                }

                if (ASN1_INTEGER_cmp(pRespCertIdSerial, pCertId) == 0)
                {
                    ASN1_GENERALIZEDTIME *thisupd;
                    ASN1_GENERALIZEDTIME *nextupd;
                    int lstatus;

                    lstatus = OCSP_single_get0_status(pOCSPSingleResponse, NULL,
                        NULL, &thisupd, &nextupd);
                    if (OCSP_check_validity(thisupd, nextupd, 0, 0) != 1)
                    {
                        ERRORPRINT("OCSP response expired%s\n", 
                            fIgnoreOCSPResponseExpired ? " (ignored)" : "");
                        if (fIgnoreOCSPResponseExpired)
                        {
                            status = lstatus;
                            break;
                        }
                    }
                    else
                    {
                        status = lstatus;
                        break;
                    }
                }
            }

            if (k < OCSP_resp_count(pOCSPBasicResponse))
            {
                break;
            }
        }

        if (status == -1)
        {
            ERRORPRINT("Failed to get OCSP response status\n");
            rval = ENStatus_Failure;
            goto out;
        }

        if (status != V_OCSP_CERTSTATUS_GOOD)
        {
            ERRORPRINT("Invalid OCSP response status %d\n", status);
            rval = ENStatus_Failure;
            goto out;
        }
    }

out:
    if (pBasicResponseStack)
    {
        sk_OCSP_BASICRESP_pop_free(pBasicResponseStack, OCSP_BASICRESP_free);
    }

    if (out)
    {
        BIO_free(out);
    }

    return rval;
}

ENStatus
ENATSC3SSPVerifyAndLoadCertificationData(ENATSC3SSPHandle_t hSSP, char *pXML, 
    int32_t xmlLength, ATSC3LLSCDT_CertificationDataNode *pParsedCDT,
    struct timespec *pMonoTimeout, int32_t fDontVerifySigningCertificate, 
    int32_t fIgnoreOCSPResponseExpired)
{
    ENATSC3SSP *pSSP = (ENATSC3SSP *)hSSP;
    STACK_OF(X509) *pCertificatesStack = NULL;
    STACK_OF(X509) *pCurrentSMTSigningCertStack = NULL;
    STACK_OF(X509) *pNextSMTSigningCertStack = NULL;
    X509_STORE *pStore = NULL;
    STACK_OF(OCSP_RESPONSE) *pOCSPResponseStack = NULL;
    int64_t ocspRefreshSeconds;
    int64_t signingTime;
    ENStatus rval = ENStatus_Success;

    if (fDontVerifySigningCertificate)
    {
        ERRORPRINT("WARN: Not verifying signing certificate chain\n");
    }

    if (fIgnoreOCSPResponseExpired)
    {
        ERRORPRINT("WARN: Ignoring OCSP response expiration status\n");
    }

    pStore = X509_STORE_new();
    if (!pStore)
    {
        ERRORPRINT("Failed to allocate X509 store\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pCertificatesStack = sk_X509_new_null();
    if (!pCertificatesStack)
    {
        ERRORPRINT("Failed to allocate X509 stack\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pCurrentSMTSigningCertStack = sk_X509_new_null();
    if (!pCurrentSMTSigningCertStack)
    {
        ERRORPRINT("Failed to allocate X509 stack\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pNextSMTSigningCertStack = sk_X509_new_null();
    if (!pNextSMTSigningCertStack)
    {
        ERRORPRINT("Failed to allocate X509 stack\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pOCSPResponseStack = sk_OCSP_RESPONSE_new_null();
    if (!pOCSPResponseStack)
    {
        ERRORPRINT("Failed to allocate OCSP response stack\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (pParsedCDT->base.presentFlags & ATSC3LLSCDTRTChildNode_ToBeSignedData)
    {
        if (pParsedCDT->toBeSignedDataNode.base.presentFlags & 
            ATSC3LLCDTToBeSignedDataNodeAttribute_OCSPRefresh)
        {
            if (pParsedCDT->toBeSignedDataNode.ocspRefreshSeconds > 10 * 86400)
            {
                ERRORPRINT("OCSPRefresh too large (%"PRId64"s)\n", 
                    pParsedCDT->toBeSignedDataNode.ocspRefreshSeconds);
                rval = ENStatus_Failure;
                goto out;
            }
            ocspRefreshSeconds = 
                pParsedCDT->toBeSignedDataNode.ocspRefreshSeconds;
        }
        else
        {
            ERRORPRINT("OCSPRefresh missing\n");
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("ToBeSignedData missing\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (ENATSC3SSPExtractCertificationDataTableCertificates(pParsedCDT, 
        pCertificatesStack, pCurrentSMTSigningCertStack, 
        pNextSMTSigningCertStack, pStore))
    {
        ERRORPRINT("Failed to extract certificates\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (pSSP->LoadCertificatesCallback)
    {
        if (pSSP->LoadCertificatesCallback(hSSP, pCertificatesStack, 
            pSSP->pCallbackUserData))
        {
            ERRORPRINT("Failed to load application-supplied certificates\n");
            rval = ENStatus_Failure;
            goto errout;
        }
    }

    if (ENATSC3SSPExtractCertificationDataTableOCSPResponses(pParsedCDT, 
        pOCSPResponseStack))
    {
        ERRORPRINT("Failed to extract certificates\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (ENATSC3SSPVerifyOCSPResponses(pOCSPResponseStack, pCertificatesStack,
        NULL, ocspRefreshSeconds, OCSP_NOVERIFY, fIgnoreOCSPResponseExpired))
    {
        ERRORPRINT("Failed to verify OCSP responses\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (ENATSC3SSPVerifyCertificationDataSignature(pSSP, pXML, xmlLength,
        pParsedCDT, pCertificatesStack, NULL, fDontVerifySigningCertificate, 
        &signingTime))
    {
        ERRORPRINT("Failed to verify certification data\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (ENATSC3SSPCertificateStoreReplaceCertificates(pSSP, pCertificatesStack,
         pOCSPResponseStack, pCurrentSMTSigningCertStack, 
         pNextSMTSigningCertStack, pStore, signingTime, pMonoTimeout))
    {
        ERRORPRINT("Failed to commit certificates\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    return rval;

errout:
    if (pOCSPResponseStack)
    {
        sk_OCSP_RESPONSE_pop_free(pOCSPResponseStack, OCSP_RESPONSE_free);
    }

    if (pCurrentSMTSigningCertStack)
    {
        sk_X509_free(pCurrentSMTSigningCertStack);
    }

    if (pNextSMTSigningCertStack)
    {
        sk_X509_free(pNextSMTSigningCertStack);
    }

    if (pCertificatesStack)
    {
        sk_X509_pop_free(pCertificatesStack, X509_free);
    }

    if (pStore)
    {
        X509_STORE_free(pStore);
    }

    goto out;
}
