/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of
 * the following permission added to Section 15 as permitted in Section 7(a):
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>

#include <libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libenatsc3utils/enxml.h>

#define numof(x) (sizeof(x) / sizeof(x[0]))

typedef struct
{
    EnXMLParserNodeBase base;
    int32_t groupDataOffset;
    int32_t nameDataOffset;
} ParsedChildNode;

typedef struct
{
    EnXMLParserNodeBase base;
    EnXMLParserDataOffsetArray childNodes;
    int32_t groupDataOffset;
} ParsedMainNode;

typedef struct
{
    ParsedMainNode mainNode;
    EnXMLParserDataBuffer dataBuffer;
} ParsedData;

#define NodeType_MainNode       1
#define NodeType_ChildNode      2

/*****************************
 * Child Node Parsing
 */

#define ChildNodeAttribute_group        0x00000001

static ENStatus
ChildNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ParsedData *pParsedData = (ParsedData*)pUserData;
    ParsedChildNode *pChildNode = (ParsedChildNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ChildNodeAttribute_group:
            // copy the group into the EnXMLDataBuffer and note the offset
            // in the ParsedChildNode structure.
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pParsedData->dataBuffer, (char *)pValue, sValue,
                &pChildNode->groupDataOffset))
            {
                printf("Failed to store type\n");
                return ENStatus_Failure;
            }
            break;
        default:
            printf("Unhandled attribute %d (%s)\n", pAttribute->id,
                pAttribute->pName);
       break;
    }

    return ENStatus_Success;
}

static ENStatus
ChildNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ParsedData *pParsedData = (ParsedData*)pUserData;
    ParsedChildNode *pChildNode = (ParsedChildNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pParsedData->dataBuffer,
        (char *)pValue, sValue, &pChildNode->nameDataOffset))
    {
        printf("Failed to store location\n");
        return ENStatus_Failure;
    }
    return ENStatus_Success;
}

// Description for child-node

EnXMLParserAttribute vChildNodeAttributes[] =
{
    {
        .id = ChildNodeAttribute_group,
        .sName = sizeof("group") - 1,
        .pName = "group",
        .handler = ChildNodeAttributeHandler,
    },
};
EnXMLParserChildNode vChildNodeChildNodes[] =
{
};
EnXMLParserNode childNodeDescription =
{
    .pName = "ChildNode",
    .pAttributes = vChildNodeAttributes,
    .nAttributes = numof(vChildNodeAttributes),
    .pChildNodes = vChildNodeChildNodes,
    .nChildNodes = numof(vChildNodeChildNodes),
    .textHandler = ChildNodeTextHandler,
};

static ENStatus
ProcessChildNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ParsedMainNode *pMainNode = (ParsedMainNode *)pParentNode;
    ParsedData *pParsedData = (ParsedData *)pUserData;
    ParsedChildNode childNode;
    int32_t childNodeOffset;

    memset(&childNode, 0, sizeof(childNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the main-node's childNodes data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pParsedData->dataBuffer,
        sizeof(childNode), &childNodeOffset))
    {
        printf("Failed to allocate space for aea node\n");
        return ENStatus_Failure;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pMainNode->childNodes,
        childNodeOffset))
    {
        printf("Failed to store childnode offset\n");
        return ENStatus_Failure;
    }

    // Initialize the childNode
    // Because the memory for dataBuffer may move due to being realloc'd we are
    // using a stack-allocated version of childNode. It will be committed to the
    // dataBuffer after it is parsed.
    childNode.base.nodeType = NodeType_ChildNode;

    if (EnXMLParserProcessNode(&childNodeDescription, reader, pUserData,
        (EnXMLParserNodeBase *)&childNode))
    {
       printf("Failed to parse child-node\n");
        return ENStatus_Failure;
    }

    // Now commit child-node to the dataBuffer
    memcpy(pParsedData->dataBuffer.pBuffer + childNodeOffset, &childNode, sizeof(childNode));

    return ENStatus_Success;
}

/**********************
 * Main Node Parsing
 ********************/

#define MainNodeAttribute_group         0x00000001
#define MainNodeAttribute_foobar        0x00000002
#define MainNodeChildNode_ChildNode     0x00000004

static ENStatus
MainNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ParsedData *pParsedData= (ParsedData*)pUserData;
    ParsedMainNode *pMainNode = (ParsedMainNode *)pParentNode;

    switch (pAttribute->id)
    {
        case MainNodeAttribute_group:
            // copy the group into the EnXMLDataBuffer and note the offset
            // in the ParsedMainNode structure.
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pParsedData->dataBuffer, (char *)pValue, sValue,
                &pMainNode->groupDataOffset))
            {
                printf("Failed to store type\n");
                return ENStatus_Failure;
            }
            break;
        default:
            printf("Unhandled attribute %d (%s)\n", pAttribute->id,
                pAttribute->pName);
       break;
    }

    return ENStatus_Success;
}

// Description for main-node

EnXMLParserAttribute vMainNodeAttributes[] =
{
    {
        .id = MainNodeAttribute_group,
        .sName = sizeof("group") - 1,
        .pName = "group",
        .handler = MainNodeAttributeHandler,
    },
    {
        .id = MainNodeAttribute_foobar,
        .sName = sizeof("foobar") - 1,
        .pName = "foobar",
        .handler = MainNodeAttributeHandler,
    },
};
EnXMLParserChildNode vMainNodeChildNodes[] =
{
    {
        .id = MainNodeChildNode_ChildNode,
        .sName = sizeof("child-node") - 1,
        .pName = "child-node",
        .handler = ProcessChildNode,
    },
};
EnXMLParserNode mainNodeDescription =
{
    .pName = "Example",
    .pAttributes = vMainNodeAttributes,
    .nAttributes = numof(vMainNodeAttributes),
    .pChildNodes = vMainNodeChildNodes,
    .nChildNodes = numof(vMainNodeChildNodes),
    .textHandler = NULL,
};

char document[] = "\
<?xml version = \"1.0\"?>\n\
<main-node group=\"animals\">\n\
    <child-node group=\"dogs\">Chloe</child-node>\n\
    <child-node group=\"cats\">Leo</child-node>\n\
    <child-node group=\"mice\">Mini</child-node>\n\
</main-node>";

int 
main(int argc, char **argv)
{
    ParsedData parsedData;
    char *pXML = document;
    int32_t sXML = strlen(document);

    memset(&parsedData, 0, sizeof(parsedData));

    // Initialize the document-wide EnXMLDataBuffer with an initial size of 8192 bytes increased in
    // 4095 byte increments
    if (EnXMLParserDataBufferInitialize(&parsedData.dataBuffer, 8192, 4096,
        "example"))
    {
        printf("Failed to initialize data buffer\n");
        return -1;
    }

    xmlTextReaderPtr reader = NULL;

    // pXML is a pointer to the XML document and sXML is the size of
    // the document in bytes
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);
    if (!reader)
    {
        printf("Failed to create xml reader\n");
        return -1;
    }

    if (xmlTextReaderRead(reader) == 0)
    {
        printf("failed to read document\n");
        return -1;
    }

    int nodeType;
    const xmlChar *pLocalName;

    nodeType = xmlTextReaderNodeType(reader);
    pLocalName = xmlTextReaderConstLocalName(reader);

    // Process the top node locally to ensure we're in fact dealing
    // with the expected document type
    if (nodeType == XML_READER_TYPE_ELEMENT &&
        !strcmp((char *)pLocalName, "main-node"))
    {
        // Since main-node may have several child-node elements, initialize an 
        // EnXMLDataOffsetArray to hold the offsets of the parsed child-node elements. 
        if (EnXMLParserDataOffsetArrayInitialize(&parsedData.mainNode.childNodes,
            &parsedData.dataBuffer, 16, 16))
        {
            printf("Failed to allocate child-node array\n");
            return -1;
        }

        // Process the main node based on its description. The handlers specified
        // in the description will be called as attributes and child nodes are 
        // encountered.
        if (EnXMLParserProcessNode(&mainNodeDescription, reader, &parsedData,
            (EnXMLParserNodeBase *)&parsedData.mainNode))
        {
            printf("Failed to parse main-mode\n");
            return -1;
        }
    }
    else
    {
        printf("Top-level object isn't main-node (%d/%s)\n", nodeType, 
            pLocalName);
        return -1;
    }
    
    // Now that we have the parsed node let's do something with it...
    printf("main-node\n");
    if (parsedData.mainNode.base.presentFlags & MainNodeAttribute_group)
    {
        printf("  group: %s\n", (char *)EnXMLParserDataBufferGetPointer(
            &parsedData.dataBuffer, parsedData.mainNode.groupDataOffset));
    }
    if (parsedData.mainNode.base.presentFlags & MainNodeChildNode_ChildNode)
    {
        int32_t i;

        for (i = 0; i < parsedData.mainNode.childNodes.nOffsetsUsed; i++)
        {
            ParsedChildNode *pChildNode;

            pChildNode = 
                (ParsedChildNode *)EnXMLParserDataOffsetArrayGetPointer(
                    &parsedData.mainNode.childNodes, i);
            printf("  child-node %d\n", i);
            if (pChildNode->base.presentFlags & ChildNodeAttribute_group)
            {
                printf("    group: %s\n", 
                    (char *)EnXMLParserDataBufferGetPointer(
                    &parsedData.dataBuffer, pChildNode->groupDataOffset));
                printf("    name: %s\n", 
                    (char *)EnXMLParserDataBufferGetPointer(
                    &parsedData.dataBuffer, pChildNode->nameDataOffset));
            }
        }
    }

    return 0;
}
