/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_status.h>

#include "enipfilterset.h"

typedef struct
{
    uint32_t addr;
    uint16_t port;
    uint16_t refcount;
} EnIPFilter;

typedef struct _EnIPFilterSet
{
    int32_t nFiltersInUse;
    int32_t nMaxFilters;
    EnIPFilter vFilters[0];
} EnIPFilterSet;

ENStatus
EnIPFilterSetAllocate(int32_t nMaxFilters, EnIPFilterSet **ppFilterSet)
{
    ENStatus rval = ENStatus_Success;

    *ppFilterSet = (EnIPFilterSet *)calloc(1, sizeof(EnIPFilterSet) +
        sizeof(EnIPFilter) * nMaxFilters);
    if (!*ppFilterSet)
    {
        rval = ENStatus_Failure;
        goto out;
    }
    (*ppFilterSet)->nMaxFilters = nMaxFilters;

out:
    return rval;
}

ENStatus
EnIPFilterSetFree(EnIPFilterSet *pFilterSet)
{
    ENStatus rval = ENStatus_Success;

    if (pFilterSet->nFiltersInUse)
    {
        fprintf(stderr, "WARN: Freeing IPFilterSet with %d filters in use\n",
            pFilterSet->nFiltersInUse);
    }

    free(pFilterSet);

    return rval;
}

void
EnIPFilterSetClear(EnIPFilterSet *pFilterSet)
{
    pFilterSet->nFiltersInUse = 0;
}

ENStatus
EnIPFilterSetAddFilter(uint32_t addr, uint16_t port, EnIPFilterSet *pFilterSet)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < pFilterSet->nFiltersInUse; i++)
    {
        if (pFilterSet->vFilters[i].addr == addr &&
            pFilterSet->vFilters[i].port == port)
        {
            pFilterSet->vFilters[i].refcount++;
            break;
        }
    }

    if (i == pFilterSet->nFiltersInUse)
    {
        if (i < pFilterSet->nMaxFilters)
        {
            pFilterSet->vFilters[i].addr = addr;
            pFilterSet->vFilters[i].port = port;
            pFilterSet->vFilters[i].refcount = 1;
            pFilterSet->nFiltersInUse++;
        }
        else
        {
            rval = ENStatus_Failure;
        }
    }

    return rval;
}

ENStatus
EnIPFilterSetRemoveFilter(uint32_t addr, uint16_t port,
    EnIPFilterSet *pFilterSet)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < pFilterSet->nFiltersInUse; i++)
    {
        if (pFilterSet->vFilters[i].addr == addr &&
            pFilterSet->vFilters[i].port == port)
        {
            pFilterSet->vFilters[i].refcount--;
            if (!pFilterSet->vFilters[i].refcount)
            {
                memmove(pFilterSet->vFilters + i, pFilterSet->vFilters + i + 1,
                    sizeof(EnIPFilter) * (pFilterSet->nFiltersInUse - i - 1));
                pFilterSet->nFiltersInUse--;
            }
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
EnIPFilterSetFilter(uint32_t addr, uint16_t port, EnIPFilterSet *pFilterSet)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < pFilterSet->nFiltersInUse; i++)
    {
        if (pFilterSet->vFilters[i].addr == addr &&
            pFilterSet->vFilters[i].port == port)
        {
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
EnIPFilterSetIter(EnIPFilterSet *pFilterSet, EnIPFilterSetIterFn Iter, 
    void *pUserData)
{
    EnIPFilter *pFilter;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pFilter = pFilterSet->vFilters; 
         i < pFilterSet->nFiltersInUse; 
         i++, pFilter++)
    {
        if (Iter(pFilter->addr, pFilter->port, pUserData))
        {
            rval = ENStatus_Failure;
        }
    }

    return rval;
}

