/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of
 * the following permission added to Section 15 as permitted in Section 7(a):
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/errno.h>

#include <libendebug/libendebug.h>

#include "libenutils_defs.h"
#include "libenutils.h"

static char *ensys = "enutils";
static char *ensubsys = "tsklt";

typedef struct _TaskletPrivateData
{
    char strFile[128];
    char strFunction[128];
    int32_t line; 

    int64_t taskletId;
    int32_t fCancelled;
    int32_t fFreeOnReturn;
    struct timespec runAtMonoTime;

    struct _Tasklet *pNext;
} TaskletPrivateData;

#define TaskletPrivateData(pTasklet) ((TaskletPrivateData *)(((uint8_t *)pTasklet)+sizeof(Tasklet)))

typedef struct _TaskletRunner
{
    int32_t fRunning;

    char strName[32];

    int32_t priority;
    int32_t taskletUserDataBytes;

    pthread_mutex_t queueLock;
    pthread_cond_t queueCond;
    pthread_cond_t emptyQueueCond;

    int32_t fSignalOnEmptyQueue;

    Tasklet *pTasklets;

    pthread_t thread;

    int64_t nextTaskletId;

    // Queue of free tasklets. These are allocated when the runner is created 
    // and persist for its lifetime.
    Tasklet *pFreeTasklets;
    pthread_mutex_t freeTaskletsLock;
} TaskletRunner;

int32_t
ENUtils_TaskletRunnerTaskletAllocate(TaskletRunnerHandle_t hTaskletRunner,
    int32_t taskletUserDataBytes, Tasklet **ppTasklet)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;
    Tasklet *pTasklet;
    int32_t rval = 0;

    if (taskletUserDataBytes > pTaskletRunner->taskletUserDataBytes)
    {
        ERRORPRINT("User data size too large (%d > %d)\n",
            taskletUserDataBytes, pTaskletRunner->taskletUserDataBytes);
        rval = 1;
        goto out;
    }

    if (pthread_mutex_lock(&pTaskletRunner->freeTaskletsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pTasklet = pTaskletRunner->pFreeTasklets;
    if (pTasklet)
    {
        pTaskletRunner->pFreeTasklets = 
            TaskletPrivateData(pTaskletRunner->pFreeTasklets)->pNext;
    }

    if (pthread_mutex_unlock(&pTaskletRunner->freeTaskletsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (!pTasklet)
    {
        pTasklet = (Tasklet *)malloc(sizeof(Tasklet) + 
            sizeof(TaskletPrivateData) +
            pTaskletRunner->taskletUserDataBytes);
        if (!pTasklet)
        {
            ERRORPRINT("Out of memory\n");
            rval = 1;
            goto out;
        }
        TaskletPrivateData(pTasklet)->fFreeOnReturn = 1;
    }

    pTasklet->pArg = (TaskletArg *)(((uint8_t *)pTasklet) + sizeof(Tasklet) + 
        sizeof(TaskletPrivateData));

    *ppTasklet = pTasklet;

out:
    return rval;
}

void
ENUtils_TaskletRunnerTaskletReturn(TaskletRunnerHandle_t hTaskletRunner,
    Tasklet *pTasklet)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;

    if (TaskletPrivateData(pTasklet)->fFreeOnReturn)
    {
        free(pTasklet);
    }
    else
    {
        if (pthread_mutex_lock(&pTaskletRunner->freeTaskletsLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        TaskletPrivateData(pTasklet)->pNext = pTaskletRunner->pFreeTasklets;
        pTaskletRunner->pFreeTasklets = pTasklet;

        if (pthread_mutex_unlock(&pTaskletRunner->freeTaskletsLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }
}

void *
TaskletMain(void *pArg)
{
    pid_t tid;
    TaskletRunner *pTaskletRunner = (TaskletRunner *)pArg;

    // XXX This is non-portable. On Linux you can call setpriority on a 
    // thread id.
    if (pTaskletRunner->priority == TASKLET_PRIORITY_LOW)
    {
        tid = syscall(SYS_gettid);
        if (setpriority(PRIO_PROCESS, tid, 10))
        {
            ERRORPRINT("Failed to lower thread priority: %s\n", 
                strerror(errno));
        }
    }

    if (pthread_mutex_lock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    while (pTaskletRunner->fRunning)
    {
        int ptret;
        Tasklet *pTasklet;

        if (pTaskletRunner->fSignalOnEmptyQueue && 
            !pTaskletRunner->pTasklets)
        {
            if (pthread_cond_signal(&pTaskletRunner->emptyQueueCond))
            {
                ERRORPRINT("Failed to signal");
            }

            pTaskletRunner->fSignalOnEmptyQueue = 0;
        }

        if (pTaskletRunner->pTasklets)
        {
            if (!TaskletPrivateData(pTaskletRunner->pTasklets)->fCancelled)
            {
                ptret = pthread_cond_timedwait(&pTaskletRunner->queueCond, 
                    &pTaskletRunner->queueLock, 
                    &TaskletPrivateData(pTaskletRunner->pTasklets)->runAtMonoTime);
                if (ptret)
                {
                    if (ptret != ETIMEDOUT)
                    {
                        ERRORPRINT("Failed to wait on cond\n");
                    }
                }
            }
        }
        else
        {
            ptret = pthread_cond_wait(&pTaskletRunner->queueCond, 
                &pTaskletRunner->queueLock);
            if (ptret)
            {
                ERRORPRINT("Failed to wait on cond\n");
            }
        }

        while (pTaskletRunner->pTasklets)
        {
            pTasklet = pTaskletRunner->pTasklets;

            if (!TaskletPrivateData(pTasklet)->fCancelled)
            {
                // Tasklets with a runAtTime.tv_sec == 0 are always run
                // immediately. Otherwise we need to check if the tasklet
                // has matured.
                if (TaskletPrivateData(pTasklet)->runAtMonoTime.tv_sec != 0)
                {
                    struct timespec mononow;

                    clock_gettime(CLOCK_MONOTONIC, &mononow);
                    if (mononow.tv_sec < 
                        TaskletPrivateData(pTasklet)->runAtMonoTime.tv_sec ||
                        (mononow.tv_sec == 
                         TaskletPrivateData(pTasklet)->runAtMonoTime.tv_sec &&
                         mononow.tv_nsec < 
                         TaskletPrivateData(pTasklet)->runAtMonoTime.tv_nsec))
                    {
                        break;
                    }
                }

                pTaskletRunner->pTasklets = TaskletPrivateData(pTasklet)->pNext;

                if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
                {
                    ERRORPRINT("Failed to unlock");
                }

                pTasklet->Function(pTaskletRunner, pTasklet->pArg);

                if (pthread_mutex_lock(&pTaskletRunner->queueLock))
                {
                    ERRORPRINT("Failed to lock");
                }
            }
            else
            {
                pTaskletRunner->pTasklets = TaskletPrivateData(pTasklet)->pNext;
            }

            ENUtils_TaskletRunnerTaskletReturn(pTaskletRunner, pTasklet);
        }
    }

    if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return NULL;
}

int32_t
ENUtils_TaskletRunnerCreate(char *pName, int32_t priority, 
    int32_t taskletUserDataBytes, int32_t nTasklets, 
    TaskletRunnerHandle_t *phTaskletRunner)
{
    pthread_attr_t threadAttr;
    struct sched_param sched_param;
    int32_t fHaveThreadAttr = 0;
    int32_t fHaveQueueLock = 0;
    int32_t fHaveQueueCond = 0;
    int32_t fHaveEmptyQueueCond = 0;
    int32_t fHaveCondattr = 0;
    int32_t fHaveFreeTaskletsLock = 0;
    pthread_condattr_t monocondattr;
    char strThreadName[16];
    TaskletRunner *pTaskletRunner = NULL;
    int ptret;
    int32_t i;
    int32_t rval = 0;

    pTaskletRunner = (TaskletRunner *)malloc(sizeof(TaskletRunner));
    if (!pTaskletRunner)
    {
        ERRORPRINT("Out of memory\n");
        rval = -1;
        goto errout;
    }

    strncpy(pTaskletRunner->strName, pName, 
        sizeof(pTaskletRunner->strName) - 1);
    pTaskletRunner->strName[sizeof(pTaskletRunner->strName) - 1] = 0;
    pTaskletRunner->pTasklets = NULL;
    pTaskletRunner->pFreeTasklets = NULL;
    pTaskletRunner->nextTaskletId = 1;
    pTaskletRunner->priority = priority;
    pTaskletRunner->fSignalOnEmptyQueue = 0;
    pTaskletRunner->taskletUserDataBytes = taskletUserDataBytes;

    if (pthread_mutex_init(&pTaskletRunner->freeTaskletsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = -1;
        goto errout;
    }
    fHaveFreeTaskletsLock = 1;

    if (pthread_mutex_init(&pTaskletRunner->queueLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = -1;
        goto errout;
    }
    fHaveQueueLock = 1;

    if (pthread_condattr_init(&monocondattr))
    {
        ERRORPRINT("Failed to init condattr\n");
        rval = -1;
        goto errout;
    }
    fHaveCondattr = 1;

    if (pthread_condattr_setclock(&monocondattr, CLOCK_MONOTONIC))
    {
        ERRORPRINT("Failed to set condattr clock\n");
        rval = -1;
        goto errout;
    }

    if (pthread_cond_init(&pTaskletRunner->queueCond, &monocondattr))
    {
        ERRORPRINT("Failed to init cond\n");
        rval = -1;
        goto errout;
    }
    fHaveQueueCond = 1;

    if (pthread_cond_init(&pTaskletRunner->emptyQueueCond, &monocondattr))
    {
        ERRORPRINT("Failed to init cond\n");
        rval = -1;
        goto errout;
    }
    fHaveEmptyQueueCond = 1;

    if (pthread_attr_init(&threadAttr))
    {
        ERRORPRINT("Failed to init thread attr\n");
        rval = -1;
        goto errout;
    }
    fHaveThreadAttr = 1;

    if (priority >= TASKLET_PRIORITY_REALTIME_LOW && 
        priority <= TASKLET_PRIORITY_REALTIME_HIGH)
    {
        if (pthread_attr_setinheritsched(&threadAttr, PTHREAD_EXPLICIT_SCHED))
        {
            ERRORPRINT("Failed to disable inherited scheduling policy\n");
        }

        if (pthread_attr_setschedpolicy(&threadAttr, SCHED_FIFO))
        {
            ERRORPRINT("Failed to set RT FIFO scheduling policy\n");
        }

        if (priority == TASKLET_PRIORITY_REALTIME_LOW)
        {
            sched_param.sched_priority = ENUtils_Scheduling_Realtime_Low;
        }
        else if (priority == TASKLET_PRIORITY_REALTIME_MEDIUM)
        {
            sched_param.sched_priority = ENUtils_Scheduling_Realtime_Medium;
        }
        else if (priority == TASKLET_PRIORITY_REALTIME_HIGH)
        {
            sched_param.sched_priority = ENUtils_Scheduling_Realtime_High;
        }
        else
        {
            ERRORPRINT("Invalid tasklet priority %d\n", priority);
            sched_param.sched_priority = ENUtils_Scheduling_Realtime_Low;
        }

        if (pthread_attr_setschedparam(&threadAttr, &sched_param))
        {
            ERRORPRINT("Failed to set RT FIFO scheduling prority\n");
        }
    }

    pTaskletRunner->fRunning = 1;

    ptret = pthread_create(&pTaskletRunner->thread, &threadAttr, TaskletMain, 
        pTaskletRunner);
    if (ptret)
    {
        ERRORPRINT("Failed to start thread at requested priority (%s), trying non-RT\n", strerror(ptret));
        if (pthread_create(&pTaskletRunner->thread, NULL, TaskletMain, 
            pTaskletRunner))
        {
            ERRORPRINT("Failed to start thread\n");
            rval = -1;
            goto errout;
        }
    }

    if (snprintf(strThreadName, sizeof(strThreadName), "tsk%s", 
        pTaskletRunner->strName) >= sizeof(strThreadName))
    {
        strThreadName[sizeof(strThreadName) - 1] = 0;
    }

    if (ENUtils_SetThreadName(pTaskletRunner->thread, strThreadName))
    {
        ERRORPRINT("WARN: Failed to set thread name\n");
    }

    if (pthread_attr_destroy(&threadAttr)) 
    { 
        ERRORPRINT("Failed to destroy thread attr\n"); 
    }
    fHaveThreadAttr = 0;

    for (i = 0; i < nTasklets; i++)
    {
        Tasklet *pTasklet;

        pTasklet = (Tasklet *)malloc(sizeof(Tasklet) + 
            sizeof(TaskletPrivateData) + pTaskletRunner->taskletUserDataBytes);
        if (!pTasklet)
        {
            ERRORPRINT("Out of memory\n");
            rval = 1;
            goto out;
        }

        TaskletPrivateData(pTasklet)->fFreeOnReturn = 0;
        TaskletPrivateData(pTasklet)->pNext = pTaskletRunner->pFreeTasklets;
        pTaskletRunner->pFreeTasklets = pTasklet;
    }

    *phTaskletRunner = pTaskletRunner;

out:
    if (fHaveCondattr)
    {
        if (pthread_condattr_destroy(&monocondattr))
        {
            ERRORPRINT("Failed to destroy condattr\n");
        }
    }

    return rval;

errout:
    if (fHaveThreadAttr)
    {
        if (pthread_attr_destroy(&threadAttr)) 
        { 
            ERRORPRINT("Failed to destroy thread attr\n"); 
        }
    }

    if (pTaskletRunner)
    {
        while (pTaskletRunner->pFreeTasklets)
        {
            Tasklet *pNext = 
                TaskletPrivateData(pTaskletRunner->pFreeTasklets)->pNext;

            free(pTaskletRunner->pFreeTasklets);

            pTaskletRunner->pFreeTasklets = pNext;
        }

        if (fHaveEmptyQueueCond)
        {
            if (pthread_cond_destroy(&pTaskletRunner->emptyQueueCond))
            {
                ERRORPRINT("Failed to destroy cond\n");
            }
        }

        if (fHaveQueueCond)
        {
            if (pthread_cond_destroy(&pTaskletRunner->queueCond))
            {
                ERRORPRINT("Failed to destroy cond\n");
            }
        }

        if (fHaveQueueLock)
        {
            if (pthread_mutex_destroy(&pTaskletRunner->queueLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (fHaveFreeTaskletsLock)
        {
            if (pthread_mutex_destroy(&pTaskletRunner->freeTaskletsLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        free(pTaskletRunner);
    }

    goto out;
}

int32_t
ENUtils_TaskletRunnerDestroy(TaskletRunnerHandle_t hTaskletRunner)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;
    int32_t rval = 0;

    pTaskletRunner->fRunning = 0;

    if (pthread_cond_signal(&pTaskletRunner->queueCond))
    {
        ERRORPRINT("Failed to signal tasklet runner\n");
    }

    if (pthread_join(pTaskletRunner->thread, NULL))
    {
        ERRORPRINT("Failed to join tasklet runner\n");
    }

    if (pthread_cond_destroy(&pTaskletRunner->emptyQueueCond))
    {
        ERRORPRINT("Failed to destroy cond\n");
    }

    if (pthread_cond_destroy(&pTaskletRunner->queueCond))
    {
        ERRORPRINT("Failed to destroy cond\n");
    }

    if (pthread_mutex_destroy(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pthread_mutex_destroy(&pTaskletRunner->freeTaskletsLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    while (pTaskletRunner->pTasklets)
    {
        Tasklet *pNext = TaskletPrivateData(pTaskletRunner->pTasklets)->pNext;

        free(pTaskletRunner->pTasklets);

        pTaskletRunner->pTasklets = pNext;
    }

    while (pTaskletRunner->pFreeTasklets)
    {
        Tasklet *pNext = 
            TaskletPrivateData(pTaskletRunner->pFreeTasklets)->pNext;

        free(pTaskletRunner->pFreeTasklets);

        pTaskletRunner->pFreeTasklets = pNext;
    }

    free(pTaskletRunner);

    return rval;
}

int32_t
ENUtils_TaskletRunnerRunAllTasklets(TaskletRunnerHandle_t hTaskletRunner)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;

    if (pthread_mutex_lock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock");
    }

    if (pthread_cond_signal(&pTaskletRunner->queueCond))
    {
        ERRORPRINT("Failed to signal\n");
    }

    pTaskletRunner->fSignalOnEmptyQueue = 1;

    if (pthread_cond_wait(&pTaskletRunner->emptyQueueCond, 
        &pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to wait\n");
    }

    if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to unlock");
    }

    return 0;
}

int32_t
ENUtils_TaskletRunnerAddTaskletWithExecutionTimeMarked(
    TaskletRunnerHandle_t hTaskletRunner, Tasklet *pTasklet, 
    struct timespec runAtMonoTime, const char *pFile, const char *pFunction, 
    int32_t line, TaskletHandle_t *phTasklet)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;
    int32_t rval = 0;

    TaskletPrivateData(pTasklet)->runAtMonoTime = runAtMonoTime;
    strncpy(TaskletPrivateData(pTasklet)->strFile, pFile, 
        sizeof(TaskletPrivateData(pTasklet)->strFile));
    TaskletPrivateData(pTasklet)->strFile[
        sizeof(TaskletPrivateData(pTasklet)->strFile) - 1] = 0;
    strncpy(TaskletPrivateData(pTasklet)->strFunction, pFunction, 
        sizeof(TaskletPrivateData(pTasklet)->strFunction));
    TaskletPrivateData(pTasklet)->strFunction[
        sizeof(TaskletPrivateData(pTasklet)->strFunction) - 1] = 0;
    TaskletPrivateData(pTasklet)->line = line;
    TaskletPrivateData(pTasklet)->pNext = NULL;

    if (pthread_mutex_lock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    TaskletPrivateData(pTasklet)->taskletId = pTaskletRunner->nextTaskletId++;
    TaskletPrivateData(pTasklet)->fCancelled = 0;

    if (phTasklet)
    {
        *phTasklet = TaskletPrivateData(pTasklet)->taskletId;
    }

    if (!pTaskletRunner->pTasklets)
    {
        pTaskletRunner->pTasklets = pTasklet;
    }
    else
    {
        Tasklet *pIter;
        Tasklet *pPrev = NULL;

        pIter = pTaskletRunner->pTasklets;
        while (pIter &&
            (TaskletPrivateData(pIter)->runAtMonoTime.tv_sec < 
            runAtMonoTime.tv_sec ||
             (TaskletPrivateData(pIter)->runAtMonoTime.tv_sec == 
             runAtMonoTime.tv_sec &&
              TaskletPrivateData(pIter)->runAtMonoTime.tv_nsec <= 
              runAtMonoTime.tv_nsec)))
        {
            pPrev = pIter;
            pIter = TaskletPrivateData(pIter)->pNext;
        }
        if (!pPrev)
        {
            // Prepend
            TaskletPrivateData(pTasklet)->pNext = pTaskletRunner->pTasklets;
            pTaskletRunner->pTasklets = pTasklet;
        }
        else if (pIter)
        {
            // Insert between pPrev and pIter
            TaskletPrivateData(pTasklet)->pNext = pIter;
            TaskletPrivateData(pPrev)->pNext = pTasklet;
        }
        else
        {
            // Append to list
            TaskletPrivateData(pPrev)->pNext = pTasklet;
        }
    }

    if (pthread_cond_signal(&pTaskletRunner->queueCond))
    {
        ERRORPRINT("Failed to signal tasklet runner\n");
    }

    if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

int32_t
ENUtils_TaskletRunnerAddTaskletMarked(TaskletRunnerHandle_t hTaskletRunner,
    Tasklet *pTasklet, const char *pFile, const char *pFunction, int32_t line, 
    TaskletHandle_t *phTasklet)
{
    struct timespec zerotime = {.tv_sec = 0, .tv_nsec = 0};
    int32_t rval;

    rval = ENUtils_TaskletRunnerAddTaskletWithExecutionTimeMarked(
        hTaskletRunner, pTasklet, zerotime, pFile, pFunction, line, phTasklet);

    return rval;
}

int32_t
ENUtils_TaskletRunnerCancelTasklet(TaskletRunnerHandle_t hTaskletRunner,
    TaskletHandle_t hTasklet)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;
    int32_t rval = 0;
    Tasklet *pIter;

    if (pthread_mutex_lock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pIter = pTaskletRunner->pTasklets;

    while (pIter)
    {
        if (TaskletPrivateData(pIter)->taskletId == hTasklet)
        {
            TaskletPrivateData(pIter)->fCancelled = 1;
            break;
        }
        pIter = TaskletPrivateData(pIter)->pNext;
    }

    if (!pIter)
    {
        rval = 1;
    }

    if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

int32_t
ENUtils_TaskletRunnerPeek(TaskletRunnerHandle_t hTaskletRunner,
    TaskletFunction Function, TaskletArg **ppArg, TaskletHandle_t *phTasklet)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;
    Tasklet *pTasklet;
    int32_t rval =  -1;

    if (pthread_mutex_lock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock");
    }

    pTasklet = pTaskletRunner->pTasklets;
    while (pTasklet)
    {
        if (pTasklet->Function == Function)
        {
            *ppArg = pTasklet->pArg;
            *phTasklet = TaskletPrivateData(pTasklet)->taskletId;
            rval = 0;
            break;
        }
        pTasklet = TaskletPrivateData(pTasklet)->pNext;
    }

    if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock");
    }

    return rval;
}

int32_t
ENUtils_TaskletRunnerCheckPending(TaskletRunnerHandle_t hTaskletRunner,
    TaskletFunction Function)
{
    TaskletRunner *pTaskletRunner = (TaskletRunner *)hTaskletRunner;
    Tasklet *pTasklet;
    int32_t rval = 0;

    if (pthread_mutex_lock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock");
    }

    pTasklet = pTaskletRunner->pTasklets;
    while (pTasklet)
    {
        if (pTasklet->Function == Function)
        {
            rval = 1;
            break;
        }

        pTasklet = TaskletPrivateData(pTasklet)->pNext;
    }

    if (pthread_mutex_unlock(&pTaskletRunner->queueLock))
    {
        ERRORPRINT("Failed to lock");
    }

    return rval;
}
